/* MciAvi resources
 *
 * Copyright (c) 2000, Eric Pouech
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "windef.h"
#include "winbase.h"
#include "mmddk.h"
#include "digitalv.h"

#define MCI_DWORD_PTR 13 /* MCI_INTEGER64 */

MCIAVI RCDATA
BEGIN
L"play\0", MCI_PLAY, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"from\0", 0x00000004L, MCI_INTEGER,
L"to\0", 0x00000008L, MCI_INTEGER,
L"repeat\0", 0x00010000L, MCI_FLAG,
L"reverse\0", 0x00020000L, MCI_FLAG,
L"window\0", 0x01000000L, MCI_FLAG,
L"fullscreen by 2\0", 0x04000000L, MCI_FLAG,
L"fullscreen\0", 0x02000000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"capability\0", MCI_GETDEVCAPS, 0, MCI_COMMAND_HEAD,
L"\0", MCI_INTEGER, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"\0", 0x00000100L, MCI_CONSTANT,
L"can record\0", 0x00000001L, MCI_INTEGER,
L"has audio\0", 0x00000002L, MCI_INTEGER,
L"has video\0", 0x00000003L, MCI_INTEGER,
L"uses files\0", 0x00000005L, MCI_INTEGER,
L"compound device\0", 0x00000006L, MCI_INTEGER,
L"device type\0", 0x00000004L, MCI_INTEGER,
L"can eject\0", 0x00000007L, MCI_INTEGER,
L"can play\0", 0x00000008L, MCI_INTEGER,
L"can save\0", 0x00000009L, MCI_INTEGER,
L"can lock\0", 0x00004000L, MCI_INTEGER,
L"can reverse\0", 0x00004004L, MCI_INTEGER,
L"can stretch input\0", 0x00004008L, MCI_INTEGER,
L"can stretch\0", 0x00004001L, MCI_INTEGER,
L"can test\0", 0x00004009L, MCI_INTEGER,
L"has still\0", 0x00004005L, MCI_INTEGER,
L"can freeze\0", 0x00004002L, MCI_INTEGER,
L"uses palettes\0", 0x00004006L, MCI_INTEGER,
L"windows\0", 0x00004003L, MCI_INTEGER,
L"maximum play rate\0", 0x0000400aL, MCI_INTEGER,
L"minimum play rate\0", 0x0000400bL, MCI_INTEGER,
L"can colorkey\0", 0x00004100L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"capture\0", MCI_CAPTURE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"as\0", 0x00010000L, MCI_STRING,
L"at\0", 0x00020000L, MCI_RECT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"close\0", MCI_CLOSE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"configure\0", MCI_CONFIGURE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"copy\0", MCI_COPY, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"from\0", 0x00000004L, MCI_INTEGER,
L"to\0", 0x00000008L, MCI_INTEGER,
L"at\0", 0x00010000L, MCI_RECT,
L"audio stream\0", 0x00020000L, MCI_INTEGER,
L"video stream\0", 0x00040000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"cue\0", MCI_CUE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"input\0", 0x00010000L, MCI_FLAG,
L"output\0", 0x00020000L, MCI_FLAG,
L"to\0", 0x00000008L, MCI_INTEGER,
L"noshow\0", 0x00040000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"cut\0", MCI_CUT, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"from\0", 0x00000004L, MCI_INTEGER,
L"to\0", 0x00000008L, MCI_INTEGER,
L"at\0", 0x00010000L, MCI_RECT,
L"audio stream\0", 0x00020000L, MCI_INTEGER,
L"video stream\0", 0x00040000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"delete\0", MCI_DELETE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"from\0", 0x00000004L, MCI_INTEGER,
L"to\0", 0x00000008L, MCI_INTEGER,
L"at\0", 0x00010000L, MCI_RECT,
L"audio stream\0", 0x00020000L, MCI_INTEGER,
L"video stream\0", 0x00040000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"freeze\0", MCI_FREEZE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"at\0", 0x00010000L, MCI_RECT,
L"outside\0", 0x00020000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"info\0", MCI_INFO, 0, MCI_COMMAND_HEAD,
L"\0", MCI_STRING, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"file\0", 0x00000200L, MCI_FLAG,
L"product\0", 0x00000100L, MCI_FLAG,
L"window text\0", 0x00010000L, MCI_FLAG,
L"usage\0", 0x00004000L, MCI_FLAG,
L"version\0", 0x00000400L, MCI_FLAG,
L"\0", 0x00020000L, MCI_CONSTANT,
L"audio algorithm\0", 0x00004004L, MCI_INTEGER,
L"audio quality\0", 0x00004001L, MCI_INTEGER,
L"still algorithm\0", 0x00004005L, MCI_INTEGER,
L"still quality\0", 0x00004002L, MCI_INTEGER,
L"video algorithm\0", 0x00004006L, MCI_INTEGER,
L"video quality\0", 0x00004003L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"list\0", MCI_LIST, 0, MCI_COMMAND_HEAD,
L"\0", MCI_STRING, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"count\0", 0x00020000L, MCI_FLAG,
L"number\0", 0x00040000L, MCI_INTEGER,
L"\0", 0x00010000L, MCI_CONSTANT,
L"audio algorithm\0", 0x00004000L, MCI_INTEGER,
L"audio quality\0", 0x00004001L, MCI_INTEGER,
L"audio stream\0", 0x00004002L, MCI_INTEGER,
L"still algorithm\0", 0x00004003L, MCI_INTEGER,
L"still quality\0", 0x00004004L, MCI_INTEGER,
L"video algorithm\0", 0x00004005L, MCI_INTEGER,
L"video quality\0", 0x00004006L, MCI_INTEGER,
L"video source\0", 0x00004008L, MCI_INTEGER,
L"video stream\0", 0x00004007L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"algorithm\0", 0x00080000L, MCI_STRING,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"load\0", MCI_LOAD, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"\0", 0x00000100L, MCI_STRING,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"monitor\0", MCI_MONITOR, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"\0", 0x00020000L, MCI_CONSTANT,
L"input\0", 0x00004000L, MCI_INTEGER,
L"file\0", 0x00004001L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"method\0", 0x00010000L, MCI_CONSTANT,
L"pre\0", 0x0000a000L, MCI_INTEGER,
L"post\0", 0x0000a001L, MCI_INTEGER,
L"direct\0", 0x0000a002L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"open\0", MCI_OPEN, 0, MCI_COMMAND_HEAD,
L"\0", MCI_INTEGER, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"type\0", 0x00002000L, MCI_STRING,
L"element\0", 0x00000200L, MCI_STRING,
L"alias\0", 0x00000400L, MCI_STRING,
L"shareable\0", 0x00000100L, MCI_FLAG,
L"style\0", 0x00010000L, MCI_CONSTANT,
L"overlapped\0", 0x00cf0000L, MCI_INTEGER,
L"popup\0", 0x80880000L, MCI_INTEGER,
L"child\0", 0x40000000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"parent\0", 0x00020000L, MCI_HWND,
L"nostatic\0", 0x00040000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"paste\0", MCI_PASTE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"to\0", 0x00000008L, MCI_INTEGER,
L"at\0", 0x00010000L, MCI_RECT,
L"insert\0", 0x00080000L, MCI_FLAG,
L"overwrite\0", 0x00100000L, MCI_FLAG,
L"audio stream\0", 0x00020000L, MCI_INTEGER,
L"video stream\0", 0x00040000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"pause\0", MCI_PAUSE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"put\0", MCI_PUT, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"at\0", 0x00010000L, MCI_RECT,
L"source\0", 0x00020000L, MCI_FLAG,
L"destination\0", 0x00040000L, MCI_FLAG,
L"frame\0", 0x00080000L, MCI_FLAG,
L"video\0", 0x00100000L, MCI_FLAG,
L"window\0", 0x00200000L, MCI_FLAG,
L"client\0", 0x00400000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"quality\0", MCI_QUALITY, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"\0", 0x00010000L, MCI_CONSTANT,
L"audio\0", 0x00004000L, MCI_INTEGER,
L"still\0", 0x00004001L, MCI_INTEGER,
L"video\0", 0x00004002L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"name\0", 0x00020000L, MCI_STRING,
L"algorithm\0", 0x00040000L, MCI_STRING,
L"dialog\0", 0x00080000L, MCI_FLAG,
L"handle\0", 0x00100000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"realize\0", MCI_REALIZE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"normal\0", 0x00010000L, MCI_FLAG,
L"background\0", 0x00020000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"record\0", MCI_RECORD, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"from\0", 0x00000004L, MCI_INTEGER,
L"to\0", 0x00000008L, MCI_INTEGER,
L"insert\0", 0x00000100L, MCI_FLAG,
L"overwrite\0", 0x00000200L, MCI_FLAG,
L"at\0", 0x00010000L, MCI_RECT,
L"hold\0", 0x00020000L, MCI_FLAG,
L"audio stream\0", 0x00040000L, MCI_INTEGER,
L"video stream\0", 0x00080000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"reserve\0", MCI_RESERVE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"in\0", 0x00010000L, MCI_STRING,
L"size\0", 0x00020000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"restore\0", MCI_RESTORE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"from\0", 0x00010000L, MCI_STRING,
L"at\0", 0x00020000L, MCI_RECT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"resume\0", MCI_RESUME, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"save\0", MCI_SAVE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"\0", 0x00000100L, MCI_STRING,
L"at\0", 0x00010000L, MCI_RECT,
L"abort\0", 0x00020000L, MCI_FLAG,
L"keepreserve\0", 0x00040000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"seek\0", MCI_SEEK, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"to start\0", 0x00000100L, MCI_FLAG,
L"to end\0", 0x00000200L, MCI_FLAG,
L"to\0", 0x00000008L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"set\0", MCI_SET, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"time format\0", 0x00000400L, MCI_CONSTANT,
L"frames\0", 0x00000003L, MCI_INTEGER,
L"milliseconds\0", 0x00000000L, MCI_INTEGER,
L"ms\0", 0x00000000L, MCI_INTEGER,
L"bytes\0", 0x00000008L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"door open\0", 0x00000100L, MCI_FLAG,
L"door closed\0", 0x00000200L, MCI_FLAG,
L"audio\0", 0x00000800L, MCI_CONSTANT,
L"all\0", 0x00000000L, MCI_INTEGER,
L"left\0", 0x00000001L, MCI_INTEGER,
L"right\0", 0x00000002L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"video\0", 0x00001000L, MCI_FLAG,
L"seek exactly\0", 0x00010000L, MCI_FLAG,
L"file format\0", 0x00080000L, MCI_CONSTANT,
L"avss\0", 0x00004000L, MCI_INTEGER,
L"avi\0", 0x00004001L, MCI_INTEGER,
L"dib\0", 0x00004002L, MCI_INTEGER,
L"rdib\0", 0x00004003L, MCI_INTEGER,
L"jpeg\0", 0x00004004L, MCI_INTEGER,
L"rjpeg\0", 0x00004005L, MCI_INTEGER,
L"jfif\0", 0x00004006L, MCI_INTEGER,
L"mpeg\0", 0x00004007L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"still\0", 0x00040000L, MCI_FLAG,
L"on\0", 0x00002000L, MCI_FLAG,
L"off\0", 0x00004000L, MCI_FLAG,
L"speed\0", 0x00020000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"setaudio\0", MCI_SETAUDIO, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"\0", 0x00800000L, MCI_CONSTANT,
L"bass\0", 0x00004001L, MCI_INTEGER,
L"treble\0", 0x00004000L, MCI_INTEGER,
L"volume\0", 0x00004002L, MCI_INTEGER,
L"stream\0", 0x00004003L, MCI_INTEGER,
L"source\0", 0x00004004L, MCI_INTEGER,
L"samplespersec\0", 0x00004005L, MCI_INTEGER,
L"bytespersec\0", 0x00004006L, MCI_INTEGER,
L"alignment\0", 0x00004007L, MCI_INTEGER,
L"bitspersample\0", 0x00004008L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"to\0", 0x01000000L, MCI_CONSTANT,
L"left\0", 0x00000001L, MCI_INTEGER,
L"right\0", 0x00000002L, MCI_INTEGER,
L"average\0", 0x00004000L, MCI_INTEGER,
L"stereo\0", 0x00000000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"over\0", 0x00010000L, MCI_INTEGER,
L"algorithm\0", 0x00040000L, MCI_STRING,
L"quality\0", 0x00080000L, MCI_STRING,
L"record\0", 0x00100000L, MCI_FLAG,
L"left\0", 0x00200000L, MCI_FLAG,
L"clocktime\0", 0x00020000L, MCI_FLAG,
L"right\0", 0x00400000L, MCI_FLAG,
L"on\0", 0x00002000L, MCI_FLAG,
L"off\0", 0x00004000L, MCI_FLAG,
L"input\0", 0x02000000L, MCI_FLAG,
L"output\0", 0x04000000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"setvideo\0", MCI_SETVIDEO, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"\0", 0x00100000L, MCI_CONSTANT,
L"frame rate\0", 0x00004008L, MCI_INTEGER,
L"brightness\0", 0x00004000L, MCI_INTEGER,
L"color\0", 0x00004001L, MCI_INTEGER,
L"contrast\0", 0x00004002L, MCI_INTEGER,
L"tint\0", 0x00004003L, MCI_INTEGER,
L"sharpness\0", 0x00004004L, MCI_INTEGER,
L"gamma\0", 0x00004005L, MCI_INTEGER,
L"palette handle\0", 0x00004007L, MCI_INTEGER,
L"stream\0", 0x00004006L, MCI_INTEGER,
L"source\0", 0x00004009L, MCI_INTEGER,
L"key index\0", 0x0000400aL, MCI_INTEGER,
L"key color\0", 0x0000400bL, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"to\0", 0x01000000L, MCI_CONSTANT,
L"ntsc\0", 0x00004000L, MCI_INTEGER,
L"rgb\0", 0x00004001L, MCI_INTEGER,
L"svideo\0", 0x00004002L, MCI_INTEGER,
L"pal\0", 0x00004003L, MCI_INTEGER,
L"secam\0", 0x00004004L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"quality\0", 0x00010000L, MCI_STRING,
L"algorithm\0", 0x00020000L, MCI_STRING,
L"number\0", 0x00080000L, MCI_INTEGER,
L"over\0", 0x00200000L, MCI_INTEGER,
L"record\0", 0x00400000L, MCI_FLAG,
L"still\0", 0x00800000L, MCI_FLAG,
L"clocktime\0", 0x00040000L, MCI_FLAG,
L"on\0", 0x00002000L, MCI_FLAG,
L"off\0", 0x00004000L, MCI_FLAG,
L"input\0", 0x02000000L, MCI_FLAG,
L"output\0", 0x04000000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"signal\0", MCI_SIGNAL, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"at\0", 0x00010000L, MCI_INTEGER,
L"every\0", 0x00020000L, MCI_INTEGER,
L"uservalue\0", 0x00040000L, MCI_INTEGER,
L"cancel\0", 0x00080000L, MCI_FLAG,
L"return position\0", 0x00100000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"status\0", MCI_STATUS, 0, MCI_COMMAND_HEAD,
L"\0", MCI_DWORD_PTR, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"\0", 0x00000100L, MCI_CONSTANT,
L"frames skipped\0", 0x00008001L, MCI_INTEGER,
L"play speed\0", 0x00008002L, MCI_INTEGER,
L"audio breaks\0", 0x00008003L, MCI_INTEGER,
L"alignment\0", 0x00004029L, MCI_INTEGER,
L"audio input\0", 0x00004000L, MCI_INTEGER,
L"audio record\0", 0x0000401aL, MCI_INTEGER,
L"audio source\0", 0x00004009L, MCI_INTEGER,
L"audio stream\0", 0x0000402dL, MCI_INTEGER,
L"bass\0", 0x0000400fL, MCI_INTEGER,
L"bitsperpel\0", 0x0000402bL, MCI_INTEGER,
L"bitspersample\0", 0x0000402aL, MCI_INTEGER,
L"brightness\0", 0x00004005L, MCI_INTEGER,
L"bytespersec\0", 0x00004028L, MCI_INTEGER,
L"color\0", 0x00004006L, MCI_INTEGER,
L"contrast\0", 0x00004007L, MCI_INTEGER,
L"current track\0", 0x00000008L, MCI_INTEGER,
L"file format\0", 0x00004008L, MCI_INTEGER,
L"still file format\0", 0x0000401dL, MCI_INTEGER,
L"file mode\0", 0x0000401fL, MCI_INTEGER,
L"file completion\0", 0x00004020L, MCI_INTEGER,
L"forward\0", 0x0000402cL, MCI_INTEGER,
L"gamma\0", 0x0000400aL, MCI_INTEGER,
L"length\0", 0x00000001L, MCI_INTEGER,
L"media present\0", 0x00000005L, MCI_INTEGER,
L"mode\0", 0x00000004L, MCI_INTEGER,
L"monitor method\0", 0x0000400cL, MCI_INTEGER,
L"monitor\0", 0x0000400bL, MCI_INTEGER,
L"number of tracks\0", 0x00000003L, MCI_INTEGER,
L"palette handle\0", 0x00004004L, MCI_INTEGER,
L"pause mode\0", 0x00004026L, MCI_INTEGER,
L"position\0", 0x00000002L, MCI_INTEGER,
L"ready\0", 0x00000007L, MCI_INTEGER,
L"frame rate\0", 0x0000400eL, MCI_INTEGER,
L"reserved size\0", 0x00004010L, MCI_INTEGER,
L"samplespersec\0", 0x00004027L, MCI_INTEGER,
L"seek exactly\0", 0x00004011L, MCI_INTEGER,
L"sharpness\0", 0x00004012L, MCI_INTEGER,
L"smpte\0", 0x00004013L, MCI_INTEGER,
L"speed\0", 0x00004003L, MCI_INTEGER,
L"time format\0", 0x00000006L, MCI_INTEGER,
L"tint\0", 0x00004015L, MCI_INTEGER,
L"treble\0", 0x00004016L, MCI_INTEGER,
L"unsaved\0", 0x00004017L, MCI_INTEGER,
L"video key color\0", 0x00004025L, MCI_INTEGER,
L"video key index\0", 0x00004024L, MCI_INTEGER,
L"video source\0", 0x0000401bL, MCI_INTEGER,
L"video source number\0", 0x0000401eL, MCI_INTEGER,
L"video record\0", 0x0000401cL, MCI_INTEGER,
L"video stream\0", 0x0000402eL, MCI_INTEGER,
L"volume\0", 0x00004019L, MCI_INTEGER,
L"window handle\0", 0x00004001L, MCI_INTEGER,
L"window visible\0", 0x00004021L, MCI_INTEGER,
L"window minimized\0", 0x00004022L, MCI_INTEGER,
L"window maximized\0", 0x00004023L, MCI_INTEGER,
L"video streams\0", 0x00004100L, MCI_INTEGER,
L"video bitrate\0", 0x00004101L, MCI_INTEGER,
L"video maxbitrate\0", 0x00004102L, MCI_INTEGER,
L"audio streams\0", 0x00004103L, MCI_INTEGER,
L"audio bitrate\0", 0x00004104L, MCI_INTEGER,
L"video brush\0", 0x00004105L, MCI_INTEGER,
L"audio\0", 0x00004014L, MCI_INTEGER,
L"video\0", 0x00004018L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"track\0", 0x00000010L, MCI_INTEGER,
L"start\0", 0x00000200L, MCI_FLAG,
L"left\0", 0x00080000L, MCI_FLAG,
L"right\0", 0x00100000L, MCI_FLAG,
L"nominal\0", 0x00020000L, MCI_FLAG,
L"record\0", 0x01000000L, MCI_FLAG,
L"input\0", 0x00400000L, MCI_FLAG,
L"output\0", 0x00800000L, MCI_FLAG,
L"disk space\0", 0x00200000L, MCI_STRING,
L"reference\0", 0x00040000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"step\0", MCI_STEP, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"reverse\0", 0x00010000L, MCI_FLAG,
L"by\0", 0x00020000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"stop\0", MCI_STOP, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"hold\0", 0x00010000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"undo\0", MCI_UNDO, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"unfreeze\0", MCI_UNFREEZE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"at\0", 0x00010000L, MCI_RECT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"update\0", MCI_UPDATE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"at\0", 0x00010000L, MCI_RECT,
L"hdc\0", 0x00020000L, MCI_HDC,
L"paint\0", 0x00040000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"where\0", MCI_WHERE, 0, MCI_COMMAND_HEAD,
L"\0", MCI_RECT, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"source\0", 0x00020000L, MCI_FLAG,
L"destination\0", 0x00040000L, MCI_FLAG,
L"frame\0", 0x00080000L, MCI_FLAG,
L"video\0", 0x00100000L, MCI_FLAG,
L"window\0", 0x00200000L, MCI_FLAG,
L"max\0", 0x00400000L, MCI_FLAG,
L"min\0", 0x00800000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"window\0", MCI_WINDOW, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"test\0", 0x00000020L, MCI_FLAG,
L"handle\0", 0x00010000L, MCI_CONSTANT,
L"default\0", 0x00000000L, MCI_HWND,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"state\0", 0x00040000L, MCI_CONSTANT,
L"hide\0", 0x00000000L, MCI_INTEGER,
L"minimize\0", 0x00000006L, MCI_INTEGER,
L"restore\0", 0x00000009L, MCI_INTEGER,
L"show maximized\0", 0x00000003L, MCI_INTEGER,
L"show min noactive\0", 0x00000007L, MCI_INTEGER,
L"show minimized\0", 0x00000002L, MCI_INTEGER,
L"show na\0", 0x00000008L, MCI_INTEGER,
L"show noactivate\0", 0x00000004L, MCI_INTEGER,
L"show normal\0", 0x00000001L, MCI_INTEGER,
L"show\0", 0x00000005L, MCI_INTEGER,
L"no action\0", 0x00000008L, MCI_INTEGER,
L"minimized\0", 0x00000002L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"text\0", 0x00080000L, MCI_STRING,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"\0", 0x00000000L, MCI_END_COMMAND_LIST
END

STRINGTABLE
BEGIN

MCI_ON_S,			"on"
MCI_OFF_S,			"off"
END
