/*
 * Resources for gdi32 test suite.
 *
 * Copyright 2010 Dmitry Timoshkov
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "windef.h"

/* @makedep: wine_test.ttf */
wine_test.ttf RCDATA wine_test.ttf

/* @makedep: wine_vdmx.ttf */
wine_vdmx.ttf RCDATA wine_vdmx.ttf

/* @makedep: vertical.ttf */
vertical.ttf RCDATA vertical.ttf

/* @makedep: wine_longname.ttf */
wine_longname.ttf RCDATA wine_longname.ttf

/* @makedep: wine_ttfnames.ttf */
wine_ttfnames.ttf RCDATA wine_ttfnames.ttf

/* @makedep: wine_ttfnames_bold.ttf */
wine_ttfnames_bold.ttf RCDATA wine_ttfnames_bold.ttf

/* @makedep: wine_langnames.ttf */
wine_langnames.ttf RCDATA wine_langnames.ttf

/* @makedep: wine_langnames2.ttf */
wine_langnames2.ttf RCDATA wine_langnames2.ttf

/* @makedep: wine_langnames3.ttf */
wine_langnames3.ttf RCDATA wine_langnames3.ttf

/* @makedep: wine_nul.ttf */
wine_nul.ttf RCDATA wine_nul.ttf

/* @makedep: wine_heavy.ttf */
wine_heavy.ttf RCDATA wine_heavy.ttf
