/*
 * COM Classes for qedit
 *
 * Copyright 2010 Alexandre Julliard
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#pragma makedep register

[
    threading(both),
    uuid(78530B75-61F9-11D2-8CAD-00A024580902)
]
coclass AMTimeline { interface IAMTimeline; interface IPersistStream; interface IAMSetErrorLog; }

[
    threading(both),
    uuid(65bd0711-24d2-4ff7-9324-ed2e5d3abafa)
]
coclass MediaDet { interface IMediaDet; }

[
    helpstring("Null Renderer"),
    threading(both),
    uuid(c1f400a4-3f08-11d3-9f0b-006008039e37)
]
coclass NullRenderer { interface IBaseFilter; }

[
    helpstring("Sample Grabber"),
    threading(both),
    uuid(c1f400a0-3f08-11d3-9f0b-006008039e37)
]
coclass SampleGrabber { interface ISampleGrabber; }
