/*
 * Copyright 2013 Austin English
 * Copyright 2023 Zhiyi Zhang for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "tasklist.h"

#pragma makedep po

LANGUAGE LANG_ENGLISH, SUBLANG_DEFAULT

STRINGTABLE
{
    STRING_USAGE,                   "Usage: tasklist [/?]\n"
    STRING_IMAGE_NAME,              "Image Name"
    STRING_PID,                     "PID"
    STRING_SESSION_NAME,            "Session Name"
    STRING_SESSION_NUMBER,          "Session#"
    STRING_MEM_USAGE,               "Mem Usage"
    STRING_K,                       "K"
    STRING_INVALID_SYNTAX,          "ERROR: Invalid syntax\n"
    STRING_FILTER_NOT_RECOGNIZED,   "ERROR: The search filter cannot be recognized.\n"
}

#define WINE_FILEDESCRIPTION_STR "Wine tasklist"
#define WINE_FILENAME_STR "tasklist.exe"
#define WINE_FILEVERSION 6,1,7601,16385
#define WINE_FILEVERSION_STR "6.1.7601.16385"
#define WINE_PRODUCTVERSION 6,1,7601,16385
#define WINE_PRODUCTVERSION_STR "6.1.7601.16385"

#include "wine/wine_common_ver.rc"
