/*
 * Copyright 2006 Jacek Caban for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#pragma makedep regtypelib
#pragma makedep register

#include "exdisp.idl"

[
    threading(apartment),
    uuid(871c5380-42a0-1069-a2ea-08002b30309d)
] coclass Internet { }

[
    helpstring("Microsoft Url History Service"),
    threading(apartment),
    uuid(3c374a40-bae4-11cf-bf7d-00aa006946ee)
] coclass CUrlHistory { interface IUrlHistoryStg2; }

[
    helpstring("Internet Shortcut"),
    threading(apartment),
    progid("InternetShortcut"),
    uuid(fbf23b40-e3f0-101b-8488-00aa003e56f8)
] coclass InternetShortcut
{
    interface IUniformResourceLocatorA;
    interface IUniformResourceLocatorW;
    interface IPersistFile;
    interface IPropertySetStorage;
}
