/*
 * Copyright 2009 Austin English
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "config.h" /* Needed to get PACKAGE_VERSION */

#define WINE_FILEDESCRIPTION_STR "Wine D3D"
#define WINE_FILENAME_STR "wined3d.dll"
#define WINE_FILEVERSION_STR PACKAGE_VERSION
#define WINE_PRODUCTVERSION_STR PACKAGE_VERSION
#define WINE_PRODUCTNAME_STR "Wine D3D"

#include "wine/wine_common_ver.rc"
