/*
 * Top level resource file for localspl
 *
 * Copyright 2006 Detlef Riekenberg
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 *
 */

#include "localspl_private.h"

#pragma makedep po

LANGUAGE LANG_ENGLISH, SUBLANG_DEFAULT

STRINGTABLE
{
    IDS_LOCALPORT "Local Port"
    IDS_LOCALMONITOR "Local Monitor"
}

STRINGTABLE
{
    IDS_FORM_LETTER "Letter"
    IDS_FORM_LETTER_SMALL "Letter Small"
    IDS_FORM_TABLOID "Tabloid"
    IDS_FORM_LEDGER "Ledger"
    IDS_FORM_LEGAL "Legal"
    IDS_FORM_STATEMENT "Statement"
    IDS_FORM_EXECUTIVE "Executive"
    IDS_FORM_A3 "A3"
    IDS_FORM_A4 "A4"
    IDS_FORM_A4_SMALL "A4 Small"
    IDS_FORM_A5 "A5"
    IDS_FORM_B4_JIS "B4 (JIS)"
    IDS_FORM_B5_JIS "B5 (JIS)"
    IDS_FORM_FOLIO "Folio"
    IDS_FORM_QUARTO "Quarto"
    IDS_FORM_10x14 "10x14"
    IDS_FORM_11x17 "11x17"
    IDS_FORM_NOTE "Note"
    IDS_FORM_ENVELOPE_9 "Envelope #9"
    IDS_FORM_ENVELOPE_10 "Envelope #10"
    IDS_FORM_ENVELOPE_11 "Envelope #11"
    IDS_FORM_ENVELOPE_12 "Envelope #12"
    IDS_FORM_ENVELOPE_14 "Envelope #14"
    IDS_FORM_C_SIZE_SHEET "C size sheet"
    IDS_FORM_D_SIZE_SHEET "D size sheet"
    IDS_FORM_E_SIZE_SHEET "E size sheet"
    IDS_FORM_ENVELOPE_DL "Envelope DL"
    IDS_FORM_ENVELOPE_C5 "Envelope C5"
    IDS_FORM_ENVELOPE_C3 "Envelope C3"
    IDS_FORM_ENVELOPE_C4 "Envelope C4"
    IDS_FORM_ENVELOPE_C6 "Envelope C6"
    IDS_FORM_ENVELOPE_C65 "Envelope C65"
    IDS_FORM_ENVELOPE_B4 "Envelope B4"
    IDS_FORM_ENVELOPE_B5 "Envelope B5"
    IDS_FORM_ENVELOPE_B6 "Envelope B6"
    IDS_FORM_ENVELOPE "Envelope"
    IDS_FORM_ENVELOPE_MONARCH "Envelope Monarch"
    IDS_FORM_6_34_ENVELOPE "6 3/4 Envelope"
    IDS_FORM_US_STD_FANFOLD "US Std Fanfold"
    IDS_FORM_GERMAN_STD_FANFOLD "German Std Fanfold"
    IDS_FORM_GERMAN_LEGAL_FANFOLD "German Legal Fanfold"
    IDS_FORM_B4_ISO "B4 (ISO)"
    IDS_FORM_JAPANESE_POSTCARD "Japanese Postcard"
    IDS_FORM_9x11 "9x11"
    IDS_FORM_10x11 "10x11"
    IDS_FORM_15x11 "15x11"
    IDS_FORM_ENVELOPE_INVITE "Envelope Invite"
    IDS_FORM_LETTER_EXTRA "Letter Extra"
    IDS_FORM_LEGAL_EXTRA "Legal Extra"
    IDS_FORM_TABLOID_EXTRA "Tabloid Extra"
    IDS_FORM_A4_EXTRA "A4 Extra"
    IDS_FORM_LETTER_TRANSVERSE "Letter Transverse"
    IDS_FORM_A4_TRANSVERSE "A4 Transverse"
    IDS_FORM_LETTER_EXTRA_TRANSVERSE "Letter Extra Transverse"
    IDS_FORM_SUPER_A "Super A"
    IDS_FORM_SUPER_B "Super B"
    IDS_FORM_LETTER_PLUS "Letter Plus"
    IDS_FORM_A4_PLUS "A4 Plus"
    IDS_FORM_A5_TRANSVERSE "A5 Transverse"
    IDS_FORM_B5_JIS_TRANSVERSE "B5 (JIS) Transverse"
    IDS_FORM_A3_EXTRA "A3 Extra"
    IDS_FORM_A5_EXTRA "A5 Extra"
    IDS_FORM_B5_ISO_EXTRA "B5 (ISO) Extra"
    IDS_FORM_A2 "A2"
    IDS_FORM_A3_TRANSVERSE "A3 Transverse"
    IDS_FORM_A3_EXTRA_TRANSVERSE "A3 Extra Transverse"
    IDS_FORM_JAPANESE_DOUBLE_POSTCARD "Japanese Double Postcard"
    IDS_FORM_A6 "A6"
    IDS_FORM_JAPANESE_ENVELOPE_KAKU_2 "Japanese Envelope Kaku #2"
    IDS_FORM_JAPANESE_ENVELOPE_KAKU_3 "Japanese Envelope Kaku #3"
    IDS_FORM_JAPANESE_ENVELOPE_CHOU_3 "Japanese Envelope Chou #3"
    IDS_FORM_JAPANESE_ENVELOPE_CHOU_4 "Japanese Envelope Chou #4"
    IDS_FORM_LETTER_ROTATED "Letter Rotated"
    IDS_FORM_A3_ROTATED "A3 Rotated"
    IDS_FORM_A4_ROTATED "A4 Rotated"
    IDS_FORM_A5_ROTATED "A5 Rotated"
    IDS_FORM_B4_JIS_ROTATED "B4 (JIS) Rotated"
    IDS_FORM_B5_JIS_ROTATED "B5 (JIS) Rotated"
    IDS_FORM_JAPANESE_POSTCARD_ROTATED "Japanese Postcard Rotated"
    IDS_FORM_DOUBLE_JAPAN_POSTCARD_ROTATED "Double Japan Postcard Rotated"
    IDS_FORM_A6_ROTATED "A6 Rotated"
    IDS_FORM_JAPAN_ENVELOPE_KAKU_2_ROTATED "Japan Envelope Kaku #2 Rotated"
    IDS_FORM_JAPAN_ENVELOPE_KAKU_3_ROTATED "Japan Envelope Kaku #3 Rotated"
    IDS_FORM_JAPAN_ENVELOPE_CHOU_3_ROTATED "Japan Envelope Chou #3 Rotated"
    IDS_FORM_JAPAN_ENVELOPE_CHOU_4_ROTATED "Japan Envelope Chou #4 Rotated"
    IDS_FORM_B6_JIS "B6 (JIS)"
    IDS_FORM_B6_JIS_ROTATED "B6 (JIS) Rotated"
    IDS_FORM_12x11 "12x11"
    IDS_FORM_JAPAN_ENVELOPE_YOU_4 "Japan Envelope You #4"
    IDS_FORM_JAPAN_ENVELOPE_YOU_4_ROTATED "Japan Envelope You #4 Rotated"
    IDS_FORM_PRC_16K "PRC 16K"
    IDS_FORM_PRC_32K "PRC 32K"
    IDS_FORM_PRC_32K_BIG "PRC 32K(Big)"
    IDS_FORM_PRC_ENVELOPE_1 "PRC Envelope #1"
    IDS_FORM_PRC_ENVELOPE_2 "PRC Envelope #2"
    IDS_FORM_PRC_ENVELOPE_3 "PRC Envelope #3"
    IDS_FORM_PRC_ENVELOPE_4 "PRC Envelope #4"
    IDS_FORM_PRC_ENVELOPE_5 "PRC Envelope #5"
    IDS_FORM_PRC_ENVELOPE_6 "PRC Envelope #6"
    IDS_FORM_PRC_ENVELOPE_7 "PRC Envelope #7"
    IDS_FORM_PRC_ENVELOPE_8 "PRC Envelope #8"
    IDS_FORM_PRC_ENVELOPE_9 "PRC Envelope #9"
    IDS_FORM_PRC_ENVELOPE_10 "PRC Envelope #10"
    IDS_FORM_PRC_16K_ROTATED "PRC 16K Rotated"
    IDS_FORM_PRC_32K_ROTATED "PRC 32K Rotated"
    IDS_FORM_PRC_32K_BIG_ROTATED "PRC 32K(Big) Rotated"
    IDS_FORM_PRC_ENVELOPE_1_ROTATED "PRC Envelope #1 Rotated"
    IDS_FORM_PRC_ENVELOPE_2_ROTATED "PRC Envelope #2 Rotated"
    IDS_FORM_PRC_ENVELOPE_3_ROTATED "PRC Envelope #3 Rotated"
    IDS_FORM_PRC_ENVELOPE_4_ROTATED "PRC Envelope #4 Rotated"
    IDS_FORM_PRC_ENVELOPE_5_ROTATED "PRC Envelope #5 Rotated"
    IDS_FORM_PRC_ENVELOPE_6_ROTATED "PRC Envelope #6 Rotated"
    IDS_FORM_PRC_ENVELOPE_7_ROTATED "PRC Envelope #7 Rotated"
    IDS_FORM_PRC_ENVELOPE_8_ROTATED "PRC Envelope #8 Rotated"
    IDS_FORM_PRC_ENVELOPE_9_ROTATED "PRC Envelope #9 Rotated"
    IDS_FORM_PRC_ENVELOPE_10_ROTATED "PRC Envelope #10 Rotated"
}

LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL

#define WINE_FILENAME_STR "localspl.dll"
#define WINE_FILEDESCRIPTION_STR "Wine Printer spooler component"

/* Same Version as WinXP_sp2 */
#define WINE_FILEVERSION      5,1,2600,2180
#define WINE_FILEVERSION_STR "5.1.2600.2180"

#define WINE_PRODUCTVERSION      5,1,2600,2180
#define WINE_PRODUCTVERSION_STR "5.1.2600.2180"

#include "wine/wine_common_ver.rc"
