/*
 * Copyright 2011 Jacek Caban for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#pragma makedep register

[
    helpstring("VBScript Regular Expression"),
    threading(apartment),
    progid("VBScript.RegExp"),
    version(5.5),
    uuid(3f4daca4-160d-11d2-a8e9-00104b365c9f)
]
coclass VBScriptRegExp {}

[
    helpstring("VB Script Language"),
    threading(both),
    progid("VBScript"),
    uuid(b54f3741-5b07-11cf-a4b0-00aa004a55e8)
]
coclass VBScript {}

[
    helpstring("VB Script Language Authoring"),
    threading(both),
    progid("VBScript Author"),
    uuid(b54f3742-5b07-11cf-a4b0-00aa004a55e8)
]
coclass VBScriptAuthor {}

[
    helpstring("VB Script Language Encoding"),
    threading(both),
    progid("VBScript.Encode"),
    uuid(b54f3743-5b07-11cf-a4b0-00aa004a55e8)
]
coclass VBScriptEncode {}
