/*
 * Copyright 2022 Nikolay Sivov for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#pragma makedep register

[
    helpstring("AVI Byte Stream Handler"),
    threading(both),
    uuid(7afa253e-f823-42f6-a5d9-714bde467412)
]
coclass AVIByteStreamPlugin { }

[
    helpstring("MF WAVE Sink Factory"),
    threading(both),
    uuid(36f99745-23c9-4c9c-8dd5-cc31ce964390)
]
coclass MFWAVESinkClassFactory { }

[
    helpstring("WAV Byte Stream Handler"),
    threading(both),
    uuid(42c9b9f5-16fc-47ef-af22-da05f7c842e3)
]
coclass WAVByteStreamPlugin {}
