/*
 * Resources for mf test suite.
 *
 * Copyright 2022 Rémi Bernon for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "windef.h"

/* @makedep: audiodata.bin */
audiodata.bin RCDATA audiodata.bin

/* @makedep: audioconvdata.bin */
audioconvdata.bin RCDATA audioconvdata.bin

/* @makedep: aacencdata.bin */
aacencdata.bin RCDATA aacencdata.bin

/* @makedep: aacdecdata.bin */
aacdecdata.bin RCDATA aacdecdata.bin

/* @makedep: wmaencdata.bin */
wmaencdata.bin RCDATA wmaencdata.bin

/* @makedep: wmadecdata.bin */
wmadecdata.bin RCDATA wmadecdata.bin

/* @makedep: wmvencdata.bin */
wmvencdata.bin RCDATA wmvencdata.bin

/* Generated with:
 *  gst-launch-1.0 filesrc location=dlls/mf/tests/audiodata.bin ! \
 *      audio/x-raw,format=F32LE,rate=22050,channels=2,layout=interleaved ! \
 *      audioconvert ! lamemp3enc ! \
 *      audio/mpeg,mpegversion=1,layer=3,rate=22050,channels=2 ! \
 *      filesink location=dlls/mf/tests/mp3encdata.bin
 */
/* @makedep: mp3encdata.bin */
mp3encdata.bin RCDATA mp3encdata.bin

/* @makedep: mp3decdata.bin */
mp3decdata.bin RCDATA mp3decdata.bin

/* Generated with:
 *  gst-launch-1.0 videotestsrc num-buffers=360 pattern=smpte100 ! \
 *      video/x-raw,format=I420,width=84,height=82,framerate=30000/1001 ! \
 *      videoflip method=clockwise ! videoconvert ! \
 *      x264enc ! filesink location=dlls/mf/tests/h264data.bin
 */
/* @makedep: h264data.bin */
h264data.bin RCDATA h264data.bin

/* Generated from running the tests on Windows */
/* @makedep: h264encdata.bin */
h264encdata.bin RCDATA h264encdata.bin

/* Generated with:
  gst-launch-1.0 videotestsrc num-buffers=30 pattern=smpte100 ! \
      video/x-raw,format=I420,width=84,height=82,framerate=30000/1001 ! \
      videoflip method=clockwise ! videoconvert ! \
      x264enc ! filesink location=dlls/mf/tests/h264data-1.bin
 */
/* @makedep: h264data-1.bin */
h264data-1.bin RCDATA h264data-1.bin

/* Generated with:
  gst-launch-1.0 videotestsrc num-buffers=30 pattern=smpte100 ! \
      video/x-raw,format=I420,width=100,height=98,framerate=60000/1001 ! \
      videoflip method=clockwise ! videoconvert ! \
      x264enc ! filesink location=dlls/mf/tests/h264data-2.bin
 */
/* @makedep: h264data-2.bin */
h264data-2.bin RCDATA h264data-2.bin

/* Generated from running the tests on Windows */
/* @makedep: nv12frame.bmp */
nv12frame.bmp RCDATA nv12frame.bmp

/* Generated from running the tests on Windows */
/* @makedep: nv12frame-2d.bmp */
nv12frame-2d.bmp RCDATA nv12frame-2d.bmp

/* Generated from running the tests on Windows */
/* @makedep: nv12frame-grabber.bmp */
nv12frame-grabber.bmp RCDATA nv12frame-grabber.bmp

/* Generated from running the tests on Windows */
/* @makedep: nv12frame-flip.bmp */
nv12frame-flip.bmp RCDATA nv12frame-flip.bmp

/* Generated from running the tests on Windows */
/* @makedep: nv12frame-flip-2d.bmp */
nv12frame-flip-2d.bmp RCDATA nv12frame-flip-2d.bmp

/* Generated from running the tests on Windows */
/* @makedep: i420frame.bmp */
i420frame.bmp RCDATA i420frame.bmp

/* Generated from running the tests on Windows */
/* @makedep: rgb32frame.bmp */
rgb32frame.bmp RCDATA rgb32frame.bmp

/* Generated from running the tests on Windows */
/* @makedep: rgb32frame-flip.bmp */
rgb32frame-flip.bmp RCDATA rgb32frame-flip.bmp

/* Generated from running the tests on Windows */
/* @makedep: rgb32frame-flip-2d.bmp */
rgb32frame-flip-2d.bmp RCDATA rgb32frame-flip-2d.bmp

/* Generated from running the tests on Windows */
/* @makedep: rgb32frame-crop.bmp */
rgb32frame-crop.bmp RCDATA rgb32frame-crop.bmp

/* Generated from running the tests on Windows */
/* @makedep: rgb32frame-crop-flip.bmp */
rgb32frame-crop-flip.bmp RCDATA rgb32frame-crop-flip.bmp

/* Generated from running the tests on Windows */
/* @makedep: rgb32frame-crop-flip-2d.bmp */
rgb32frame-crop-flip-2d.bmp RCDATA rgb32frame-crop-flip-2d.bmp

/* Generated from running the tests on Windows */
/* @makedep: abgr32frame-crop.bmp */
abgr32frame-crop.bmp RCDATA abgr32frame-crop.bmp

/* Generated from running the tests on Windows */
/* @makedep: rgb32frame-grabber.bmp */
rgb32frame-grabber.bmp RCDATA rgb32frame-grabber.bmp

/* Generated from running the tests on Windows */
/* @makedep: rgb32frame-extra-width.bmp */
rgb32frame-extra-width.bmp RCDATA rgb32frame-extra-width.bmp

/* Generated from running the tests on Windows */
/* @makedep: nv12frame-extra-width.bmp */
nv12frame-extra-width.bmp RCDATA nv12frame-extra-width.bmp

/* Generated from running the tests on Windows */
/* @makedep: nv12frame-extra-width-2d.bmp */
nv12frame-extra-width-2d.bmp RCDATA nv12frame-extra-width-2d.bmp

/* Generated from running the tests on Windows */
/* @makedep: rgb555frame.bmp */
rgb555frame.bmp RCDATA rgb555frame.bmp

/* Generated from running the tests on Windows */
/* @makedep: rgb555frame-flip.bmp */
rgb555frame-flip.bmp RCDATA rgb555frame-flip.bmp

/* Generated from running the tests on Windows */
/* @makedep: iv50frame.bin */
iv50frame.bin RCDATA iv50frame.bin

/* Generated from running the tests on Windows */
/* @makedep: rgb24frame.bmp */
rgb24frame.bmp RCDATA rgb24frame.bmp

/* Generated with:
 *  gst-launch-1.0 videotestsrc num-buffers=60 pattern=smpte100 ! \
 *      video/x-raw,format=I420,width=64,height=64,framerate=30000/1001 ! \
 *      videoflip method=clockwise ! videoconvert ! \
 *      x264enc ! mp4mux ! filesink location=dlls/mf/tests/test.mp4
 */
/* @makedep: test.mp4 */
test.mp4 RCDATA test.mp4
