/*
 * Resources for shlwapi
 *
 * Copyright 2004 Jon Griffiths
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "resource.h"

#pragma makedep po

LANGUAGE LANG_ENGLISH, SUBLANG_DEFAULT

STRINGTABLE
{
    IDS_BYTES_FORMAT    "%d bytes"
    IDS_TIME_INTERVAL_HOURS    "#msgctxt#time unit: hours# hr"
    IDS_TIME_INTERVAL_MINUTES  "#msgctxt#time unit: minutes# min"
    IDS_TIME_INTERVAL_SECONDS  "#msgctxt#time unit: seconds# sec"
}

IDD_ERR_DIALOG DIALOG 0, 0, 220, 60
STYLE DS_MODALFRAME | DS_CENTER | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Error"
FONT 8, "MS Shell Dlg"
{
 LTEXT "", IDS_ERR_USER_MSG2, 15, 5, 28, 20
 LTEXT "", IDS_ERR_USER_MSG, 15, 5, 210, 8
 CHECKBOX "Don't show me th&is message again", IDC_ERR_DONT_SHOW, 5, 20, 210, 10, BS_AUTOCHECKBOX | WS_GROUP | WS_TABSTOP
 PUSHBUTTON "OK", IDOK, 105, 40, 50, 14, WS_GROUP | WS_TABSTOP
 PUSHBUTTON "Cancel", IDCANCEL, 160, 40, 50, 14, WS_GROUP | WS_TABSTOP
 PUSHBUTTON "&Yes", IDYES, 105, 40, 50, 14, WS_GROUP | WS_TABSTOP
 PUSHBUTTON "&No", IDNO, 160, 40, 50, 14, WS_GROUP | WS_TABSTOP
}

LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL

#define WINE_FILENAME_STR "shlwapi.dll"
#define WINE_FILEVERSION 6,0,2800,1692
#define WINE_FILEVERSION_STR "6.0.2800.1692"
#define WINE_PRODUCTVERSION 6,0,2800,1692
#define WINE_PRODUCTVERSION_STR "6.0.2800.1692"

#include "wine/wine_common_ver.rc"
