/*
 * Resource definitions
 *
 * Copyright 2004 Ferenc Wagner
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#define IDI_WINE 1

#define IDD_STATUS 100
#define IDD_ABOUT  101
#define IDD_TAG    102
#define IDD_EMAIL  103

#define IDC_STATIC -1

#define IDC_ST0 1000
#define IDC_PB0 1001
#define IDC_ST1 1002
#define IDC_PB1 1003
#define IDC_ST2 1004
#define IDC_PB2 1005

#define IDC_DIR 2000
#define IDC_OUT 2001
#define IDC_TAG 2002
#define IDC_EMAIL 2002

#define IDC_SB  3000

#define IDC_EDIT  4000
#define IDC_ABOUT 4001

#define IDS_BUILD_ID 1
