/*
 *
 * Copyright 2011 Alistair Leslie-Hughes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

import "basetsd.h";
import "oaidl.idl";

#ifndef __WIDL__
#define threading(model)
#define progid(str)
#define vi_progid(str)
#endif

[
    uuid(A8D2C5FE-CDCB-4B9D-BDE5-A27343FF54BC),
]
interface IMMCVersionInfo : IUnknown
{
   HRESULT GetMMCVersion([out] long * pVersionMajor, [out] long * pVersionMinor);
}


[
   uuid(9EECDD85-B715-4188-8A72-61159ABDB8C4),
   version(1.0),
]
library MMCVersionLib
{
   importlib("stdole2.tlb");

   [
      helpstring("MMCVersionInfo 1.0 Object"),
      uuid(D6FEDB1D-CF21-4BD9-AF3B-C5468E9C6684),
      threading(apartment),
      progid("NODEMGR.MMCVersionInfo.1"),
      vi_progid("NODEMGR.MMCVersionInfo")
   ]
   coclass MMCVersionInfo
   {
       [default] interface IMMCVersionInfo;
   }

}
