/*
 * Copyright 2008 Jacek Caban for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

/* @makedep: api.js */
api.js 40 "api.js"

/* @makedep: cc.js */
cc.js 40 "cc.js"

/* @makedep: lang.js */
lang.js 40 "lang.js"

/* @makedep: regexp.js */
regexp.js 40 "regexp.js"

/* @makedep: sunspider-regexp-dna.js */
dna.js 40 "sunspider-regexp-dna.js"

/* @makedep: sunspider-string-base64.js */
base64.js 40 "sunspider-string-base64.js"

/* @makedep: sunspider-string-validate-input.js */
validateinput.js 40 "sunspider-string-validate-input.js"
