/*
 * Copyright (C) 2022 Esme Povirk
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

import "unknwn.idl";
import "objidl.idl";
import "oaidl.idl";

#pragma makedep header

/* Undocumented interface used by
 * https://github.com/dotnet/symreader/blob/fe43beb65a8567834c717e2478ba36f1e98db8b5/src/Microsoft.DiaSymReader/Writer/ISymUnmanagedWriter.cs */
[
    object,
    uuid(98ecee1e-752d-11d3-8d56-00c04f680b2b),
    pointer_default(unique)
]
interface IPdbWriter : IUnknown
{
    HRESULT SetPath([in] const WCHAR *fullpath, [in] IStream *stream, [in] BOOL fullbuild);
    HRESULT OpenMod([in] const WCHAR *modulename, [in] const WCHAR *filename);
    HRESULT CloseMod();
    HRESULT GetPath([in] DWORD ccData, [out] DWORD *pccData, [out, size_is(ccData), length_is(*pccData)] WCHAR *path);
    HRESULT GetSignatureAge([out] DWORD *timestamp, [out] DWORD *age);
}
