/*
 * IDL to test COM object proxy/stub functions
 *
 * Copyright 2018 Zebediah Figura
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#pragma makedep proxy

import "oleidl.idl";

[
    uuid(deadbeef-0001-44c7-850f-2a0f465c0c6c),
    object
]
interface ITest1 : IPersist
{
    int square(int x);
}

[
    uuid(deadbeef-ffff-44c7-850f-2a0f465c0c6c)
]
coclass test_ps { interface IPSFactoryBuffer; }
