/*
 * Copyright 2024 Rémi Bernon for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 *
 */

#include "resource.h"

#pragma makedep po

LANGUAGE LANG_ENGLISH, SUBLANG_DEFAULT

STRINGTABLE
BEGIN
    IDS_CPL_NAME        "Display Settings"
    IDS_CPL_INFO        "Configure monitors and display devices."
END

IDD_DESKTOP DIALOG 0, 0, 320, 300
STYLE WS_CAPTION | WS_CHILD | WS_DISABLED
CAPTION "Desktop"
FONT 8, "Ms Shell Dlg"
{
    GROUPBOX        "Virtual Desktop", IDC_VIRTUAL_DESKTOP, 10, 10, 300, 120
    COMBOBOX        IDC_DISPLAY_SETTINGS_LIST, 10, 136, 160, 60, CBS_DROPDOWNLIST | CBS_HASSTRINGS
    PUSHBUTTON      "&Reset", IDC_DISPLAY_SETTINGS_RESET, 180, 135, 60, 15
    PUSHBUTTON      "&Apply", IDC_DISPLAY_SETTINGS_APPLY, 250, 135, 60, 15
    AUTOCHECKBOX    "Emulate display mode changes (requires restart)", IDC_EMULATE_MODESET, 10, 155, 300, 10
}

LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL

/* @makedep: desk.manifest */
124 RT_MANIFEST desk.manifest

/* @makedep: desk.ico */
ICO_MAIN ICON "desk.ico"

#define WINE_FILENAME_STR "desk.cpl"
#define WINE_FILEDESCRIPTION_STR "Display Settings Panel"

#include "wine/wine_common_ver.rc"
