/*
 * Top level resource file for localui
 *
 * Copyright 2007 Detlef Riekenberg
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 *
 */

#include "localui.h"

#pragma makedep po

LANGUAGE LANG_ENGLISH, SUBLANG_DEFAULT

STRINGTABLE
{
    IDS_LOCALPORT "Local Port"
    IDS_INVALIDNAME "'%s' is not a valid port name"
    IDS_PORTEXISTS "Port %s already exists"
    IDS_NOTHINGTOCONFIG "This port has no options to configure"
}

ADDPORT_DIALOG DIALOG 6, 18, 245, 47
STYLE DS_CONTEXTHELP | DS_MODALFRAME | DS_SETFONT | DS_SETFOREGROUND | WS_POPUPWINDOW | WS_VISIBLE | WS_CAPTION
CAPTION "Add a Local Port"
FONT 8, "MS Shell Dlg"
BEGIN
    LTEXT "&Enter the port name to add:", -1, 7, 13, 194, 13, WS_VISIBLE
    EDITTEXT ADDPORT_EDIT, 6, 28, 174, 12, WS_VISIBLE | ES_AUTOHSCROLL
    DEFPUSHBUTTON "OK", IDOK, 188, 10, 50, 14, WS_VISIBLE
    PUSHBUTTON "Cancel", IDCANCEL, 188, 27, 50, 14, WS_VISIBLE
END


LPTCONFIG_DIALOG DIALOG 6, 18, 220, 47
STYLE DS_CONTEXTHELP | DS_MODALFRAME | DS_SETFONT | DS_SETFOREGROUND | WS_POPUPWINDOW | WS_VISIBLE | WS_CAPTION
CAPTION "Configure LPT Port"
FONT 8, "MS Shell Dlg"
BEGIN
    GROUPBOX "Timeout (seconds)", LPTCONFIG_GROUP, 6, 6, 150, 35, BS_GROUPBOX
    LTEXT "&Transmission Retry:", -1, 14, 22, 90, 13, WS_VISIBLE
    EDITTEXT LPTCONFIG_EDIT, 112, 20, 32, 13, WS_VISIBLE | ES_NUMBER
    DEFPUSHBUTTON "OK", IDOK, 164, 10, 50, 14, WS_VISIBLE
    PUSHBUTTON "Cancel", IDCANCEL, 164, 27, 50, 14, WS_VISIBLE
END

LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL

#define WINE_FILENAME_STR "localui.dll"
#define WINE_FILEDESCRIPTION_STR "User Interface for the Local Monitor"

/* Same Version as WinXP_sp2 */
#define WINE_FILEVERSION      5,1,2600,2180
#define WINE_FILEVERSION_STR "5.1.2600.2180"

#define WINE_PRODUCTVERSION      5,1,2600,2180
#define WINE_PRODUCTVERSION_STR "5.1.2600.2180"

#include "wine/wine_common_ver.rc"
