/*
 * Copyright 2019 Nikolay Sivov for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

import "wtypes.idl";
#include "wine/orpc.idl"

cpp_quote("#define IRPCSS_PROTSEQ {'n','c','a','l','r','p','c',0}")
cpp_quote("#define IRPCSS_ENDPOINT {'i','r','p','c','s','s',0}")

[
    uuid(85da4974-edc7-40ff-bad4-9c4525a8d044),
    version(0.0),
]
interface Irpcss
{
    HRESULT irpcss_server_register(
        [in] handle_t handle,
        [in] const GUID *clsid,
        [in] unsigned int flags,
        [in] PMInterfacePointer object,
        [out] unsigned int *cookie);

    HRESULT irpcss_server_revoke(
        [in] handle_t handle,
        [in] unsigned int cookie);

    HRESULT irpcss_get_class_object(
        [in] handle_t handle,
        [in] const GUID *clsid,
        [out] PMInterfacePointer *object);

    HRESULT irpcss_get_thread_seq_id(
        [in] handle_t handle,
        [out] DWORD *sequence_id);
}
