/*
 * Copyright (C) 2023 Maxim Karasev <mxkrsv@etersoft.ru>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "resources.h"

#pragma makedep po

LANGUAGE LANG_ENGLISH, SUBLANG_DEFAULT

STRINGTABLE
{
    STRING_USAGE, "Usage: klist <tickets | tgt | purge | get [service principal name]>\n"
    STRING_UNKNOWN_ERROR, "Unknown error\n"
    STRING_START_TIME, "Start Time"
    STRING_END_TIME, "End Time"
    STRING_RENEW_TIME, "Renew Time"
    STRING_TICKET_FLAGS, "Ticket Flags"
    STRING_CACHED_TICKETS, "Cached Tickets"
    STRING_SERVER, "Server"
    STRING_ENCRYPTION_TYPE, "KerbTicket Encryption Type"
    STRING_LOGON_ID, "Current LogonId is"
    STRING_CLIENT, "Client"
}
