/* Multimedia resources (MMSYSTEM.DLL & WINMM.DLL)
 *
 * Copyright (c) 1999, Eric Pouech
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 *
 * Note: Win 9x (and of course Win 3.x) store their resources in MMSYSTEM.DLL
 *       Wine stores all resources in WINMM.DLL (as NT does)
 *       This shall be transparent to most apps (unless some look directly in
 *       MMSYSTEM...)
 */

#include "windef.h"
#include "mmddk.h"

#define MCI_DWORD_PTR 13 /* MCI_INTEGER64 */

LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL

CORE RCDATA
BEGIN
L"open\0", MCI_OPEN, 0, MCI_COMMAND_HEAD,
L"\0", MCI_INTEGER, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"type\0", 0x00002000L, MCI_STRING,
L"element\0", 0x00000200L, MCI_STRING,
L"alias\0", 0x00000400L, MCI_STRING,
L"shareable\0", 0x00000100L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"close\0", MCI_CLOSE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"play\0", MCI_PLAY, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"from\0", 0x00000004L, MCI_INTEGER,
L"to\0", 0x00000008L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"record\0", MCI_RECORD, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"from\0", 0x00000004L, MCI_INTEGER,
L"to\0", 0x00000008L, MCI_INTEGER,
L"insert\0", 0x00000100L, MCI_FLAG,
L"overwrite\0", 0x00000200L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"seek\0", MCI_SEEK, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"to start\0", 0x00000100L, MCI_FLAG,
L"to end\0", 0x00000200L, MCI_FLAG,
L"to\0", 0x00000008L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"stop\0", MCI_STOP, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"pause\0", MCI_PAUSE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"status\0", MCI_STATUS, 0, MCI_COMMAND_HEAD,
L"\0", MCI_DWORD_PTR, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_CONSTANT,
L"position\0", 0x00000002L, MCI_INTEGER,
L"length\0", 0x00000001L, MCI_INTEGER,
L"number of tracks\0", 0x00000003L, MCI_INTEGER,
L"ready\0", 0x00000007L, MCI_INTEGER,
L"mode\0", 0x00000004L, MCI_INTEGER,
L"time format\0", 0x00000006L, MCI_INTEGER,
L"current track\0", 0x00000008L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"track\0", 0x00000010L, MCI_INTEGER,
L"start\0", 0x00000200L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"capability\0", MCI_GETDEVCAPS, 0, MCI_COMMAND_HEAD,
L"\0", MCI_INTEGER, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_CONSTANT,
L"can record\0", 0x00000001L, MCI_INTEGER,
L"has audio\0", 0x00000002L, MCI_INTEGER,
L"has video\0", 0x00000003L, MCI_INTEGER,
L"uses files\0", 0x00000005L, MCI_INTEGER,
L"compound device\0", 0x00000006L, MCI_INTEGER,
L"device type\0", 0x00000004L, MCI_INTEGER,
L"can eject\0", 0x00000007L, MCI_INTEGER,
L"can play\0", 0x00000008L, MCI_INTEGER,
L"can save\0", 0x00000009L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"info\0", MCI_INFO, 0, MCI_COMMAND_HEAD,
L"\0", MCI_STRING, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"product\0", 0x00000100L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"set\0", MCI_SET, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"time format\0", 0x00000400L, MCI_CONSTANT,
L"milliseconds\0", 0x00000000L, MCI_INTEGER,
L"ms\0", 0x00000000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"door open\0", 0x00000100L, MCI_FLAG,
L"door closed\0", 0x00000200L, MCI_FLAG,
L"audio\0", 0x00000800L, MCI_CONSTANT,
L"all\0", 0x00000000L, MCI_INTEGER,
L"left\0", 0x00000001L, MCI_INTEGER,
L"right\0", 0x00000002L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"video\0", 0x00001000L, MCI_FLAG,
L"on\0", 0x00002000L, MCI_FLAG,
L"off\0", 0x00004000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"sysinfo\0", MCI_SYSINFO, 0, MCI_COMMAND_HEAD,
L"\0", MCI_STRING, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"quantity\0", 0x00000100L, MCI_FLAG,
L"open\0", 0x00000200L, MCI_FLAG,
L"installname\0", 0x00000800L, MCI_FLAG,
L"name\0", 0x00000400L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"break\0", MCI_BREAK, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"on\0", 0x00000100L, MCI_INTEGER,
L"off\0", 0x00000400L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"sound\0", MCI_SOUND, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"save\0", MCI_SAVE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_STRING,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"load\0", MCI_LOAD, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_STRING,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"resume\0", MCI_RESUME, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"\0", 0x00000000L, MCI_END_COMMAND_LIST
END

VIDEODISC RCDATA
BEGIN
L"play\0", MCI_PLAY, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"from\0", 0x00000004L, MCI_INTEGER,
L"to\0", 0x00000008L, MCI_INTEGER,
L"fast\0", 0x00020000L, MCI_FLAG,
L"slow\0", 0x00100000L, MCI_FLAG,
L"speed\0", 0x00040000L, MCI_INTEGER,
L"reverse\0", 0x00010000L, MCI_FLAG,
L"scan\0", 0x00080000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"seek\0", MCI_SEEK, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"reverse\0", 0x00010000L, MCI_FLAG,
L"to start\0", 0x00000100L, MCI_FLAG,
L"to end\0", 0x00000200L, MCI_FLAG,
L"to\0", 0x00000008L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"spin\0", MCI_SPIN, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"up\0", 0x00010000L, MCI_FLAG,
L"down\0", 0x00020000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"step\0", MCI_STEP, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"reverse\0", 0x00020000L, MCI_FLAG,
L"by\0", 0x00010000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"set\0", MCI_SET, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"time format\0", 0x00000400L, MCI_CONSTANT,
L"milliseconds\0", 0x00000000L, MCI_INTEGER,
L"ms\0", 0x00000000L, MCI_INTEGER,
L"frames\0", 0x00000003L, MCI_INTEGER,
L"hms\0", 0x00000001L, MCI_INTEGER,
L"track\0", 0x00004001L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"door open\0", 0x00000100L, MCI_FLAG,
L"door closed\0", 0x00000200L, MCI_FLAG,
L"audio\0", 0x00000800L, MCI_CONSTANT,
L"all\0", 0x00000000L, MCI_INTEGER,
L"left\0", 0x00000001L, MCI_INTEGER,
L"right\0", 0x00000002L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"video\0", 0x00001000L, MCI_FLAG,
L"on\0", 0x00002000L, MCI_FLAG,
L"off\0", 0x00004000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"status\0", MCI_STATUS, 0, MCI_COMMAND_HEAD,
L"\0", MCI_DWORD_PTR, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_CONSTANT,
L"position\0", 0x00000002L, MCI_INTEGER,
L"length\0", 0x00000001L, MCI_INTEGER,
L"number of tracks\0", 0x00000003L, MCI_INTEGER,
L"mode\0", 0x00000004L, MCI_INTEGER,
L"media present\0", 0x00000005L, MCI_INTEGER,
L"speed\0", 0x00004002L, MCI_INTEGER,
L"forward\0", 0x00004003L, MCI_INTEGER,
L"media type\0", 0x00004004L, MCI_INTEGER,
L"ready\0", 0x00000007L, MCI_INTEGER,
L"side\0", 0x00004005L, MCI_INTEGER,
L"disc size\0", 0x00004006L, MCI_INTEGER,
L"time format\0", 0x00000006L, MCI_INTEGER,
L"current track\0", 0x00000008L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"track\0", 0x00000010L, MCI_INTEGER,
L"start\0", 0x00000200L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"capability\0", MCI_GETDEVCAPS, 0, MCI_COMMAND_HEAD,
L"\0", MCI_INTEGER, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"clv\0", 0x00010000L, MCI_FLAG,
L"cav\0", 0x00020000L, MCI_FLAG,
L"\0", 0x00000100L, MCI_CONSTANT,
L"can record\0", 0x00000001L, MCI_INTEGER,
L"has audio\0", 0x00000002L, MCI_INTEGER,
L"has video\0", 0x00000003L, MCI_INTEGER,
L"device type\0", 0x00000004L, MCI_INTEGER,
L"uses files\0", 0x00000005L, MCI_INTEGER,
L"compound device\0", 0x00000006L, MCI_INTEGER,
L"can eject\0", 0x00000007L, MCI_INTEGER,
L"can reverse\0", 0x00004002L, MCI_INTEGER,
L"can play\0", 0x00000008L, MCI_INTEGER,
L"can save\0", 0x00000009L, MCI_INTEGER,
L"fast play rate\0", 0x00004003L, MCI_INTEGER,
L"slow play rate\0", 0x00004004L, MCI_INTEGER,
L"normal play rate\0", 0x00004005L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"escape\0", MCI_ESCAPE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_STRING,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"info\0", MCI_INFO, 0, MCI_COMMAND_HEAD,
L"\0", MCI_STRING, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"product\0", 0x00000100L, MCI_FLAG,
L"identity\0", 0x00000800L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"\0", 0x00000000L, MCI_END_COMMAND_LIST
END

WAVEAUDIO RCDATA
BEGIN
L"open\0", MCI_OPEN, 0, MCI_COMMAND_HEAD,
L"\0", MCI_INTEGER, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"type\0", 0x00002000L, MCI_STRING,
L"element\0", 0x00000200L, MCI_STRING,
L"alias\0", 0x00000400L, MCI_STRING,
L"shareable\0", 0x00000100L, MCI_FLAG,
L"buffer\0", 0x00010000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"status\0", MCI_STATUS, 0, MCI_COMMAND_HEAD,
L"\0", MCI_DWORD_PTR, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_CONSTANT,
L"position\0", 0x00000002L, MCI_INTEGER,
L"length\0", 0x00000001L, MCI_INTEGER,
L"number of tracks\0", 0x00000003L, MCI_INTEGER,
L"media present\0", 0x00000005L, MCI_INTEGER,
L"mode\0", 0x00000004L, MCI_INTEGER,
L"format tag\0", 0x00004001L, MCI_INTEGER,
L"channels\0", 0x00004002L, MCI_INTEGER,
L"samplespersec\0", 0x00004003L, MCI_INTEGER,
L"bytespersec\0", 0x00004004L, MCI_INTEGER,
L"alignment\0", 0x00004005L, MCI_INTEGER,
L"bitspersample\0", 0x00004006L, MCI_INTEGER,
L"input\0", 0x00400000L, MCI_INTEGER,
L"output\0", 0x00800000L, MCI_INTEGER,
L"level\0", 0x00004007L, MCI_INTEGER,
L"ready\0", 0x00000007L, MCI_INTEGER,
L"time format\0", 0x00000006L, MCI_INTEGER,
L"current track\0", 0x00000008L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"track\0", 0x00000010L, MCI_INTEGER,
L"start\0", 0x00000200L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"set\0", MCI_SET, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"any input\0", 0x04000000L, MCI_FLAG,
L"any output\0", 0x08000000L, MCI_FLAG,
L"time format\0", 0x00000400L, MCI_CONSTANT,
L"milliseconds\0", 0x00000000L, MCI_INTEGER,
L"ms\0", 0x00000000L, MCI_INTEGER,
L"bytes\0", 0x00000008L, MCI_INTEGER,
L"samples\0", 0x00000009L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"audio\0", 0x00000800L, MCI_CONSTANT,
L"all\0", 0x00000000L, MCI_INTEGER,
L"left\0", 0x00000001L, MCI_INTEGER,
L"right\0", 0x00000002L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"video\0", 0x00001000L, MCI_FLAG,
L"on\0", 0x00002000L, MCI_FLAG,
L"off\0", 0x00004000L, MCI_FLAG,
L"door open\0", 0x00000100L, MCI_FLAG,
L"door closed\0", 0x00000200L, MCI_FLAG,
L"input\0", 0x00400000L, MCI_INTEGER,
L"output\0", 0x00800000L, MCI_INTEGER,
L"format tag\0", 0x00010000L, MCI_CONSTANT,
L"pcm\0", 0x00000001L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"channels\0", 0x00020000L, MCI_INTEGER,
L"samplespersec\0", 0x00040000L, MCI_INTEGER,
L"bytespersec\0", 0x00080000L, MCI_INTEGER,
L"alignment\0", 0x00100000L, MCI_INTEGER,
L"bitspersample\0", 0x00200000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"capability\0", MCI_GETDEVCAPS, 0, MCI_COMMAND_HEAD,
L"\0", MCI_INTEGER, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_CONSTANT,
L"can record\0", 0x00000001L, MCI_INTEGER,
L"has audio\0", 0x00000002L, MCI_INTEGER,
L"has video\0", 0x00000003L, MCI_INTEGER,
L"device type\0", 0x00000004L, MCI_INTEGER,
L"uses files\0", 0x00000005L, MCI_INTEGER,
L"compound device\0", 0x00000006L, MCI_INTEGER,
L"can eject\0", 0x00000007L, MCI_INTEGER,
L"can play\0", 0x00000008L, MCI_INTEGER,
L"can save\0", 0x00000009L, MCI_INTEGER,
L"inputs\0", 0x00004001L, MCI_INTEGER,
L"outputs\0", 0x00004002L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"info\0", MCI_INFO, 0, MCI_COMMAND_HEAD,
L"\0", MCI_STRING, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"product\0", 0x00000100L, MCI_FLAG,
L"input\0", 0x00400000L, MCI_FLAG,
L"output\0", 0x00800000L, MCI_FLAG,
L"file\0", 0x00000200L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"cue\0", MCI_CUE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"input\0", 0x00400000L, MCI_FLAG,
L"output\0", 0x00800000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"delete\0", MCI_DELETE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"from\0", 0x00000004L, MCI_INTEGER,
L"to\0", 0x00000008L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"\0", 0x00000000L, MCI_END_COMMAND_LIST
END

CDAUDIO RCDATA
BEGIN
L"status\0", MCI_STATUS, 0, MCI_COMMAND_HEAD,
L"\0", MCI_DWORD_PTR, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_CONSTANT,
L"position\0", 0x00000002L, MCI_INTEGER,
L"length\0", 0x00000001L, MCI_INTEGER,
L"number of tracks\0", 0x00000003L, MCI_INTEGER,
L"ready\0", 0x00000007L, MCI_INTEGER,
L"mode\0", 0x00000004L, MCI_INTEGER,
L"media present\0", 0x00000005L, MCI_INTEGER,
L"time format\0", 0x00000006L, MCI_INTEGER,
L"current track\0", 0x00000008L, MCI_INTEGER,
L"type\0", 0x00004001L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"track\0", 0x00000010L, MCI_INTEGER,
L"start\0", 0x00000200L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"set\0", MCI_SET, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"time format\0", 0x00000400L, MCI_CONSTANT,
L"msf\0", 0x00000002L, MCI_INTEGER,
L"tmsf\0", 0x0000000aL, MCI_INTEGER,
L"milliseconds\0", 0x00000000L, MCI_INTEGER,
L"ms\0", 0x00000000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"audio\0", 0x00000800L, MCI_CONSTANT,
L"all\0", 0x00000000L, MCI_INTEGER,
L"left\0", 0x00000001L, MCI_INTEGER,
L"right\0", 0x00000002L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"video\0", 0x00001000L, MCI_FLAG,
L"on\0", 0x00002000L, MCI_FLAG,
L"off\0", 0x00004000L, MCI_FLAG,
L"door open\0", 0x00000100L, MCI_FLAG,
L"door closed\0", 0x00000200L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"info\0", MCI_INFO, 0, MCI_COMMAND_HEAD,
L"\0", MCI_STRING, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"product\0", 0x00000100L, MCI_FLAG,
L"upc\0", 0x00000400L, MCI_FLAG,
L"identity\0", 0x00000800L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"\0", 0x00000000L, MCI_END_COMMAND_LIST
END

SEQUENCER RCDATA
BEGIN
L"status\0", MCI_STATUS, 0, MCI_COMMAND_HEAD,
L"\0", MCI_DWORD_PTR, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_CONSTANT,
L"position\0", 0x00000002L, MCI_INTEGER,
L"mode\0", 0x00000004L, MCI_INTEGER,
L"length\0", 0x00000001L, MCI_INTEGER,
L"number of tracks\0", 0x00000003L, MCI_INTEGER,
L"media present\0", 0x00000005L, MCI_INTEGER,
L"ready\0", 0x00000007L, MCI_INTEGER,
L"tempo\0", 0x00004002L, MCI_INTEGER,
L"port\0", 0x00004003L, MCI_INTEGER,
L"slave\0", 0x00004007L, MCI_INTEGER,
L"master\0", 0x00004008L, MCI_INTEGER,
L"offset\0", 0x00004009L, MCI_INTEGER,
L"division type\0", 0x0000400aL, MCI_INTEGER,
L"time format\0", 0x00000006L, MCI_INTEGER,
L"current track\0", 0x00000008L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"track\0", 0x00000010L, MCI_INTEGER,
L"start\0", 0x00000200L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"set\0", MCI_SET, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"time format\0", 0x00000400L, MCI_CONSTANT,
L"milliseconds\0", 0x00000000L, MCI_INTEGER,
L"ms\0", 0x00000000L, MCI_INTEGER,
L"smpte 30 drop\0", 0x00000007L, MCI_INTEGER,
L"smpte 30\0", 0x00000006L, MCI_INTEGER,
L"smpte 25\0", 0x00000005L, MCI_INTEGER,
L"smpte 24\0", 0x00000004L, MCI_INTEGER,
L"song pointer\0", 0x00004001L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"audio\0", 0x00000800L, MCI_CONSTANT,
L"all\0", 0x00000000L, MCI_INTEGER,
L"left\0", 0x00000001L, MCI_INTEGER,
L"right\0", 0x00000002L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"video\0", 0x00001000L, MCI_FLAG,
L"on\0", 0x00002000L, MCI_FLAG,
L"off\0", 0x00004000L, MCI_FLAG,
L"tempo\0", 0x00010000L, MCI_INTEGER,
L"port\0", 0x00020000L, MCI_CONSTANT,
L"none\0", 0x0000fffdL, MCI_INTEGER,
L"mapper\0", 0x0000ffffL, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"slave\0", 0x00040000L, MCI_CONSTANT,
L"smpte\0", 0x00004004L, MCI_INTEGER,
L"midi\0", 0x00004003L, MCI_INTEGER,
L"none\0", 0x0000fffdL, MCI_INTEGER,
L"file\0", 0x00004002L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"master\0", 0x00080000L, MCI_CONSTANT,
L"smpte\0", 0x00004004L, MCI_INTEGER,
L"midi\0", 0x00004003L, MCI_INTEGER,
L"none\0", 0x0000fffdL, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"offset\0", 0x01000000L, MCI_INTEGER,
L"door open\0", 0x00000100L, MCI_FLAG,
L"door closed\0", 0x00000200L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"info\0", MCI_INFO, 0, MCI_COMMAND_HEAD,
L"\0", MCI_STRING, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"product\0", 0x00000100L, MCI_FLAG,
L"file\0", 0x00000200L, MCI_FLAG,
L"name\0", 0x00001000L, MCI_FLAG,
L"copyright\0", 0x00002000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"\0", 0x00000000L, MCI_END_COMMAND_LIST
END

ANIMATION RCDATA
BEGIN
L"open\0", MCI_OPEN, 0, MCI_COMMAND_HEAD,
L"\0", MCI_INTEGER, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"type\0", 0x00002000L, MCI_STRING,
L"element\0", 0x00000200L, MCI_STRING,
L"alias\0", 0x00000400L, MCI_STRING,
L"shareable\0", 0x00000100L, MCI_FLAG,
L"style\0", 0x00010000L, MCI_CONSTANT,
L"overlapped\0", 0x00cf0000L, MCI_INTEGER,
L"popup\0", 0x80880000L, MCI_INTEGER,
L"child\0", 0x40000000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"parent\0", 0x00020000L, MCI_HWND,
L"nostatic\0", 0x00040000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"play\0", MCI_PLAY, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"from\0", 0x00000004L, MCI_INTEGER,
L"to\0", 0x00000008L, MCI_INTEGER,
L"fast\0", 0x00040000L, MCI_FLAG,
L"slow\0", 0x00080000L, MCI_FLAG,
L"scan\0", 0x00100000L, MCI_FLAG,
L"reverse\0", 0x00020000L, MCI_FLAG,
L"speed\0", 0x00010000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"step\0", MCI_STEP, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"reverse\0", 0x00010000L, MCI_FLAG,
L"by\0", 0x00020000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"set\0", MCI_SET, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"time format\0", 0x00000400L, MCI_CONSTANT,
L"milliseconds\0", 0x00000000L, MCI_INTEGER,
L"ms\0", 0x00000000L, MCI_INTEGER,
L"frames\0", 0x00000003L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"audio\0", 0x00000800L, MCI_CONSTANT,
L"all\0", 0x00000000L, MCI_INTEGER,
L"left\0", 0x00000001L, MCI_INTEGER,
L"right\0", 0x00000002L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"video\0", 0x00001000L, MCI_FLAG,
L"on\0", 0x00002000L, MCI_FLAG,
L"off\0", 0x00004000L, MCI_FLAG,
L"door open\0", 0x00000100L, MCI_FLAG,
L"door closed\0", 0x00000200L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"status\0", MCI_STATUS, 0, MCI_COMMAND_HEAD,
L"\0", MCI_DWORD_PTR, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_CONSTANT,
L"position\0", 0x00000002L, MCI_INTEGER,
L"length\0", 0x00000001L, MCI_INTEGER,
L"number of tracks\0", 0x00000003L, MCI_INTEGER,
L"mode\0", 0x00000004L, MCI_INTEGER,
L"ready\0", 0x00000007L, MCI_INTEGER,
L"speed\0", 0x00004001L, MCI_INTEGER,
L"forward\0", 0x00004002L, MCI_INTEGER,
L"window handle\0", 0x00004003L, MCI_INTEGER,
L"palette handle\0", 0x00004004L, MCI_INTEGER,
L"media present\0", 0x00000005L, MCI_INTEGER,
L"time format\0", 0x00000006L, MCI_INTEGER,
L"current track\0", 0x00000008L, MCI_INTEGER,
L"stretch\0", 0x00004005L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"track\0", 0x00000010L, MCI_INTEGER,
L"start\0", 0x00000200L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"info\0", MCI_INFO, 0, MCI_COMMAND_HEAD,
L"\0", MCI_STRING, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"product\0", 0x00000100L, MCI_FLAG,
L"file\0", 0x00000200L, MCI_FLAG,
L"window text\0", 0x00010000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"capability\0", MCI_GETDEVCAPS, 0, MCI_COMMAND_HEAD,
L"\0", MCI_INTEGER, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_CONSTANT,
L"can record\0", 0x00000001L, MCI_INTEGER,
L"has audio\0", 0x00000002L, MCI_INTEGER,
L"has video\0", 0x00000003L, MCI_INTEGER,
L"device type\0", 0x00000004L, MCI_INTEGER,
L"uses files\0", 0x00000005L, MCI_INTEGER,
L"compound device\0", 0x00000006L, MCI_INTEGER,
L"can eject\0", 0x00000007L, MCI_INTEGER,
L"can play\0", 0x00000008L, MCI_INTEGER,
L"can save\0", 0x00000009L, MCI_INTEGER,
L"can reverse\0", 0x00004001L, MCI_INTEGER,
L"fast play rate\0", 0x00004002L, MCI_INTEGER,
L"slow play rate\0", 0x00004003L, MCI_INTEGER,
L"normal play rate\0", 0x00004004L, MCI_INTEGER,
L"uses palettes\0", 0x00004006L, MCI_INTEGER,
L"can stretch\0", 0x00004007L, MCI_INTEGER,
L"windows\0", 0x00004008L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"realize\0", MCI_REALIZE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"normal\0", 0x00010000L, MCI_FLAG,
L"background\0", 0x00020000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"window\0", MCI_WINDOW, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"handle\0", 0x00010000L, MCI_CONSTANT,
L"default\0", 0x00000000L, MCI_HWND,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"state\0", 0x00040000L, MCI_CONSTANT,
L"hide\0", 0x00000000L, MCI_INTEGER,
L"minimize\0", 0x00000006L, MCI_INTEGER,
L"show\0", 0x00000005L, MCI_INTEGER,
L"maximized\0", 0x00000003L, MCI_INTEGER,
L"minimized\0", 0x00000002L, MCI_INTEGER,
L"iconic\0", 0x00000007L, MCI_INTEGER,
L"no action\0", 0x00000008L, MCI_INTEGER,
L"no activate\0", 0x00000004L, MCI_INTEGER,
L"normal\0", 0x00000001L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"text\0", 0x00080000L, MCI_STRING,
L"stretch\0", 0x00100000L, MCI_FLAG,
L"fixed\0", 0x00200000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"put\0", MCI_PUT, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"at\0", 0x00010000L, MCI_RECT,
L"source\0", 0x00020000L, MCI_FLAG,
L"destination\0", 0x00040000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"where\0", MCI_WHERE, 0, MCI_COMMAND_HEAD,
L"\0", MCI_RECT, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"source\0", 0x00020000L, MCI_FLAG,
L"destination\0", 0x00040000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"update\0", MCI_UPDATE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"at\0", 0x00010000L, MCI_RECT,
L"hdc\0", 0x00020000L, MCI_HDC,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"\0", 0x00000000L, MCI_END_COMMAND_LIST
END

OVERLAY RCDATA
BEGIN
L"open\0", MCI_OPEN, 0, MCI_COMMAND_HEAD,
L"\0", MCI_INTEGER, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"type\0", 0x00002000L, MCI_STRING,
L"element\0", 0x00000200L, MCI_STRING,
L"alias\0", 0x00000400L, MCI_STRING,
L"shareable\0", 0x00000100L, MCI_FLAG,
L"style\0", 0x00010000L, MCI_CONSTANT,
L"overlapped\0", 0x00cf0000L, MCI_INTEGER,
L"popup\0", 0x80880000L, MCI_INTEGER,
L"child\0", 0x40000000L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"parent\0", 0x00020000L, MCI_HWND,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"status\0", MCI_STATUS, 0, MCI_COMMAND_HEAD,
L"\0", MCI_DWORD_PTR, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_CONSTANT,
L"position\0", 0x00000002L, MCI_INTEGER,
L"length\0", 0x00000001L, MCI_INTEGER,
L"number of tracks\0", 0x00000003L, MCI_INTEGER,
L"mode\0", 0x00000004L, MCI_INTEGER,
L"ready\0", 0x00000007L, MCI_INTEGER,
L"window handle\0", 0x00004001L, MCI_INTEGER,
L"media present\0", 0x00000005L, MCI_INTEGER,
L"stretch\0", 0x00004002L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"track\0", 0x00000010L, MCI_INTEGER,
L"start\0", 0x00000200L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"info\0", MCI_INFO, 0, MCI_COMMAND_HEAD,
L"\0", MCI_STRING, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"product\0", 0x00000100L, MCI_FLAG,
L"file\0", 0x00000200L, MCI_FLAG,
L"window text\0", 0x00010000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"capability\0", MCI_GETDEVCAPS, 0, MCI_COMMAND_HEAD,
L"\0", MCI_INTEGER, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_CONSTANT,
L"can record\0", 0x00000001L, MCI_INTEGER,
L"has audio\0", 0x00000002L, MCI_INTEGER,
L"has video\0", 0x00000003L, MCI_INTEGER,
L"device type\0", 0x00000004L, MCI_INTEGER,
L"uses files\0", 0x00000005L, MCI_INTEGER,
L"compound device\0", 0x00000006L, MCI_INTEGER,
L"can eject\0", 0x00000007L, MCI_INTEGER,
L"can play\0", 0x00000008L, MCI_INTEGER,
L"can save\0", 0x00000009L, MCI_INTEGER,
L"can stretch\0", 0x00004001L, MCI_INTEGER,
L"can freeze\0", 0x00004002L, MCI_INTEGER,
L"windows\0", 0x00004003L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"window\0", MCI_WINDOW, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"handle\0", 0x00010000L, MCI_CONSTANT,
L"default\0", 0x00000000L, MCI_HWND,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"state\0", 0x00040000L, MCI_CONSTANT,
L"hide\0", 0x00000000L, MCI_INTEGER,
L"minimize\0", 0x00000006L, MCI_INTEGER,
L"show\0", 0x00000005L, MCI_INTEGER,
L"maximized\0", 0x00000003L, MCI_INTEGER,
L"minimized\0", 0x00000002L, MCI_INTEGER,
L"iconic\0", 0x00000007L, MCI_INTEGER,
L"no action\0", 0x00000008L, MCI_INTEGER,
L"no activate\0", 0x00000004L, MCI_INTEGER,
L"normal\0", 0x00000001L, MCI_INTEGER,
L"\0", 0x00000000L, MCI_END_CONSTANT,
L"text\0", 0x00080000L, MCI_STRING,
L"stretch\0", 0x00100000L, MCI_FLAG,
L"fixed\0", 0x00200000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"put\0", MCI_PUT, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"at\0", 0x00010000L, MCI_RECT,
L"source\0", 0x00020000L, MCI_FLAG,
L"destination\0", 0x00040000L, MCI_FLAG,
L"frame\0", 0x00080000L, MCI_FLAG,
L"video\0", 0x00100000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"where\0", MCI_WHERE, 0, MCI_COMMAND_HEAD,
L"\0", MCI_RECT, 0, MCI_RETURN,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"source\0", 0x00020000L, MCI_FLAG,
L"destination\0", 0x00040000L, MCI_FLAG,
L"frame\0", 0x00080000L, MCI_FLAG,
L"video\0", 0x00100000L, MCI_FLAG,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"save\0", MCI_SAVE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_STRING,
L"at\0", 0x00010000L, MCI_RECT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"load\0", MCI_LOAD, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"\0", 0x00000100L, MCI_STRING,
L"at\0", 0x00010000L, MCI_RECT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"freeze\0", MCI_FREEZE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"at\0", 0x00010000L, MCI_RECT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"unfreeze\0", MCI_UNFREEZE, 0, MCI_COMMAND_HEAD,
L"notify\0", 0x00000001L, MCI_FLAG,
L"wait\0", 0x00000002L, MCI_FLAG,
L"at\0", 0x00010000L, MCI_RECT,
L"\0", 0x00000000L, MCI_END_COMMAND,
L"\0", 0x00000000L, MCI_END_COMMAND_LIST
END

STRINGTABLE
BEGIN

MCI_DEVTYPE_VCR,		"vcr"
MCI_DEVTYPE_VIDEODISC,		"videodisc"
MCI_DEVTYPE_OVERLAY,		"overlay"
MCI_DEVTYPE_CD_AUDIO,		"cdaudio"
MCI_DEVTYPE_DAT,		"dat"
MCI_DEVTYPE_SCANNER,		"scanner"
MCI_DEVTYPE_ANIMATION,		"animation"
MCI_DEVTYPE_DIGITAL_VIDEO,	"digitalvideo"
MCI_DEVTYPE_OTHER,		"other"
MCI_DEVTYPE_WAVEFORM_AUDIO,	"waveaudio"
MCI_DEVTYPE_SEQUENCER,		"sequencer"

MCI_MODE_NOT_READY,		"not ready"
MCI_MODE_STOP,			"stopped"
MCI_MODE_PLAY,			"playing"
MCI_MODE_RECORD,		"recording"
MCI_MODE_SEEK,			"seeking"
MCI_MODE_PAUSE,			"paused"
MCI_MODE_OPEN,			"open"
MCI_FALSE,			"false"
MCI_TRUE,			"true"
MCI_FORMAT_MILLISECONDS_S, 	"milliseconds"
MCI_FORMAT_HMS_S,		"hms"
MCI_FORMAT_MSF_S, 		"msf"
MCI_FORMAT_FRAMES_S, 		"frames"
MCI_FORMAT_SMPTE_24_S,		"smpte 24"
MCI_FORMAT_SMPTE_25_S,		"smpte 25"
MCI_FORMAT_SMPTE_30_S,		"smpte 30"
MCI_FORMAT_SMPTE_30DROP_S,	"smpte 30 drop"
MCI_FORMAT_BYTES_S, 		"bytes"
MCI_FORMAT_SAMPLES_S,		"samples"
MCI_FORMAT_TMSF_S,		"tmsf"

MCI_VD_MODE_PARK,		"parked"
MCI_VD_MEDIA_CLV,		"CLV"
MCI_VD_MEDIA_CAV,		"CAV"
MCI_VD_MEDIA_OTHER		"other"

MCI_VD_FORMAT_TRACK_S, 		"track"

MCI_SEQ_DIV_PPQN,		"PPQN"
MCI_SEQ_DIV_SMPTE_24		"SMPTE 24 Frame"
MCI_SEQ_DIV_SMPTE_25		"SMPTE 25 Frame"
MCI_SEQ_DIV_SMPTE_30DROP	"SMPTE 30 Drop Frame"
MCI_SEQ_DIV_SMPTE_30		"SMPTE 30 Frame"

MCI_SEQ_FILE_S, 		"file"
MCI_SEQ_MIDI_S,			"midi"
MCI_SEQ_SMPTE_S, 		"smpte"
MCI_SEQ_FORMAT_SONGPTR_S, 	"song pointer"
MCI_SEQ_NONE_S,			"none"
MCI_SEQ_MAPPER_S, 		"mapper"
WAVE_FORMAT_PCM_S, 		"pcm"
WAVE_MAPPER_S, 			"mapper"
END
