/*
 * Top level resource file for avifil32.dll
 *
 * Copyright 2002 Michael Günnewig
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "avifile_private.h"

#pragma makedep po

LANGUAGE LANG_ENGLISH, SUBLANG_DEFAULT

STRINGTABLE
{
  IDS_WAVESTREAMFORMAT	"Waveform: %s"
  IDS_WAVEFILETYPE	"Waveform"
  IDS_ALLMULTIMEDIA	"All multimedia files"
  IDS_ALLFILES		"All files (*.*)"
  IDS_VIDEO		"video"
  IDS_AUDIO		"audio"
  IDS_AVIFILETYPE	"Wine AVI-default-filehandler"
  IDS_UNCOMPRESSED      "uncompressed"
}

IDD_SAVEOPTIONS DIALOG  43, 37, 226, 82
STYLE DS_MODALFRAME | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Compress options"
FONT 8, "MS Shell Dlg"
BEGIN
    LTEXT           "&Choose a stream:",-1,5,5,154,10
    COMBOBOX        IDC_STREAM,5,18,154,61,CBS_DROPDOWNLIST | WS_VSCROLL |
                    WS_TABSTOP
    PUSHBUTTON      "&Options...",IDC_OPTIONS,170,17,50,14
    AUTOCHECKBOX    "&Interleave every",IDC_INTERLEAVE,5,42,85,11,WS_TABSTOP
    EDITTEXT        IDC_INTERLEAVEEVERY,91,41,32,12,ES_AUTOHSCROLL
    LTEXT           "frames",-1,129,43,36,9
    LTEXT           "Current format:",-1,5,56,73,9
    LTEXT           "",IDC_FORMATTEXT,80,56,88,26
    DEFPUSHBUTTON   "OK",IDOK,170,42,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,170,61,50,14
END

LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL

/* @makedep: avifile.rgs */
1 WINE_REGISTRY avifile.rgs

#define WINE_FILEDESCRIPTION_STR "Wine AVI file support library"
#define WINE_FILENAME_STR "avifil32.dll"
#define WINE_FILEVERSION 4,0,3,1998
#define WINE_FILEVERSION_STR "4.03.1998"
#define WINE_EXTRAVALUES VALUE "OLESelfRegister",""

#include "wine/wine_common_ver.rc"
