/* Capstone Disassembly Engine, http://www.capstone-engine.org */
/* This is auto-gen data for Capstone disassembly engine (www.capstone-engine.org) */
/* By Nguyen Anh Quynh <aquynh@gmail.com>, 2013-2019 */


{
	AArch64_RDFFR_P, ARM64_INS_RDFFR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RDFFR_PPz, ARM64_INS_RDFFR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ABS_ZPmZ_B, ARM64_INS_ABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ABS_ZPmZ_D, ARM64_INS_ABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ABS_ZPmZ_H, ARM64_INS_ABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ABS_ZPmZ_S, ARM64_INS_ABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ABSv16i8, ARM64_INS_ABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ABSv1i64, ARM64_INS_ABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ABSv2i32, ARM64_INS_ABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ABSv2i64, ARM64_INS_ABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ABSv4i16, ARM64_INS_ABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ABSv4i32, ARM64_INS_ABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ABSv8i16, ARM64_INS_ABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ABSv8i8, ARM64_INS_ABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADCLB_ZZZ_D, ARM64_INS_ADCLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADCLB_ZZZ_S, ARM64_INS_ADCLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADCLT_ZZZ_D, ARM64_INS_ADCLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADCLT_ZZZ_S, ARM64_INS_ADCLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADCSWr, ARM64_INS_ADCS,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADCSXr, ARM64_INS_ADCS,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADCWr, ARM64_INS_ADC,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADCXr, ARM64_INS_ADC,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDG, ARM64_INS_ADDG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDHA_MPPZ_D, ARM64_INS_ADDHA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDHA_MPPZ_S, ARM64_INS_ADDHA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDHNB_ZZZ_B, ARM64_INS_ADDHNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDHNB_ZZZ_H, ARM64_INS_ADDHNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDHNB_ZZZ_S, ARM64_INS_ADDHNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDHNT_ZZZ_B, ARM64_INS_ADDHNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDHNT_ZZZ_H, ARM64_INS_ADDHNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDHNT_ZZZ_S, ARM64_INS_ADDHNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDHNv2i64_v2i32, ARM64_INS_ADDHN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDHNv2i64_v4i32, ARM64_INS_ADDHN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDHNv4i32_v4i16, ARM64_INS_ADDHN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDHNv4i32_v8i16, ARM64_INS_ADDHN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDHNv8i16_v16i8, ARM64_INS_ADDHN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDHNv8i16_v8i8, ARM64_INS_ADDHN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDPL_XXI, ARM64_INS_ADDPL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDP_ZPmZ_B, ARM64_INS_ADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDP_ZPmZ_D, ARM64_INS_ADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDP_ZPmZ_H, ARM64_INS_ADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDP_ZPmZ_S, ARM64_INS_ADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDPv16i8, ARM64_INS_ADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDPv2i32, ARM64_INS_ADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDPv2i64, ARM64_INS_ADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDPv2i64p, ARM64_INS_ADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDPv4i16, ARM64_INS_ADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDPv4i32, ARM64_INS_ADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDPv8i16, ARM64_INS_ADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDPv8i8, ARM64_INS_ADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDSWri, ARM64_INS_ADDS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDSWrs, ARM64_INS_ADDS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDSWrx, ARM64_INS_ADDS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDSXri, ARM64_INS_ADDS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDSXrs, ARM64_INS_ADDS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDSXrx, ARM64_INS_ADDS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDSXrx64, ARM64_INS_ADDS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDVA_MPPZ_D, ARM64_INS_ADDVA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDVA_MPPZ_S, ARM64_INS_ADDVA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDVL_XXI, ARM64_INS_ADDVL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDVv16i8v, ARM64_INS_ADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDVv4i16v, ARM64_INS_ADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDVv4i32v, ARM64_INS_ADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDVv8i16v, ARM64_INS_ADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDVv8i8v, ARM64_INS_ADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDWri, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDWrs, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDWrx, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDXri, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDXrs, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDXrx, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDXrx64, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADD_ZI_B, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADD_ZI_D, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADD_ZI_H, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADD_ZI_S, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADD_ZPmZ_B, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADD_ZPmZ_D, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADD_ZPmZ_H, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADD_ZPmZ_S, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADD_ZZZ_B, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADD_ZZZ_D, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADD_ZZZ_H, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADD_ZZZ_S, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADDv16i8, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDv1i64, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDv2i32, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDv2i64, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDv4i16, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDv4i32, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDv8i16, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADDv8i8, ARM64_INS_ADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ADR, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADRP, ARM64_INS_ADRP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_LSL_ZZZ_D_0, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_LSL_ZZZ_D_1, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_LSL_ZZZ_D_2, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_LSL_ZZZ_D_3, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_LSL_ZZZ_S_0, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_LSL_ZZZ_S_1, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_LSL_ZZZ_S_2, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_LSL_ZZZ_S_3, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_SXTW_ZZZ_D_0, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_SXTW_ZZZ_D_1, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_SXTW_ZZZ_D_2, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_SXTW_ZZZ_D_3, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_UXTW_ZZZ_D_0, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_UXTW_ZZZ_D_1, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_UXTW_ZZZ_D_2, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ADR_UXTW_ZZZ_D_3, ARM64_INS_ADR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_AESD_ZZZ_B, ARM64_INS_AESD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_AESDrr, ARM64_INS_AESD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_AESE_ZZZ_B, ARM64_INS_AESE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_AESErr, ARM64_INS_AESE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_AESIMC_ZZ_B, ARM64_INS_AESIMC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_AESIMCrr, ARM64_INS_AESIMC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_AESMC_ZZ_B, ARM64_INS_AESMC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_AESMCrr, ARM64_INS_AESMC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_ANDSWri, ARM64_INS_ANDS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ANDSWrs, ARM64_INS_ANDS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ANDSXri, ARM64_INS_ANDS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ANDSXrs, ARM64_INS_ANDS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ANDS_PPzPP, ARM64_INS_ANDS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ANDV_VPZ_B, ARM64_INS_ANDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ANDV_VPZ_D, ARM64_INS_ANDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ANDV_VPZ_H, ARM64_INS_ANDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ANDV_VPZ_S, ARM64_INS_ANDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ANDWri, ARM64_INS_AND,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ANDWrs, ARM64_INS_AND,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ANDXri, ARM64_INS_AND,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ANDXrs, ARM64_INS_AND,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_AND_PPzPP, ARM64_INS_AND,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_AND_ZI, ARM64_INS_AND,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_AND_ZPmZ_B, ARM64_INS_AND,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_AND_ZPmZ_D, ARM64_INS_AND,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_AND_ZPmZ_H, ARM64_INS_AND,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_AND_ZPmZ_S, ARM64_INS_AND,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_AND_ZZZ, ARM64_INS_AND,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ANDv16i8, ARM64_INS_AND,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ANDv8i8, ARM64_INS_AND,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ASRD_ZPmI_B, ARM64_INS_ASRD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASRD_ZPmI_D, ARM64_INS_ASRD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASRD_ZPmI_H, ARM64_INS_ASRD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASRD_ZPmI_S, ARM64_INS_ASRD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASRR_ZPmZ_B, ARM64_INS_ASRR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASRR_ZPmZ_D, ARM64_INS_ASRR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASRR_ZPmZ_H, ARM64_INS_ASRR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASRR_ZPmZ_S, ARM64_INS_ASRR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASRVWr, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASRVXr, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_WIDE_ZPmZ_B, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_WIDE_ZPmZ_H, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_WIDE_ZPmZ_S, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_WIDE_ZZZ_B, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_WIDE_ZZZ_H, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_WIDE_ZZZ_S, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_ZPmI_B, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_ZPmI_D, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_ZPmI_H, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_ZPmI_S, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_ZPmZ_B, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_ZPmZ_D, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_ZPmZ_H, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_ZPmZ_S, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_ZZI_B, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_ZZI_D, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_ZZI_H, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ASR_ZZI_S, ARM64_INS_ASR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_AUTDA, ARM64_INS_AUTDA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_AUTDB, ARM64_INS_AUTDB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_AUTDZA, ARM64_INS_AUTDZA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_AUTDZB, ARM64_INS_AUTDZB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_AUTIA, ARM64_INS_AUTIA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_AUTIA1716, ARM64_INS_AUTIA1716,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_X16, ARM64_REG_X17, 0 }, { ARM64_REG_X17, 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_AUTIASP, ARM64_INS_AUTIASP,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_LR, ARM64_REG_SP, 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_AUTIAZ, ARM64_INS_AUTIAZ,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_LR, 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_AUTIB, ARM64_INS_AUTIB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_AUTIB1716, ARM64_INS_AUTIB1716,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_X16, ARM64_REG_X17, 0 }, { ARM64_REG_X17, 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_AUTIBSP, ARM64_INS_AUTIBSP,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_LR, ARM64_REG_SP, 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_AUTIBZ, ARM64_INS_AUTIBZ,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_LR, 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_AUTIZA, ARM64_INS_AUTIZA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_AUTIZB, ARM64_INS_AUTIZB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_AXFLAG, ARM64_INS_AXFLAG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_B, ARM64_INS_B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, 0 }, 1, 0
#endif
},

{
	AArch64_BCAX, ARM64_INS_BCAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BCAX_ZZZZ, ARM64_INS_BCAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BCcc, ARM64_INS_BC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BDEP_ZZZ_B, ARM64_INS_BDEP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BDEP_ZZZ_D, ARM64_INS_BDEP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BDEP_ZZZ_H, ARM64_INS_BDEP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BDEP_ZZZ_S, ARM64_INS_BDEP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BEXT_ZZZ_B, ARM64_INS_BEXT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BEXT_ZZZ_D, ARM64_INS_BEXT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BEXT_ZZZ_H, ARM64_INS_BEXT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BEXT_ZZZ_S, ARM64_INS_BEXT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BF16DOTlanev4bf16, ARM64_INS_BFDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BF16DOTlanev8bf16, ARM64_INS_BFDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFCVT, ARM64_INS_BFCVT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFCVTN, ARM64_INS_BFCVTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFCVTN2, ARM64_INS_BFCVTN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFCVTNT_ZPmZ, ARM64_INS_BFCVTNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFCVT_ZPmZ, ARM64_INS_BFCVT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFDOT_ZZI, ARM64_INS_BFDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFDOT_ZZZ, ARM64_INS_BFDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFDOTv4bf16, ARM64_INS_BFDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFDOTv8bf16, ARM64_INS_BFDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFMLALB, ARM64_INS_BFMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFMLALBIdx, ARM64_INS_BFMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFMLALT, ARM64_INS_BFMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFMLALTIdx, ARM64_INS_BFMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFMMLA, ARM64_INS_BFMMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFMMLA_B_ZZI, ARM64_INS_BFMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFMMLA_B_ZZZ, ARM64_INS_BFMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFMMLA_T_ZZI, ARM64_INS_BFMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFMMLA_T_ZZZ, ARM64_INS_BFMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFMMLA_ZZZ, ARM64_INS_BFMMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFMWri, ARM64_INS_BFM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BFMXri, ARM64_INS_BFM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BGRP_ZZZ_B, ARM64_INS_BGRP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BGRP_ZZZ_D, ARM64_INS_BGRP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BGRP_ZZZ_H, ARM64_INS_BGRP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BGRP_ZZZ_S, ARM64_INS_BGRP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BICSWrs, ARM64_INS_BICS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BICSXrs, ARM64_INS_BICS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BICS_PPzPP, ARM64_INS_BICS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BICWrs, ARM64_INS_BIC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BICXrs, ARM64_INS_BIC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BIC_PPzPP, ARM64_INS_BIC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BIC_ZPmZ_B, ARM64_INS_BIC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BIC_ZPmZ_D, ARM64_INS_BIC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BIC_ZPmZ_H, ARM64_INS_BIC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BIC_ZPmZ_S, ARM64_INS_BIC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BIC_ZZZ, ARM64_INS_BIC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BICv16i8, ARM64_INS_BIC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_BICv2i32, ARM64_INS_BIC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_BICv4i16, ARM64_INS_BIC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_BICv4i32, ARM64_INS_BIC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_BICv8i16, ARM64_INS_BIC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_BICv8i8, ARM64_INS_BIC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_BIFv16i8, ARM64_INS_BIF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_BIFv8i8, ARM64_INS_BIF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_BITv16i8, ARM64_INS_BIT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_BITv8i8, ARM64_INS_BIT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_BL, ARM64_INS_BL,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_CALL, ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, 0 }, 1, 0
#endif
},

{
	AArch64_BLR, ARM64_INS_BLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_CALL, ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, 0 }, 1, 1
#endif
},

{
	AArch64_BLRAA, ARM64_INS_BLRAA,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_CALL, ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_BLRAAZ, ARM64_INS_BLRAAZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_CALL, ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_BLRAB, ARM64_INS_BLRAB,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_CALL, ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_BLRABZ, ARM64_INS_BLRABZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_CALL, ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_BR, ARM64_INS_BR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_JUMP, 0 }, 1, 1
#endif
},

{
	AArch64_BRAA, ARM64_INS_BRAA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_JUMP, ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_BRAAZ, ARM64_INS_BRAAZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_JUMP, ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_BRAB, ARM64_INS_BRAB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_JUMP, ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_BRABZ, ARM64_INS_BRABZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_JUMP, ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_BRB_IALL, ARM64_INS_BRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BRB_INJ, ARM64_INS_BRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BRK, ARM64_INS_BRK,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BRKAS_PPzP, ARM64_INS_BRKAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BRKA_PPmP, ARM64_INS_BRKA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BRKA_PPzP, ARM64_INS_BRKA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BRKBS_PPzP, ARM64_INS_BRKBS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BRKB_PPmP, ARM64_INS_BRKB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BRKB_PPzP, ARM64_INS_BRKB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BRKNS_PPzP, ARM64_INS_BRKNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BRKN_PPzP, ARM64_INS_BRKN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BRKPAS_PPzPP, ARM64_INS_BRKPAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BRKPA_PPzPP, ARM64_INS_BRKPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BRKPBS_PPzPP, ARM64_INS_BRKPBS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BRKPB_PPzPP, ARM64_INS_BRKPB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BSL1N_ZZZZ, ARM64_INS_BSL1N,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BSL2N_ZZZZ, ARM64_INS_BSL2N,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BSL_ZZZZ, ARM64_INS_BSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_BSLv16i8, ARM64_INS_BSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_BSLv8i8, ARM64_INS_BSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_Bcc, ARM64_INS_B,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, 0 }, 1, 0
#endif
},

{
	AArch64_CADD_ZZI_B, ARM64_INS_CADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CADD_ZZI_D, ARM64_INS_CADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CADD_ZZI_H, ARM64_INS_CADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CADD_ZZI_S, ARM64_INS_CADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASAB, ARM64_INS_CASAB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASAH, ARM64_INS_CASAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASALB, ARM64_INS_CASALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASALH, ARM64_INS_CASALH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASALW, ARM64_INS_CASAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASALX, ARM64_INS_CASAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASAW, ARM64_INS_CASA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASAX, ARM64_INS_CASA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASB, ARM64_INS_CASB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASH, ARM64_INS_CASH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASLB, ARM64_INS_CASLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASLH, ARM64_INS_CASLH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASLW, ARM64_INS_CASL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASLX, ARM64_INS_CASL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASPALW, ARM64_INS_CASPAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASPALX, ARM64_INS_CASPAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASPAW, ARM64_INS_CASPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASPAX, ARM64_INS_CASPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASPLW, ARM64_INS_CASPL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASPLX, ARM64_INS_CASPL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASPW, ARM64_INS_CASP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASPX, ARM64_INS_CASP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASW, ARM64_INS_CAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CASX, ARM64_INS_CAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CBNZW, ARM64_INS_CBNZ,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, 0 }, 1, 0
#endif
},

{
	AArch64_CBNZX, ARM64_INS_CBNZ,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, 0 }, 1, 0
#endif
},

{
	AArch64_CBZW, ARM64_INS_CBZ,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, 0 }, 1, 0
#endif
},

{
	AArch64_CBZX, ARM64_INS_CBZ,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, 0 }, 1, 0
#endif
},

{
	AArch64_CCMNWi, ARM64_INS_CCMN,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CCMNWr, ARM64_INS_CCMN,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CCMNXi, ARM64_INS_CCMN,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CCMNXr, ARM64_INS_CCMN,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CCMPWi, ARM64_INS_CCMP,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CCMPWr, ARM64_INS_CCMP,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CCMPXi, ARM64_INS_CCMP,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CCMPXr, ARM64_INS_CCMP,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CDOT_ZZZI_D, ARM64_INS_CDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CDOT_ZZZI_S, ARM64_INS_CDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CDOT_ZZZ_D, ARM64_INS_CDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CDOT_ZZZ_S, ARM64_INS_CDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CFINV, ARM64_INS_CFINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTA_RPZ_B, ARM64_INS_CLASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTA_RPZ_D, ARM64_INS_CLASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTA_RPZ_H, ARM64_INS_CLASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTA_RPZ_S, ARM64_INS_CLASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTA_VPZ_B, ARM64_INS_CLASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTA_VPZ_D, ARM64_INS_CLASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTA_VPZ_H, ARM64_INS_CLASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTA_VPZ_S, ARM64_INS_CLASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTA_ZPZ_B, ARM64_INS_CLASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTA_ZPZ_D, ARM64_INS_CLASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTA_ZPZ_H, ARM64_INS_CLASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTA_ZPZ_S, ARM64_INS_CLASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTB_RPZ_B, ARM64_INS_CLASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTB_RPZ_D, ARM64_INS_CLASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTB_RPZ_H, ARM64_INS_CLASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTB_RPZ_S, ARM64_INS_CLASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTB_VPZ_B, ARM64_INS_CLASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTB_VPZ_D, ARM64_INS_CLASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTB_VPZ_H, ARM64_INS_CLASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTB_VPZ_S, ARM64_INS_CLASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTB_ZPZ_B, ARM64_INS_CLASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTB_ZPZ_D, ARM64_INS_CLASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTB_ZPZ_H, ARM64_INS_CLASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLASTB_ZPZ_S, ARM64_INS_CLASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLREX, ARM64_INS_CLREX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLSWr, ARM64_INS_CLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLSXr, ARM64_INS_CLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLS_ZPmZ_B, ARM64_INS_CLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLS_ZPmZ_D, ARM64_INS_CLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLS_ZPmZ_H, ARM64_INS_CLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLS_ZPmZ_S, ARM64_INS_CLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLSv16i8, ARM64_INS_CLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CLSv2i32, ARM64_INS_CLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CLSv4i16, ARM64_INS_CLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CLSv4i32, ARM64_INS_CLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CLSv8i16, ARM64_INS_CLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CLSv8i8, ARM64_INS_CLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CLZWr, ARM64_INS_CLZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLZXr, ARM64_INS_CLZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLZ_ZPmZ_B, ARM64_INS_CLZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLZ_ZPmZ_D, ARM64_INS_CLZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLZ_ZPmZ_H, ARM64_INS_CLZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLZ_ZPmZ_S, ARM64_INS_CLZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CLZv16i8, ARM64_INS_CLZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CLZv2i32, ARM64_INS_CLZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CLZv4i16, ARM64_INS_CLZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CLZv4i32, ARM64_INS_CLZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CLZv8i16, ARM64_INS_CLZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CLZv8i8, ARM64_INS_CLZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv16i8, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv16i8rz, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv1i64, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv1i64rz, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv2i32, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv2i32rz, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv2i64, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv2i64rz, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv4i16, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv4i16rz, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv4i32, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv4i32rz, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv8i16, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv8i16rz, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv8i8, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMEQv8i8rz, ARM64_INS_CMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv16i8, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv16i8rz, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv1i64, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv1i64rz, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv2i32, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv2i32rz, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv2i64, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv2i64rz, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv4i16, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv4i16rz, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv4i32, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv4i32rz, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv8i16, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv8i16rz, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv8i8, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGEv8i8rz, ARM64_INS_CMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv16i8, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv16i8rz, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv1i64, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv1i64rz, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv2i32, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv2i32rz, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv2i64, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv2i64rz, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv4i16, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv4i16rz, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv4i32, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv4i32rz, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv8i16, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv8i16rz, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv8i8, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMGTv8i8rz, ARM64_INS_CMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHIv16i8, ARM64_INS_CMHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHIv1i64, ARM64_INS_CMHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHIv2i32, ARM64_INS_CMHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHIv2i64, ARM64_INS_CMHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHIv4i16, ARM64_INS_CMHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHIv4i32, ARM64_INS_CMHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHIv8i16, ARM64_INS_CMHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHIv8i8, ARM64_INS_CMHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHSv16i8, ARM64_INS_CMHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHSv1i64, ARM64_INS_CMHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHSv2i32, ARM64_INS_CMHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHSv2i64, ARM64_INS_CMHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHSv4i16, ARM64_INS_CMHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHSv4i32, ARM64_INS_CMHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHSv8i16, ARM64_INS_CMHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMHSv8i8, ARM64_INS_CMHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLA_ZZZI_H, ARM64_INS_CMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMLA_ZZZI_S, ARM64_INS_CMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMLA_ZZZ_B, ARM64_INS_CMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMLA_ZZZ_D, ARM64_INS_CMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMLA_ZZZ_H, ARM64_INS_CMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMLA_ZZZ_S, ARM64_INS_CMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMLEv16i8rz, ARM64_INS_CMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLEv1i64rz, ARM64_INS_CMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLEv2i32rz, ARM64_INS_CMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLEv2i64rz, ARM64_INS_CMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLEv4i16rz, ARM64_INS_CMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLEv4i32rz, ARM64_INS_CMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLEv8i16rz, ARM64_INS_CMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLEv8i8rz, ARM64_INS_CMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLTv16i8rz, ARM64_INS_CMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLTv1i64rz, ARM64_INS_CMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLTv2i32rz, ARM64_INS_CMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLTv2i64rz, ARM64_INS_CMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLTv4i16rz, ARM64_INS_CMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLTv4i32rz, ARM64_INS_CMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLTv8i16rz, ARM64_INS_CMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMLTv8i8rz, ARM64_INS_CMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMPEQ_PPzZI_B, ARM64_INS_CMPEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPEQ_PPzZI_D, ARM64_INS_CMPEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPEQ_PPzZI_H, ARM64_INS_CMPEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPEQ_PPzZI_S, ARM64_INS_CMPEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPEQ_PPzZZ_B, ARM64_INS_CMPEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPEQ_PPzZZ_D, ARM64_INS_CMPEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPEQ_PPzZZ_H, ARM64_INS_CMPEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPEQ_PPzZZ_S, ARM64_INS_CMPEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPEQ_WIDE_PPzZZ_B, ARM64_INS_CMPEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPEQ_WIDE_PPzZZ_H, ARM64_INS_CMPEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPEQ_WIDE_PPzZZ_S, ARM64_INS_CMPEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGE_PPzZI_B, ARM64_INS_CMPGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGE_PPzZI_D, ARM64_INS_CMPGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGE_PPzZI_H, ARM64_INS_CMPGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGE_PPzZI_S, ARM64_INS_CMPGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGE_PPzZZ_B, ARM64_INS_CMPGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGE_PPzZZ_D, ARM64_INS_CMPGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGE_PPzZZ_H, ARM64_INS_CMPGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGE_PPzZZ_S, ARM64_INS_CMPGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGE_WIDE_PPzZZ_B, ARM64_INS_CMPGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGE_WIDE_PPzZZ_H, ARM64_INS_CMPGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGE_WIDE_PPzZZ_S, ARM64_INS_CMPGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGT_PPzZI_B, ARM64_INS_CMPGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGT_PPzZI_D, ARM64_INS_CMPGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGT_PPzZI_H, ARM64_INS_CMPGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGT_PPzZI_S, ARM64_INS_CMPGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGT_PPzZZ_B, ARM64_INS_CMPGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGT_PPzZZ_D, ARM64_INS_CMPGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGT_PPzZZ_H, ARM64_INS_CMPGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGT_PPzZZ_S, ARM64_INS_CMPGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGT_WIDE_PPzZZ_B, ARM64_INS_CMPGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGT_WIDE_PPzZZ_H, ARM64_INS_CMPGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPGT_WIDE_PPzZZ_S, ARM64_INS_CMPGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHI_PPzZI_B, ARM64_INS_CMPHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHI_PPzZI_D, ARM64_INS_CMPHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHI_PPzZI_H, ARM64_INS_CMPHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHI_PPzZI_S, ARM64_INS_CMPHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHI_PPzZZ_B, ARM64_INS_CMPHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHI_PPzZZ_D, ARM64_INS_CMPHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHI_PPzZZ_H, ARM64_INS_CMPHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHI_PPzZZ_S, ARM64_INS_CMPHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHI_WIDE_PPzZZ_B, ARM64_INS_CMPHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHI_WIDE_PPzZZ_H, ARM64_INS_CMPHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHI_WIDE_PPzZZ_S, ARM64_INS_CMPHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHS_PPzZI_B, ARM64_INS_CMPHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHS_PPzZI_D, ARM64_INS_CMPHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHS_PPzZI_H, ARM64_INS_CMPHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHS_PPzZI_S, ARM64_INS_CMPHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHS_PPzZZ_B, ARM64_INS_CMPHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHS_PPzZZ_D, ARM64_INS_CMPHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHS_PPzZZ_H, ARM64_INS_CMPHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHS_PPzZZ_S, ARM64_INS_CMPHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHS_WIDE_PPzZZ_B, ARM64_INS_CMPHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHS_WIDE_PPzZZ_H, ARM64_INS_CMPHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPHS_WIDE_PPzZZ_S, ARM64_INS_CMPHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLE_PPzZI_B, ARM64_INS_CMPLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLE_PPzZI_D, ARM64_INS_CMPLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLE_PPzZI_H, ARM64_INS_CMPLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLE_PPzZI_S, ARM64_INS_CMPLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLE_WIDE_PPzZZ_B, ARM64_INS_CMPLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLE_WIDE_PPzZZ_H, ARM64_INS_CMPLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLE_WIDE_PPzZZ_S, ARM64_INS_CMPLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLO_PPzZI_B, ARM64_INS_CMPLO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLO_PPzZI_D, ARM64_INS_CMPLO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLO_PPzZI_H, ARM64_INS_CMPLO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLO_PPzZI_S, ARM64_INS_CMPLO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLO_WIDE_PPzZZ_B, ARM64_INS_CMPLO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLO_WIDE_PPzZZ_H, ARM64_INS_CMPLO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLO_WIDE_PPzZZ_S, ARM64_INS_CMPLO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLS_PPzZI_B, ARM64_INS_CMPLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLS_PPzZI_D, ARM64_INS_CMPLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLS_PPzZI_H, ARM64_INS_CMPLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLS_PPzZI_S, ARM64_INS_CMPLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLS_WIDE_PPzZZ_B, ARM64_INS_CMPLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLS_WIDE_PPzZZ_H, ARM64_INS_CMPLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLS_WIDE_PPzZZ_S, ARM64_INS_CMPLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLT_PPzZI_B, ARM64_INS_CMPLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLT_PPzZI_D, ARM64_INS_CMPLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLT_PPzZI_H, ARM64_INS_CMPLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLT_PPzZI_S, ARM64_INS_CMPLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLT_WIDE_PPzZZ_B, ARM64_INS_CMPLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLT_WIDE_PPzZZ_H, ARM64_INS_CMPLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPLT_WIDE_PPzZZ_S, ARM64_INS_CMPLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPNE_PPzZI_B, ARM64_INS_CMPNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPNE_PPzZI_D, ARM64_INS_CMPNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPNE_PPzZI_H, ARM64_INS_CMPNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPNE_PPzZI_S, ARM64_INS_CMPNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPNE_PPzZZ_B, ARM64_INS_CMPNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPNE_PPzZZ_D, ARM64_INS_CMPNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPNE_PPzZZ_H, ARM64_INS_CMPNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPNE_PPzZZ_S, ARM64_INS_CMPNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPNE_WIDE_PPzZZ_B, ARM64_INS_CMPNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPNE_WIDE_PPzZZ_H, ARM64_INS_CMPNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMPNE_WIDE_PPzZZ_S, ARM64_INS_CMPNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CMTSTv16i8, ARM64_INS_CMTST,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMTSTv1i64, ARM64_INS_CMTST,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMTSTv2i32, ARM64_INS_CMTST,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMTSTv2i64, ARM64_INS_CMTST,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMTSTv4i16, ARM64_INS_CMTST,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMTSTv4i32, ARM64_INS_CMTST,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMTSTv8i16, ARM64_INS_CMTST,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CMTSTv8i8, ARM64_INS_CMTST,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CNOT_ZPmZ_B, ARM64_INS_CNOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNOT_ZPmZ_D, ARM64_INS_CNOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNOT_ZPmZ_H, ARM64_INS_CNOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNOT_ZPmZ_S, ARM64_INS_CNOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNTB_XPiI, ARM64_INS_CNTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNTD_XPiI, ARM64_INS_CNTD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNTH_XPiI, ARM64_INS_CNTH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNTP_XPP_B, ARM64_INS_CNTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNTP_XPP_D, ARM64_INS_CNTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNTP_XPP_H, ARM64_INS_CNTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNTP_XPP_S, ARM64_INS_CNTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNTW_XPiI, ARM64_INS_CNTW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNT_ZPmZ_B, ARM64_INS_CNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNT_ZPmZ_D, ARM64_INS_CNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNT_ZPmZ_H, ARM64_INS_CNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNT_ZPmZ_S, ARM64_INS_CNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CNTv16i8, ARM64_INS_CNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_CNTv8i8, ARM64_INS_CNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_COMPACT_ZPZ_D, ARM64_INS_COMPACT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_COMPACT_ZPZ_S, ARM64_INS_COMPACT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYE, ARM64_INS_CPYE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYEN, ARM64_INS_CPYEN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYERN, ARM64_INS_CPYERN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYERT, ARM64_INS_CPYERT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYERTN, ARM64_INS_CPYERTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYERTRN, ARM64_INS_CPYERTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYERTWN, ARM64_INS_CPYERTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYET, ARM64_INS_CPYET,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYETN, ARM64_INS_CPYETN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYETRN, ARM64_INS_CPYETRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYETWN, ARM64_INS_CPYETWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYEWN, ARM64_INS_CPYEWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYEWT, ARM64_INS_CPYEWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYEWTN, ARM64_INS_CPYEWTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYEWTRN, ARM64_INS_CPYEWTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYEWTWN, ARM64_INS_CPYEWTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFE, ARM64_INS_CPYFE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFEN, ARM64_INS_CPYFEN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFERN, ARM64_INS_CPYFERN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFERT, ARM64_INS_CPYFERT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFERTN, ARM64_INS_CPYFERTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFERTRN, ARM64_INS_CPYFERTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFERTWN, ARM64_INS_CPYFERTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFET, ARM64_INS_CPYFET,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFETN, ARM64_INS_CPYFETN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFETRN, ARM64_INS_CPYFETRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFETWN, ARM64_INS_CPYFETWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFEWN, ARM64_INS_CPYFEWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFEWT, ARM64_INS_CPYFEWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFEWTN, ARM64_INS_CPYFEWTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFEWTRN, ARM64_INS_CPYFEWTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFEWTWN, ARM64_INS_CPYFEWTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFM, ARM64_INS_CPYFM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMN, ARM64_INS_CPYFMN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMRN, ARM64_INS_CPYFMRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMRT, ARM64_INS_CPYFMRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMRTN, ARM64_INS_CPYFMRTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMRTRN, ARM64_INS_CPYFMRTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMRTWN, ARM64_INS_CPYFMRTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMT, ARM64_INS_CPYFMT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMTN, ARM64_INS_CPYFMTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMTRN, ARM64_INS_CPYFMTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMTWN, ARM64_INS_CPYFMTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMWN, ARM64_INS_CPYFMWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMWT, ARM64_INS_CPYFMWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMWTN, ARM64_INS_CPYFMWTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMWTRN, ARM64_INS_CPYFMWTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFMWTWN, ARM64_INS_CPYFMWTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFP, ARM64_INS_CPYFP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPN, ARM64_INS_CPYFPN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPRN, ARM64_INS_CPYFPRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPRT, ARM64_INS_CPYFPRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPRTN, ARM64_INS_CPYFPRTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPRTRN, ARM64_INS_CPYFPRTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPRTWN, ARM64_INS_CPYFPRTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPT, ARM64_INS_CPYFPT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPTN, ARM64_INS_CPYFPTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPTRN, ARM64_INS_CPYFPTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPTWN, ARM64_INS_CPYFPTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPWN, ARM64_INS_CPYFPWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPWT, ARM64_INS_CPYFPWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPWTN, ARM64_INS_CPYFPWTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPWTRN, ARM64_INS_CPYFPWTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYFPWTWN, ARM64_INS_CPYFPWTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYM, ARM64_INS_CPYM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMN, ARM64_INS_CPYMN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMRN, ARM64_INS_CPYMRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMRT, ARM64_INS_CPYMRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMRTN, ARM64_INS_CPYMRTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMRTRN, ARM64_INS_CPYMRTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMRTWN, ARM64_INS_CPYMRTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMT, ARM64_INS_CPYMT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMTN, ARM64_INS_CPYMTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMTRN, ARM64_INS_CPYMTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMTWN, ARM64_INS_CPYMTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMWN, ARM64_INS_CPYMWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMWT, ARM64_INS_CPYMWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMWTN, ARM64_INS_CPYMWTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMWTRN, ARM64_INS_CPYMWTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYMWTWN, ARM64_INS_CPYMWTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYP, ARM64_INS_CPYP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPN, ARM64_INS_CPYPN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPRN, ARM64_INS_CPYPRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPRT, ARM64_INS_CPYPRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPRTN, ARM64_INS_CPYPRTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPRTRN, ARM64_INS_CPYPRTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPRTWN, ARM64_INS_CPYPRTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPT, ARM64_INS_CPYPT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPTN, ARM64_INS_CPYPTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPTRN, ARM64_INS_CPYPTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPTWN, ARM64_INS_CPYPTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPWN, ARM64_INS_CPYPWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPWT, ARM64_INS_CPYPWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPWTN, ARM64_INS_CPYPWTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPWTRN, ARM64_INS_CPYPWTRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPYPWTWN, ARM64_INS_CPYPWTWN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPmI_B, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPmI_D, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPmI_H, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPmI_S, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPmR_B, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPmR_D, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPmR_H, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPmR_S, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPmV_B, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPmV_D, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPmV_H, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPmV_S, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPzI_B, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPzI_D, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPzI_H, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CPY_ZPzI_S, ARM64_INS_CPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CRC32Brr, ARM64_INS_CRC32B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRC, 0 }, 0, 0
#endif
},

{
	AArch64_CRC32CBrr, ARM64_INS_CRC32CB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRC, 0 }, 0, 0
#endif
},

{
	AArch64_CRC32CHrr, ARM64_INS_CRC32CH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRC, 0 }, 0, 0
#endif
},

{
	AArch64_CRC32CWrr, ARM64_INS_CRC32CW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRC, 0 }, 0, 0
#endif
},

{
	AArch64_CRC32CXrr, ARM64_INS_CRC32CX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRC, 0 }, 0, 0
#endif
},

{
	AArch64_CRC32Hrr, ARM64_INS_CRC32H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRC, 0 }, 0, 0
#endif
},

{
	AArch64_CRC32Wrr, ARM64_INS_CRC32W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRC, 0 }, 0, 0
#endif
},

{
	AArch64_CRC32Xrr, ARM64_INS_CRC32X,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRC, 0 }, 0, 0
#endif
},

{
	AArch64_CSELWr, ARM64_INS_CSEL,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CSELXr, ARM64_INS_CSEL,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CSINCWr, ARM64_INS_CSINC,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CSINCXr, ARM64_INS_CSINC,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CSINVWr, ARM64_INS_CSINV,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CSINVXr, ARM64_INS_CSINV,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CSNEGWr, ARM64_INS_CSNEG,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CSNEGXr, ARM64_INS_CSNEG,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CTERMEQ_WW, ARM64_INS_CTERMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CTERMEQ_XX, ARM64_INS_CTERMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CTERMNE_WW, ARM64_INS_CTERMNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_CTERMNE_XX, ARM64_INS_CTERMNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DCPS1, ARM64_INS_DCPS1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DCPS2, ARM64_INS_DCPS2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DCPS3, ARM64_INS_DCPS3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DECB_XPiI, ARM64_INS_DECB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DECD_XPiI, ARM64_INS_DECD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DECD_ZPiI, ARM64_INS_DECD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DECH_XPiI, ARM64_INS_DECH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DECH_ZPiI, ARM64_INS_DECH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DECP_XP_B, ARM64_INS_DECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DECP_XP_D, ARM64_INS_DECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DECP_XP_H, ARM64_INS_DECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DECP_XP_S, ARM64_INS_DECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DECP_ZP_D, ARM64_INS_DECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DECP_ZP_H, ARM64_INS_DECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DECP_ZP_S, ARM64_INS_DECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DECW_XPiI, ARM64_INS_DECW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DECW_ZPiI, ARM64_INS_DECW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DMB, ARM64_INS_DMB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DRPS, ARM64_INS_DRPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DSB, ARM64_INS_DSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DSBnXS, ARM64_INS_DSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUPM_ZI, ARM64_INS_DUPM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUP_ZI_B, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUP_ZI_D, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUP_ZI_H, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUP_ZI_S, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUP_ZR_B, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUP_ZR_D, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUP_ZR_H, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUP_ZR_S, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUP_ZZI_B, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUP_ZZI_D, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUP_ZZI_H, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUP_ZZI_Q, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUP_ZZI_S, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUPi16, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUPi32, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUPi64, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUPi8, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_DUPv16i8gpr, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_DUPv16i8lane, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_DUPv2i32gpr, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_DUPv2i32lane, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_DUPv2i64gpr, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_DUPv2i64lane, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_DUPv4i16gpr, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_DUPv4i16lane, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_DUPv4i32gpr, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_DUPv4i32lane, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_DUPv8i16gpr, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_DUPv8i16lane, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_DUPv8i8gpr, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_DUPv8i8lane, ARM64_INS_DUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_EONWrs, ARM64_INS_EON,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EONXrs, ARM64_INS_EON,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EOR3, ARM64_INS_EOR3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EOR3_ZZZZ, ARM64_INS_EOR3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORBT_ZZZ_B, ARM64_INS_EORBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORBT_ZZZ_D, ARM64_INS_EORBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORBT_ZZZ_H, ARM64_INS_EORBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORBT_ZZZ_S, ARM64_INS_EORBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORS_PPzPP, ARM64_INS_EORS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORTB_ZZZ_B, ARM64_INS_EORTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORTB_ZZZ_D, ARM64_INS_EORTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORTB_ZZZ_H, ARM64_INS_EORTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORTB_ZZZ_S, ARM64_INS_EORTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORV_VPZ_B, ARM64_INS_EORV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORV_VPZ_D, ARM64_INS_EORV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORV_VPZ_H, ARM64_INS_EORV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORV_VPZ_S, ARM64_INS_EORV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORWri, ARM64_INS_EOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORWrs, ARM64_INS_EOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORXri, ARM64_INS_EOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORXrs, ARM64_INS_EOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EOR_PPzPP, ARM64_INS_EOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EOR_ZI, ARM64_INS_EOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EOR_ZPmZ_B, ARM64_INS_EOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EOR_ZPmZ_D, ARM64_INS_EOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EOR_ZPmZ_H, ARM64_INS_EOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EOR_ZPmZ_S, ARM64_INS_EOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EOR_ZZZ, ARM64_INS_EOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EORv16i8, ARM64_INS_EOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_EORv8i8, ARM64_INS_EOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ERET, ARM64_INS_ERET,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ERETAA, ARM64_INS_ERETAA,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_LR, ARM64_REG_SP, 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_ERETAB, ARM64_INS_ERETAB,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_LR, ARM64_REG_SP, 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_EXTRACT_ZPMXI_H_B, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EXTRACT_ZPMXI_H_D, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EXTRACT_ZPMXI_H_H, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EXTRACT_ZPMXI_H_Q, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EXTRACT_ZPMXI_H_S, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EXTRACT_ZPMXI_V_B, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EXTRACT_ZPMXI_V_D, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EXTRACT_ZPMXI_V_H, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EXTRACT_ZPMXI_V_Q, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EXTRACT_ZPMXI_V_S, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EXTRWrri, ARM64_INS_EXTR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EXTRXrri, ARM64_INS_EXTR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EXT_ZZI, ARM64_INS_EXT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EXT_ZZI_B, ARM64_INS_EXT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_EXTv16i8, ARM64_INS_EXT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_EXTv8i8, ARM64_INS_EXT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FABD16, ARM64_INS_FABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FABD32, ARM64_INS_FABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FABD64, ARM64_INS_FABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FABD_ZPmZ_D, ARM64_INS_FABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FABD_ZPmZ_H, ARM64_INS_FABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FABD_ZPmZ_S, ARM64_INS_FABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FABDv2f32, ARM64_INS_FABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FABDv2f64, ARM64_INS_FABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FABDv4f16, ARM64_INS_FABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FABDv4f32, ARM64_INS_FABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FABDv8f16, ARM64_INS_FABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FABSDr, ARM64_INS_FABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FABSHr, ARM64_INS_FABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FABSSr, ARM64_INS_FABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FABS_ZPmZ_D, ARM64_INS_FABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FABS_ZPmZ_H, ARM64_INS_FABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FABS_ZPmZ_S, ARM64_INS_FABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FABSv2f32, ARM64_INS_FABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FABSv2f64, ARM64_INS_FABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FABSv4f16, ARM64_INS_FABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FABSv4f32, ARM64_INS_FABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FABSv8f16, ARM64_INS_FABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FACGE16, ARM64_INS_FACGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FACGE32, ARM64_INS_FACGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FACGE64, ARM64_INS_FACGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FACGE_PPzZZ_D, ARM64_INS_FACGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FACGE_PPzZZ_H, ARM64_INS_FACGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FACGE_PPzZZ_S, ARM64_INS_FACGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FACGEv2f32, ARM64_INS_FACGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FACGEv2f64, ARM64_INS_FACGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FACGEv4f16, ARM64_INS_FACGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FACGEv4f32, ARM64_INS_FACGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FACGEv8f16, ARM64_INS_FACGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FACGT16, ARM64_INS_FACGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FACGT32, ARM64_INS_FACGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FACGT64, ARM64_INS_FACGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FACGT_PPzZZ_D, ARM64_INS_FACGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FACGT_PPzZZ_H, ARM64_INS_FACGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FACGT_PPzZZ_S, ARM64_INS_FACGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FACGTv2f32, ARM64_INS_FACGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FACGTv2f64, ARM64_INS_FACGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FACGTv4f16, ARM64_INS_FACGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FACGTv4f32, ARM64_INS_FACGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FACGTv8f16, ARM64_INS_FACGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDA_VPZ_D, ARM64_INS_FADDA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDA_VPZ_H, ARM64_INS_FADDA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDA_VPZ_S, ARM64_INS_FADDA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDDrr, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FADDHrr, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDP_ZPmZZ_D, ARM64_INS_FADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDP_ZPmZZ_H, ARM64_INS_FADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDP_ZPmZZ_S, ARM64_INS_FADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDPv2f32, ARM64_INS_FADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FADDPv2f64, ARM64_INS_FADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FADDPv2i16p, ARM64_INS_FADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDPv2i32p, ARM64_INS_FADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FADDPv2i64p, ARM64_INS_FADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FADDPv4f16, ARM64_INS_FADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDPv4f32, ARM64_INS_FADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FADDPv8f16, ARM64_INS_FADDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDSrr, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FADDV_VPZ_D, ARM64_INS_FADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDV_VPZ_H, ARM64_INS_FADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDV_VPZ_S, ARM64_INS_FADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADD_ZPmI_D, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADD_ZPmI_H, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADD_ZPmI_S, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADD_ZPmZ_D, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADD_ZPmZ_H, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADD_ZPmZ_S, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADD_ZZZ_D, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADD_ZZZ_H, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADD_ZZZ_S, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDv2f32, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FADDv2f64, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FADDv4f16, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FADDv4f32, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FADDv8f16, ARM64_INS_FADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCADD_ZPmZ_D, ARM64_INS_FCADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCADD_ZPmZ_H, ARM64_INS_FCADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCADD_ZPmZ_S, ARM64_INS_FCADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCADDv2f32, ARM64_INS_FCADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCADDv2f64, ARM64_INS_FCADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCADDv4f16, ARM64_INS_FCADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCADDv4f32, ARM64_INS_FCADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCADDv8f16, ARM64_INS_FCADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCCMPDrr, ARM64_INS_FCCMP,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCCMPEDrr, ARM64_INS_FCCMPE,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCCMPEHrr, ARM64_INS_FCCMPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCCMPESrr, ARM64_INS_FCCMPE,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCCMPHrr, ARM64_INS_FCCMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCCMPSrr, ARM64_INS_FCCMP,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQ16, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQ32, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQ64, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQ_PPzZ0_D, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQ_PPzZ0_H, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQ_PPzZ0_S, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQ_PPzZZ_D, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQ_PPzZZ_H, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQ_PPzZZ_S, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQv1i16rz, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQv1i32rz, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQv1i64rz, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQv2f32, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQv2f64, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQv2i32rz, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQv2i64rz, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQv4f16, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQv4f32, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQv4i16rz, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQv4i32rz, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQv8f16, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMEQv8i16rz, ARM64_INS_FCMEQ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGE16, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGE32, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGE64, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGE_PPzZ0_D, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGE_PPzZ0_H, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGE_PPzZ0_S, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGE_PPzZZ_D, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGE_PPzZZ_H, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGE_PPzZZ_S, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGEv1i16rz, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGEv1i32rz, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGEv1i64rz, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGEv2f32, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGEv2f64, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGEv2i32rz, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGEv2i64rz, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGEv4f16, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGEv4f32, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGEv4i16rz, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGEv4i32rz, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGEv8f16, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGEv8i16rz, ARM64_INS_FCMGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGT16, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGT32, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGT64, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGT_PPzZ0_D, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGT_PPzZ0_H, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGT_PPzZ0_S, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGT_PPzZZ_D, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGT_PPzZZ_H, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGT_PPzZZ_S, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGTv1i16rz, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGTv1i32rz, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGTv1i64rz, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGTv2f32, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGTv2f64, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGTv2i32rz, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGTv2i64rz, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGTv4f16, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGTv4f32, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGTv4i16rz, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGTv4i32rz, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMGTv8f16, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMGTv8i16rz, ARM64_INS_FCMGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLA_ZPmZZ_D, ARM64_INS_FCMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLA_ZPmZZ_H, ARM64_INS_FCMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLA_ZPmZZ_S, ARM64_INS_FCMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLA_ZZZI_H, ARM64_INS_FCMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLA_ZZZI_S, ARM64_INS_FCMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLAv2f32, ARM64_INS_FCMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLAv2f64, ARM64_INS_FCMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLAv4f16, ARM64_INS_FCMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLAv4f16_indexed, ARM64_INS_FCMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLAv4f32, ARM64_INS_FCMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLAv4f32_indexed, ARM64_INS_FCMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLAv8f16, ARM64_INS_FCMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLAv8f16_indexed, ARM64_INS_FCMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLE_PPzZ0_D, ARM64_INS_FCMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLE_PPzZ0_H, ARM64_INS_FCMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLE_PPzZ0_S, ARM64_INS_FCMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLEv1i16rz, ARM64_INS_FCMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLEv1i32rz, ARM64_INS_FCMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMLEv1i64rz, ARM64_INS_FCMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMLEv2i32rz, ARM64_INS_FCMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMLEv2i64rz, ARM64_INS_FCMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMLEv4i16rz, ARM64_INS_FCMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLEv4i32rz, ARM64_INS_FCMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMLEv8i16rz, ARM64_INS_FCMLE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLT_PPzZ0_D, ARM64_INS_FCMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLT_PPzZ0_H, ARM64_INS_FCMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLT_PPzZ0_S, ARM64_INS_FCMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLTv1i16rz, ARM64_INS_FCMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLTv1i32rz, ARM64_INS_FCMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMLTv1i64rz, ARM64_INS_FCMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMLTv2i32rz, ARM64_INS_FCMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMLTv2i64rz, ARM64_INS_FCMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMLTv4i16rz, ARM64_INS_FCMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMLTv4i32rz, ARM64_INS_FCMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCMLTv8i16rz, ARM64_INS_FCMLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMNE_PPzZ0_D, ARM64_INS_FCMNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMNE_PPzZ0_H, ARM64_INS_FCMNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMNE_PPzZ0_S, ARM64_INS_FCMNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMNE_PPzZZ_D, ARM64_INS_FCMNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMNE_PPzZZ_H, ARM64_INS_FCMNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMNE_PPzZZ_S, ARM64_INS_FCMNE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMPDri, ARM64_INS_FCMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCMPDrr, ARM64_INS_FCMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCMPEDri, ARM64_INS_FCMPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCMPEDrr, ARM64_INS_FCMPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCMPEHri, ARM64_INS_FCMPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMPEHrr, ARM64_INS_FCMPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMPESri, ARM64_INS_FCMPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCMPESrr, ARM64_INS_FCMPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCMPHri, ARM64_INS_FCMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMPHrr, ARM64_INS_FCMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMPSri, ARM64_INS_FCMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCMPSrr, ARM64_INS_FCMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCMUO_PPzZZ_D, ARM64_INS_FCMUO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMUO_PPzZZ_H, ARM64_INS_FCMUO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCMUO_PPzZZ_S, ARM64_INS_FCMUO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCPY_ZPmI_D, ARM64_INS_FCPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCPY_ZPmI_H, ARM64_INS_FCPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCPY_ZPmI_S, ARM64_INS_FCPY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCSELDrrr, ARM64_INS_FCSEL,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCSELHrrr, ARM64_INS_FCSEL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCSELSrrr, ARM64_INS_FCSEL,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTASUWDr, ARM64_INS_FCVTAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTASUWHr, ARM64_INS_FCVTAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTASUWSr, ARM64_INS_FCVTAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTASUXDr, ARM64_INS_FCVTAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTASUXHr, ARM64_INS_FCVTAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTASUXSr, ARM64_INS_FCVTAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTASv1f16, ARM64_INS_FCVTAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTASv1i32, ARM64_INS_FCVTAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTASv1i64, ARM64_INS_FCVTAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTASv2f32, ARM64_INS_FCVTAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTASv2f64, ARM64_INS_FCVTAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTASv4f16, ARM64_INS_FCVTAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTASv4f32, ARM64_INS_FCVTAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTASv8f16, ARM64_INS_FCVTAS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTAUUWDr, ARM64_INS_FCVTAU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTAUUWHr, ARM64_INS_FCVTAU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTAUUWSr, ARM64_INS_FCVTAU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTAUUXDr, ARM64_INS_FCVTAU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTAUUXHr, ARM64_INS_FCVTAU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTAUUXSr, ARM64_INS_FCVTAU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTAUv1f16, ARM64_INS_FCVTAU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTAUv1i32, ARM64_INS_FCVTAU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTAUv1i64, ARM64_INS_FCVTAU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTAUv2f32, ARM64_INS_FCVTAU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTAUv2f64, ARM64_INS_FCVTAU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTAUv4f16, ARM64_INS_FCVTAU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTAUv4f32, ARM64_INS_FCVTAU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTAUv8f16, ARM64_INS_FCVTAU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTDHr, ARM64_INS_FCVT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTDSr, ARM64_INS_FCVT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTHDr, ARM64_INS_FCVT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTHSr, ARM64_INS_FCVT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTLT_ZPmZ_HtoS, ARM64_INS_FCVTLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTLT_ZPmZ_StoD, ARM64_INS_FCVTLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTLv2i32, ARM64_INS_FCVTL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTLv4i16, ARM64_INS_FCVTL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTLv4i32, ARM64_INS_FCVTL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTLv8i16, ARM64_INS_FCVTL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMSUWDr, ARM64_INS_FCVTMS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMSUWHr, ARM64_INS_FCVTMS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMSUWSr, ARM64_INS_FCVTMS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMSUXDr, ARM64_INS_FCVTMS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMSUXHr, ARM64_INS_FCVTMS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMSUXSr, ARM64_INS_FCVTMS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMSv1f16, ARM64_INS_FCVTMS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMSv1i32, ARM64_INS_FCVTMS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMSv1i64, ARM64_INS_FCVTMS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMSv2f32, ARM64_INS_FCVTMS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMSv2f64, ARM64_INS_FCVTMS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMSv4f16, ARM64_INS_FCVTMS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMSv4f32, ARM64_INS_FCVTMS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMSv8f16, ARM64_INS_FCVTMS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMUUWDr, ARM64_INS_FCVTMU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMUUWHr, ARM64_INS_FCVTMU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMUUWSr, ARM64_INS_FCVTMU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMUUXDr, ARM64_INS_FCVTMU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMUUXHr, ARM64_INS_FCVTMU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMUUXSr, ARM64_INS_FCVTMU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMUv1f16, ARM64_INS_FCVTMU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMUv1i32, ARM64_INS_FCVTMU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMUv1i64, ARM64_INS_FCVTMU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMUv2f32, ARM64_INS_FCVTMU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMUv2f64, ARM64_INS_FCVTMU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMUv4f16, ARM64_INS_FCVTMU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMUv4f32, ARM64_INS_FCVTMU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTMUv8f16, ARM64_INS_FCVTMU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNSUWDr, ARM64_INS_FCVTNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNSUWHr, ARM64_INS_FCVTNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNSUWSr, ARM64_INS_FCVTNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNSUXDr, ARM64_INS_FCVTNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNSUXHr, ARM64_INS_FCVTNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNSUXSr, ARM64_INS_FCVTNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNSv1f16, ARM64_INS_FCVTNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNSv1i32, ARM64_INS_FCVTNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNSv1i64, ARM64_INS_FCVTNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNSv2f32, ARM64_INS_FCVTNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNSv2f64, ARM64_INS_FCVTNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNSv4f16, ARM64_INS_FCVTNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNSv4f32, ARM64_INS_FCVTNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNSv8f16, ARM64_INS_FCVTNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNT_ZPmZ_DtoS, ARM64_INS_FCVTNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNT_ZPmZ_StoH, ARM64_INS_FCVTNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNUUWDr, ARM64_INS_FCVTNU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNUUWHr, ARM64_INS_FCVTNU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNUUWSr, ARM64_INS_FCVTNU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNUUXDr, ARM64_INS_FCVTNU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNUUXHr, ARM64_INS_FCVTNU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNUUXSr, ARM64_INS_FCVTNU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNUv1f16, ARM64_INS_FCVTNU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNUv1i32, ARM64_INS_FCVTNU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNUv1i64, ARM64_INS_FCVTNU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNUv2f32, ARM64_INS_FCVTNU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNUv2f64, ARM64_INS_FCVTNU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNUv4f16, ARM64_INS_FCVTNU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNUv4f32, ARM64_INS_FCVTNU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNUv8f16, ARM64_INS_FCVTNU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNv2i32, ARM64_INS_FCVTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNv4i16, ARM64_INS_FCVTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNv4i32, ARM64_INS_FCVTN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTNv8i16, ARM64_INS_FCVTN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPSUWDr, ARM64_INS_FCVTPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPSUWHr, ARM64_INS_FCVTPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPSUWSr, ARM64_INS_FCVTPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPSUXDr, ARM64_INS_FCVTPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPSUXHr, ARM64_INS_FCVTPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPSUXSr, ARM64_INS_FCVTPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPSv1f16, ARM64_INS_FCVTPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPSv1i32, ARM64_INS_FCVTPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPSv1i64, ARM64_INS_FCVTPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPSv2f32, ARM64_INS_FCVTPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPSv2f64, ARM64_INS_FCVTPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPSv4f16, ARM64_INS_FCVTPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPSv4f32, ARM64_INS_FCVTPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPSv8f16, ARM64_INS_FCVTPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPUUWDr, ARM64_INS_FCVTPU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPUUWHr, ARM64_INS_FCVTPU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPUUWSr, ARM64_INS_FCVTPU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPUUXDr, ARM64_INS_FCVTPU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPUUXHr, ARM64_INS_FCVTPU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPUUXSr, ARM64_INS_FCVTPU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPUv1f16, ARM64_INS_FCVTPU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPUv1i32, ARM64_INS_FCVTPU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPUv1i64, ARM64_INS_FCVTPU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPUv2f32, ARM64_INS_FCVTPU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPUv2f64, ARM64_INS_FCVTPU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPUv4f16, ARM64_INS_FCVTPU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPUv4f32, ARM64_INS_FCVTPU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTPUv8f16, ARM64_INS_FCVTPU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTSDr, ARM64_INS_FCVT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTSHr, ARM64_INS_FCVT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTXNT_ZPmZ_DtoS, ARM64_INS_FCVTXNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTXNv1i64, ARM64_INS_FCVTXN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTXNv2f32, ARM64_INS_FCVTXN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTXNv4f32, ARM64_INS_FCVTXN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTX_ZPmZ_DtoS, ARM64_INS_FCVTX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSSWDri, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSSWHri, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSSWSri, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSSXDri, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSSXHri, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSSXSri, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSUWDr, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSUWHr, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSUWSr, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSUXDr, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSUXHr, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSUXSr, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZS_ZPmZ_DtoD, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZS_ZPmZ_DtoS, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZS_ZPmZ_HtoD, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZS_ZPmZ_HtoH, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZS_ZPmZ_HtoS, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZS_ZPmZ_StoD, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZS_ZPmZ_StoS, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSd, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSh, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSs, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSv1f16, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSv1i32, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSv1i64, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSv2f32, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSv2f64, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSv2i32_shift, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSv2i64_shift, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSv4f16, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSv4f32, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSv4i16_shift, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSv4i32_shift, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSv8f16, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZSv8i16_shift, ARM64_INS_FCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUSWDri, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUSWHri, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUSWSri, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUSXDri, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUSXHri, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUSXSri, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUUWDr, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUUWHr, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUUWSr, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUUXDr, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUUXHr, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUUXSr, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZU_ZPmZ_DtoD, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZU_ZPmZ_DtoS, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZU_ZPmZ_HtoD, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZU_ZPmZ_HtoH, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZU_ZPmZ_HtoS, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZU_ZPmZ_StoD, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZU_ZPmZ_StoS, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUd, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUh, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUs, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUv1f16, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUv1i32, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUv1i64, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUv2f32, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUv2f64, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUv2i32_shift, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUv2i64_shift, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUv4f16, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUv4f32, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUv4i16_shift, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUv4i32_shift, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUv8f16, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVTZUv8i16_shift, ARM64_INS_FCVTZU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVT_ZPmZ_DtoH, ARM64_INS_FCVT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVT_ZPmZ_DtoS, ARM64_INS_FCVT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVT_ZPmZ_HtoD, ARM64_INS_FCVT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVT_ZPmZ_HtoS, ARM64_INS_FCVT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVT_ZPmZ_StoD, ARM64_INS_FCVT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FCVT_ZPmZ_StoH, ARM64_INS_FCVT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FDIVDrr, ARM64_INS_FDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FDIVHrr, ARM64_INS_FDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FDIVR_ZPmZ_D, ARM64_INS_FDIVR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FDIVR_ZPmZ_H, ARM64_INS_FDIVR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FDIVR_ZPmZ_S, ARM64_INS_FDIVR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FDIVSrr, ARM64_INS_FDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FDIV_ZPmZ_D, ARM64_INS_FDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FDIV_ZPmZ_H, ARM64_INS_FDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FDIV_ZPmZ_S, ARM64_INS_FDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FDIVv2f32, ARM64_INS_FDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FDIVv2f64, ARM64_INS_FDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FDIVv4f16, ARM64_INS_FDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FDIVv4f32, ARM64_INS_FDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FDIVv8f16, ARM64_INS_FDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FDUP_ZI_D, ARM64_INS_FDUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FDUP_ZI_H, ARM64_INS_FDUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FDUP_ZI_S, ARM64_INS_FDUP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FEXPA_ZZ_D, ARM64_INS_FEXPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FEXPA_ZZ_H, ARM64_INS_FEXPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FEXPA_ZZ_S, ARM64_INS_FEXPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FJCVTZS, ARM64_INS_FJCVTZS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FLOGB_ZPmZ_D, ARM64_INS_FLOGB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FLOGB_ZPmZ_H, ARM64_INS_FLOGB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FLOGB_ZPmZ_S, ARM64_INS_FLOGB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMADDDrrr, ARM64_INS_FMADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMADDHrrr, ARM64_INS_FMADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMADDSrrr, ARM64_INS_FMADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMAD_ZPmZZ_D, ARM64_INS_FMAD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAD_ZPmZZ_H, ARM64_INS_FMAD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAD_ZPmZZ_S, ARM64_INS_FMAD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXDrr, ARM64_INS_FMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXHrr, ARM64_INS_FMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMDrr, ARM64_INS_FMAXNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMHrr, ARM64_INS_FMAXNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMP_ZPmZZ_D, ARM64_INS_FMAXNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMP_ZPmZZ_H, ARM64_INS_FMAXNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMP_ZPmZZ_S, ARM64_INS_FMAXNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMPv2f32, ARM64_INS_FMAXNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMPv2f64, ARM64_INS_FMAXNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMPv2i16p, ARM64_INS_FMAXNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMPv2i32p, ARM64_INS_FMAXNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMPv2i64p, ARM64_INS_FMAXNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMPv4f16, ARM64_INS_FMAXNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMPv4f32, ARM64_INS_FMAXNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMPv8f16, ARM64_INS_FMAXNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMSrr, ARM64_INS_FMAXNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMV_VPZ_D, ARM64_INS_FMAXNMV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMV_VPZ_H, ARM64_INS_FMAXNMV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMV_VPZ_S, ARM64_INS_FMAXNMV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMVv4i16v, ARM64_INS_FMAXNMV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMVv4i32v, ARM64_INS_FMAXNMV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMVv8i16v, ARM64_INS_FMAXNMV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNM_ZPmI_D, ARM64_INS_FMAXNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNM_ZPmI_H, ARM64_INS_FMAXNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNM_ZPmI_S, ARM64_INS_FMAXNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNM_ZPmZ_D, ARM64_INS_FMAXNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNM_ZPmZ_H, ARM64_INS_FMAXNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNM_ZPmZ_S, ARM64_INS_FMAXNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMv2f32, ARM64_INS_FMAXNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMv2f64, ARM64_INS_FMAXNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMv4f16, ARM64_INS_FMAXNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMv4f32, ARM64_INS_FMAXNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXNMv8f16, ARM64_INS_FMAXNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXP_ZPmZZ_D, ARM64_INS_FMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXP_ZPmZZ_H, ARM64_INS_FMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXP_ZPmZZ_S, ARM64_INS_FMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXPv2f32, ARM64_INS_FMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXPv2f64, ARM64_INS_FMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXPv2i16p, ARM64_INS_FMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXPv2i32p, ARM64_INS_FMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXPv2i64p, ARM64_INS_FMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXPv4f16, ARM64_INS_FMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXPv4f32, ARM64_INS_FMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXPv8f16, ARM64_INS_FMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXSrr, ARM64_INS_FMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXV_VPZ_D, ARM64_INS_FMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXV_VPZ_H, ARM64_INS_FMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXV_VPZ_S, ARM64_INS_FMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXVv4i16v, ARM64_INS_FMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXVv4i32v, ARM64_INS_FMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXVv8i16v, ARM64_INS_FMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAX_ZPmI_D, ARM64_INS_FMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAX_ZPmI_H, ARM64_INS_FMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAX_ZPmI_S, ARM64_INS_FMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAX_ZPmZ_D, ARM64_INS_FMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAX_ZPmZ_H, ARM64_INS_FMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAX_ZPmZ_S, ARM64_INS_FMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXv2f32, ARM64_INS_FMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXv2f64, ARM64_INS_FMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXv4f16, ARM64_INS_FMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMAXv4f32, ARM64_INS_FMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMAXv8f16, ARM64_INS_FMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINDrr, ARM64_INS_FMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMINHrr, ARM64_INS_FMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMDrr, ARM64_INS_FMINNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMHrr, ARM64_INS_FMINNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMP_ZPmZZ_D, ARM64_INS_FMINNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMP_ZPmZZ_H, ARM64_INS_FMINNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMP_ZPmZZ_S, ARM64_INS_FMINNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMPv2f32, ARM64_INS_FMINNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMPv2f64, ARM64_INS_FMINNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMPv2i16p, ARM64_INS_FMINNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMPv2i32p, ARM64_INS_FMINNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMPv2i64p, ARM64_INS_FMINNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMPv4f16, ARM64_INS_FMINNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMPv4f32, ARM64_INS_FMINNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMPv8f16, ARM64_INS_FMINNMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMSrr, ARM64_INS_FMINNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMV_VPZ_D, ARM64_INS_FMINNMV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMV_VPZ_H, ARM64_INS_FMINNMV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMV_VPZ_S, ARM64_INS_FMINNMV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMVv4i16v, ARM64_INS_FMINNMV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMVv4i32v, ARM64_INS_FMINNMV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMVv8i16v, ARM64_INS_FMINNMV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNM_ZPmI_D, ARM64_INS_FMINNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNM_ZPmI_H, ARM64_INS_FMINNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNM_ZPmI_S, ARM64_INS_FMINNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNM_ZPmZ_D, ARM64_INS_FMINNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNM_ZPmZ_H, ARM64_INS_FMINNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNM_ZPmZ_S, ARM64_INS_FMINNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMv2f32, ARM64_INS_FMINNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMv2f64, ARM64_INS_FMINNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMv4f16, ARM64_INS_FMINNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMv4f32, ARM64_INS_FMINNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINNMv8f16, ARM64_INS_FMINNM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINP_ZPmZZ_D, ARM64_INS_FMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINP_ZPmZZ_H, ARM64_INS_FMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINP_ZPmZZ_S, ARM64_INS_FMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINPv2f32, ARM64_INS_FMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINPv2f64, ARM64_INS_FMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINPv2i16p, ARM64_INS_FMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINPv2i32p, ARM64_INS_FMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINPv2i64p, ARM64_INS_FMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINPv4f16, ARM64_INS_FMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINPv4f32, ARM64_INS_FMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINPv8f16, ARM64_INS_FMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINSrr, ARM64_INS_FMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMINV_VPZ_D, ARM64_INS_FMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINV_VPZ_H, ARM64_INS_FMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINV_VPZ_S, ARM64_INS_FMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINVv4i16v, ARM64_INS_FMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINVv4i32v, ARM64_INS_FMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINVv8i16v, ARM64_INS_FMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMIN_ZPmI_D, ARM64_INS_FMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMIN_ZPmI_H, ARM64_INS_FMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMIN_ZPmI_S, ARM64_INS_FMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMIN_ZPmZ_D, ARM64_INS_FMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMIN_ZPmZ_H, ARM64_INS_FMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMIN_ZPmZ_S, ARM64_INS_FMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINv2f32, ARM64_INS_FMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINv2f64, ARM64_INS_FMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINv4f16, ARM64_INS_FMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMINv4f32, ARM64_INS_FMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMINv8f16, ARM64_INS_FMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLAL2lanev4f16, ARM64_INS_FMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLAL2lanev8f16, ARM64_INS_FMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLAL2v4f16, ARM64_INS_FMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLAL2v8f16, ARM64_INS_FMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLALB_ZZZI_SHH, ARM64_INS_FMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLALB_ZZZ_SHH, ARM64_INS_FMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLALT_ZZZI_SHH, ARM64_INS_FMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLALT_ZZZ_SHH, ARM64_INS_FMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLALlanev4f16, ARM64_INS_FMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLALlanev8f16, ARM64_INS_FMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLALv4f16, ARM64_INS_FMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLALv8f16, ARM64_INS_FMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLA_ZPmZZ_D, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLA_ZPmZZ_H, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLA_ZPmZZ_S, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLA_ZZZI_D, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLA_ZZZI_H, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLA_ZZZI_S, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLAv1i16_indexed, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLAv1i32_indexed, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLAv1i64_indexed, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLAv2f32, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLAv2f64, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLAv2i32_indexed, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLAv2i64_indexed, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLAv4f16, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLAv4f32, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLAv4i16_indexed, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLAv4i32_indexed, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLAv8f16, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLAv8i16_indexed, ARM64_INS_FMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSL2lanev4f16, ARM64_INS_FMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSL2lanev8f16, ARM64_INS_FMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSL2v4f16, ARM64_INS_FMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSL2v8f16, ARM64_INS_FMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSLB_ZZZI_SHH, ARM64_INS_FMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSLB_ZZZ_SHH, ARM64_INS_FMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSLT_ZZZI_SHH, ARM64_INS_FMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSLT_ZZZ_SHH, ARM64_INS_FMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSLlanev4f16, ARM64_INS_FMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSLlanev8f16, ARM64_INS_FMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSLv4f16, ARM64_INS_FMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSLv8f16, ARM64_INS_FMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLS_ZPmZZ_D, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLS_ZPmZZ_H, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLS_ZPmZZ_S, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLS_ZZZI_D, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLS_ZZZI_H, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLS_ZZZI_S, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSv1i16_indexed, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSv1i32_indexed, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLSv1i64_indexed, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLSv2f32, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLSv2f64, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLSv2i32_indexed, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLSv2i64_indexed, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLSv4f16, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSv4f32, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLSv4i16_indexed, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSv4i32_indexed, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMLSv8f16, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMLSv8i16_indexed, ARM64_INS_FMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMMLA_ZZZ_D, ARM64_INS_FMMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMMLA_ZZZ_S, ARM64_INS_FMMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMOPA_MPPZZ_D, ARM64_INS_FMOPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMOPA_MPPZZ_S, ARM64_INS_FMOPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMOPS_MPPZZ_D, ARM64_INS_FMOPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMOPS_MPPZZ_S, ARM64_INS_FMOPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMOVDXHighr, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMOVDXr, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMOVDi, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMOVDr, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMOVHWr, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMOVHXr, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMOVHi, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMOVHr, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMOVSWr, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMOVSi, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMOVSr, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMOVWHr, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMOVWSr, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMOVXDHighr, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMOVXDr, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMOVXHr, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMOVv2f32_ns, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMOVv2f64_ns, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMOVv4f16_ns, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMOVv4f32_ns, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMOVv8f16_ns, ARM64_INS_FMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMSB_ZPmZZ_D, ARM64_INS_FMSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMSB_ZPmZZ_H, ARM64_INS_FMSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMSB_ZPmZZ_S, ARM64_INS_FMSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMSUBDrrr, ARM64_INS_FMSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMSUBHrrr, ARM64_INS_FMSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMSUBSrrr, ARM64_INS_FMSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMULDrr, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMULHrr, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMULSrr, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FMULX16, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMULX32, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULX64, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULX_ZPmZ_D, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMULX_ZPmZ_H, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMULX_ZPmZ_S, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMULXv1i16_indexed, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMULXv1i32_indexed, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULXv1i64_indexed, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULXv2f32, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULXv2f64, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULXv2i32_indexed, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULXv2i64_indexed, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULXv4f16, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMULXv4f32, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULXv4i16_indexed, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMULXv4i32_indexed, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULXv8f16, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMULXv8i16_indexed, ARM64_INS_FMULX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMUL_ZPmI_D, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMUL_ZPmI_H, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMUL_ZPmI_S, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMUL_ZPmZ_D, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMUL_ZPmZ_H, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMUL_ZPmZ_S, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMUL_ZZZI_D, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMUL_ZZZI_H, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMUL_ZZZI_S, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMUL_ZZZ_D, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMUL_ZZZ_H, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMUL_ZZZ_S, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMULv1i16_indexed, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMULv1i32_indexed, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULv1i64_indexed, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULv2f32, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULv2f64, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULv2i32_indexed, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULv2i64_indexed, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULv4f16, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMULv4f32, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULv4i16_indexed, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMULv4i32_indexed, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FMULv8f16, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FMULv8i16_indexed, ARM64_INS_FMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNEGDr, ARM64_INS_FNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FNEGHr, ARM64_INS_FNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNEGSr, ARM64_INS_FNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FNEG_ZPmZ_D, ARM64_INS_FNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNEG_ZPmZ_H, ARM64_INS_FNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNEG_ZPmZ_S, ARM64_INS_FNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNEGv2f32, ARM64_INS_FNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FNEGv2f64, ARM64_INS_FNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FNEGv4f16, ARM64_INS_FNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNEGv4f32, ARM64_INS_FNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FNEGv8f16, ARM64_INS_FNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMADDDrrr, ARM64_INS_FNMADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FNMADDHrrr, ARM64_INS_FNMADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMADDSrrr, ARM64_INS_FNMADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FNMAD_ZPmZZ_D, ARM64_INS_FNMAD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMAD_ZPmZZ_H, ARM64_INS_FNMAD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMAD_ZPmZZ_S, ARM64_INS_FNMAD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMLA_ZPmZZ_D, ARM64_INS_FNMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMLA_ZPmZZ_H, ARM64_INS_FNMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMLA_ZPmZZ_S, ARM64_INS_FNMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMLS_ZPmZZ_D, ARM64_INS_FNMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMLS_ZPmZZ_H, ARM64_INS_FNMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMLS_ZPmZZ_S, ARM64_INS_FNMLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMSB_ZPmZZ_D, ARM64_INS_FNMSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMSB_ZPmZZ_H, ARM64_INS_FNMSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMSB_ZPmZZ_S, ARM64_INS_FNMSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMSUBDrrr, ARM64_INS_FNMSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FNMSUBHrrr, ARM64_INS_FNMSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMSUBSrrr, ARM64_INS_FNMSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FNMULDrr, ARM64_INS_FNMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FNMULHrr, ARM64_INS_FNMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FNMULSrr, ARM64_INS_FNMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRECPE_ZZ_D, ARM64_INS_FRECPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPE_ZZ_H, ARM64_INS_FRECPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPE_ZZ_S, ARM64_INS_FRECPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPEv1f16, ARM64_INS_FRECPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPEv1i32, ARM64_INS_FRECPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRECPEv1i64, ARM64_INS_FRECPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRECPEv2f32, ARM64_INS_FRECPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRECPEv2f64, ARM64_INS_FRECPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRECPEv4f16, ARM64_INS_FRECPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPEv4f32, ARM64_INS_FRECPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRECPEv8f16, ARM64_INS_FRECPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPS16, ARM64_INS_FRECPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPS32, ARM64_INS_FRECPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRECPS64, ARM64_INS_FRECPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRECPS_ZZZ_D, ARM64_INS_FRECPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPS_ZZZ_H, ARM64_INS_FRECPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPS_ZZZ_S, ARM64_INS_FRECPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPSv2f32, ARM64_INS_FRECPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRECPSv2f64, ARM64_INS_FRECPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRECPSv4f16, ARM64_INS_FRECPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPSv4f32, ARM64_INS_FRECPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRECPSv8f16, ARM64_INS_FRECPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPX_ZPmZ_D, ARM64_INS_FRECPX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPX_ZPmZ_H, ARM64_INS_FRECPX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPX_ZPmZ_S, ARM64_INS_FRECPX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPXv1f16, ARM64_INS_FRECPX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRECPXv1i32, ARM64_INS_FRECPX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRECPXv1i64, ARM64_INS_FRECPX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINT32XDr, ARM64_INS_FRINT32X,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT32XSr, ARM64_INS_FRINT32X,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT32Xv2f32, ARM64_INS_FRINT32X,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT32Xv2f64, ARM64_INS_FRINT32X,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT32Xv4f32, ARM64_INS_FRINT32X,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT32ZDr, ARM64_INS_FRINT32Z,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT32ZSr, ARM64_INS_FRINT32Z,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT32Zv2f32, ARM64_INS_FRINT32Z,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT32Zv2f64, ARM64_INS_FRINT32Z,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT32Zv4f32, ARM64_INS_FRINT32Z,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT64XDr, ARM64_INS_FRINT64X,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT64XSr, ARM64_INS_FRINT64X,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT64Xv2f32, ARM64_INS_FRINT64X,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT64Xv2f64, ARM64_INS_FRINT64X,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT64Xv4f32, ARM64_INS_FRINT64X,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT64ZDr, ARM64_INS_FRINT64Z,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT64ZSr, ARM64_INS_FRINT64Z,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT64Zv2f32, ARM64_INS_FRINT64Z,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT64Zv2f64, ARM64_INS_FRINT64Z,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINT64Zv4f32, ARM64_INS_FRINT64Z,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTADr, ARM64_INS_FRINTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTAHr, ARM64_INS_FRINTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTASr, ARM64_INS_FRINTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTA_ZPmZ_D, ARM64_INS_FRINTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTA_ZPmZ_H, ARM64_INS_FRINTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTA_ZPmZ_S, ARM64_INS_FRINTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTAv2f32, ARM64_INS_FRINTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTAv2f64, ARM64_INS_FRINTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTAv4f16, ARM64_INS_FRINTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTAv4f32, ARM64_INS_FRINTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTAv8f16, ARM64_INS_FRINTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTIDr, ARM64_INS_FRINTI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTIHr, ARM64_INS_FRINTI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTISr, ARM64_INS_FRINTI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTI_ZPmZ_D, ARM64_INS_FRINTI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTI_ZPmZ_H, ARM64_INS_FRINTI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTI_ZPmZ_S, ARM64_INS_FRINTI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTIv2f32, ARM64_INS_FRINTI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTIv2f64, ARM64_INS_FRINTI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTIv4f16, ARM64_INS_FRINTI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTIv4f32, ARM64_INS_FRINTI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTIv8f16, ARM64_INS_FRINTI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTMDr, ARM64_INS_FRINTM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTMHr, ARM64_INS_FRINTM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTMSr, ARM64_INS_FRINTM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTM_ZPmZ_D, ARM64_INS_FRINTM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTM_ZPmZ_H, ARM64_INS_FRINTM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTM_ZPmZ_S, ARM64_INS_FRINTM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTMv2f32, ARM64_INS_FRINTM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTMv2f64, ARM64_INS_FRINTM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTMv4f16, ARM64_INS_FRINTM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTMv4f32, ARM64_INS_FRINTM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTMv8f16, ARM64_INS_FRINTM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTNDr, ARM64_INS_FRINTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTNHr, ARM64_INS_FRINTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTNSr, ARM64_INS_FRINTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTN_ZPmZ_D, ARM64_INS_FRINTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTN_ZPmZ_H, ARM64_INS_FRINTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTN_ZPmZ_S, ARM64_INS_FRINTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTNv2f32, ARM64_INS_FRINTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTNv2f64, ARM64_INS_FRINTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTNv4f16, ARM64_INS_FRINTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTNv4f32, ARM64_INS_FRINTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTNv8f16, ARM64_INS_FRINTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTPDr, ARM64_INS_FRINTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTPHr, ARM64_INS_FRINTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTPSr, ARM64_INS_FRINTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTP_ZPmZ_D, ARM64_INS_FRINTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTP_ZPmZ_H, ARM64_INS_FRINTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTP_ZPmZ_S, ARM64_INS_FRINTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTPv2f32, ARM64_INS_FRINTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTPv2f64, ARM64_INS_FRINTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTPv4f16, ARM64_INS_FRINTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTPv4f32, ARM64_INS_FRINTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTPv8f16, ARM64_INS_FRINTP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTXDr, ARM64_INS_FRINTX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTXHr, ARM64_INS_FRINTX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTXSr, ARM64_INS_FRINTX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTX_ZPmZ_D, ARM64_INS_FRINTX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTX_ZPmZ_H, ARM64_INS_FRINTX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTX_ZPmZ_S, ARM64_INS_FRINTX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTXv2f32, ARM64_INS_FRINTX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTXv2f64, ARM64_INS_FRINTX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTXv4f16, ARM64_INS_FRINTX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTXv4f32, ARM64_INS_FRINTX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTXv8f16, ARM64_INS_FRINTX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTZDr, ARM64_INS_FRINTZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTZHr, ARM64_INS_FRINTZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTZSr, ARM64_INS_FRINTZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTZ_ZPmZ_D, ARM64_INS_FRINTZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTZ_ZPmZ_H, ARM64_INS_FRINTZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTZ_ZPmZ_S, ARM64_INS_FRINTZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTZv2f32, ARM64_INS_FRINTZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTZv2f64, ARM64_INS_FRINTZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTZv4f16, ARM64_INS_FRINTZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRINTZv4f32, ARM64_INS_FRINTZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRINTZv8f16, ARM64_INS_FRINTZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTE_ZZ_D, ARM64_INS_FRSQRTE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTE_ZZ_H, ARM64_INS_FRSQRTE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTE_ZZ_S, ARM64_INS_FRSQRTE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTEv1f16, ARM64_INS_FRSQRTE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTEv1i32, ARM64_INS_FRSQRTE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTEv1i64, ARM64_INS_FRSQRTE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTEv2f32, ARM64_INS_FRSQRTE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTEv2f64, ARM64_INS_FRSQRTE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTEv4f16, ARM64_INS_FRSQRTE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTEv4f32, ARM64_INS_FRSQRTE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTEv8f16, ARM64_INS_FRSQRTE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTS16, ARM64_INS_FRSQRTS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTS32, ARM64_INS_FRSQRTS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTS64, ARM64_INS_FRSQRTS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTS_ZZZ_D, ARM64_INS_FRSQRTS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTS_ZZZ_H, ARM64_INS_FRSQRTS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTS_ZZZ_S, ARM64_INS_FRSQRTS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTSv2f32, ARM64_INS_FRSQRTS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTSv2f64, ARM64_INS_FRSQRTS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTSv4f16, ARM64_INS_FRSQRTS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTSv4f32, ARM64_INS_FRSQRTS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FRSQRTSv8f16, ARM64_INS_FRSQRTS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSCALE_ZPmZ_D, ARM64_INS_FSCALE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSCALE_ZPmZ_H, ARM64_INS_FSCALE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSCALE_ZPmZ_S, ARM64_INS_FSCALE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSQRTDr, ARM64_INS_FSQRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FSQRTHr, ARM64_INS_FSQRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSQRTSr, ARM64_INS_FSQRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FSQRT_ZPmZ_D, ARM64_INS_FSQRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSQRT_ZPmZ_H, ARM64_INS_FSQRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSQRT_ZPmZ_S, ARM64_INS_FSQRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSQRTv2f32, ARM64_INS_FSQRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FSQRTv2f64, ARM64_INS_FSQRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FSQRTv4f16, ARM64_INS_FSQRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSQRTv4f32, ARM64_INS_FSQRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FSQRTv8f16, ARM64_INS_FSQRT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUBDrr, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FSUBHrr, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUBR_ZPmI_D, ARM64_INS_FSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUBR_ZPmI_H, ARM64_INS_FSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUBR_ZPmI_S, ARM64_INS_FSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUBR_ZPmZ_D, ARM64_INS_FSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUBR_ZPmZ_H, ARM64_INS_FSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUBR_ZPmZ_S, ARM64_INS_FSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUBSrr, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_FSUB_ZPmI_D, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUB_ZPmI_H, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUB_ZPmI_S, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUB_ZPmZ_D, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUB_ZPmZ_H, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUB_ZPmZ_S, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUB_ZZZ_D, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUB_ZZZ_H, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUB_ZZZ_S, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUBv2f32, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FSUBv2f64, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FSUBv4f16, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FSUBv4f32, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_FSUBv8f16, ARM64_INS_FSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FTMAD_ZZI_D, ARM64_INS_FTMAD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FTMAD_ZZI_H, ARM64_INS_FTMAD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FTMAD_ZZI_S, ARM64_INS_FTMAD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FTSMUL_ZZZ_D, ARM64_INS_FTSMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FTSMUL_ZZZ_H, ARM64_INS_FTSMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FTSMUL_ZZZ_S, ARM64_INS_FTSMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FTSSEL_ZZZ_D, ARM64_INS_FTSSEL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FTSSEL_ZZZ_H, ARM64_INS_FTSSEL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_FTSSEL_ZZZ_S, ARM64_INS_FTSSEL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1B_D_IMM_REAL, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1B_D_REAL, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1B_D_SXTW_REAL, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1B_D_UXTW_REAL, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1B_S_IMM_REAL, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1B_S_SXTW_REAL, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1B_S_UXTW_REAL, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1D_IMM_REAL, ARM64_INS_LD1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1D_REAL, ARM64_INS_LD1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1D_SCALED_REAL, ARM64_INS_LD1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1D_SXTW_REAL, ARM64_INS_LD1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1D_SXTW_SCALED_REAL, ARM64_INS_LD1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1D_UXTW_REAL, ARM64_INS_LD1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1D_UXTW_SCALED_REAL, ARM64_INS_LD1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1H_D_IMM_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1H_D_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1H_D_SCALED_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1H_D_SXTW_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1H_D_SXTW_SCALED_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1H_D_UXTW_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1H_D_UXTW_SCALED_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1H_S_IMM_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1H_S_SXTW_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1H_S_SXTW_SCALED_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1H_S_UXTW_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1H_S_UXTW_SCALED_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SB_D_IMM_REAL, ARM64_INS_LD1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SB_D_REAL, ARM64_INS_LD1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SB_D_SXTW_REAL, ARM64_INS_LD1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SB_D_UXTW_REAL, ARM64_INS_LD1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SB_S_IMM_REAL, ARM64_INS_LD1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SB_S_SXTW_REAL, ARM64_INS_LD1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SB_S_UXTW_REAL, ARM64_INS_LD1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SH_D_IMM_REAL, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SH_D_REAL, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SH_D_SCALED_REAL, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SH_D_SXTW_REAL, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SH_D_SXTW_SCALED_REAL, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SH_D_UXTW_REAL, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SH_D_UXTW_SCALED_REAL, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SH_S_IMM_REAL, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SH_S_SXTW_REAL, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SH_S_SXTW_SCALED_REAL, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SH_S_UXTW_REAL, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SH_S_UXTW_SCALED_REAL, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SW_D_IMM_REAL, ARM64_INS_LD1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SW_D_REAL, ARM64_INS_LD1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SW_D_SCALED_REAL, ARM64_INS_LD1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SW_D_SXTW_REAL, ARM64_INS_LD1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SW_D_SXTW_SCALED_REAL, ARM64_INS_LD1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SW_D_UXTW_REAL, ARM64_INS_LD1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1SW_D_UXTW_SCALED_REAL, ARM64_INS_LD1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1W_D_IMM_REAL, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1W_D_REAL, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1W_D_SCALED_REAL, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1W_D_SXTW_REAL, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1W_D_SXTW_SCALED_REAL, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1W_D_UXTW_REAL, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1W_D_UXTW_SCALED_REAL, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1W_IMM_REAL, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1W_SXTW_REAL, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1W_SXTW_SCALED_REAL, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1W_UXTW_REAL, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLD1W_UXTW_SCALED_REAL, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1B_D_IMM_REAL, ARM64_INS_LDFF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1B_D_REAL, ARM64_INS_LDFF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1B_D_SXTW_REAL, ARM64_INS_LDFF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1B_D_UXTW_REAL, ARM64_INS_LDFF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1B_S_IMM_REAL, ARM64_INS_LDFF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1B_S_SXTW_REAL, ARM64_INS_LDFF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1B_S_UXTW_REAL, ARM64_INS_LDFF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1D_IMM_REAL, ARM64_INS_LDFF1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1D_REAL, ARM64_INS_LDFF1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1D_SCALED_REAL, ARM64_INS_LDFF1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1D_SXTW_REAL, ARM64_INS_LDFF1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1D_SXTW_SCALED_REAL, ARM64_INS_LDFF1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1D_UXTW_REAL, ARM64_INS_LDFF1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1D_UXTW_SCALED_REAL, ARM64_INS_LDFF1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1H_D_IMM_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1H_D_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1H_D_SCALED_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1H_D_SXTW_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1H_D_SXTW_SCALED_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1H_D_UXTW_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1H_D_UXTW_SCALED_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1H_S_IMM_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1H_S_SXTW_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1H_S_SXTW_SCALED_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1H_S_UXTW_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1H_S_UXTW_SCALED_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SB_D_IMM_REAL, ARM64_INS_LDFF1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SB_D_REAL, ARM64_INS_LDFF1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SB_D_SXTW_REAL, ARM64_INS_LDFF1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SB_D_UXTW_REAL, ARM64_INS_LDFF1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SB_S_IMM_REAL, ARM64_INS_LDFF1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SB_S_SXTW_REAL, ARM64_INS_LDFF1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SB_S_UXTW_REAL, ARM64_INS_LDFF1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SH_D_IMM_REAL, ARM64_INS_LDFF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SH_D_REAL, ARM64_INS_LDFF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SH_D_SCALED_REAL, ARM64_INS_LDFF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SH_D_SXTW_REAL, ARM64_INS_LDFF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SH_D_SXTW_SCALED_REAL, ARM64_INS_LDFF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SH_D_UXTW_REAL, ARM64_INS_LDFF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SH_D_UXTW_SCALED_REAL, ARM64_INS_LDFF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SH_S_IMM_REAL, ARM64_INS_LDFF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SH_S_SXTW_REAL, ARM64_INS_LDFF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SH_S_SXTW_SCALED_REAL, ARM64_INS_LDFF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SH_S_UXTW_REAL, ARM64_INS_LDFF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SH_S_UXTW_SCALED_REAL, ARM64_INS_LDFF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SW_D_IMM_REAL, ARM64_INS_LDFF1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SW_D_REAL, ARM64_INS_LDFF1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SW_D_SCALED_REAL, ARM64_INS_LDFF1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SW_D_SXTW_REAL, ARM64_INS_LDFF1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SW_D_SXTW_SCALED_REAL, ARM64_INS_LDFF1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SW_D_UXTW_REAL, ARM64_INS_LDFF1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1SW_D_UXTW_SCALED_REAL, ARM64_INS_LDFF1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1W_D_IMM_REAL, ARM64_INS_LDFF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1W_D_REAL, ARM64_INS_LDFF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1W_D_SCALED_REAL, ARM64_INS_LDFF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1W_D_SXTW_REAL, ARM64_INS_LDFF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1W_D_SXTW_SCALED_REAL, ARM64_INS_LDFF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1W_D_UXTW_REAL, ARM64_INS_LDFF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1W_D_UXTW_SCALED_REAL, ARM64_INS_LDFF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1W_IMM_REAL, ARM64_INS_LDFF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1W_SXTW_REAL, ARM64_INS_LDFF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1W_SXTW_SCALED_REAL, ARM64_INS_LDFF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1W_UXTW_REAL, ARM64_INS_LDFF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GLDFF1W_UXTW_SCALED_REAL, ARM64_INS_LDFF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_GMI, ARM64_INS_GMI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_HINT, ARM64_INS_HINT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_HISTCNT_ZPzZZ_D, ARM64_INS_HISTCNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_HISTCNT_ZPzZZ_S, ARM64_INS_HISTCNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_HISTSEG_ZZZ, ARM64_INS_HISTSEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_HLT, ARM64_INS_HLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_HVC, ARM64_INS_HVC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INCB_XPiI, ARM64_INS_INCB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INCD_XPiI, ARM64_INS_INCD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INCD_ZPiI, ARM64_INS_INCD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INCH_XPiI, ARM64_INS_INCH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INCH_ZPiI, ARM64_INS_INCH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INCP_XP_B, ARM64_INS_INCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INCP_XP_D, ARM64_INS_INCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INCP_XP_H, ARM64_INS_INCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INCP_XP_S, ARM64_INS_INCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INCP_ZP_D, ARM64_INS_INCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INCP_ZP_H, ARM64_INS_INCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INCP_ZP_S, ARM64_INS_INCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INCW_XPiI, ARM64_INS_INCW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INCW_ZPiI, ARM64_INS_INCW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_II_B, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_II_D, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_II_H, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_II_S, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_IR_B, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_IR_D, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_IR_H, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_IR_S, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_RI_B, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_RI_D, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_RI_H, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_RI_S, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_RR_B, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_RR_D, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_RR_H, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INDEX_RR_S, ARM64_INS_INDEX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSERT_MXIPZ_H_B, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSERT_MXIPZ_H_D, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSERT_MXIPZ_H_H, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSERT_MXIPZ_H_Q, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSERT_MXIPZ_H_S, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSERT_MXIPZ_V_B, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSERT_MXIPZ_V_D, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSERT_MXIPZ_V_H, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSERT_MXIPZ_V_Q, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSERT_MXIPZ_V_S, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSR_ZR_B, ARM64_INS_INSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSR_ZR_D, ARM64_INS_INSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSR_ZR_H, ARM64_INS_INSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSR_ZR_S, ARM64_INS_INSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSR_ZV_B, ARM64_INS_INSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSR_ZV_D, ARM64_INS_INSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSR_ZV_H, ARM64_INS_INSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSR_ZV_S, ARM64_INS_INSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_INSvi16gpr, ARM64_INS_INS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_INSvi16lane, ARM64_INS_INS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_INSvi32gpr, ARM64_INS_INS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_INSvi32lane, ARM64_INS_INS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_INSvi64gpr, ARM64_INS_INS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_INSvi64lane, ARM64_INS_INS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_INSvi8gpr, ARM64_INS_INS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_INSvi8lane, ARM64_INS_INS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_IRG, ARM64_INS_IRG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ISB, ARM64_INS_ISB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTA_RPZ_B, ARM64_INS_LASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTA_RPZ_D, ARM64_INS_LASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTA_RPZ_H, ARM64_INS_LASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTA_RPZ_S, ARM64_INS_LASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTA_VPZ_B, ARM64_INS_LASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTA_VPZ_D, ARM64_INS_LASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTA_VPZ_H, ARM64_INS_LASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTA_VPZ_S, ARM64_INS_LASTA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTB_RPZ_B, ARM64_INS_LASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTB_RPZ_D, ARM64_INS_LASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTB_RPZ_H, ARM64_INS_LASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTB_RPZ_S, ARM64_INS_LASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTB_VPZ_B, ARM64_INS_LASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTB_VPZ_D, ARM64_INS_LASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTB_VPZ_H, ARM64_INS_LASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LASTB_VPZ_S, ARM64_INS_LASTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1B, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1B_D, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1B_D_IMM_REAL, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1B_H, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1B_H_IMM_REAL, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1B_IMM_REAL, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1B_S, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1B_S_IMM_REAL, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1D, ARM64_INS_LD1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1D_IMM_REAL, ARM64_INS_LD1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv16b, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv16b_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv1d, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv1d_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv2d, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv2d_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv2s, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv2s_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv4h, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv4h_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv4s, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv4s_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv8b, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv8b_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv8h, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Fourv8h_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1H, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1H_D, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1H_D_IMM_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1H_IMM_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1H_S, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1H_S_IMM_REAL, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev16b, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev16b_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev1d, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev1d_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev2d, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev2d_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev2s, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev2s_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev4h, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev4h_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev4s, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev4s_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev8b, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev8b_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev8h, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Onev8h_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1RB_D_IMM, ARM64_INS_LD1RB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RB_H_IMM, ARM64_INS_LD1RB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RB_IMM, ARM64_INS_LD1RB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RB_S_IMM, ARM64_INS_LD1RB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RD_IMM, ARM64_INS_LD1RD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RH_D_IMM, ARM64_INS_LD1RH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RH_IMM, ARM64_INS_LD1RH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RH_S_IMM, ARM64_INS_LD1RH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RO_B, ARM64_INS_LD1ROB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RO_B_IMM, ARM64_INS_LD1ROB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RO_D, ARM64_INS_LD1ROD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RO_D_IMM, ARM64_INS_LD1ROD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RO_H, ARM64_INS_LD1ROH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RO_H_IMM, ARM64_INS_LD1ROH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RO_W, ARM64_INS_LD1ROW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RO_W_IMM, ARM64_INS_LD1ROW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RQ_B, ARM64_INS_LD1RQB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RQ_B_IMM, ARM64_INS_LD1RQB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RQ_D, ARM64_INS_LD1RQD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RQ_D_IMM, ARM64_INS_LD1RQD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RQ_H, ARM64_INS_LD1RQH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RQ_H_IMM, ARM64_INS_LD1RQH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RQ_W, ARM64_INS_LD1RQW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RQ_W_IMM, ARM64_INS_LD1RQW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RSB_D_IMM, ARM64_INS_LD1RSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RSB_H_IMM, ARM64_INS_LD1RSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RSB_S_IMM, ARM64_INS_LD1RSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RSH_D_IMM, ARM64_INS_LD1RSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RSH_S_IMM, ARM64_INS_LD1RSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RSW_IMM, ARM64_INS_LD1RSW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RW_D_IMM, ARM64_INS_LD1RW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1RW_IMM, ARM64_INS_LD1RW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv16b, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv16b_POST, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv1d, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv1d_POST, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv2d, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv2d_POST, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv2s, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv2s_POST, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv4h, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv4h_POST, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv4s, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv4s_POST, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv8b, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv8b_POST, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv8h, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Rv8h_POST, ARM64_INS_LD1R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1SB_D, ARM64_INS_LD1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1SB_D_IMM_REAL, ARM64_INS_LD1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1SB_H, ARM64_INS_LD1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1SB_H_IMM_REAL, ARM64_INS_LD1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1SB_S, ARM64_INS_LD1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1SB_S_IMM_REAL, ARM64_INS_LD1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1SH_D, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1SH_D_IMM_REAL, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1SH_S, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1SH_S_IMM_REAL, ARM64_INS_LD1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1SW_D, ARM64_INS_LD1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1SW_D_IMM_REAL, ARM64_INS_LD1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev16b, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev16b_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev1d, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev1d_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev2d, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev2d_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev2s, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev2s_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev4h, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev4h_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev4s, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev4s_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev8b, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev8b_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev8h, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Threev8h_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov16b, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov16b_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov1d, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov1d_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov2d, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov2d_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov2s, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov2s_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov4h, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov4h_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov4s, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov4s_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov8b, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov8b_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov8h, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1Twov8h_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1W, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1W_D, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1W_D_IMM_REAL, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1W_IMM_REAL, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1_MXIPXX_H_B, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1_MXIPXX_H_D, ARM64_INS_LD1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1_MXIPXX_H_H, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1_MXIPXX_H_Q, ARM64_INS_LD1Q,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1_MXIPXX_H_S, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1_MXIPXX_V_B, ARM64_INS_LD1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1_MXIPXX_V_D, ARM64_INS_LD1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1_MXIPXX_V_H, ARM64_INS_LD1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1_MXIPXX_V_Q, ARM64_INS_LD1Q,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1_MXIPXX_V_S, ARM64_INS_LD1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD1i16, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1i16_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1i32, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1i32_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1i64, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1i64_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1i8, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD1i8_POST, ARM64_INS_LD1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2B, ARM64_INS_LD2B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD2B_IMM, ARM64_INS_LD2B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD2D, ARM64_INS_LD2D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD2D_IMM, ARM64_INS_LD2D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD2H, ARM64_INS_LD2H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD2H_IMM, ARM64_INS_LD2H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv16b, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv16b_POST, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv1d, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv1d_POST, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv2d, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv2d_POST, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv2s, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv2s_POST, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv4h, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv4h_POST, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv4s, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv4s_POST, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv8b, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv8b_POST, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv8h, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Rv8h_POST, ARM64_INS_LD2R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Twov16b, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Twov16b_POST, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Twov2d, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Twov2d_POST, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Twov2s, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Twov2s_POST, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Twov4h, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Twov4h_POST, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Twov4s, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Twov4s_POST, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Twov8b, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Twov8b_POST, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Twov8h, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2Twov8h_POST, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2W, ARM64_INS_LD2W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD2W_IMM, ARM64_INS_LD2W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD2i16, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2i16_POST, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2i32, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2i32_POST, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2i64, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2i64_POST, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2i8, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD2i8_POST, ARM64_INS_LD2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3B, ARM64_INS_LD3B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD3B_IMM, ARM64_INS_LD3B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD3D, ARM64_INS_LD3D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD3D_IMM, ARM64_INS_LD3D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD3H, ARM64_INS_LD3H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD3H_IMM, ARM64_INS_LD3H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv16b, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv16b_POST, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv1d, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv1d_POST, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv2d, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv2d_POST, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv2s, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv2s_POST, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv4h, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv4h_POST, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv4s, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv4s_POST, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv8b, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv8b_POST, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv8h, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Rv8h_POST, ARM64_INS_LD3R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Threev16b, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Threev16b_POST, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Threev2d, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Threev2d_POST, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Threev2s, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Threev2s_POST, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Threev4h, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Threev4h_POST, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Threev4s, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Threev4s_POST, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Threev8b, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Threev8b_POST, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Threev8h, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3Threev8h_POST, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3W, ARM64_INS_LD3W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD3W_IMM, ARM64_INS_LD3W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD3i16, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3i16_POST, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3i32, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3i32_POST, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3i64, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3i64_POST, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3i8, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD3i8_POST, ARM64_INS_LD3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4B, ARM64_INS_LD4B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD4B_IMM, ARM64_INS_LD4B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD4D, ARM64_INS_LD4D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD4D_IMM, ARM64_INS_LD4D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD4Fourv16b, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Fourv16b_POST, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Fourv2d, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Fourv2d_POST, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Fourv2s, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Fourv2s_POST, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Fourv4h, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Fourv4h_POST, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Fourv4s, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Fourv4s_POST, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Fourv8b, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Fourv8b_POST, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Fourv8h, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Fourv8h_POST, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4H, ARM64_INS_LD4H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD4H_IMM, ARM64_INS_LD4H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv16b, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv16b_POST, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv1d, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv1d_POST, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv2d, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv2d_POST, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv2s, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv2s_POST, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv4h, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv4h_POST, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv4s, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv4s_POST, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv8b, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv8b_POST, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv8h, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4Rv8h_POST, ARM64_INS_LD4R,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4W, ARM64_INS_LD4W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD4W_IMM, ARM64_INS_LD4W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LD4i16, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4i16_POST, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4i32, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4i32_POST, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4i64, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4i64_POST, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4i8, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD4i8_POST, ARM64_INS_LD4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_LD64B, ARM64_INS_LD64B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDAB, ARM64_INS_LDADDAB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDAH, ARM64_INS_LDADDAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDALB, ARM64_INS_LDADDALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDALH, ARM64_INS_LDADDALH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDALW, ARM64_INS_LDADDAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDALX, ARM64_INS_LDADDAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDAW, ARM64_INS_LDADDA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDAX, ARM64_INS_LDADDA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDB, ARM64_INS_LDADDB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDH, ARM64_INS_LDADDH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDLB, ARM64_INS_LDADDLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDLH, ARM64_INS_LDADDLH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDLW, ARM64_INS_LDADDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDLX, ARM64_INS_LDADDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDW, ARM64_INS_LDADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDADDX, ARM64_INS_LDADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAPRB, ARM64_INS_LDAPRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAPRH, ARM64_INS_LDAPRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAPRW, ARM64_INS_LDAPR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAPRX, ARM64_INS_LDAPR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAPURBi, ARM64_INS_LDAPURB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAPURHi, ARM64_INS_LDAPURH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAPURSBWi, ARM64_INS_LDAPURSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAPURSBXi, ARM64_INS_LDAPURSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAPURSHWi, ARM64_INS_LDAPURSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAPURSHXi, ARM64_INS_LDAPURSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAPURSWi, ARM64_INS_LDAPURSW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAPURXi, ARM64_INS_LDAPUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAPURi, ARM64_INS_LDAPUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDARB, ARM64_INS_LDARB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDARH, ARM64_INS_LDARH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDARW, ARM64_INS_LDAR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDARX, ARM64_INS_LDAR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAXPW, ARM64_INS_LDAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAXPX, ARM64_INS_LDAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAXRB, ARM64_INS_LDAXRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAXRH, ARM64_INS_LDAXRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAXRW, ARM64_INS_LDAXR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDAXRX, ARM64_INS_LDAXR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRAB, ARM64_INS_LDCLRAB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRAH, ARM64_INS_LDCLRAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRALB, ARM64_INS_LDCLRALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRALH, ARM64_INS_LDCLRALH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRALW, ARM64_INS_LDCLRAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRALX, ARM64_INS_LDCLRAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRAW, ARM64_INS_LDCLRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRAX, ARM64_INS_LDCLRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRB, ARM64_INS_LDCLRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRH, ARM64_INS_LDCLRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRLB, ARM64_INS_LDCLRLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRLH, ARM64_INS_LDCLRLH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRLW, ARM64_INS_LDCLRL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRLX, ARM64_INS_LDCLRL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRW, ARM64_INS_LDCLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDCLRX, ARM64_INS_LDCLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORAB, ARM64_INS_LDEORAB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORAH, ARM64_INS_LDEORAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORALB, ARM64_INS_LDEORALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORALH, ARM64_INS_LDEORALH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORALW, ARM64_INS_LDEORAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORALX, ARM64_INS_LDEORAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORAW, ARM64_INS_LDEORA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORAX, ARM64_INS_LDEORA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORB, ARM64_INS_LDEORB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORH, ARM64_INS_LDEORH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORLB, ARM64_INS_LDEORLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORLH, ARM64_INS_LDEORLH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORLW, ARM64_INS_LDEORL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORLX, ARM64_INS_LDEORL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORW, ARM64_INS_LDEOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDEORX, ARM64_INS_LDEOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1B_D_REAL, ARM64_INS_LDFF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1B_H_REAL, ARM64_INS_LDFF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1B_REAL, ARM64_INS_LDFF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1B_S_REAL, ARM64_INS_LDFF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1D_REAL, ARM64_INS_LDFF1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1H_D_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1H_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1H_S_REAL, ARM64_INS_LDFF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1SB_D_REAL, ARM64_INS_LDFF1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1SB_H_REAL, ARM64_INS_LDFF1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1SB_S_REAL, ARM64_INS_LDFF1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1SH_D_REAL, ARM64_INS_LDFF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1SH_S_REAL, ARM64_INS_LDFF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1SW_D_REAL, ARM64_INS_LDFF1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1W_D_REAL, ARM64_INS_LDFF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDFF1W_REAL, ARM64_INS_LDFF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDG, ARM64_INS_LDG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDGM, ARM64_INS_LDGM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDLARB, ARM64_INS_LDLARB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDLARH, ARM64_INS_LDLARH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDLARW, ARM64_INS_LDLAR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDLARX, ARM64_INS_LDLAR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1B_D_IMM_REAL, ARM64_INS_LDNF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1B_H_IMM_REAL, ARM64_INS_LDNF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1B_IMM_REAL, ARM64_INS_LDNF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1B_S_IMM_REAL, ARM64_INS_LDNF1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1D_IMM_REAL, ARM64_INS_LDNF1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1H_D_IMM_REAL, ARM64_INS_LDNF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1H_IMM_REAL, ARM64_INS_LDNF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1H_S_IMM_REAL, ARM64_INS_LDNF1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1SB_D_IMM_REAL, ARM64_INS_LDNF1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1SB_H_IMM_REAL, ARM64_INS_LDNF1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1SB_S_IMM_REAL, ARM64_INS_LDNF1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1SH_D_IMM_REAL, ARM64_INS_LDNF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1SH_S_IMM_REAL, ARM64_INS_LDNF1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1SW_D_IMM_REAL, ARM64_INS_LDNF1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1W_D_IMM_REAL, ARM64_INS_LDNF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNF1W_IMM_REAL, ARM64_INS_LDNF1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNPDi, ARM64_INS_LDNP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNPQi, ARM64_INS_LDNP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNPSi, ARM64_INS_LDNP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNPWi, ARM64_INS_LDNP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNPXi, ARM64_INS_LDNP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1B_ZRI, ARM64_INS_LDNT1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1B_ZRR, ARM64_INS_LDNT1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1B_ZZR_D_REAL, ARM64_INS_LDNT1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1B_ZZR_S_REAL, ARM64_INS_LDNT1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1D_ZRI, ARM64_INS_LDNT1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1D_ZRR, ARM64_INS_LDNT1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1D_ZZR_D_REAL, ARM64_INS_LDNT1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1H_ZRI, ARM64_INS_LDNT1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1H_ZRR, ARM64_INS_LDNT1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1H_ZZR_D_REAL, ARM64_INS_LDNT1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1H_ZZR_S_REAL, ARM64_INS_LDNT1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1SB_ZZR_D_REAL, ARM64_INS_LDNT1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1SB_ZZR_S_REAL, ARM64_INS_LDNT1SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1SH_ZZR_D_REAL, ARM64_INS_LDNT1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1SH_ZZR_S_REAL, ARM64_INS_LDNT1SH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1SW_ZZR_D_REAL, ARM64_INS_LDNT1SW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1W_ZRI, ARM64_INS_LDNT1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1W_ZRR, ARM64_INS_LDNT1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1W_ZZR_D_REAL, ARM64_INS_LDNT1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDNT1W_ZZR_S_REAL, ARM64_INS_LDNT1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPDi, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPDpost, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPDpre, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPQi, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPQpost, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPQpre, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPSWi, ARM64_INS_LDPSW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPSWpost, ARM64_INS_LDPSW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPSWpre, ARM64_INS_LDPSW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPSi, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPSpost, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPSpre, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPWi, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPWpost, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPWpre, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPXi, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPXpost, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDPXpre, ARM64_INS_LDP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRAAindexed, ARM64_INS_LDRAA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRAAwriteback, ARM64_INS_LDRAA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRABindexed, ARM64_INS_LDRAB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRABwriteback, ARM64_INS_LDRAB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRBBpost, ARM64_INS_LDRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRBBpre, ARM64_INS_LDRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRBBroW, ARM64_INS_LDRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRBBroX, ARM64_INS_LDRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRBBui, ARM64_INS_LDRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRBpost, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRBpre, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRBroW, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRBroX, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRBui, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRDl, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRDpost, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRDpre, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRDroW, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRDroX, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRDui, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRHHpost, ARM64_INS_LDRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRHHpre, ARM64_INS_LDRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRHHroW, ARM64_INS_LDRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRHHroX, ARM64_INS_LDRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRHHui, ARM64_INS_LDRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRHpost, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRHpre, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRHroW, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRHroX, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRHui, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRQl, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRQpost, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRQpre, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRQroW, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRQroX, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRQui, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSBWpost, ARM64_INS_LDRSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSBWpre, ARM64_INS_LDRSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSBWroW, ARM64_INS_LDRSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSBWroX, ARM64_INS_LDRSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSBWui, ARM64_INS_LDRSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSBXpost, ARM64_INS_LDRSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSBXpre, ARM64_INS_LDRSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSBXroW, ARM64_INS_LDRSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSBXroX, ARM64_INS_LDRSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSBXui, ARM64_INS_LDRSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSHWpost, ARM64_INS_LDRSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSHWpre, ARM64_INS_LDRSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSHWroW, ARM64_INS_LDRSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSHWroX, ARM64_INS_LDRSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSHWui, ARM64_INS_LDRSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSHXpost, ARM64_INS_LDRSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSHXpre, ARM64_INS_LDRSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSHXroW, ARM64_INS_LDRSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSHXroX, ARM64_INS_LDRSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSHXui, ARM64_INS_LDRSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSWl, ARM64_INS_LDRSW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSWpost, ARM64_INS_LDRSW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSWpre, ARM64_INS_LDRSW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSWroW, ARM64_INS_LDRSW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSWroX, ARM64_INS_LDRSW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSWui, ARM64_INS_LDRSW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSl, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSpost, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSpre, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSroW, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSroX, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRSui, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRWl, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRWpost, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRWpre, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRWroW, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRWroX, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRWui, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRXl, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRXpost, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRXpre, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRXroW, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRXroX, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDRXui, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDR_PXI, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDR_ZA, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDR_ZXI, ARM64_INS_LDR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETAB, ARM64_INS_LDSETAB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETAH, ARM64_INS_LDSETAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETALB, ARM64_INS_LDSETALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETALH, ARM64_INS_LDSETALH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETALW, ARM64_INS_LDSETAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETALX, ARM64_INS_LDSETAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETAW, ARM64_INS_LDSETA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETAX, ARM64_INS_LDSETA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETB, ARM64_INS_LDSETB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETH, ARM64_INS_LDSETH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETLB, ARM64_INS_LDSETLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETLH, ARM64_INS_LDSETLH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETLW, ARM64_INS_LDSETL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETLX, ARM64_INS_LDSETL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETW, ARM64_INS_LDSET,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSETX, ARM64_INS_LDSET,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXAB, ARM64_INS_LDSMAXAB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXAH, ARM64_INS_LDSMAXAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXALB, ARM64_INS_LDSMAXALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXALH, ARM64_INS_LDSMAXALH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXALW, ARM64_INS_LDSMAXAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXALX, ARM64_INS_LDSMAXAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXAW, ARM64_INS_LDSMAXA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXAX, ARM64_INS_LDSMAXA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXB, ARM64_INS_LDSMAXB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXH, ARM64_INS_LDSMAXH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXLB, ARM64_INS_LDSMAXLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXLH, ARM64_INS_LDSMAXLH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXLW, ARM64_INS_LDSMAXL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXLX, ARM64_INS_LDSMAXL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXW, ARM64_INS_LDSMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMAXX, ARM64_INS_LDSMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINAB, ARM64_INS_LDSMINAB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINAH, ARM64_INS_LDSMINAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINALB, ARM64_INS_LDSMINALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINALH, ARM64_INS_LDSMINALH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINALW, ARM64_INS_LDSMINAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINALX, ARM64_INS_LDSMINAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINAW, ARM64_INS_LDSMINA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINAX, ARM64_INS_LDSMINA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINB, ARM64_INS_LDSMINB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINH, ARM64_INS_LDSMINH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINLB, ARM64_INS_LDSMINLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINLH, ARM64_INS_LDSMINLH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINLW, ARM64_INS_LDSMINL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINLX, ARM64_INS_LDSMINL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINW, ARM64_INS_LDSMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDSMINX, ARM64_INS_LDSMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDTRBi, ARM64_INS_LDTRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDTRHi, ARM64_INS_LDTRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDTRSBWi, ARM64_INS_LDTRSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDTRSBXi, ARM64_INS_LDTRSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDTRSHWi, ARM64_INS_LDTRSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDTRSHXi, ARM64_INS_LDTRSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDTRSWi, ARM64_INS_LDTRSW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDTRWi, ARM64_INS_LDTR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDTRXi, ARM64_INS_LDTR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXAB, ARM64_INS_LDUMAXAB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXAH, ARM64_INS_LDUMAXAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXALB, ARM64_INS_LDUMAXALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXALH, ARM64_INS_LDUMAXALH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXALW, ARM64_INS_LDUMAXAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXALX, ARM64_INS_LDUMAXAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXAW, ARM64_INS_LDUMAXA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXAX, ARM64_INS_LDUMAXA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXB, ARM64_INS_LDUMAXB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXH, ARM64_INS_LDUMAXH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXLB, ARM64_INS_LDUMAXLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXLH, ARM64_INS_LDUMAXLH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXLW, ARM64_INS_LDUMAXL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXLX, ARM64_INS_LDUMAXL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXW, ARM64_INS_LDUMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMAXX, ARM64_INS_LDUMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINAB, ARM64_INS_LDUMINAB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINAH, ARM64_INS_LDUMINAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINALB, ARM64_INS_LDUMINALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINALH, ARM64_INS_LDUMINALH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINALW, ARM64_INS_LDUMINAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINALX, ARM64_INS_LDUMINAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINAW, ARM64_INS_LDUMINA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINAX, ARM64_INS_LDUMINA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINB, ARM64_INS_LDUMINB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINH, ARM64_INS_LDUMINH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINLB, ARM64_INS_LDUMINLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINLH, ARM64_INS_LDUMINLH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINLW, ARM64_INS_LDUMINL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINLX, ARM64_INS_LDUMINL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINW, ARM64_INS_LDUMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDUMINX, ARM64_INS_LDUMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDURBBi, ARM64_INS_LDURB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDURBi, ARM64_INS_LDUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDURDi, ARM64_INS_LDUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDURHHi, ARM64_INS_LDURH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDURHi, ARM64_INS_LDUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDURQi, ARM64_INS_LDUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDURSBWi, ARM64_INS_LDURSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDURSBXi, ARM64_INS_LDURSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDURSHWi, ARM64_INS_LDURSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDURSHXi, ARM64_INS_LDURSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDURSWi, ARM64_INS_LDURSW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDURSi, ARM64_INS_LDUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDURWi, ARM64_INS_LDUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDURXi, ARM64_INS_LDUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDXPW, ARM64_INS_LDXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDXPX, ARM64_INS_LDXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDXRB, ARM64_INS_LDXRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDXRH, ARM64_INS_LDXRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDXRW, ARM64_INS_LDXR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LDXRX, ARM64_INS_LDXR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSLR_ZPmZ_B, ARM64_INS_LSLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSLR_ZPmZ_D, ARM64_INS_LSLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSLR_ZPmZ_H, ARM64_INS_LSLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSLR_ZPmZ_S, ARM64_INS_LSLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSLVWr, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSLVXr, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_WIDE_ZPmZ_B, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_WIDE_ZPmZ_H, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_WIDE_ZPmZ_S, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_WIDE_ZZZ_B, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_WIDE_ZZZ_H, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_WIDE_ZZZ_S, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_ZPmI_B, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_ZPmI_D, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_ZPmI_H, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_ZPmI_S, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_ZPmZ_B, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_ZPmZ_D, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_ZPmZ_H, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_ZPmZ_S, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_ZZI_B, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_ZZI_D, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_ZZI_H, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSL_ZZI_S, ARM64_INS_LSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSRR_ZPmZ_B, ARM64_INS_LSRR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSRR_ZPmZ_D, ARM64_INS_LSRR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSRR_ZPmZ_H, ARM64_INS_LSRR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSRR_ZPmZ_S, ARM64_INS_LSRR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSRVWr, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSRVXr, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_WIDE_ZPmZ_B, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_WIDE_ZPmZ_H, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_WIDE_ZPmZ_S, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_WIDE_ZZZ_B, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_WIDE_ZZZ_H, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_WIDE_ZZZ_S, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_ZPmI_B, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_ZPmI_D, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_ZPmI_H, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_ZPmI_S, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_ZPmZ_B, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_ZPmZ_D, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_ZPmZ_H, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_ZPmZ_S, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_ZZI_B, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_ZZI_D, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_ZZI_H, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_LSR_ZZI_S, ARM64_INS_LSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MADDWrrr, ARM64_INS_MADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MADDXrrr, ARM64_INS_MADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MAD_ZPmZZ_B, ARM64_INS_MAD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MAD_ZPmZZ_D, ARM64_INS_MAD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MAD_ZPmZZ_H, ARM64_INS_MAD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MAD_ZPmZZ_S, ARM64_INS_MAD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MATCH_PPzZZ_B, ARM64_INS_MATCH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MATCH_PPzZZ_H, ARM64_INS_MATCH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLA_ZPmZZ_B, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLA_ZPmZZ_D, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLA_ZPmZZ_H, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLA_ZPmZZ_S, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLA_ZZZI_D, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLA_ZZZI_H, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLA_ZZZI_S, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLAv16i8, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLAv2i32, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLAv2i32_indexed, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLAv4i16, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLAv4i16_indexed, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLAv4i32, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLAv4i32_indexed, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLAv8i16, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLAv8i16_indexed, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLAv8i8, ARM64_INS_MLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLS_ZPmZZ_B, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLS_ZPmZZ_D, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLS_ZPmZZ_H, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLS_ZPmZZ_S, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLS_ZZZI_D, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLS_ZZZI_H, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLS_ZZZI_S, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MLSv16i8, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLSv2i32, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLSv2i32_indexed, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLSv4i16, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLSv4i16_indexed, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLSv4i32, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLSv4i32_indexed, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLSv8i16, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLSv8i16_indexed, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MLSv8i8, ARM64_INS_MLS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MOPSSETGE, ARM64_INS_SETGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOPSSETGEN, ARM64_INS_SETGEN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOPSSETGET, ARM64_INS_SETGET,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOPSSETGETN, ARM64_INS_SETGETN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVID, ARM64_INS_MOVI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MOVIv16b_ns, ARM64_INS_MOVI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MOVIv2d_ns, ARM64_INS_MOVI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MOVIv2i32, ARM64_INS_MOVI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MOVIv2s_msl, ARM64_INS_MOVI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MOVIv4i16, ARM64_INS_MOVI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MOVIv4i32, ARM64_INS_MOVI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MOVIv4s_msl, ARM64_INS_MOVI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MOVIv8b_ns, ARM64_INS_MOVI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MOVIv8i16, ARM64_INS_MOVI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MOVKWi, ARM64_INS_MOVK,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVKXi, ARM64_INS_MOVK,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVNWi, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVNXi, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVPRFX_ZPmZ_B, ARM64_INS_MOVPRFX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVPRFX_ZPmZ_D, ARM64_INS_MOVPRFX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVPRFX_ZPmZ_H, ARM64_INS_MOVPRFX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVPRFX_ZPmZ_S, ARM64_INS_MOVPRFX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVPRFX_ZPzZ_B, ARM64_INS_MOVPRFX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVPRFX_ZPzZ_D, ARM64_INS_MOVPRFX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVPRFX_ZPzZ_H, ARM64_INS_MOVPRFX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVPRFX_ZPzZ_S, ARM64_INS_MOVPRFX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVPRFX_ZZ, ARM64_INS_MOVPRFX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVZWi, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MOVZXi, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MRS, ARM64_INS_MRS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PRIVILEGE, 0 }, 0, 0
#endif
},

{
	AArch64_MSB_ZPmZZ_B, ARM64_INS_MSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MSB_ZPmZZ_D, ARM64_INS_MSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MSB_ZPmZZ_H, ARM64_INS_MSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MSB_ZPmZZ_S, ARM64_INS_MSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MSR, ARM64_INS_MSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PRIVILEGE, 0 }, 0, 0
#endif
},

{
	AArch64_MSRpstateImm1, ARM64_INS_MSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { ARM64_GRP_PRIVILEGE, 0 }, 0, 0
#endif
},

{
	AArch64_MSRpstateImm4, ARM64_INS_MSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { ARM64_GRP_PRIVILEGE, 0 }, 0, 0
#endif
},

{
	AArch64_MSRpstatesvcrImm1, ARM64_INS_MSR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MSUBWrrr, ARM64_INS_MSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MSUBXrrr, ARM64_INS_MSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZI_B, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZI_D, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZI_H, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZI_S, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZPmZ_B, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZPmZ_D, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZPmZ_H, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZPmZ_S, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZZZI_D, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZZZI_H, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZZZI_S, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZZZ_B, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZZZ_D, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZZZ_H, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MUL_ZZZ_S, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_MULv16i8, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MULv2i32, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MULv2i32_indexed, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MULv4i16, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MULv4i16_indexed, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MULv4i32, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MULv4i32_indexed, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MULv8i16, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MULv8i16_indexed, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MULv8i8, ARM64_INS_MUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MVNIv2i32, ARM64_INS_MVNI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MVNIv2s_msl, ARM64_INS_MVNI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MVNIv4i16, ARM64_INS_MVNI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MVNIv4i32, ARM64_INS_MVNI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MVNIv4s_msl, ARM64_INS_MVNI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_MVNIv8i16, ARM64_INS_MVNI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_NANDS_PPzPP, ARM64_INS_NANDS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NAND_PPzPP, ARM64_INS_NAND,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NBSL_ZZZZ, ARM64_INS_NBSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NEG_ZPmZ_B, ARM64_INS_NEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NEG_ZPmZ_D, ARM64_INS_NEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NEG_ZPmZ_H, ARM64_INS_NEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NEG_ZPmZ_S, ARM64_INS_NEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NEGv16i8, ARM64_INS_NEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_NEGv1i64, ARM64_INS_NEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_NEGv2i32, ARM64_INS_NEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_NEGv2i64, ARM64_INS_NEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_NEGv4i16, ARM64_INS_NEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_NEGv4i32, ARM64_INS_NEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_NEGv8i16, ARM64_INS_NEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_NEGv8i8, ARM64_INS_NEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_NMATCH_PPzZZ_B, ARM64_INS_NMATCH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NMATCH_PPzZZ_H, ARM64_INS_NMATCH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NORS_PPzPP, ARM64_INS_NORS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NOR_PPzPP, ARM64_INS_NOR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NOT_ZPmZ_B, ARM64_INS_NOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NOT_ZPmZ_D, ARM64_INS_NOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NOT_ZPmZ_H, ARM64_INS_NOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NOT_ZPmZ_S, ARM64_INS_NOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_NOTv16i8, ARM64_INS_NOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_NOTv8i8, ARM64_INS_NOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ORNS_PPzPP, ARM64_INS_ORNS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORNWrs, ARM64_INS_ORN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORNXrs, ARM64_INS_ORN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORN_PPzPP, ARM64_INS_ORN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORNv16i8, ARM64_INS_ORN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ORNv8i8, ARM64_INS_ORN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ORRS_PPzPP, ARM64_INS_ORRS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORRWri, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORRWrs, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORRXri, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORRXrs, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORR_PPzPP, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORR_ZI, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORR_ZPmZ_B, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORR_ZPmZ_D, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORR_ZPmZ_H, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORR_ZPmZ_S, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORR_ZZZ, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORRv16i8, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ORRv2i32, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ORRv4i16, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ORRv4i32, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ORRv8i16, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ORRv8i8, ARM64_INS_ORR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ORV_VPZ_B, ARM64_INS_ORV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORV_VPZ_D, ARM64_INS_ORV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORV_VPZ_H, ARM64_INS_ORV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ORV_VPZ_S, ARM64_INS_ORV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PACDA, ARM64_INS_PACDA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PACDB, ARM64_INS_PACDB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PACDZA, ARM64_INS_PACDZA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PACDZB, ARM64_INS_PACDZB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PACGA, ARM64_INS_PACGA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PACIA, ARM64_INS_PACIA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PACIA1716, ARM64_INS_PACIA1716,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_X17, 0 }, { ARM64_REG_X16, ARM64_REG_X17, 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PACIASP, ARM64_INS_PACIASP,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_SP, ARM64_REG_LR, 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PACIAZ, ARM64_INS_PACIAZ,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_LR, 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PACIB, ARM64_INS_PACIB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PACIB1716, ARM64_INS_PACIB1716,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_X17, 0 }, { ARM64_REG_X16, ARM64_REG_X17, 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PACIBSP, ARM64_INS_PACIBSP,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_SP, ARM64_REG_LR, 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PACIBZ, ARM64_INS_PACIBZ,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_LR, 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PACIZA, ARM64_INS_PACIZA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PACIZB, ARM64_INS_PACIZB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_PFALSE, ARM64_INS_PFALSE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PFIRST_B, ARM64_INS_PFIRST,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PMULLB_ZZZ_D, ARM64_INS_PMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PMULLB_ZZZ_H, ARM64_INS_PMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PMULLB_ZZZ_Q, ARM64_INS_PMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PMULLT_ZZZ_D, ARM64_INS_PMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PMULLT_ZZZ_H, ARM64_INS_PMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PMULLT_ZZZ_Q, ARM64_INS_PMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PMULLv16i8, ARM64_INS_PMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_PMULLv1i64, ARM64_INS_PMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_PMULLv2i64, ARM64_INS_PMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_PMULLv8i8, ARM64_INS_PMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_PMUL_ZZZ_B, ARM64_INS_PMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PMULv16i8, ARM64_INS_PMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_PMULv8i8, ARM64_INS_PMUL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_PNEXT_B, ARM64_INS_PNEXT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PNEXT_D, ARM64_INS_PNEXT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PNEXT_H, ARM64_INS_PNEXT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PNEXT_S, ARM64_INS_PNEXT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFB_D_PZI, ARM64_INS_PRFB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFB_D_SCALED, ARM64_INS_PRFB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFB_D_SXTW_SCALED, ARM64_INS_PRFB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFB_D_UXTW_SCALED, ARM64_INS_PRFB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFB_PRI, ARM64_INS_PRFB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFB_PRR, ARM64_INS_PRFB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFB_S_PZI, ARM64_INS_PRFB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFB_S_SXTW_SCALED, ARM64_INS_PRFB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFB_S_UXTW_SCALED, ARM64_INS_PRFB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFD_D_PZI, ARM64_INS_PRFD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFD_D_SCALED, ARM64_INS_PRFD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFD_D_SXTW_SCALED, ARM64_INS_PRFD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFD_D_UXTW_SCALED, ARM64_INS_PRFD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFD_PRI, ARM64_INS_PRFD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFD_PRR, ARM64_INS_PRFD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFD_S_PZI, ARM64_INS_PRFD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFD_S_SXTW_SCALED, ARM64_INS_PRFD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFD_S_UXTW_SCALED, ARM64_INS_PRFD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFH_D_PZI, ARM64_INS_PRFH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFH_D_SCALED, ARM64_INS_PRFH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFH_D_SXTW_SCALED, ARM64_INS_PRFH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFH_D_UXTW_SCALED, ARM64_INS_PRFH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFH_PRI, ARM64_INS_PRFH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFH_PRR, ARM64_INS_PRFH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFH_S_PZI, ARM64_INS_PRFH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFH_S_SXTW_SCALED, ARM64_INS_PRFH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFH_S_UXTW_SCALED, ARM64_INS_PRFH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFMl, ARM64_INS_PRFM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFMroW, ARM64_INS_PRFM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFMroX, ARM64_INS_PRFM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFMui, ARM64_INS_PRFM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFS_PRR, ARM64_INS_PRFW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFUMi, ARM64_INS_PRFUM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFW_D_PZI, ARM64_INS_PRFW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFW_D_SCALED, ARM64_INS_PRFW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFW_D_SXTW_SCALED, ARM64_INS_PRFW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFW_D_UXTW_SCALED, ARM64_INS_PRFW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFW_PRI, ARM64_INS_PRFW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFW_S_PZI, ARM64_INS_PRFW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFW_S_SXTW_SCALED, ARM64_INS_PRFW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PRFW_S_UXTW_SCALED, ARM64_INS_PRFW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PSEL_PPPRI_B, ARM64_INS_PSEL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PSEL_PPPRI_D, ARM64_INS_PSEL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PSEL_PPPRI_H, ARM64_INS_PSEL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PSEL_PPPRI_S, ARM64_INS_PSEL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PTEST_PP, ARM64_INS_PTEST,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PTRUES_B, ARM64_INS_PTRUES,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PTRUES_D, ARM64_INS_PTRUES,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PTRUES_H, ARM64_INS_PTRUES,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PTRUES_S, ARM64_INS_PTRUES,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PTRUE_B, ARM64_INS_PTRUE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PTRUE_D, ARM64_INS_PTRUE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PTRUE_H, ARM64_INS_PTRUE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PTRUE_S, ARM64_INS_PTRUE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PUNPKHI_PP, ARM64_INS_PUNPKHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_PUNPKLO_PP, ARM64_INS_PUNPKLO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RADDHNB_ZZZ_B, ARM64_INS_RADDHNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RADDHNB_ZZZ_H, ARM64_INS_RADDHNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RADDHNB_ZZZ_S, ARM64_INS_RADDHNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RADDHNT_ZZZ_B, ARM64_INS_RADDHNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RADDHNT_ZZZ_H, ARM64_INS_RADDHNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RADDHNT_ZZZ_S, ARM64_INS_RADDHNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RADDHNv2i64_v2i32, ARM64_INS_RADDHN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RADDHNv2i64_v4i32, ARM64_INS_RADDHN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RADDHNv4i32_v4i16, ARM64_INS_RADDHN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RADDHNv4i32_v8i16, ARM64_INS_RADDHN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RADDHNv8i16_v16i8, ARM64_INS_RADDHN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RADDHNv8i16_v8i8, ARM64_INS_RADDHN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RAX1, ARM64_INS_RAX1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RAX1_ZZZ_D, ARM64_INS_RAX1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RBITWr, ARM64_INS_RBIT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RBITXr, ARM64_INS_RBIT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RBIT_ZPmZ_B, ARM64_INS_RBIT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RBIT_ZPmZ_D, ARM64_INS_RBIT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RBIT_ZPmZ_H, ARM64_INS_RBIT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RBIT_ZPmZ_S, ARM64_INS_RBIT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RBITv16i8, ARM64_INS_RBIT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RBITv8i8, ARM64_INS_RBIT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RDFFRS_PPz, ARM64_INS_RDFFRS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RDFFR_PPz_REAL, ARM64_INS_RDFFR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RDFFR_P_REAL, ARM64_INS_RDFFR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RDVLI_XI, ARM64_INS_RDVL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RET, ARM64_INS_RET,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_RET, 0 }, 0, 0
#endif
},

{
	AArch64_RETAA, ARM64_INS_RETAA,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_LR, ARM64_REG_SP, 0 }, { 0 }, { ARM64_GRP_PAC, ARM64_GRP_RET, 0 }, 0, 0
#endif
},

{
	AArch64_RETAB, ARM64_INS_RETAB,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_LR, ARM64_REG_SP, 0 }, { 0 }, { ARM64_GRP_PAC, ARM64_GRP_RET, 0 }, 0, 0
#endif
},

{
	AArch64_REV16Wr, ARM64_INS_REV16,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REV16Xr, ARM64_INS_REV16,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REV16v16i8, ARM64_INS_REV16,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_REV16v8i8, ARM64_INS_REV16,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_REV32Xr, ARM64_INS_REV32,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REV32v16i8, ARM64_INS_REV32,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_REV32v4i16, ARM64_INS_REV32,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_REV32v8i16, ARM64_INS_REV32,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_REV32v8i8, ARM64_INS_REV32,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_REV64v16i8, ARM64_INS_REV64,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_REV64v2i32, ARM64_INS_REV64,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_REV64v4i16, ARM64_INS_REV64,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_REV64v4i32, ARM64_INS_REV64,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_REV64v8i16, ARM64_INS_REV64,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_REV64v8i8, ARM64_INS_REV64,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_REVB_ZPmZ_D, ARM64_INS_REVB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REVB_ZPmZ_H, ARM64_INS_REVB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REVB_ZPmZ_S, ARM64_INS_REVB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REVD_ZPmZ, ARM64_INS_REVD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REVH_ZPmZ_D, ARM64_INS_REVH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REVH_ZPmZ_S, ARM64_INS_REVH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REVW_ZPmZ_D, ARM64_INS_REVW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REVWr, ARM64_INS_REV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REVXr, ARM64_INS_REV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REV_PP_B, ARM64_INS_REV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REV_PP_D, ARM64_INS_REV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REV_PP_H, ARM64_INS_REV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REV_PP_S, ARM64_INS_REV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REV_ZZ_B, ARM64_INS_REV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REV_ZZ_D, ARM64_INS_REV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REV_ZZ_H, ARM64_INS_REV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_REV_ZZ_S, ARM64_INS_REV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RMIF, ARM64_INS_RMIF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RORVWr, ARM64_INS_ROR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RORVXr, ARM64_INS_ROR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RSHRNB_ZZI_B, ARM64_INS_RSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RSHRNB_ZZI_H, ARM64_INS_RSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RSHRNB_ZZI_S, ARM64_INS_RSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RSHRNT_ZZI_B, ARM64_INS_RSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RSHRNT_ZZI_H, ARM64_INS_RSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RSHRNT_ZZI_S, ARM64_INS_RSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RSHRNv16i8_shift, ARM64_INS_RSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RSHRNv2i32_shift, ARM64_INS_RSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RSHRNv4i16_shift, ARM64_INS_RSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RSHRNv4i32_shift, ARM64_INS_RSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RSHRNv8i16_shift, ARM64_INS_RSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RSHRNv8i8_shift, ARM64_INS_RSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RSUBHNB_ZZZ_B, ARM64_INS_RSUBHNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RSUBHNB_ZZZ_H, ARM64_INS_RSUBHNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RSUBHNB_ZZZ_S, ARM64_INS_RSUBHNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RSUBHNT_ZZZ_B, ARM64_INS_RSUBHNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RSUBHNT_ZZZ_H, ARM64_INS_RSUBHNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RSUBHNT_ZZZ_S, ARM64_INS_RSUBHNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_RSUBHNv2i64_v2i32, ARM64_INS_RSUBHN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RSUBHNv2i64_v4i32, ARM64_INS_RSUBHN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RSUBHNv4i32_v4i16, ARM64_INS_RSUBHN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RSUBHNv4i32_v8i16, ARM64_INS_RSUBHN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RSUBHNv8i16_v16i8, ARM64_INS_RSUBHN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_RSUBHNv8i16_v8i8, ARM64_INS_RSUBHN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABALB_ZZZ_D, ARM64_INS_SABALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABALB_ZZZ_H, ARM64_INS_SABALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABALB_ZZZ_S, ARM64_INS_SABALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABALT_ZZZ_D, ARM64_INS_SABALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABALT_ZZZ_H, ARM64_INS_SABALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABALT_ZZZ_S, ARM64_INS_SABALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABALv16i8_v8i16, ARM64_INS_SABAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABALv2i32_v2i64, ARM64_INS_SABAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABALv4i16_v4i32, ARM64_INS_SABAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABALv4i32_v2i64, ARM64_INS_SABAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABALv8i16_v4i32, ARM64_INS_SABAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABALv8i8_v8i16, ARM64_INS_SABAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABA_ZZZ_B, ARM64_INS_SABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABA_ZZZ_D, ARM64_INS_SABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABA_ZZZ_H, ARM64_INS_SABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABA_ZZZ_S, ARM64_INS_SABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABAv16i8, ARM64_INS_SABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABAv2i32, ARM64_INS_SABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABAv4i16, ARM64_INS_SABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABAv4i32, ARM64_INS_SABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABAv8i16, ARM64_INS_SABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABAv8i8, ARM64_INS_SABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABDLB_ZZZ_D, ARM64_INS_SABDLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABDLB_ZZZ_H, ARM64_INS_SABDLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABDLB_ZZZ_S, ARM64_INS_SABDLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABDLT_ZZZ_D, ARM64_INS_SABDLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABDLT_ZZZ_H, ARM64_INS_SABDLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABDLT_ZZZ_S, ARM64_INS_SABDLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABDLv16i8_v8i16, ARM64_INS_SABDL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABDLv2i32_v2i64, ARM64_INS_SABDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABDLv4i16_v4i32, ARM64_INS_SABDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABDLv4i32_v2i64, ARM64_INS_SABDL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABDLv8i16_v4i32, ARM64_INS_SABDL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABDLv8i8_v8i16, ARM64_INS_SABDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABD_ZPmZ_B, ARM64_INS_SABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABD_ZPmZ_D, ARM64_INS_SABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABD_ZPmZ_H, ARM64_INS_SABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABD_ZPmZ_S, ARM64_INS_SABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SABDv16i8, ARM64_INS_SABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABDv2i32, ARM64_INS_SABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABDv4i16, ARM64_INS_SABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABDv4i32, ARM64_INS_SABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABDv8i16, ARM64_INS_SABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SABDv8i8, ARM64_INS_SABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADALP_ZPmZ_D, ARM64_INS_SADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADALP_ZPmZ_H, ARM64_INS_SADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADALP_ZPmZ_S, ARM64_INS_SADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADALPv16i8_v8i16, ARM64_INS_SADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADALPv2i32_v1i64, ARM64_INS_SADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADALPv4i16_v2i32, ARM64_INS_SADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADALPv4i32_v2i64, ARM64_INS_SADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADALPv8i16_v4i32, ARM64_INS_SADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADALPv8i8_v4i16, ARM64_INS_SADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLBT_ZZZ_D, ARM64_INS_SADDLBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDLBT_ZZZ_H, ARM64_INS_SADDLBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDLBT_ZZZ_S, ARM64_INS_SADDLBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDLB_ZZZ_D, ARM64_INS_SADDLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDLB_ZZZ_H, ARM64_INS_SADDLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDLB_ZZZ_S, ARM64_INS_SADDLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDLPv16i8_v8i16, ARM64_INS_SADDLP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLPv2i32_v1i64, ARM64_INS_SADDLP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLPv4i16_v2i32, ARM64_INS_SADDLP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLPv4i32_v2i64, ARM64_INS_SADDLP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLPv8i16_v4i32, ARM64_INS_SADDLP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLPv8i8_v4i16, ARM64_INS_SADDLP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLT_ZZZ_D, ARM64_INS_SADDLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDLT_ZZZ_H, ARM64_INS_SADDLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDLT_ZZZ_S, ARM64_INS_SADDLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDLVv16i8v, ARM64_INS_SADDLV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLVv4i16v, ARM64_INS_SADDLV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLVv4i32v, ARM64_INS_SADDLV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLVv8i16v, ARM64_INS_SADDLV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLVv8i8v, ARM64_INS_SADDLV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLv16i8_v8i16, ARM64_INS_SADDL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLv2i32_v2i64, ARM64_INS_SADDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLv4i16_v4i32, ARM64_INS_SADDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLv4i32_v2i64, ARM64_INS_SADDL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLv8i16_v4i32, ARM64_INS_SADDL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDLv8i8_v8i16, ARM64_INS_SADDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDV_VPZ_B, ARM64_INS_SADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDV_VPZ_H, ARM64_INS_SADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDV_VPZ_S, ARM64_INS_SADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDWB_ZZZ_D, ARM64_INS_SADDWB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDWB_ZZZ_H, ARM64_INS_SADDWB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDWB_ZZZ_S, ARM64_INS_SADDWB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDWT_ZZZ_D, ARM64_INS_SADDWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDWT_ZZZ_H, ARM64_INS_SADDWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDWT_ZZZ_S, ARM64_INS_SADDWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SADDWv16i8_v8i16, ARM64_INS_SADDW2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDWv2i32_v2i64, ARM64_INS_SADDW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDWv4i16_v4i32, ARM64_INS_SADDW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDWv4i32_v2i64, ARM64_INS_SADDW2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDWv8i16_v4i32, ARM64_INS_SADDW2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SADDWv8i8_v8i16, ARM64_INS_SADDW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SB, ARM64_INS_SB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SBCLB_ZZZ_D, ARM64_INS_SBCLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SBCLB_ZZZ_S, ARM64_INS_SBCLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SBCLT_ZZZ_D, ARM64_INS_SBCLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SBCLT_ZZZ_S, ARM64_INS_SBCLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SBCSWr, ARM64_INS_SBCS,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SBCSXr, ARM64_INS_SBCS,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SBCWr, ARM64_INS_SBC,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SBCXr, ARM64_INS_SBC,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_NZCV, 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SBFMWri, ARM64_INS_SBFM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SBFMXri, ARM64_INS_SBFM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCLAMP_ZZZ_B, ARM64_INS_SCLAMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCLAMP_ZZZ_D, ARM64_INS_SCLAMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCLAMP_ZZZ_H, ARM64_INS_SCLAMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCLAMP_ZZZ_S, ARM64_INS_SCLAMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFSWDri, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFSWHri, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFSWSri, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFSXDri, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFSXHri, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFSXSri, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFUWDri, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFUWHri, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFUWSri, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFUXDri, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFUXHri, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFUXSri, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTF_ZPmZ_DtoD, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTF_ZPmZ_DtoH, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTF_ZPmZ_DtoS, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTF_ZPmZ_HtoH, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTF_ZPmZ_StoD, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTF_ZPmZ_StoH, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTF_ZPmZ_StoS, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFd, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFh, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFs, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFv1i16, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFv1i32, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFv1i64, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFv2f32, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFv2f64, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFv2i32_shift, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFv2i64_shift, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFv4f16, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFv4f32, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFv4i16_shift, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFv4i32_shift, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFv8f16, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SCVTFv8i16_shift, ARM64_INS_SCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SDIVR_ZPmZ_D, ARM64_INS_SDIVR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SDIVR_ZPmZ_S, ARM64_INS_SDIVR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SDIVWr, ARM64_INS_SDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SDIVXr, ARM64_INS_SDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SDIV_ZPmZ_D, ARM64_INS_SDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SDIV_ZPmZ_S, ARM64_INS_SDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SDOT_ZZZI_D, ARM64_INS_SDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SDOT_ZZZI_S, ARM64_INS_SDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SDOT_ZZZ_D, ARM64_INS_SDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SDOT_ZZZ_S, ARM64_INS_SDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SDOTlanev16i8, ARM64_INS_SDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SDOTlanev8i8, ARM64_INS_SDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SDOTv16i8, ARM64_INS_SDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SDOTv8i8, ARM64_INS_SDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SEL_PPPP, ARM64_INS_SEL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SEL_ZPZZ_B, ARM64_INS_SEL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SEL_ZPZZ_D, ARM64_INS_SEL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SEL_ZPZZ_H, ARM64_INS_SEL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SEL_ZPZZ_S, ARM64_INS_SEL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETE, ARM64_INS_SETE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETEN, ARM64_INS_SETEN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETET, ARM64_INS_SETET,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETETN, ARM64_INS_SETETN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETF16, ARM64_INS_SETF16,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETF8, ARM64_INS_SETF8,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETFFR, ARM64_INS_SETFFR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETGM, ARM64_INS_SETGM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETGMN, ARM64_INS_SETGMN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETGMT, ARM64_INS_SETGMT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETGMTN, ARM64_INS_SETGMTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETGP, ARM64_INS_SETGP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETGPN, ARM64_INS_SETGPN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETGPT, ARM64_INS_SETGPT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETGPTN, ARM64_INS_SETGPTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETM, ARM64_INS_SETM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETMN, ARM64_INS_SETMN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETMT, ARM64_INS_SETMT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETMTN, ARM64_INS_SETMTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETP, ARM64_INS_SETP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETPN, ARM64_INS_SETPN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETPT, ARM64_INS_SETPT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SETPTN, ARM64_INS_SETPTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHA1Crrr, ARM64_INS_SHA1C,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_SHA1Hrr, ARM64_INS_SHA1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_SHA1Mrrr, ARM64_INS_SHA1M,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_SHA1Prrr, ARM64_INS_SHA1P,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_SHA1SU0rrr, ARM64_INS_SHA1SU0,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_SHA1SU1rr, ARM64_INS_SHA1SU1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_SHA256H2rrr, ARM64_INS_SHA256H2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_SHA256Hrrr, ARM64_INS_SHA256H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_SHA256SU0rr, ARM64_INS_SHA256SU0,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_SHA256SU1rrr, ARM64_INS_SHA256SU1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_CRYPTO, 0 }, 0, 0
#endif
},

{
	AArch64_SHA512H, ARM64_INS_SHA512H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHA512H2, ARM64_INS_SHA512H2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHA512SU0, ARM64_INS_SHA512SU0,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHA512SU1, ARM64_INS_SHA512SU1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHADD_ZPmZ_B, ARM64_INS_SHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHADD_ZPmZ_D, ARM64_INS_SHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHADD_ZPmZ_H, ARM64_INS_SHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHADD_ZPmZ_S, ARM64_INS_SHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHADDv16i8, ARM64_INS_SHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHADDv2i32, ARM64_INS_SHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHADDv4i16, ARM64_INS_SHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHADDv4i32, ARM64_INS_SHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHADDv8i16, ARM64_INS_SHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHADDv8i8, ARM64_INS_SHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHLLv16i8, ARM64_INS_SHLL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHLLv2i32, ARM64_INS_SHLL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHLLv4i16, ARM64_INS_SHLL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHLLv4i32, ARM64_INS_SHLL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHLLv8i16, ARM64_INS_SHLL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHLLv8i8, ARM64_INS_SHLL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHLd, ARM64_INS_SHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHLv16i8_shift, ARM64_INS_SHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHLv2i32_shift, ARM64_INS_SHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHLv2i64_shift, ARM64_INS_SHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHLv4i16_shift, ARM64_INS_SHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHLv4i32_shift, ARM64_INS_SHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHLv8i16_shift, ARM64_INS_SHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHLv8i8_shift, ARM64_INS_SHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHRNB_ZZI_B, ARM64_INS_SHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHRNB_ZZI_H, ARM64_INS_SHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHRNB_ZZI_S, ARM64_INS_SHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHRNT_ZZI_B, ARM64_INS_SHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHRNT_ZZI_H, ARM64_INS_SHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHRNT_ZZI_S, ARM64_INS_SHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHRNv16i8_shift, ARM64_INS_SHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHRNv2i32_shift, ARM64_INS_SHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHRNv4i16_shift, ARM64_INS_SHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHRNv4i32_shift, ARM64_INS_SHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHRNv8i16_shift, ARM64_INS_SHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHRNv8i8_shift, ARM64_INS_SHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHSUBR_ZPmZ_B, ARM64_INS_SHSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHSUBR_ZPmZ_D, ARM64_INS_SHSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHSUBR_ZPmZ_H, ARM64_INS_SHSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHSUBR_ZPmZ_S, ARM64_INS_SHSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHSUB_ZPmZ_B, ARM64_INS_SHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHSUB_ZPmZ_D, ARM64_INS_SHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHSUB_ZPmZ_H, ARM64_INS_SHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHSUB_ZPmZ_S, ARM64_INS_SHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SHSUBv16i8, ARM64_INS_SHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHSUBv2i32, ARM64_INS_SHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHSUBv4i16, ARM64_INS_SHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHSUBv4i32, ARM64_INS_SHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHSUBv8i16, ARM64_INS_SHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SHSUBv8i8, ARM64_INS_SHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SLI_ZZI_B, ARM64_INS_SLI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SLI_ZZI_D, ARM64_INS_SLI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SLI_ZZI_H, ARM64_INS_SLI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SLI_ZZI_S, ARM64_INS_SLI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SLId, ARM64_INS_SLI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SLIv16i8_shift, ARM64_INS_SLI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SLIv2i32_shift, ARM64_INS_SLI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SLIv2i64_shift, ARM64_INS_SLI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SLIv4i16_shift, ARM64_INS_SLI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SLIv4i32_shift, ARM64_INS_SLI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SLIv8i16_shift, ARM64_INS_SLI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SLIv8i8_shift, ARM64_INS_SLI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SM3PARTW1, ARM64_INS_SM3PARTW1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SM3PARTW2, ARM64_INS_SM3PARTW2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SM3SS1, ARM64_INS_SM3SS1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SM3TT1A, ARM64_INS_SM3TT1A,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SM3TT1B, ARM64_INS_SM3TT1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SM3TT2A, ARM64_INS_SM3TT2A,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SM3TT2B, ARM64_INS_SM3TT2B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SM4E, ARM64_INS_SM4E,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SM4EKEY_ZZZ_S, ARM64_INS_SM4EKEY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SM4ENCKEY, ARM64_INS_SM4EKEY,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SM4E_ZZZ_S, ARM64_INS_SM4E,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMADDLrrr, ARM64_INS_SMADDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAXP_ZPmZ_B, ARM64_INS_SMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAXP_ZPmZ_D, ARM64_INS_SMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAXP_ZPmZ_H, ARM64_INS_SMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAXP_ZPmZ_S, ARM64_INS_SMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAXPv16i8, ARM64_INS_SMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXPv2i32, ARM64_INS_SMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXPv4i16, ARM64_INS_SMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXPv4i32, ARM64_INS_SMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXPv8i16, ARM64_INS_SMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXPv8i8, ARM64_INS_SMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXV_VPZ_B, ARM64_INS_SMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAXV_VPZ_D, ARM64_INS_SMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAXV_VPZ_H, ARM64_INS_SMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAXV_VPZ_S, ARM64_INS_SMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAXVv16i8v, ARM64_INS_SMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXVv4i16v, ARM64_INS_SMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXVv4i32v, ARM64_INS_SMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXVv8i16v, ARM64_INS_SMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXVv8i8v, ARM64_INS_SMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAX_ZI_B, ARM64_INS_SMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAX_ZI_D, ARM64_INS_SMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAX_ZI_H, ARM64_INS_SMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAX_ZI_S, ARM64_INS_SMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAX_ZPmZ_B, ARM64_INS_SMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAX_ZPmZ_D, ARM64_INS_SMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAX_ZPmZ_H, ARM64_INS_SMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAX_ZPmZ_S, ARM64_INS_SMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMAXv16i8, ARM64_INS_SMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXv2i32, ARM64_INS_SMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXv4i16, ARM64_INS_SMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXv4i32, ARM64_INS_SMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXv8i16, ARM64_INS_SMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMAXv8i8, ARM64_INS_SMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMC, ARM64_INS_SMC,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMINP_ZPmZ_B, ARM64_INS_SMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMINP_ZPmZ_D, ARM64_INS_SMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMINP_ZPmZ_H, ARM64_INS_SMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMINP_ZPmZ_S, ARM64_INS_SMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMINPv16i8, ARM64_INS_SMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINPv2i32, ARM64_INS_SMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINPv4i16, ARM64_INS_SMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINPv4i32, ARM64_INS_SMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINPv8i16, ARM64_INS_SMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINPv8i8, ARM64_INS_SMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINV_VPZ_B, ARM64_INS_SMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMINV_VPZ_D, ARM64_INS_SMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMINV_VPZ_H, ARM64_INS_SMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMINV_VPZ_S, ARM64_INS_SMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMINVv16i8v, ARM64_INS_SMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINVv4i16v, ARM64_INS_SMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINVv4i32v, ARM64_INS_SMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINVv8i16v, ARM64_INS_SMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINVv8i8v, ARM64_INS_SMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMIN_ZI_B, ARM64_INS_SMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMIN_ZI_D, ARM64_INS_SMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMIN_ZI_H, ARM64_INS_SMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMIN_ZI_S, ARM64_INS_SMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMIN_ZPmZ_B, ARM64_INS_SMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMIN_ZPmZ_D, ARM64_INS_SMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMIN_ZPmZ_H, ARM64_INS_SMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMIN_ZPmZ_S, ARM64_INS_SMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMINv16i8, ARM64_INS_SMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINv2i32, ARM64_INS_SMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINv4i16, ARM64_INS_SMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINv4i32, ARM64_INS_SMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINv8i16, ARM64_INS_SMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMINv8i8, ARM64_INS_SMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLALB_ZZZI_D, ARM64_INS_SMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLALB_ZZZI_S, ARM64_INS_SMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLALB_ZZZ_D, ARM64_INS_SMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLALB_ZZZ_H, ARM64_INS_SMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLALB_ZZZ_S, ARM64_INS_SMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLALT_ZZZI_D, ARM64_INS_SMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLALT_ZZZI_S, ARM64_INS_SMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLALT_ZZZ_D, ARM64_INS_SMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLALT_ZZZ_H, ARM64_INS_SMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLALT_ZZZ_S, ARM64_INS_SMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLALv16i8_v8i16, ARM64_INS_SMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLALv2i32_indexed, ARM64_INS_SMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLALv2i32_v2i64, ARM64_INS_SMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLALv4i16_indexed, ARM64_INS_SMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLALv4i16_v4i32, ARM64_INS_SMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLALv4i32_indexed, ARM64_INS_SMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLALv4i32_v2i64, ARM64_INS_SMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLALv8i16_indexed, ARM64_INS_SMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLALv8i16_v4i32, ARM64_INS_SMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLALv8i8_v8i16, ARM64_INS_SMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLB_ZZZI_D, ARM64_INS_SMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLB_ZZZI_S, ARM64_INS_SMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLB_ZZZ_D, ARM64_INS_SMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLB_ZZZ_H, ARM64_INS_SMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLB_ZZZ_S, ARM64_INS_SMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLT_ZZZI_D, ARM64_INS_SMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLT_ZZZI_S, ARM64_INS_SMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLT_ZZZ_D, ARM64_INS_SMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLT_ZZZ_H, ARM64_INS_SMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLT_ZZZ_S, ARM64_INS_SMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLv16i8_v8i16, ARM64_INS_SMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLv2i32_indexed, ARM64_INS_SMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLv2i32_v2i64, ARM64_INS_SMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLv4i16_indexed, ARM64_INS_SMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLv4i16_v4i32, ARM64_INS_SMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLv4i32_indexed, ARM64_INS_SMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLv4i32_v2i64, ARM64_INS_SMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLv8i16_indexed, ARM64_INS_SMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLv8i16_v4i32, ARM64_INS_SMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMLSLv8i8_v8i16, ARM64_INS_SMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMMLA, ARM64_INS_SMMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMMLA_ZZZ, ARM64_INS_SMMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMOPA_MPPZZ_D, ARM64_INS_SMOPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMOPA_MPPZZ_S, ARM64_INS_SMOPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMOPS_MPPZZ_D, ARM64_INS_SMOPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMOPS_MPPZZ_S, ARM64_INS_SMOPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMOVvi16to32, ARM64_INS_SMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMOVvi16to32_idx0, ARM64_INS_SMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMOVvi16to64, ARM64_INS_SMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMOVvi16to64_idx0, ARM64_INS_SMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMOVvi32to64, ARM64_INS_SMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMOVvi32to64_idx0, ARM64_INS_SMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMOVvi8to32, ARM64_INS_SMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMOVvi8to32_idx0, ARM64_INS_SMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMOVvi8to64, ARM64_INS_SMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMOVvi8to64_idx0, ARM64_INS_SMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMSUBLrrr, ARM64_INS_SMSUBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULH_ZPmZ_B, ARM64_INS_SMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULH_ZPmZ_D, ARM64_INS_SMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULH_ZPmZ_H, ARM64_INS_SMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULH_ZPmZ_S, ARM64_INS_SMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULH_ZZZ_B, ARM64_INS_SMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULH_ZZZ_D, ARM64_INS_SMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULH_ZZZ_H, ARM64_INS_SMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULH_ZZZ_S, ARM64_INS_SMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULHrr, ARM64_INS_SMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULLB_ZZZI_D, ARM64_INS_SMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULLB_ZZZI_S, ARM64_INS_SMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULLB_ZZZ_D, ARM64_INS_SMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULLB_ZZZ_H, ARM64_INS_SMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULLB_ZZZ_S, ARM64_INS_SMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULLT_ZZZI_D, ARM64_INS_SMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULLT_ZZZI_S, ARM64_INS_SMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULLT_ZZZ_D, ARM64_INS_SMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULLT_ZZZ_H, ARM64_INS_SMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULLT_ZZZ_S, ARM64_INS_SMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SMULLv16i8_v8i16, ARM64_INS_SMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMULLv2i32_indexed, ARM64_INS_SMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMULLv2i32_v2i64, ARM64_INS_SMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMULLv4i16_indexed, ARM64_INS_SMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMULLv4i16_v4i32, ARM64_INS_SMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMULLv4i32_indexed, ARM64_INS_SMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMULLv4i32_v2i64, ARM64_INS_SMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMULLv8i16_indexed, ARM64_INS_SMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMULLv8i16_v4i32, ARM64_INS_SMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SMULLv8i8_v8i16, ARM64_INS_SMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SPLICE_ZPZZ_B, ARM64_INS_SPLICE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SPLICE_ZPZZ_D, ARM64_INS_SPLICE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SPLICE_ZPZZ_H, ARM64_INS_SPLICE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SPLICE_ZPZZ_S, ARM64_INS_SPLICE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SPLICE_ZPZ_B, ARM64_INS_SPLICE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SPLICE_ZPZ_D, ARM64_INS_SPLICE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SPLICE_ZPZ_H, ARM64_INS_SPLICE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SPLICE_ZPZ_S, ARM64_INS_SPLICE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQABS_ZPmZ_B, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQABS_ZPmZ_D, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQABS_ZPmZ_H, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQABS_ZPmZ_S, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQABSv16i8, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQABSv1i16, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQABSv1i32, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQABSv1i64, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQABSv1i8, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQABSv2i32, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQABSv2i64, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQABSv4i16, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQABSv4i32, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQABSv8i16, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQABSv8i8, ARM64_INS_SQABS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQADD_ZI_B, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQADD_ZI_D, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQADD_ZI_H, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQADD_ZI_S, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQADD_ZPmZ_B, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQADD_ZPmZ_D, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQADD_ZPmZ_H, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQADD_ZPmZ_S, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQADD_ZZZ_B, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQADD_ZZZ_D, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQADD_ZZZ_H, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQADD_ZZZ_S, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQADDv16i8, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQADDv1i16, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQADDv1i32, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQADDv1i64, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQADDv1i8, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQADDv2i32, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQADDv2i64, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQADDv4i16, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQADDv4i32, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQADDv8i16, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQADDv8i8, ARM64_INS_SQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQCADD_ZZI_B, ARM64_INS_SQCADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQCADD_ZZI_D, ARM64_INS_SQCADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQCADD_ZZI_H, ARM64_INS_SQCADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQCADD_ZZI_S, ARM64_INS_SQCADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECB_XPiI, ARM64_INS_SQDECB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECB_XPiWdI, ARM64_INS_SQDECB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECD_XPiI, ARM64_INS_SQDECD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECD_XPiWdI, ARM64_INS_SQDECD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECD_ZPiI, ARM64_INS_SQDECD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECH_XPiI, ARM64_INS_SQDECH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECH_XPiWdI, ARM64_INS_SQDECH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECH_ZPiI, ARM64_INS_SQDECH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECP_XPWd_B, ARM64_INS_SQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECP_XPWd_D, ARM64_INS_SQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECP_XPWd_H, ARM64_INS_SQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECP_XPWd_S, ARM64_INS_SQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECP_XP_B, ARM64_INS_SQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECP_XP_D, ARM64_INS_SQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECP_XP_H, ARM64_INS_SQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECP_XP_S, ARM64_INS_SQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECP_ZP_D, ARM64_INS_SQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECP_ZP_H, ARM64_INS_SQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECP_ZP_S, ARM64_INS_SQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECW_XPiI, ARM64_INS_SQDECW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECW_XPiWdI, ARM64_INS_SQDECW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDECW_ZPiI, ARM64_INS_SQDECW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALBT_ZZZ_D, ARM64_INS_SQDMLALBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALBT_ZZZ_H, ARM64_INS_SQDMLALBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALBT_ZZZ_S, ARM64_INS_SQDMLALBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALB_ZZZI_D, ARM64_INS_SQDMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALB_ZZZI_S, ARM64_INS_SQDMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALB_ZZZ_D, ARM64_INS_SQDMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALB_ZZZ_H, ARM64_INS_SQDMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALB_ZZZ_S, ARM64_INS_SQDMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALT_ZZZI_D, ARM64_INS_SQDMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALT_ZZZI_S, ARM64_INS_SQDMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALT_ZZZ_D, ARM64_INS_SQDMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALT_ZZZ_H, ARM64_INS_SQDMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALT_ZZZ_S, ARM64_INS_SQDMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALi16, ARM64_INS_SQDMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALi32, ARM64_INS_SQDMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALv1i32_indexed, ARM64_INS_SQDMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALv1i64_indexed, ARM64_INS_SQDMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALv2i32_indexed, ARM64_INS_SQDMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALv2i32_v2i64, ARM64_INS_SQDMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALv4i16_indexed, ARM64_INS_SQDMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALv4i16_v4i32, ARM64_INS_SQDMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALv4i32_indexed, ARM64_INS_SQDMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALv4i32_v2i64, ARM64_INS_SQDMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALv8i16_indexed, ARM64_INS_SQDMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLALv8i16_v4i32, ARM64_INS_SQDMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLBT_ZZZ_D, ARM64_INS_SQDMLSLBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLBT_ZZZ_H, ARM64_INS_SQDMLSLBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLBT_ZZZ_S, ARM64_INS_SQDMLSLBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLB_ZZZI_D, ARM64_INS_SQDMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLB_ZZZI_S, ARM64_INS_SQDMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLB_ZZZ_D, ARM64_INS_SQDMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLB_ZZZ_H, ARM64_INS_SQDMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLB_ZZZ_S, ARM64_INS_SQDMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLT_ZZZI_D, ARM64_INS_SQDMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLT_ZZZI_S, ARM64_INS_SQDMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLT_ZZZ_D, ARM64_INS_SQDMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLT_ZZZ_H, ARM64_INS_SQDMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLT_ZZZ_S, ARM64_INS_SQDMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLi16, ARM64_INS_SQDMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLi32, ARM64_INS_SQDMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLv1i32_indexed, ARM64_INS_SQDMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLv1i64_indexed, ARM64_INS_SQDMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLv2i32_indexed, ARM64_INS_SQDMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLv2i32_v2i64, ARM64_INS_SQDMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLv4i16_indexed, ARM64_INS_SQDMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLv4i16_v4i32, ARM64_INS_SQDMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLv4i32_indexed, ARM64_INS_SQDMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLv4i32_v2i64, ARM64_INS_SQDMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLv8i16_indexed, ARM64_INS_SQDMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMLSLv8i16_v4i32, ARM64_INS_SQDMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULH_ZZZI_D, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULH_ZZZI_H, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULH_ZZZI_S, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULH_ZZZ_B, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULH_ZZZ_D, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULH_ZZZ_H, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULH_ZZZ_S, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULHv1i16, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULHv1i16_indexed, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULHv1i32, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULHv1i32_indexed, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULHv2i32, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULHv2i32_indexed, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULHv4i16, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULHv4i16_indexed, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULHv4i32, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULHv4i32_indexed, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULHv8i16, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULHv8i16_indexed, ARM64_INS_SQDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLB_ZZZI_D, ARM64_INS_SQDMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLB_ZZZI_S, ARM64_INS_SQDMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLB_ZZZ_D, ARM64_INS_SQDMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLB_ZZZ_H, ARM64_INS_SQDMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLB_ZZZ_S, ARM64_INS_SQDMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLT_ZZZI_D, ARM64_INS_SQDMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLT_ZZZI_S, ARM64_INS_SQDMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLT_ZZZ_D, ARM64_INS_SQDMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLT_ZZZ_H, ARM64_INS_SQDMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLT_ZZZ_S, ARM64_INS_SQDMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLi16, ARM64_INS_SQDMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLi32, ARM64_INS_SQDMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLv1i32_indexed, ARM64_INS_SQDMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLv1i64_indexed, ARM64_INS_SQDMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLv2i32_indexed, ARM64_INS_SQDMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLv2i32_v2i64, ARM64_INS_SQDMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLv4i16_indexed, ARM64_INS_SQDMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLv4i16_v4i32, ARM64_INS_SQDMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLv4i32_indexed, ARM64_INS_SQDMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLv4i32_v2i64, ARM64_INS_SQDMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLv8i16_indexed, ARM64_INS_SQDMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQDMULLv8i16_v4i32, ARM64_INS_SQDMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQINCB_XPiI, ARM64_INS_SQINCB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCB_XPiWdI, ARM64_INS_SQINCB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCD_XPiI, ARM64_INS_SQINCD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCD_XPiWdI, ARM64_INS_SQINCD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCD_ZPiI, ARM64_INS_SQINCD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCH_XPiI, ARM64_INS_SQINCH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCH_XPiWdI, ARM64_INS_SQINCH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCH_ZPiI, ARM64_INS_SQINCH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCP_XPWd_B, ARM64_INS_SQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCP_XPWd_D, ARM64_INS_SQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCP_XPWd_H, ARM64_INS_SQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCP_XPWd_S, ARM64_INS_SQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCP_XP_B, ARM64_INS_SQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCP_XP_D, ARM64_INS_SQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCP_XP_H, ARM64_INS_SQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCP_XP_S, ARM64_INS_SQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCP_ZP_D, ARM64_INS_SQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCP_ZP_H, ARM64_INS_SQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCP_ZP_S, ARM64_INS_SQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCW_XPiI, ARM64_INS_SQINCW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCW_XPiWdI, ARM64_INS_SQINCW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQINCW_ZPiI, ARM64_INS_SQINCW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQNEG_ZPmZ_B, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQNEG_ZPmZ_D, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQNEG_ZPmZ_H, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQNEG_ZPmZ_S, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQNEGv16i8, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQNEGv1i16, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQNEGv1i32, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQNEGv1i64, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQNEGv1i8, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQNEGv2i32, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQNEGv2i64, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQNEGv4i16, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQNEGv4i32, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQNEGv8i16, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQNEGv8i8, ARM64_INS_SQNEG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRDCMLAH_ZZZI_H, ARM64_INS_SQRDCMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDCMLAH_ZZZI_S, ARM64_INS_SQRDCMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDCMLAH_ZZZ_B, ARM64_INS_SQRDCMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDCMLAH_ZZZ_D, ARM64_INS_SQRDCMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDCMLAH_ZZZ_H, ARM64_INS_SQRDCMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDCMLAH_ZZZ_S, ARM64_INS_SQRDCMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAH_ZZZI_D, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAH_ZZZI_H, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAH_ZZZI_S, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAH_ZZZ_B, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAH_ZZZ_D, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAH_ZZZ_H, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAH_ZZZ_S, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAHi16_indexed, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAHi32_indexed, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAHv1i16, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAHv1i32, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAHv2i32, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAHv2i32_indexed, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAHv4i16, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAHv4i16_indexed, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAHv4i32, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAHv4i32_indexed, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAHv8i16, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLAHv8i16_indexed, ARM64_INS_SQRDMLAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSH_ZZZI_D, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSH_ZZZI_H, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSH_ZZZI_S, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSH_ZZZ_B, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSH_ZZZ_D, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSH_ZZZ_H, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSH_ZZZ_S, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSHi16_indexed, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSHi32_indexed, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSHv1i16, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSHv1i32, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSHv2i32, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSHv2i32_indexed, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSHv4i16, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSHv4i16_indexed, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSHv4i32, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSHv4i32_indexed, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSHv8i16, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMLSHv8i16_indexed, ARM64_INS_SQRDMLSH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULH_ZZZI_D, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULH_ZZZI_H, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULH_ZZZI_S, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULH_ZZZ_B, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULH_ZZZ_D, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULH_ZZZ_H, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULH_ZZZ_S, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULHv1i16, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULHv1i16_indexed, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULHv1i32, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULHv1i32_indexed, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULHv2i32, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULHv2i32_indexed, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULHv4i16, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULHv4i16_indexed, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULHv4i32, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULHv4i32_indexed, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULHv8i16, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRDMULHv8i16_indexed, ARM64_INS_SQRDMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLR_ZPmZ_B, ARM64_INS_SQRSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLR_ZPmZ_D, ARM64_INS_SQRSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLR_ZPmZ_H, ARM64_INS_SQRSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLR_ZPmZ_S, ARM64_INS_SQRSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHL_ZPmZ_B, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHL_ZPmZ_D, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHL_ZPmZ_H, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHL_ZPmZ_S, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLv16i8, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLv1i16, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLv1i32, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLv1i64, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLv1i8, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLv2i32, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLv2i64, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLv4i16, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLv4i32, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLv8i16, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHLv8i8, ARM64_INS_SQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNB_ZZI_B, ARM64_INS_SQRSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNB_ZZI_H, ARM64_INS_SQRSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNB_ZZI_S, ARM64_INS_SQRSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNT_ZZI_B, ARM64_INS_SQRSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNT_ZZI_H, ARM64_INS_SQRSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNT_ZZI_S, ARM64_INS_SQRSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNb, ARM64_INS_SQRSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNh, ARM64_INS_SQRSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNs, ARM64_INS_SQRSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNv16i8_shift, ARM64_INS_SQRSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNv2i32_shift, ARM64_INS_SQRSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNv4i16_shift, ARM64_INS_SQRSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNv4i32_shift, ARM64_INS_SQRSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNv8i16_shift, ARM64_INS_SQRSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRNv8i8_shift, ARM64_INS_SQRSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNB_ZZI_B, ARM64_INS_SQRSHRUNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNB_ZZI_H, ARM64_INS_SQRSHRUNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNB_ZZI_S, ARM64_INS_SQRSHRUNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNT_ZZI_B, ARM64_INS_SQRSHRUNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNT_ZZI_H, ARM64_INS_SQRSHRUNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNT_ZZI_S, ARM64_INS_SQRSHRUNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNb, ARM64_INS_SQRSHRUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNh, ARM64_INS_SQRSHRUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNs, ARM64_INS_SQRSHRUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNv16i8_shift, ARM64_INS_SQRSHRUN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNv2i32_shift, ARM64_INS_SQRSHRUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNv4i16_shift, ARM64_INS_SQRSHRUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNv4i32_shift, ARM64_INS_SQRSHRUN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNv8i16_shift, ARM64_INS_SQRSHRUN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQRSHRUNv8i8_shift, ARM64_INS_SQRSHRUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLR_ZPmZ_B, ARM64_INS_SQSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLR_ZPmZ_D, ARM64_INS_SQSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLR_ZPmZ_H, ARM64_INS_SQSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLR_ZPmZ_S, ARM64_INS_SQSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLU_ZPmI_B, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLU_ZPmI_D, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLU_ZPmI_H, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLU_ZPmI_S, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLUb, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLUd, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLUh, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLUs, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLUv16i8_shift, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLUv2i32_shift, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLUv2i64_shift, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLUv4i16_shift, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLUv4i32_shift, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLUv8i16_shift, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLUv8i8_shift, ARM64_INS_SQSHLU,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHL_ZPmI_B, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHL_ZPmI_D, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHL_ZPmI_H, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHL_ZPmI_S, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHL_ZPmZ_B, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHL_ZPmZ_D, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHL_ZPmZ_H, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHL_ZPmZ_S, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLb, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLd, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLh, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLs, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv16i8, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv16i8_shift, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv1i16, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv1i32, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv1i64, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv1i8, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv2i32, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv2i32_shift, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv2i64, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv2i64_shift, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv4i16, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv4i16_shift, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv4i32, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv4i32_shift, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv8i16, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv8i16_shift, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv8i8, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHLv8i8_shift, ARM64_INS_SQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNB_ZZI_B, ARM64_INS_SQSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNB_ZZI_H, ARM64_INS_SQSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNB_ZZI_S, ARM64_INS_SQSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNT_ZZI_B, ARM64_INS_SQSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNT_ZZI_H, ARM64_INS_SQSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNT_ZZI_S, ARM64_INS_SQSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNb, ARM64_INS_SQSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNh, ARM64_INS_SQSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNs, ARM64_INS_SQSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNv16i8_shift, ARM64_INS_SQSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNv2i32_shift, ARM64_INS_SQSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNv4i16_shift, ARM64_INS_SQSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNv4i32_shift, ARM64_INS_SQSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNv8i16_shift, ARM64_INS_SQSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRNv8i8_shift, ARM64_INS_SQSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNB_ZZI_B, ARM64_INS_SQSHRUNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNB_ZZI_H, ARM64_INS_SQSHRUNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNB_ZZI_S, ARM64_INS_SQSHRUNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNT_ZZI_B, ARM64_INS_SQSHRUNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNT_ZZI_H, ARM64_INS_SQSHRUNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNT_ZZI_S, ARM64_INS_SQSHRUNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNb, ARM64_INS_SQSHRUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNh, ARM64_INS_SQSHRUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNs, ARM64_INS_SQSHRUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNv16i8_shift, ARM64_INS_SQSHRUN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNv2i32_shift, ARM64_INS_SQSHRUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNv4i16_shift, ARM64_INS_SQSHRUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNv4i32_shift, ARM64_INS_SQSHRUN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNv8i16_shift, ARM64_INS_SQSHRUN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSHRUNv8i8_shift, ARM64_INS_SQSHRUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBR_ZPmZ_B, ARM64_INS_SQSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBR_ZPmZ_D, ARM64_INS_SQSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBR_ZPmZ_H, ARM64_INS_SQSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBR_ZPmZ_S, ARM64_INS_SQSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUB_ZI_B, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUB_ZI_D, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUB_ZI_H, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUB_ZI_S, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUB_ZPmZ_B, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUB_ZPmZ_D, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUB_ZPmZ_H, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUB_ZPmZ_S, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUB_ZZZ_B, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUB_ZZZ_D, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUB_ZZZ_H, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUB_ZZZ_S, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBv16i8, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBv1i16, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBv1i32, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBv1i64, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBv1i8, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBv2i32, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBv2i64, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBv4i16, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBv4i32, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBv8i16, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQSUBv8i8, ARM64_INS_SQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNB_ZZ_B, ARM64_INS_SQXTNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNB_ZZ_H, ARM64_INS_SQXTNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNB_ZZ_S, ARM64_INS_SQXTNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNT_ZZ_B, ARM64_INS_SQXTNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNT_ZZ_H, ARM64_INS_SQXTNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNT_ZZ_S, ARM64_INS_SQXTNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNv16i8, ARM64_INS_SQXTN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNv1i16, ARM64_INS_SQXTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNv1i32, ARM64_INS_SQXTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNv1i8, ARM64_INS_SQXTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNv2i32, ARM64_INS_SQXTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNv4i16, ARM64_INS_SQXTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNv4i32, ARM64_INS_SQXTN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNv8i16, ARM64_INS_SQXTN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTNv8i8, ARM64_INS_SQXTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNB_ZZ_B, ARM64_INS_SQXTUNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNB_ZZ_H, ARM64_INS_SQXTUNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNB_ZZ_S, ARM64_INS_SQXTUNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNT_ZZ_B, ARM64_INS_SQXTUNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNT_ZZ_H, ARM64_INS_SQXTUNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNT_ZZ_S, ARM64_INS_SQXTUNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNv16i8, ARM64_INS_SQXTUN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNv1i16, ARM64_INS_SQXTUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNv1i32, ARM64_INS_SQXTUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNv1i8, ARM64_INS_SQXTUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNv2i32, ARM64_INS_SQXTUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNv4i16, ARM64_INS_SQXTUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNv4i32, ARM64_INS_SQXTUN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNv8i16, ARM64_INS_SQXTUN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SQXTUNv8i8, ARM64_INS_SQXTUN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRHADD_ZPmZ_B, ARM64_INS_SRHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRHADD_ZPmZ_D, ARM64_INS_SRHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRHADD_ZPmZ_H, ARM64_INS_SRHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRHADD_ZPmZ_S, ARM64_INS_SRHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRHADDv16i8, ARM64_INS_SRHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRHADDv2i32, ARM64_INS_SRHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRHADDv4i16, ARM64_INS_SRHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRHADDv4i32, ARM64_INS_SRHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRHADDv8i16, ARM64_INS_SRHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRHADDv8i8, ARM64_INS_SRHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRI_ZZI_B, ARM64_INS_SRI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRI_ZZI_D, ARM64_INS_SRI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRI_ZZI_H, ARM64_INS_SRI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRI_ZZI_S, ARM64_INS_SRI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRId, ARM64_INS_SRI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRIv16i8_shift, ARM64_INS_SRI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRIv2i32_shift, ARM64_INS_SRI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRIv2i64_shift, ARM64_INS_SRI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRIv4i16_shift, ARM64_INS_SRI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRIv4i32_shift, ARM64_INS_SRI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRIv8i16_shift, ARM64_INS_SRI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRIv8i8_shift, ARM64_INS_SRI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHLR_ZPmZ_B, ARM64_INS_SRSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSHLR_ZPmZ_D, ARM64_INS_SRSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSHLR_ZPmZ_H, ARM64_INS_SRSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSHLR_ZPmZ_S, ARM64_INS_SRSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSHL_ZPmZ_B, ARM64_INS_SRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSHL_ZPmZ_D, ARM64_INS_SRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSHL_ZPmZ_H, ARM64_INS_SRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSHL_ZPmZ_S, ARM64_INS_SRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSHLv16i8, ARM64_INS_SRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHLv1i64, ARM64_INS_SRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHLv2i32, ARM64_INS_SRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHLv2i64, ARM64_INS_SRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHLv4i16, ARM64_INS_SRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHLv4i32, ARM64_INS_SRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHLv8i16, ARM64_INS_SRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHLv8i8, ARM64_INS_SRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHR_ZPmI_B, ARM64_INS_SRSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSHR_ZPmI_D, ARM64_INS_SRSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSHR_ZPmI_H, ARM64_INS_SRSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSHR_ZPmI_S, ARM64_INS_SRSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSHRd, ARM64_INS_SRSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHRv16i8_shift, ARM64_INS_SRSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHRv2i32_shift, ARM64_INS_SRSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHRv2i64_shift, ARM64_INS_SRSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHRv4i16_shift, ARM64_INS_SRSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHRv4i32_shift, ARM64_INS_SRSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHRv8i16_shift, ARM64_INS_SRSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSHRv8i8_shift, ARM64_INS_SRSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSRA_ZZI_B, ARM64_INS_SRSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSRA_ZZI_D, ARM64_INS_SRSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSRA_ZZI_H, ARM64_INS_SRSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSRA_ZZI_S, ARM64_INS_SRSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SRSRAd, ARM64_INS_SRSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSRAv16i8_shift, ARM64_INS_SRSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSRAv2i32_shift, ARM64_INS_SRSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSRAv2i64_shift, ARM64_INS_SRSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSRAv4i16_shift, ARM64_INS_SRSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSRAv4i32_shift, ARM64_INS_SRSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSRAv8i16_shift, ARM64_INS_SRSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SRSRAv8i8_shift, ARM64_INS_SRSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHLLB_ZZI_D, ARM64_INS_SSHLLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSHLLB_ZZI_H, ARM64_INS_SSHLLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSHLLB_ZZI_S, ARM64_INS_SSHLLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSHLLT_ZZI_D, ARM64_INS_SSHLLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSHLLT_ZZI_H, ARM64_INS_SSHLLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSHLLT_ZZI_S, ARM64_INS_SSHLLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSHLLv16i8_shift, ARM64_INS_SSHLL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHLLv2i32_shift, ARM64_INS_SSHLL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHLLv4i16_shift, ARM64_INS_SSHLL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHLLv4i32_shift, ARM64_INS_SSHLL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHLLv8i16_shift, ARM64_INS_SSHLL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHLLv8i8_shift, ARM64_INS_SSHLL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHLv16i8, ARM64_INS_SSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHLv1i64, ARM64_INS_SSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHLv2i32, ARM64_INS_SSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHLv2i64, ARM64_INS_SSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHLv4i16, ARM64_INS_SSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHLv4i32, ARM64_INS_SSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHLv8i16, ARM64_INS_SSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHLv8i8, ARM64_INS_SSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHRd, ARM64_INS_SSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHRv16i8_shift, ARM64_INS_SSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHRv2i32_shift, ARM64_INS_SSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHRv2i64_shift, ARM64_INS_SSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHRv4i16_shift, ARM64_INS_SSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHRv4i32_shift, ARM64_INS_SSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHRv8i16_shift, ARM64_INS_SSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSHRv8i8_shift, ARM64_INS_SSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSRA_ZZI_B, ARM64_INS_SSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSRA_ZZI_D, ARM64_INS_SSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSRA_ZZI_H, ARM64_INS_SSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSRA_ZZI_S, ARM64_INS_SSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSRAd, ARM64_INS_SSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSRAv16i8_shift, ARM64_INS_SSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSRAv2i32_shift, ARM64_INS_SSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSRAv2i64_shift, ARM64_INS_SSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSRAv4i16_shift, ARM64_INS_SSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSRAv4i32_shift, ARM64_INS_SSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSRAv8i16_shift, ARM64_INS_SSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSRAv8i8_shift, ARM64_INS_SSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SST1B_D_IMM, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1B_D_REAL, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1B_D_SXTW, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1B_D_UXTW, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1B_S_IMM, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1B_S_SXTW, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1B_S_UXTW, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1D_IMM, ARM64_INS_ST1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1D_REAL, ARM64_INS_ST1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1D_SCALED_SCALED_REAL, ARM64_INS_ST1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1D_SXTW, ARM64_INS_ST1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1D_SXTW_SCALED, ARM64_INS_ST1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1D_UXTW, ARM64_INS_ST1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1D_UXTW_SCALED, ARM64_INS_ST1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1H_D_IMM, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1H_D_REAL, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1H_D_SCALED_SCALED_REAL, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1H_D_SXTW, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1H_D_SXTW_SCALED, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1H_D_UXTW, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1H_D_UXTW_SCALED, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1H_S_IMM, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1H_S_SXTW, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1H_S_SXTW_SCALED, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1H_S_UXTW, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1H_S_UXTW_SCALED, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1W_D_IMM, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1W_D_REAL, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1W_D_SCALED_SCALED_REAL, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1W_D_SXTW, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1W_D_SXTW_SCALED, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1W_D_UXTW, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1W_D_UXTW_SCALED, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1W_IMM, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1W_SXTW, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1W_SXTW_SCALED, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1W_UXTW, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SST1W_UXTW_SCALED, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLBT_ZZZ_D, ARM64_INS_SSUBLBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLBT_ZZZ_H, ARM64_INS_SSUBLBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLBT_ZZZ_S, ARM64_INS_SSUBLBT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLB_ZZZ_D, ARM64_INS_SSUBLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLB_ZZZ_H, ARM64_INS_SSUBLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLB_ZZZ_S, ARM64_INS_SSUBLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLTB_ZZZ_D, ARM64_INS_SSUBLTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLTB_ZZZ_H, ARM64_INS_SSUBLTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLTB_ZZZ_S, ARM64_INS_SSUBLTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLT_ZZZ_D, ARM64_INS_SSUBLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLT_ZZZ_H, ARM64_INS_SSUBLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLT_ZZZ_S, ARM64_INS_SSUBLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLv16i8_v8i16, ARM64_INS_SSUBL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLv2i32_v2i64, ARM64_INS_SSUBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLv4i16_v4i32, ARM64_INS_SSUBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLv4i32_v2i64, ARM64_INS_SSUBL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLv8i16_v4i32, ARM64_INS_SSUBL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSUBLv8i8_v8i16, ARM64_INS_SSUBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSUBWB_ZZZ_D, ARM64_INS_SSUBWB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBWB_ZZZ_H, ARM64_INS_SSUBWB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBWB_ZZZ_S, ARM64_INS_SSUBWB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBWT_ZZZ_D, ARM64_INS_SSUBWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBWT_ZZZ_H, ARM64_INS_SSUBWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBWT_ZZZ_S, ARM64_INS_SSUBWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SSUBWv16i8_v8i16, ARM64_INS_SSUBW2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSUBWv2i32_v2i64, ARM64_INS_SSUBW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSUBWv4i16_v4i32, ARM64_INS_SSUBW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSUBWv4i32_v2i64, ARM64_INS_SSUBW2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSUBWv8i16_v4i32, ARM64_INS_SSUBW2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SSUBWv8i8_v8i16, ARM64_INS_SSUBW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1B, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1B_D, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1B_D_IMM, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1B_H, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1B_H_IMM, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1B_IMM, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1B_S, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1B_S_IMM, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1D, ARM64_INS_ST1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1D_IMM, ARM64_INS_ST1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv16b, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv16b_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv1d, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv1d_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv2d, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv2d_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv2s, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv2s_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv4h, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv4h_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv4s, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv4s_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv8b, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv8b_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv8h, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Fourv8h_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1H, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1H_D, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1H_D_IMM, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1H_IMM, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1H_S, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1H_S_IMM, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev16b, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev16b_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev1d, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev1d_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev2d, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev2d_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev2s, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev2s_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev4h, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev4h_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev4s, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev4s_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev8b, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev8b_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev8h, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Onev8h_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev16b, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev16b_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev1d, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev1d_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev2d, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev2d_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev2s, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev2s_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev4h, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev4h_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev4s, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev4s_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev8b, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev8b_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev8h, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Threev8h_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov16b, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov16b_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov1d, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov1d_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov2d, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov2d_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov2s, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov2s_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov4h, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov4h_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov4s, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov4s_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov8b, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov8b_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov8h, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1Twov8h_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1W, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1W_D, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1W_D_IMM, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1W_IMM, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1_MXIPXX_H_B, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1_MXIPXX_H_D, ARM64_INS_ST1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1_MXIPXX_H_H, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1_MXIPXX_H_Q, ARM64_INS_ST1Q,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1_MXIPXX_H_S, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1_MXIPXX_V_B, ARM64_INS_ST1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1_MXIPXX_V_D, ARM64_INS_ST1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1_MXIPXX_V_H, ARM64_INS_ST1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1_MXIPXX_V_Q, ARM64_INS_ST1Q,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1_MXIPXX_V_S, ARM64_INS_ST1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST1i16, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1i16_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1i32, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1i32_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1i64, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1i64_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1i8, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST1i8_POST, ARM64_INS_ST1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2B, ARM64_INS_ST2B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST2B_IMM, ARM64_INS_ST2B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST2D, ARM64_INS_ST2D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST2D_IMM, ARM64_INS_ST2D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST2GOffset, ARM64_INS_ST2G,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST2GPostIndex, ARM64_INS_ST2G,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST2GPreIndex, ARM64_INS_ST2G,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST2H, ARM64_INS_ST2H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST2H_IMM, ARM64_INS_ST2H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST2Twov16b, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2Twov16b_POST, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2Twov2d, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2Twov2d_POST, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2Twov2s, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2Twov2s_POST, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2Twov4h, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2Twov4h_POST, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2Twov4s, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2Twov4s_POST, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2Twov8b, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2Twov8b_POST, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2Twov8h, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2Twov8h_POST, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2W, ARM64_INS_ST2W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST2W_IMM, ARM64_INS_ST2W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST2i16, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2i16_POST, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2i32, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2i32_POST, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2i64, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2i64_POST, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2i8, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST2i8_POST, ARM64_INS_ST2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3B, ARM64_INS_ST3B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST3B_IMM, ARM64_INS_ST3B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST3D, ARM64_INS_ST3D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST3D_IMM, ARM64_INS_ST3D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST3H, ARM64_INS_ST3H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST3H_IMM, ARM64_INS_ST3H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST3Threev16b, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3Threev16b_POST, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3Threev2d, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3Threev2d_POST, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3Threev2s, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3Threev2s_POST, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3Threev4h, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3Threev4h_POST, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3Threev4s, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3Threev4s_POST, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3Threev8b, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3Threev8b_POST, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3Threev8h, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3Threev8h_POST, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3W, ARM64_INS_ST3W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST3W_IMM, ARM64_INS_ST3W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST3i16, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3i16_POST, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3i32, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3i32_POST, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3i64, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3i64_POST, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3i8, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST3i8_POST, ARM64_INS_ST3,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4B, ARM64_INS_ST4B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST4B_IMM, ARM64_INS_ST4B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST4D, ARM64_INS_ST4D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST4D_IMM, ARM64_INS_ST4D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST4Fourv16b, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4Fourv16b_POST, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4Fourv2d, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4Fourv2d_POST, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4Fourv2s, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4Fourv2s_POST, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4Fourv4h, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4Fourv4h_POST, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4Fourv4s, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4Fourv4s_POST, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4Fourv8b, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4Fourv8b_POST, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4Fourv8h, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4Fourv8h_POST, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4H, ARM64_INS_ST4H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST4H_IMM, ARM64_INS_ST4H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST4W, ARM64_INS_ST4W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST4W_IMM, ARM64_INS_ST4W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST4i16, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4i16_POST, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4i32, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4i32_POST, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4i64, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4i64_POST, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4i8, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST4i8_POST, ARM64_INS_ST4,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ST64B, ARM64_INS_ST64B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST64BV, ARM64_INS_ST64BV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ST64BV0, ARM64_INS_ST64BV0,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STGM, ARM64_INS_STGM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STGOffset, ARM64_INS_STG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STGPi, ARM64_INS_STGP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STGPostIndex, ARM64_INS_STG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STGPpost, ARM64_INS_STGP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STGPpre, ARM64_INS_STGP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STGPreIndex, ARM64_INS_STG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLLRB, ARM64_INS_STLLRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLLRH, ARM64_INS_STLLRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLLRW, ARM64_INS_STLLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLLRX, ARM64_INS_STLLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLRB, ARM64_INS_STLRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLRH, ARM64_INS_STLRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLRW, ARM64_INS_STLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLRX, ARM64_INS_STLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLURBi, ARM64_INS_STLURB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLURHi, ARM64_INS_STLURH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLURWi, ARM64_INS_STLUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLURXi, ARM64_INS_STLUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLXPW, ARM64_INS_STLXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLXPX, ARM64_INS_STLXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLXRB, ARM64_INS_STLXRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLXRH, ARM64_INS_STLXRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLXRW, ARM64_INS_STLXR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STLXRX, ARM64_INS_STLXR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNPDi, ARM64_INS_STNP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNPQi, ARM64_INS_STNP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNPSi, ARM64_INS_STNP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNPWi, ARM64_INS_STNP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNPXi, ARM64_INS_STNP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1B_ZRI, ARM64_INS_STNT1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1B_ZRR, ARM64_INS_STNT1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1B_ZZR_D_REAL, ARM64_INS_STNT1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1B_ZZR_S_REAL, ARM64_INS_STNT1B,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1D_ZRI, ARM64_INS_STNT1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1D_ZRR, ARM64_INS_STNT1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1D_ZZR_D_REAL, ARM64_INS_STNT1D,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1H_ZRI, ARM64_INS_STNT1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1H_ZRR, ARM64_INS_STNT1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1H_ZZR_D_REAL, ARM64_INS_STNT1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1H_ZZR_S_REAL, ARM64_INS_STNT1H,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1W_ZRI, ARM64_INS_STNT1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1W_ZRR, ARM64_INS_STNT1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1W_ZZR_D_REAL, ARM64_INS_STNT1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STNT1W_ZZR_S_REAL, ARM64_INS_STNT1W,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPDi, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPDpost, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPDpre, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPQi, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPQpost, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPQpre, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPSi, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPSpost, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPSpre, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPWi, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPWpost, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPWpre, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPXi, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPXpost, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STPXpre, ARM64_INS_STP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRBBpost, ARM64_INS_STRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRBBpre, ARM64_INS_STRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRBBroW, ARM64_INS_STRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRBBroX, ARM64_INS_STRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRBBui, ARM64_INS_STRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRBpost, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRBpre, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRBroW, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRBroX, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRBui, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRDpost, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRDpre, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRDroW, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRDroX, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRDui, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRHHpost, ARM64_INS_STRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRHHpre, ARM64_INS_STRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRHHroW, ARM64_INS_STRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRHHroX, ARM64_INS_STRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRHHui, ARM64_INS_STRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRHpost, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRHpre, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRHroW, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRHroX, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRHui, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRQpost, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRQpre, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRQroW, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRQroX, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRQui, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRSpost, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRSpre, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRSroW, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRSroX, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRSui, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRWpost, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRWpre, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRWroW, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRWroX, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRWui, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRXpost, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRXpre, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRXroW, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRXroX, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STRXui, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STR_PXI, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STR_ZA, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STR_ZXI, ARM64_INS_STR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STTRBi, ARM64_INS_STTRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STTRHi, ARM64_INS_STTRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STTRWi, ARM64_INS_STTR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STTRXi, ARM64_INS_STTR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STURBBi, ARM64_INS_STURB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STURBi, ARM64_INS_STUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STURDi, ARM64_INS_STUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STURHHi, ARM64_INS_STURH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STURHi, ARM64_INS_STUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STURQi, ARM64_INS_STUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STURSi, ARM64_INS_STUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STURWi, ARM64_INS_STUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STURXi, ARM64_INS_STUR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STXPW, ARM64_INS_STXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STXPX, ARM64_INS_STXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STXRB, ARM64_INS_STXRB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STXRH, ARM64_INS_STXRH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STXRW, ARM64_INS_STXR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STXRX, ARM64_INS_STXR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STZ2GOffset, ARM64_INS_STZ2G,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STZ2GPostIndex, ARM64_INS_STZ2G,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STZ2GPreIndex, ARM64_INS_STZ2G,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STZGM, ARM64_INS_STZGM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STZGOffset, ARM64_INS_STZG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STZGPostIndex, ARM64_INS_STZG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_STZGPreIndex, ARM64_INS_STZG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBG, ARM64_INS_SUBG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBHNB_ZZZ_B, ARM64_INS_SUBHNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBHNB_ZZZ_H, ARM64_INS_SUBHNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBHNB_ZZZ_S, ARM64_INS_SUBHNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBHNT_ZZZ_B, ARM64_INS_SUBHNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBHNT_ZZZ_H, ARM64_INS_SUBHNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBHNT_ZZZ_S, ARM64_INS_SUBHNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBHNv2i64_v2i32, ARM64_INS_SUBHN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUBHNv2i64_v4i32, ARM64_INS_SUBHN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUBHNv4i32_v4i16, ARM64_INS_SUBHN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUBHNv4i32_v8i16, ARM64_INS_SUBHN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUBHNv8i16_v16i8, ARM64_INS_SUBHN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUBHNv8i16_v8i8, ARM64_INS_SUBHN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUBP, ARM64_INS_SUBP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBPS, ARM64_INS_SUBPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBR_ZI_B, ARM64_INS_SUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBR_ZI_D, ARM64_INS_SUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBR_ZI_H, ARM64_INS_SUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBR_ZI_S, ARM64_INS_SUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBR_ZPmZ_B, ARM64_INS_SUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBR_ZPmZ_D, ARM64_INS_SUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBR_ZPmZ_H, ARM64_INS_SUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBR_ZPmZ_S, ARM64_INS_SUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBSWri, ARM64_INS_SUBS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBSWrs, ARM64_INS_SUBS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBSWrx, ARM64_INS_SUBS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBSXri, ARM64_INS_SUBS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBSXrs, ARM64_INS_SUBS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBSXrx, ARM64_INS_SUBS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBSXrx64, ARM64_INS_SUBS,
#ifndef CAPSTONE_DIET
	{ 0 }, { ARM64_REG_NZCV, 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBWri, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBWrs, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBWrx, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBXri, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBXrs, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBXrx, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBXrx64, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUB_ZI_B, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUB_ZI_D, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUB_ZI_H, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUB_ZI_S, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUB_ZPmZ_B, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUB_ZPmZ_D, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUB_ZPmZ_H, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUB_ZPmZ_S, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUB_ZZZ_B, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUB_ZZZ_D, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUB_ZZZ_H, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUB_ZZZ_S, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUBv16i8, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUBv1i64, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUBv2i32, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUBv2i64, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUBv4i16, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUBv4i32, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUBv8i16, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUBv8i8, ARM64_INS_SUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUDOT_ZZZI, ARM64_INS_SUDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUDOTlanev16i8, ARM64_INS_SUDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUDOTlanev8i8, ARM64_INS_SUDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUMOPA_MPPZZ_D, ARM64_INS_SUMOPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUMOPA_MPPZZ_S, ARM64_INS_SUMOPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUMOPS_MPPZZ_D, ARM64_INS_SUMOPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUMOPS_MPPZZ_S, ARM64_INS_SUMOPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUNPKHI_ZZ_D, ARM64_INS_SUNPKHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUNPKHI_ZZ_H, ARM64_INS_SUNPKHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUNPKHI_ZZ_S, ARM64_INS_SUNPKHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUNPKLO_ZZ_D, ARM64_INS_SUNPKLO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUNPKLO_ZZ_H, ARM64_INS_SUNPKLO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUNPKLO_ZZ_S, ARM64_INS_SUNPKLO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUQADD_ZPmZ_B, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUQADD_ZPmZ_D, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUQADD_ZPmZ_H, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUQADD_ZPmZ_S, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SUQADDv16i8, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUQADDv1i16, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUQADDv1i32, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUQADDv1i64, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUQADDv1i8, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUQADDv2i32, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUQADDv2i64, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUQADDv4i16, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUQADDv4i32, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUQADDv8i16, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SUQADDv8i8, ARM64_INS_SUQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_SVC, ARM64_INS_SVC,
#ifndef CAPSTONE_DIET
	{ 0, 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_INT, 0 }, 0, 0
#endif
},

{
	AArch64_SWPAB, ARM64_INS_SWPAB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPAH, ARM64_INS_SWPAH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPALB, ARM64_INS_SWPALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPALH, ARM64_INS_SWPALH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPALW, ARM64_INS_SWPAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPALX, ARM64_INS_SWPAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPAW, ARM64_INS_SWPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPAX, ARM64_INS_SWPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPB, ARM64_INS_SWPB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPH, ARM64_INS_SWPH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPLB, ARM64_INS_SWPLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPLH, ARM64_INS_SWPLH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPLW, ARM64_INS_SWPL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPLX, ARM64_INS_SWPL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPW, ARM64_INS_SWP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SWPX, ARM64_INS_SWP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SXTB_ZPmZ_D, ARM64_INS_SXTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SXTB_ZPmZ_H, ARM64_INS_SXTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SXTB_ZPmZ_S, ARM64_INS_SXTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SXTH_ZPmZ_D, ARM64_INS_SXTH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SXTH_ZPmZ_S, ARM64_INS_SXTH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SXTW_ZPmZ_D, ARM64_INS_SXTW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SYSLxt, ARM64_INS_SYSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_SYSxt, ARM64_INS_SYS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TBL_ZZZZ_B, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TBL_ZZZZ_D, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TBL_ZZZZ_H, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TBL_ZZZZ_S, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TBL_ZZZ_B, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TBL_ZZZ_D, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TBL_ZZZ_H, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TBL_ZZZ_S, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TBLv16i8Four, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBLv16i8One, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBLv16i8Three, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBLv16i8Two, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBLv8i8Four, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBLv8i8One, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBLv8i8Three, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBLv8i8Two, ARM64_INS_TBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBNZW, ARM64_INS_TBNZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, 0 }, 1, 0
#endif
},

{
	AArch64_TBNZX, ARM64_INS_TBNZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, 0 }, 1, 0
#endif
},

{
	AArch64_TBX_ZZZ_B, ARM64_INS_TBX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TBX_ZZZ_D, ARM64_INS_TBX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TBX_ZZZ_H, ARM64_INS_TBX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TBX_ZZZ_S, ARM64_INS_TBX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TBXv16i8Four, ARM64_INS_TBX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBXv16i8One, ARM64_INS_TBX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBXv16i8Three, ARM64_INS_TBX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBXv16i8Two, ARM64_INS_TBX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBXv8i8Four, ARM64_INS_TBX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBXv8i8One, ARM64_INS_TBX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBXv8i8Three, ARM64_INS_TBX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBXv8i8Two, ARM64_INS_TBX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TBZW, ARM64_INS_TBZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, 0 }, 1, 0
#endif
},

{
	AArch64_TBZX, ARM64_INS_TBZ,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_JUMP, ARM64_GRP_BRANCH_RELATIVE, 0 }, 1, 0
#endif
},

{
	AArch64_TCANCEL, ARM64_INS_TCANCEL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TCOMMIT, ARM64_INS_TCOMMIT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN1_PPP_B, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN1_PPP_D, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN1_PPP_H, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN1_PPP_S, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN1_ZZZ_B, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN1_ZZZ_D, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN1_ZZZ_H, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN1_ZZZ_Q, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN1_ZZZ_S, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN1v16i8, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TRN1v2i32, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TRN1v2i64, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TRN1v4i16, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TRN1v4i32, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TRN1v8i16, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TRN1v8i8, ARM64_INS_TRN1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TRN2_PPP_B, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN2_PPP_D, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN2_PPP_H, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN2_PPP_S, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN2_ZZZ_B, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN2_ZZZ_D, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN2_ZZZ_H, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN2_ZZZ_Q, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN2_ZZZ_S, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TRN2v16i8, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TRN2v2i32, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TRN2v2i64, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TRN2v4i16, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TRN2v4i32, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TRN2v8i16, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TRN2v8i8, ARM64_INS_TRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_TSB, ARM64_INS_TSB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TSTART, ARM64_INS_TSTART,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_TTEST, ARM64_INS_TTEST,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABALB_ZZZ_D, ARM64_INS_UABALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABALB_ZZZ_H, ARM64_INS_UABALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABALB_ZZZ_S, ARM64_INS_UABALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABALT_ZZZ_D, ARM64_INS_UABALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABALT_ZZZ_H, ARM64_INS_UABALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABALT_ZZZ_S, ARM64_INS_UABALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABALv16i8_v8i16, ARM64_INS_UABAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABALv2i32_v2i64, ARM64_INS_UABAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABALv4i16_v4i32, ARM64_INS_UABAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABALv4i32_v2i64, ARM64_INS_UABAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABALv8i16_v4i32, ARM64_INS_UABAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABALv8i8_v8i16, ARM64_INS_UABAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABA_ZZZ_B, ARM64_INS_UABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABA_ZZZ_D, ARM64_INS_UABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABA_ZZZ_H, ARM64_INS_UABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABA_ZZZ_S, ARM64_INS_UABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABAv16i8, ARM64_INS_UABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABAv2i32, ARM64_INS_UABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABAv4i16, ARM64_INS_UABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABAv4i32, ARM64_INS_UABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABAv8i16, ARM64_INS_UABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABAv8i8, ARM64_INS_UABA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABDLB_ZZZ_D, ARM64_INS_UABDLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABDLB_ZZZ_H, ARM64_INS_UABDLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABDLB_ZZZ_S, ARM64_INS_UABDLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABDLT_ZZZ_D, ARM64_INS_UABDLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABDLT_ZZZ_H, ARM64_INS_UABDLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABDLT_ZZZ_S, ARM64_INS_UABDLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABDLv16i8_v8i16, ARM64_INS_UABDL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABDLv2i32_v2i64, ARM64_INS_UABDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABDLv4i16_v4i32, ARM64_INS_UABDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABDLv4i32_v2i64, ARM64_INS_UABDL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABDLv8i16_v4i32, ARM64_INS_UABDL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABDLv8i8_v8i16, ARM64_INS_UABDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABD_ZPmZ_B, ARM64_INS_UABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABD_ZPmZ_D, ARM64_INS_UABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABD_ZPmZ_H, ARM64_INS_UABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABD_ZPmZ_S, ARM64_INS_UABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UABDv16i8, ARM64_INS_UABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABDv2i32, ARM64_INS_UABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABDv4i16, ARM64_INS_UABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABDv4i32, ARM64_INS_UABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABDv8i16, ARM64_INS_UABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UABDv8i8, ARM64_INS_UABD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADALP_ZPmZ_D, ARM64_INS_UADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADALP_ZPmZ_H, ARM64_INS_UADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADALP_ZPmZ_S, ARM64_INS_UADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADALPv16i8_v8i16, ARM64_INS_UADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADALPv2i32_v1i64, ARM64_INS_UADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADALPv4i16_v2i32, ARM64_INS_UADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADALPv4i32_v2i64, ARM64_INS_UADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADALPv8i16_v4i32, ARM64_INS_UADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADALPv8i8_v4i16, ARM64_INS_UADALP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLB_ZZZ_D, ARM64_INS_UADDLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDLB_ZZZ_H, ARM64_INS_UADDLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDLB_ZZZ_S, ARM64_INS_UADDLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDLPv16i8_v8i16, ARM64_INS_UADDLP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLPv2i32_v1i64, ARM64_INS_UADDLP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLPv4i16_v2i32, ARM64_INS_UADDLP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLPv4i32_v2i64, ARM64_INS_UADDLP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLPv8i16_v4i32, ARM64_INS_UADDLP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLPv8i8_v4i16, ARM64_INS_UADDLP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLT_ZZZ_D, ARM64_INS_UADDLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDLT_ZZZ_H, ARM64_INS_UADDLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDLT_ZZZ_S, ARM64_INS_UADDLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDLVv16i8v, ARM64_INS_UADDLV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLVv4i16v, ARM64_INS_UADDLV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLVv4i32v, ARM64_INS_UADDLV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLVv8i16v, ARM64_INS_UADDLV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLVv8i8v, ARM64_INS_UADDLV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLv16i8_v8i16, ARM64_INS_UADDL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLv2i32_v2i64, ARM64_INS_UADDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLv4i16_v4i32, ARM64_INS_UADDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLv4i32_v2i64, ARM64_INS_UADDL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLv8i16_v4i32, ARM64_INS_UADDL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDLv8i8_v8i16, ARM64_INS_UADDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDV_VPZ_B, ARM64_INS_UADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDV_VPZ_D, ARM64_INS_UADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDV_VPZ_H, ARM64_INS_UADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDV_VPZ_S, ARM64_INS_UADDV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDWB_ZZZ_D, ARM64_INS_UADDWB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDWB_ZZZ_H, ARM64_INS_UADDWB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDWB_ZZZ_S, ARM64_INS_UADDWB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDWT_ZZZ_D, ARM64_INS_UADDWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDWT_ZZZ_H, ARM64_INS_UADDWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDWT_ZZZ_S, ARM64_INS_UADDWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UADDWv16i8_v8i16, ARM64_INS_UADDW2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDWv2i32_v2i64, ARM64_INS_UADDW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDWv4i16_v4i32, ARM64_INS_UADDW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDWv4i32_v2i64, ARM64_INS_UADDW2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDWv8i16_v4i32, ARM64_INS_UADDW2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UADDWv8i8_v8i16, ARM64_INS_UADDW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UBFMWri, ARM64_INS_UBFM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UBFMXri, ARM64_INS_UBFM,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCLAMP_ZZZ_B, ARM64_INS_UCLAMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCLAMP_ZZZ_D, ARM64_INS_UCLAMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCLAMP_ZZZ_H, ARM64_INS_UCLAMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCLAMP_ZZZ_S, ARM64_INS_UCLAMP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFSWDri, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFSWHri, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFSWSri, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFSXDri, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFSXHri, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFSXSri, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFUWDri, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFUWHri, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFUWSri, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFUXDri, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFUXHri, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFUXSri, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_FPARMV8, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTF_ZPmZ_DtoD, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTF_ZPmZ_DtoH, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTF_ZPmZ_DtoS, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTF_ZPmZ_HtoH, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTF_ZPmZ_StoD, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTF_ZPmZ_StoH, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTF_ZPmZ_StoS, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFd, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFh, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFs, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFv1i16, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFv1i32, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFv1i64, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFv2f32, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFv2f64, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFv2i32_shift, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFv2i64_shift, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFv4f16, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFv4f32, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFv4i16_shift, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFv4i32_shift, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFv8f16, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UCVTFv8i16_shift, ARM64_INS_UCVTF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDF, ARM64_INS_UDF,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDIVR_ZPmZ_D, ARM64_INS_UDIVR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDIVR_ZPmZ_S, ARM64_INS_UDIVR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDIVWr, ARM64_INS_UDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDIVXr, ARM64_INS_UDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDIV_ZPmZ_D, ARM64_INS_UDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDIV_ZPmZ_S, ARM64_INS_UDIV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDOT_ZZZI_D, ARM64_INS_UDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDOT_ZZZI_S, ARM64_INS_UDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDOT_ZZZ_D, ARM64_INS_UDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDOT_ZZZ_S, ARM64_INS_UDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDOTlanev16i8, ARM64_INS_UDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDOTlanev8i8, ARM64_INS_UDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDOTv16i8, ARM64_INS_UDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UDOTv8i8, ARM64_INS_UDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UHADD_ZPmZ_B, ARM64_INS_UHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UHADD_ZPmZ_D, ARM64_INS_UHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UHADD_ZPmZ_H, ARM64_INS_UHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UHADD_ZPmZ_S, ARM64_INS_UHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UHADDv16i8, ARM64_INS_UHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UHADDv2i32, ARM64_INS_UHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UHADDv4i16, ARM64_INS_UHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UHADDv4i32, ARM64_INS_UHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UHADDv8i16, ARM64_INS_UHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UHADDv8i8, ARM64_INS_UHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UHSUBR_ZPmZ_B, ARM64_INS_UHSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UHSUBR_ZPmZ_D, ARM64_INS_UHSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UHSUBR_ZPmZ_H, ARM64_INS_UHSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UHSUBR_ZPmZ_S, ARM64_INS_UHSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UHSUB_ZPmZ_B, ARM64_INS_UHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UHSUB_ZPmZ_D, ARM64_INS_UHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UHSUB_ZPmZ_H, ARM64_INS_UHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UHSUB_ZPmZ_S, ARM64_INS_UHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UHSUBv16i8, ARM64_INS_UHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UHSUBv2i32, ARM64_INS_UHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UHSUBv4i16, ARM64_INS_UHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UHSUBv4i32, ARM64_INS_UHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UHSUBv8i16, ARM64_INS_UHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UHSUBv8i8, ARM64_INS_UHSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMADDLrrr, ARM64_INS_UMADDL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAXP_ZPmZ_B, ARM64_INS_UMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAXP_ZPmZ_D, ARM64_INS_UMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAXP_ZPmZ_H, ARM64_INS_UMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAXP_ZPmZ_S, ARM64_INS_UMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAXPv16i8, ARM64_INS_UMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXPv2i32, ARM64_INS_UMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXPv4i16, ARM64_INS_UMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXPv4i32, ARM64_INS_UMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXPv8i16, ARM64_INS_UMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXPv8i8, ARM64_INS_UMAXP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXV_VPZ_B, ARM64_INS_UMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAXV_VPZ_D, ARM64_INS_UMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAXV_VPZ_H, ARM64_INS_UMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAXV_VPZ_S, ARM64_INS_UMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAXVv16i8v, ARM64_INS_UMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXVv4i16v, ARM64_INS_UMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXVv4i32v, ARM64_INS_UMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXVv8i16v, ARM64_INS_UMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXVv8i8v, ARM64_INS_UMAXV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAX_ZI_B, ARM64_INS_UMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAX_ZI_D, ARM64_INS_UMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAX_ZI_H, ARM64_INS_UMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAX_ZI_S, ARM64_INS_UMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAX_ZPmZ_B, ARM64_INS_UMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAX_ZPmZ_D, ARM64_INS_UMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAX_ZPmZ_H, ARM64_INS_UMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAX_ZPmZ_S, ARM64_INS_UMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMAXv16i8, ARM64_INS_UMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXv2i32, ARM64_INS_UMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXv4i16, ARM64_INS_UMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXv4i32, ARM64_INS_UMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXv8i16, ARM64_INS_UMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMAXv8i8, ARM64_INS_UMAX,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINP_ZPmZ_B, ARM64_INS_UMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMINP_ZPmZ_D, ARM64_INS_UMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMINP_ZPmZ_H, ARM64_INS_UMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMINP_ZPmZ_S, ARM64_INS_UMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMINPv16i8, ARM64_INS_UMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINPv2i32, ARM64_INS_UMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINPv4i16, ARM64_INS_UMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINPv4i32, ARM64_INS_UMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINPv8i16, ARM64_INS_UMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINPv8i8, ARM64_INS_UMINP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINV_VPZ_B, ARM64_INS_UMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMINV_VPZ_D, ARM64_INS_UMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMINV_VPZ_H, ARM64_INS_UMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMINV_VPZ_S, ARM64_INS_UMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMINVv16i8v, ARM64_INS_UMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINVv4i16v, ARM64_INS_UMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINVv4i32v, ARM64_INS_UMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINVv8i16v, ARM64_INS_UMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINVv8i8v, ARM64_INS_UMINV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMIN_ZI_B, ARM64_INS_UMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMIN_ZI_D, ARM64_INS_UMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMIN_ZI_H, ARM64_INS_UMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMIN_ZI_S, ARM64_INS_UMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMIN_ZPmZ_B, ARM64_INS_UMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMIN_ZPmZ_D, ARM64_INS_UMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMIN_ZPmZ_H, ARM64_INS_UMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMIN_ZPmZ_S, ARM64_INS_UMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMINv16i8, ARM64_INS_UMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINv2i32, ARM64_INS_UMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINv4i16, ARM64_INS_UMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINv4i32, ARM64_INS_UMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINv8i16, ARM64_INS_UMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMINv8i8, ARM64_INS_UMIN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLALB_ZZZI_D, ARM64_INS_UMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLALB_ZZZI_S, ARM64_INS_UMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLALB_ZZZ_D, ARM64_INS_UMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLALB_ZZZ_H, ARM64_INS_UMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLALB_ZZZ_S, ARM64_INS_UMLALB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLALT_ZZZI_D, ARM64_INS_UMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLALT_ZZZI_S, ARM64_INS_UMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLALT_ZZZ_D, ARM64_INS_UMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLALT_ZZZ_H, ARM64_INS_UMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLALT_ZZZ_S, ARM64_INS_UMLALT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLALv16i8_v8i16, ARM64_INS_UMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLALv2i32_indexed, ARM64_INS_UMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLALv2i32_v2i64, ARM64_INS_UMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLALv4i16_indexed, ARM64_INS_UMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLALv4i16_v4i32, ARM64_INS_UMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLALv4i32_indexed, ARM64_INS_UMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLALv4i32_v2i64, ARM64_INS_UMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLALv8i16_indexed, ARM64_INS_UMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLALv8i16_v4i32, ARM64_INS_UMLAL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLALv8i8_v8i16, ARM64_INS_UMLAL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLB_ZZZI_D, ARM64_INS_UMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLB_ZZZI_S, ARM64_INS_UMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLB_ZZZ_D, ARM64_INS_UMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLB_ZZZ_H, ARM64_INS_UMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLB_ZZZ_S, ARM64_INS_UMLSLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLT_ZZZI_D, ARM64_INS_UMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLT_ZZZI_S, ARM64_INS_UMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLT_ZZZ_D, ARM64_INS_UMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLT_ZZZ_H, ARM64_INS_UMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLT_ZZZ_S, ARM64_INS_UMLSLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLv16i8_v8i16, ARM64_INS_UMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLv2i32_indexed, ARM64_INS_UMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLv2i32_v2i64, ARM64_INS_UMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLv4i16_indexed, ARM64_INS_UMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLv4i16_v4i32, ARM64_INS_UMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLv4i32_indexed, ARM64_INS_UMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLv4i32_v2i64, ARM64_INS_UMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLv8i16_indexed, ARM64_INS_UMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLv8i16_v4i32, ARM64_INS_UMLSL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMLSLv8i8_v8i16, ARM64_INS_UMLSL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMMLA, ARM64_INS_UMMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMMLA_ZZZ, ARM64_INS_UMMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMOPA_MPPZZ_D, ARM64_INS_UMOPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMOPA_MPPZZ_S, ARM64_INS_UMOPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMOPS_MPPZZ_D, ARM64_INS_UMOPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMOPS_MPPZZ_S, ARM64_INS_UMOPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMOVvi16, ARM64_INS_UMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMOVvi16_idx0, ARM64_INS_UMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMOVvi32, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMOVvi32_idx0, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMOVvi64, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMOVvi64_idx0, ARM64_INS_MOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMOVvi8, ARM64_INS_UMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMOVvi8_idx0, ARM64_INS_UMOV,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMSUBLrrr, ARM64_INS_UMSUBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULH_ZPmZ_B, ARM64_INS_UMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULH_ZPmZ_D, ARM64_INS_UMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULH_ZPmZ_H, ARM64_INS_UMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULH_ZPmZ_S, ARM64_INS_UMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULH_ZZZ_B, ARM64_INS_UMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULH_ZZZ_D, ARM64_INS_UMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULH_ZZZ_H, ARM64_INS_UMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULH_ZZZ_S, ARM64_INS_UMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULHrr, ARM64_INS_UMULH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULLB_ZZZI_D, ARM64_INS_UMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULLB_ZZZI_S, ARM64_INS_UMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULLB_ZZZ_D, ARM64_INS_UMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULLB_ZZZ_H, ARM64_INS_UMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULLB_ZZZ_S, ARM64_INS_UMULLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULLT_ZZZI_D, ARM64_INS_UMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULLT_ZZZI_S, ARM64_INS_UMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULLT_ZZZ_D, ARM64_INS_UMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULLT_ZZZ_H, ARM64_INS_UMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULLT_ZZZ_S, ARM64_INS_UMULLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UMULLv16i8_v8i16, ARM64_INS_UMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMULLv2i32_indexed, ARM64_INS_UMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMULLv2i32_v2i64, ARM64_INS_UMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMULLv4i16_indexed, ARM64_INS_UMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMULLv4i16_v4i32, ARM64_INS_UMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMULLv4i32_indexed, ARM64_INS_UMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMULLv4i32_v2i64, ARM64_INS_UMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMULLv8i16_indexed, ARM64_INS_UMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMULLv8i16_v4i32, ARM64_INS_UMULL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UMULLv8i8_v8i16, ARM64_INS_UMULL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQADD_ZI_B, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQADD_ZI_D, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQADD_ZI_H, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQADD_ZI_S, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQADD_ZPmZ_B, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQADD_ZPmZ_D, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQADD_ZPmZ_H, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQADD_ZPmZ_S, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQADD_ZZZ_B, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQADD_ZZZ_D, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQADD_ZZZ_H, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQADD_ZZZ_S, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQADDv16i8, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQADDv1i16, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQADDv1i32, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQADDv1i64, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQADDv1i8, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQADDv2i32, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQADDv2i64, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQADDv4i16, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQADDv4i32, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQADDv8i16, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQADDv8i8, ARM64_INS_UQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQDECB_WPiI, ARM64_INS_UQDECB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECB_XPiI, ARM64_INS_UQDECB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECD_WPiI, ARM64_INS_UQDECD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECD_XPiI, ARM64_INS_UQDECD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECD_ZPiI, ARM64_INS_UQDECD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECH_WPiI, ARM64_INS_UQDECH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECH_XPiI, ARM64_INS_UQDECH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECH_ZPiI, ARM64_INS_UQDECH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECP_WP_B, ARM64_INS_UQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECP_WP_D, ARM64_INS_UQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECP_WP_H, ARM64_INS_UQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECP_WP_S, ARM64_INS_UQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECP_XP_B, ARM64_INS_UQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECP_XP_D, ARM64_INS_UQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECP_XP_H, ARM64_INS_UQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECP_XP_S, ARM64_INS_UQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECP_ZP_D, ARM64_INS_UQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECP_ZP_H, ARM64_INS_UQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECP_ZP_S, ARM64_INS_UQDECP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECW_WPiI, ARM64_INS_UQDECW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECW_XPiI, ARM64_INS_UQDECW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQDECW_ZPiI, ARM64_INS_UQDECW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCB_WPiI, ARM64_INS_UQINCB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCB_XPiI, ARM64_INS_UQINCB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCD_WPiI, ARM64_INS_UQINCD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCD_XPiI, ARM64_INS_UQINCD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCD_ZPiI, ARM64_INS_UQINCD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCH_WPiI, ARM64_INS_UQINCH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCH_XPiI, ARM64_INS_UQINCH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCH_ZPiI, ARM64_INS_UQINCH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCP_WP_B, ARM64_INS_UQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCP_WP_D, ARM64_INS_UQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCP_WP_H, ARM64_INS_UQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCP_WP_S, ARM64_INS_UQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCP_XP_B, ARM64_INS_UQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCP_XP_D, ARM64_INS_UQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCP_XP_H, ARM64_INS_UQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCP_XP_S, ARM64_INS_UQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCP_ZP_D, ARM64_INS_UQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCP_ZP_H, ARM64_INS_UQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCP_ZP_S, ARM64_INS_UQINCP,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCW_WPiI, ARM64_INS_UQINCW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCW_XPiI, ARM64_INS_UQINCW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQINCW_ZPiI, ARM64_INS_UQINCW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLR_ZPmZ_B, ARM64_INS_UQRSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLR_ZPmZ_D, ARM64_INS_UQRSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLR_ZPmZ_H, ARM64_INS_UQRSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLR_ZPmZ_S, ARM64_INS_UQRSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHL_ZPmZ_B, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHL_ZPmZ_D, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHL_ZPmZ_H, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHL_ZPmZ_S, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLv16i8, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLv1i16, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLv1i32, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLv1i64, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLv1i8, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLv2i32, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLv2i64, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLv4i16, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLv4i32, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLv8i16, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHLv8i8, ARM64_INS_UQRSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNB_ZZI_B, ARM64_INS_UQRSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNB_ZZI_H, ARM64_INS_UQRSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNB_ZZI_S, ARM64_INS_UQRSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNT_ZZI_B, ARM64_INS_UQRSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNT_ZZI_H, ARM64_INS_UQRSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNT_ZZI_S, ARM64_INS_UQRSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNb, ARM64_INS_UQRSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNh, ARM64_INS_UQRSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNs, ARM64_INS_UQRSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNv16i8_shift, ARM64_INS_UQRSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNv2i32_shift, ARM64_INS_UQRSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNv4i16_shift, ARM64_INS_UQRSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNv4i32_shift, ARM64_INS_UQRSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNv8i16_shift, ARM64_INS_UQRSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQRSHRNv8i8_shift, ARM64_INS_UQRSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLR_ZPmZ_B, ARM64_INS_UQSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLR_ZPmZ_D, ARM64_INS_UQSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLR_ZPmZ_H, ARM64_INS_UQSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLR_ZPmZ_S, ARM64_INS_UQSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHL_ZPmI_B, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHL_ZPmI_D, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHL_ZPmI_H, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHL_ZPmI_S, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHL_ZPmZ_B, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHL_ZPmZ_D, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHL_ZPmZ_H, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHL_ZPmZ_S, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLb, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLd, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLh, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLs, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv16i8, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv16i8_shift, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv1i16, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv1i32, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv1i64, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv1i8, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv2i32, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv2i32_shift, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv2i64, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv2i64_shift, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv4i16, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv4i16_shift, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv4i32, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv4i32_shift, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv8i16, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv8i16_shift, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv8i8, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHLv8i8_shift, ARM64_INS_UQSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNB_ZZI_B, ARM64_INS_UQSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNB_ZZI_H, ARM64_INS_UQSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNB_ZZI_S, ARM64_INS_UQSHRNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNT_ZZI_B, ARM64_INS_UQSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNT_ZZI_H, ARM64_INS_UQSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNT_ZZI_S, ARM64_INS_UQSHRNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNb, ARM64_INS_UQSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNh, ARM64_INS_UQSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNs, ARM64_INS_UQSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNv16i8_shift, ARM64_INS_UQSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNv2i32_shift, ARM64_INS_UQSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNv4i16_shift, ARM64_INS_UQSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNv4i32_shift, ARM64_INS_UQSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNv8i16_shift, ARM64_INS_UQSHRN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSHRNv8i8_shift, ARM64_INS_UQSHRN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBR_ZPmZ_B, ARM64_INS_UQSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBR_ZPmZ_D, ARM64_INS_UQSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBR_ZPmZ_H, ARM64_INS_UQSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBR_ZPmZ_S, ARM64_INS_UQSUBR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUB_ZI_B, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUB_ZI_D, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUB_ZI_H, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUB_ZI_S, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUB_ZPmZ_B, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUB_ZPmZ_D, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUB_ZPmZ_H, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUB_ZPmZ_S, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUB_ZZZ_B, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUB_ZZZ_D, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUB_ZZZ_H, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUB_ZZZ_S, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBv16i8, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBv1i16, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBv1i32, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBv1i64, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBv1i8, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBv2i32, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBv2i64, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBv4i16, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBv4i32, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBv8i16, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQSUBv8i8, ARM64_INS_UQSUB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNB_ZZ_B, ARM64_INS_UQXTNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNB_ZZ_H, ARM64_INS_UQXTNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNB_ZZ_S, ARM64_INS_UQXTNB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNT_ZZ_B, ARM64_INS_UQXTNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNT_ZZ_H, ARM64_INS_UQXTNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNT_ZZ_S, ARM64_INS_UQXTNT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNv16i8, ARM64_INS_UQXTN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNv1i16, ARM64_INS_UQXTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNv1i32, ARM64_INS_UQXTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNv1i8, ARM64_INS_UQXTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNv2i32, ARM64_INS_UQXTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNv4i16, ARM64_INS_UQXTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNv4i32, ARM64_INS_UQXTN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNv8i16, ARM64_INS_UQXTN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UQXTNv8i8, ARM64_INS_UQXTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URECPE_ZPmZ_S, ARM64_INS_URECPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URECPEv2i32, ARM64_INS_URECPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URECPEv4i32, ARM64_INS_URECPE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URHADD_ZPmZ_B, ARM64_INS_URHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URHADD_ZPmZ_D, ARM64_INS_URHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URHADD_ZPmZ_H, ARM64_INS_URHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URHADD_ZPmZ_S, ARM64_INS_URHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URHADDv16i8, ARM64_INS_URHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URHADDv2i32, ARM64_INS_URHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URHADDv4i16, ARM64_INS_URHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URHADDv4i32, ARM64_INS_URHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URHADDv8i16, ARM64_INS_URHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URHADDv8i8, ARM64_INS_URHADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHLR_ZPmZ_B, ARM64_INS_URSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSHLR_ZPmZ_D, ARM64_INS_URSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSHLR_ZPmZ_H, ARM64_INS_URSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSHLR_ZPmZ_S, ARM64_INS_URSHLR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSHL_ZPmZ_B, ARM64_INS_URSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSHL_ZPmZ_D, ARM64_INS_URSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSHL_ZPmZ_H, ARM64_INS_URSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSHL_ZPmZ_S, ARM64_INS_URSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSHLv16i8, ARM64_INS_URSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHLv1i64, ARM64_INS_URSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHLv2i32, ARM64_INS_URSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHLv2i64, ARM64_INS_URSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHLv4i16, ARM64_INS_URSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHLv4i32, ARM64_INS_URSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHLv8i16, ARM64_INS_URSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHLv8i8, ARM64_INS_URSHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHR_ZPmI_B, ARM64_INS_URSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSHR_ZPmI_D, ARM64_INS_URSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSHR_ZPmI_H, ARM64_INS_URSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSHR_ZPmI_S, ARM64_INS_URSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSHRd, ARM64_INS_URSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHRv16i8_shift, ARM64_INS_URSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHRv2i32_shift, ARM64_INS_URSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHRv2i64_shift, ARM64_INS_URSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHRv4i16_shift, ARM64_INS_URSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHRv4i32_shift, ARM64_INS_URSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHRv8i16_shift, ARM64_INS_URSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSHRv8i8_shift, ARM64_INS_URSHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSQRTE_ZPmZ_S, ARM64_INS_URSQRTE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSQRTEv2i32, ARM64_INS_URSQRTE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSQRTEv4i32, ARM64_INS_URSQRTE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSRA_ZZI_B, ARM64_INS_URSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSRA_ZZI_D, ARM64_INS_URSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSRA_ZZI_H, ARM64_INS_URSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSRA_ZZI_S, ARM64_INS_URSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_URSRAd, ARM64_INS_URSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSRAv16i8_shift, ARM64_INS_URSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSRAv2i32_shift, ARM64_INS_URSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSRAv2i64_shift, ARM64_INS_URSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSRAv4i16_shift, ARM64_INS_URSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSRAv4i32_shift, ARM64_INS_URSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSRAv8i16_shift, ARM64_INS_URSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_URSRAv8i8_shift, ARM64_INS_URSRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USDOT_ZZZ, ARM64_INS_USDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USDOT_ZZZI, ARM64_INS_USDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USDOTlanev16i8, ARM64_INS_USDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USDOTlanev8i8, ARM64_INS_USDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USDOTv16i8, ARM64_INS_USDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USDOTv8i8, ARM64_INS_USDOT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USHLLB_ZZI_D, ARM64_INS_USHLLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USHLLB_ZZI_H, ARM64_INS_USHLLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USHLLB_ZZI_S, ARM64_INS_USHLLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USHLLT_ZZI_D, ARM64_INS_USHLLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USHLLT_ZZI_H, ARM64_INS_USHLLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USHLLT_ZZI_S, ARM64_INS_USHLLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USHLLv16i8_shift, ARM64_INS_USHLL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHLLv2i32_shift, ARM64_INS_USHLL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHLLv4i16_shift, ARM64_INS_USHLL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHLLv4i32_shift, ARM64_INS_USHLL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHLLv8i16_shift, ARM64_INS_USHLL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHLLv8i8_shift, ARM64_INS_USHLL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHLv16i8, ARM64_INS_USHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHLv1i64, ARM64_INS_USHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHLv2i32, ARM64_INS_USHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHLv2i64, ARM64_INS_USHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHLv4i16, ARM64_INS_USHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHLv4i32, ARM64_INS_USHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHLv8i16, ARM64_INS_USHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHLv8i8, ARM64_INS_USHL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHRd, ARM64_INS_USHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHRv16i8_shift, ARM64_INS_USHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHRv2i32_shift, ARM64_INS_USHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHRv2i64_shift, ARM64_INS_USHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHRv4i16_shift, ARM64_INS_USHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHRv4i32_shift, ARM64_INS_USHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHRv8i16_shift, ARM64_INS_USHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USHRv8i8_shift, ARM64_INS_USHR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USMMLA, ARM64_INS_USMMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USMMLA_ZZZ, ARM64_INS_USMMLA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USMOPA_MPPZZ_D, ARM64_INS_USMOPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USMOPA_MPPZZ_S, ARM64_INS_USMOPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USMOPS_MPPZZ_D, ARM64_INS_USMOPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USMOPS_MPPZZ_S, ARM64_INS_USMOPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USQADD_ZPmZ_B, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USQADD_ZPmZ_D, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USQADD_ZPmZ_H, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USQADD_ZPmZ_S, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USQADDv16i8, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USQADDv1i16, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USQADDv1i32, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USQADDv1i64, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USQADDv1i8, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USQADDv2i32, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USQADDv2i64, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USQADDv4i16, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USQADDv4i32, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USQADDv8i16, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USQADDv8i8, ARM64_INS_USQADD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USRA_ZZI_B, ARM64_INS_USRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USRA_ZZI_D, ARM64_INS_USRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USRA_ZZI_H, ARM64_INS_USRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USRA_ZZI_S, ARM64_INS_USRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USRAd, ARM64_INS_USRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USRAv16i8_shift, ARM64_INS_USRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USRAv2i32_shift, ARM64_INS_USRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USRAv2i64_shift, ARM64_INS_USRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USRAv4i16_shift, ARM64_INS_USRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USRAv4i32_shift, ARM64_INS_USRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USRAv8i16_shift, ARM64_INS_USRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USRAv8i8_shift, ARM64_INS_USRA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USUBLB_ZZZ_D, ARM64_INS_USUBLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USUBLB_ZZZ_H, ARM64_INS_USUBLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USUBLB_ZZZ_S, ARM64_INS_USUBLB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USUBLT_ZZZ_D, ARM64_INS_USUBLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USUBLT_ZZZ_H, ARM64_INS_USUBLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USUBLT_ZZZ_S, ARM64_INS_USUBLT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USUBLv16i8_v8i16, ARM64_INS_USUBL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USUBLv2i32_v2i64, ARM64_INS_USUBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USUBLv4i16_v4i32, ARM64_INS_USUBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USUBLv4i32_v2i64, ARM64_INS_USUBL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USUBLv8i16_v4i32, ARM64_INS_USUBL2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USUBLv8i8_v8i16, ARM64_INS_USUBL,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USUBWB_ZZZ_D, ARM64_INS_USUBWB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USUBWB_ZZZ_H, ARM64_INS_USUBWB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USUBWB_ZZZ_S, ARM64_INS_USUBWB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USUBWT_ZZZ_D, ARM64_INS_USUBWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USUBWT_ZZZ_H, ARM64_INS_USUBWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USUBWT_ZZZ_S, ARM64_INS_USUBWT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_USUBWv16i8_v8i16, ARM64_INS_USUBW2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USUBWv2i32_v2i64, ARM64_INS_USUBW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USUBWv4i16_v4i32, ARM64_INS_USUBW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USUBWv4i32_v2i64, ARM64_INS_USUBW2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USUBWv8i16_v4i32, ARM64_INS_USUBW2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_USUBWv8i8_v8i16, ARM64_INS_USUBW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UUNPKHI_ZZ_D, ARM64_INS_UUNPKHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UUNPKHI_ZZ_H, ARM64_INS_UUNPKHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UUNPKHI_ZZ_S, ARM64_INS_UUNPKHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UUNPKLO_ZZ_D, ARM64_INS_UUNPKLO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UUNPKLO_ZZ_H, ARM64_INS_UUNPKLO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UUNPKLO_ZZ_S, ARM64_INS_UUNPKLO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UXTB_ZPmZ_D, ARM64_INS_UXTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UXTB_ZPmZ_H, ARM64_INS_UXTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UXTB_ZPmZ_S, ARM64_INS_UXTB,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UXTH_ZPmZ_D, ARM64_INS_UXTH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UXTH_ZPmZ_S, ARM64_INS_UXTH,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UXTW_ZPmZ_D, ARM64_INS_UXTW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP1_PPP_B, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP1_PPP_D, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP1_PPP_H, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP1_PPP_S, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP1_ZZZ_B, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP1_ZZZ_D, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP1_ZZZ_H, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP1_ZZZ_Q, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP1_ZZZ_S, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP1v16i8, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UZP1v2i32, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UZP1v2i64, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UZP1v4i16, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UZP1v4i32, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UZP1v8i16, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UZP1v8i8, ARM64_INS_UZP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UZP2_PPP_B, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP2_PPP_D, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP2_PPP_H, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP2_PPP_S, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP2_ZZZ_B, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP2_ZZZ_D, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP2_ZZZ_H, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP2_ZZZ_Q, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP2_ZZZ_S, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_UZP2v16i8, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UZP2v2i32, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UZP2v2i64, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UZP2v4i16, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UZP2v4i32, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UZP2v8i16, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_UZP2v8i8, ARM64_INS_UZP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_WFET, ARM64_INS_WFET,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WFIT, ARM64_INS_WFIT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGE_PWW_B, ARM64_INS_WHILEGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGE_PWW_D, ARM64_INS_WHILEGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGE_PWW_H, ARM64_INS_WHILEGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGE_PWW_S, ARM64_INS_WHILEGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGE_PXX_B, ARM64_INS_WHILEGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGE_PXX_D, ARM64_INS_WHILEGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGE_PXX_H, ARM64_INS_WHILEGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGE_PXX_S, ARM64_INS_WHILEGE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGT_PWW_B, ARM64_INS_WHILEGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGT_PWW_D, ARM64_INS_WHILEGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGT_PWW_H, ARM64_INS_WHILEGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGT_PWW_S, ARM64_INS_WHILEGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGT_PXX_B, ARM64_INS_WHILEGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGT_PXX_D, ARM64_INS_WHILEGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGT_PXX_H, ARM64_INS_WHILEGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEGT_PXX_S, ARM64_INS_WHILEGT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHI_PWW_B, ARM64_INS_WHILEHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHI_PWW_D, ARM64_INS_WHILEHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHI_PWW_H, ARM64_INS_WHILEHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHI_PWW_S, ARM64_INS_WHILEHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHI_PXX_B, ARM64_INS_WHILEHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHI_PXX_D, ARM64_INS_WHILEHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHI_PXX_H, ARM64_INS_WHILEHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHI_PXX_S, ARM64_INS_WHILEHI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHS_PWW_B, ARM64_INS_WHILEHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHS_PWW_D, ARM64_INS_WHILEHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHS_PWW_H, ARM64_INS_WHILEHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHS_PWW_S, ARM64_INS_WHILEHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHS_PXX_B, ARM64_INS_WHILEHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHS_PXX_D, ARM64_INS_WHILEHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHS_PXX_H, ARM64_INS_WHILEHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEHS_PXX_S, ARM64_INS_WHILEHS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELE_PWW_B, ARM64_INS_WHILELE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELE_PWW_D, ARM64_INS_WHILELE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELE_PWW_H, ARM64_INS_WHILELE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELE_PWW_S, ARM64_INS_WHILELE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELE_PXX_B, ARM64_INS_WHILELE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELE_PXX_D, ARM64_INS_WHILELE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELE_PXX_H, ARM64_INS_WHILELE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELE_PXX_S, ARM64_INS_WHILELE,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELO_PWW_B, ARM64_INS_WHILELO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELO_PWW_D, ARM64_INS_WHILELO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELO_PWW_H, ARM64_INS_WHILELO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELO_PWW_S, ARM64_INS_WHILELO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELO_PXX_B, ARM64_INS_WHILELO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELO_PXX_D, ARM64_INS_WHILELO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELO_PXX_H, ARM64_INS_WHILELO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELO_PXX_S, ARM64_INS_WHILELO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELS_PWW_B, ARM64_INS_WHILELS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELS_PWW_D, ARM64_INS_WHILELS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELS_PWW_H, ARM64_INS_WHILELS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELS_PWW_S, ARM64_INS_WHILELS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELS_PXX_B, ARM64_INS_WHILELS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELS_PXX_D, ARM64_INS_WHILELS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELS_PXX_H, ARM64_INS_WHILELS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELS_PXX_S, ARM64_INS_WHILELS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELT_PWW_B, ARM64_INS_WHILELT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELT_PWW_D, ARM64_INS_WHILELT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELT_PWW_H, ARM64_INS_WHILELT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELT_PWW_S, ARM64_INS_WHILELT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELT_PXX_B, ARM64_INS_WHILELT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELT_PXX_D, ARM64_INS_WHILELT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELT_PXX_H, ARM64_INS_WHILELT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILELT_PXX_S, ARM64_INS_WHILELT,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILERW_PXX_B, ARM64_INS_WHILERW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILERW_PXX_D, ARM64_INS_WHILERW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILERW_PXX_H, ARM64_INS_WHILERW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILERW_PXX_S, ARM64_INS_WHILERW,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEWR_PXX_B, ARM64_INS_WHILEWR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEWR_PXX_D, ARM64_INS_WHILEWR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEWR_PXX_H, ARM64_INS_WHILEWR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WHILEWR_PXX_S, ARM64_INS_WHILEWR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_WRFFR, ARM64_INS_WRFFR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_XAFLAG, ARM64_INS_XAFLAG,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_XAR, ARM64_INS_XAR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_XAR_ZZZI_B, ARM64_INS_XAR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_XAR_ZZZI_D, ARM64_INS_XAR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_XAR_ZZZI_H, ARM64_INS_XAR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_XAR_ZZZI_S, ARM64_INS_XAR,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_XPACD, ARM64_INS_XPACD,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_XPACI, ARM64_INS_XPACI,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_XPACLRI, ARM64_INS_XPACLRI,
#ifndef CAPSTONE_DIET
	{ ARM64_REG_LR, 0 }, { ARM64_REG_LR, 0 }, { ARM64_GRP_PAC, 0 }, 0, 0
#endif
},

{
	AArch64_XTNv16i8, ARM64_INS_XTN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_XTNv2i32, ARM64_INS_XTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_XTNv4i16, ARM64_INS_XTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_XTNv4i32, ARM64_INS_XTN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_XTNv8i16, ARM64_INS_XTN2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_XTNv8i8, ARM64_INS_XTN,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ZERO_M, ARM64_INS_ZERO,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1_PPP_B, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1_PPP_D, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1_PPP_H, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1_PPP_S, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1_ZZZ_B, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1_ZZZ_D, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1_ZZZ_H, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1_ZZZ_Q, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1_ZZZ_S, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1v16i8, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1v2i32, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1v2i64, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1v4i16, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1v4i32, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1v8i16, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ZIP1v8i8, ARM64_INS_ZIP1,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2_PPP_B, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2_PPP_D, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2_PPP_H, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2_PPP_S, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2_ZZZ_B, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2_ZZZ_D, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2_ZZZ_H, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2_ZZZ_Q, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2_ZZZ_S, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2v16i8, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2v2i32, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2v2i64, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2v4i16, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2v4i32, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2v8i16, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_ZIP2v8i8, ARM64_INS_ZIP2,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { ARM64_GRP_NEON, 0 }, 0, 0
#endif
},

{
	AArch64_anonymous_13987, ARM64_INS_BFMOPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_anonymous_13988, ARM64_INS_FMOPS,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_anonymous_5384, ARM64_INS_BFMOPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},

{
	AArch64_anonymous_5385, ARM64_INS_FMOPA,
#ifndef CAPSTONE_DIET
	{ 0 }, { 0 }, { 0 }, 0, 0
#endif
},
