/* Capstone Disassembly Engine, http://www.capstone-engine.org */
/* This is auto-gen data for Capstone disassembly engine (www.capstone-engine.org) */
/* By Nguyen Anh Quynh <aquynh@gmail.com>, 2013-2019 */


{	/* ARM_ASRi, ARM_INS_ASR: asr */
	{ 0 }
},

{	/* ARM_ASRr, ARM_INS_ASR: asr */
	{ 0 }
},

{	/* ARM_ITasm, ARM_INS_IT: it */
	{ 0 }
},

{	/* ARM_LDRBT_POST, ARM_INS_LDRBT: ldrbt */
	{ 0 }
},

{	/* ARM_LDRConstPool, ARM_INS_LDR: ldr */
	{ 0 }
},

{	/* ARM_LDRT_POST, ARM_INS_LDRT: ldrt */
	{ 0 }
},

{	/* ARM_LSLi, ARM_INS_LSL: lsl */
	{ 0 }
},

{	/* ARM_LSLr, ARM_INS_LSL: lsl */
	{ 0 }
},

{	/* ARM_LSRi, ARM_INS_LSR: lsr */
	{ 0 }
},

{	/* ARM_LSRr, ARM_INS_LSR: lsr */
	{ 0 }
},

{	/* ARM_RORi, ARM_INS_ROR: ror */
	{ 0 }
},

{	/* ARM_RORr, ARM_INS_ROR: ror */
	{ 0 }
},

{	/* ARM_RRXi, ARM_INS_RRX: rrx */
	{ 0 }
},

{	/* ARM_STRBT_POST, ARM_INS_STRBT: strbt */
	{ 0 }
},

{	/* ARM_STRT_POST, ARM_INS_STRT: strt */
	{ 0 }
},

{	/* ARM_VLD1LNdAsm_16, ARM_INS_VLD1: vld1 */
	{ 0 }
},

{	/* ARM_VLD1LNdAsm_32, ARM_INS_VLD1: vld1 */
	{ 0 }
},

{	/* ARM_VLD1LNdAsm_8, ARM_INS_VLD1: vld1 */
	{ 0 }
},

{	/* ARM_VLD1LNdWB_fixed_Asm_16, ARM_INS_VLD1: vld1 */
	{ 0 }
},

{	/* ARM_VLD1LNdWB_fixed_Asm_32, ARM_INS_VLD1: vld1 */
	{ 0 }
},

{	/* ARM_VLD1LNdWB_fixed_Asm_8, ARM_INS_VLD1: vld1 */
	{ 0 }
},

{	/* ARM_VLD1LNdWB_register_Asm_16, ARM_INS_VLD1: vld1 */
	{ 0 }
},

{	/* ARM_VLD1LNdWB_register_Asm_32, ARM_INS_VLD1: vld1 */
	{ 0 }
},

{	/* ARM_VLD1LNdWB_register_Asm_8, ARM_INS_VLD1: vld1 */
	{ 0 }
},

{	/* ARM_VLD2LNdAsm_16, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD2LNdAsm_32, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD2LNdAsm_8, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD2LNdWB_fixed_Asm_16, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD2LNdWB_fixed_Asm_32, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD2LNdWB_fixed_Asm_8, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD2LNdWB_register_Asm_16, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD2LNdWB_register_Asm_32, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD2LNdWB_register_Asm_8, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD2LNqAsm_16, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD2LNqAsm_32, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD2LNqWB_fixed_Asm_16, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD2LNqWB_fixed_Asm_32, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD2LNqWB_register_Asm_16, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD2LNqWB_register_Asm_32, ARM_INS_VLD2: vld2 */
	{ 0 }
},

{	/* ARM_VLD3DUPdAsm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPdAsm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPdAsm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPdWB_fixed_Asm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPdWB_fixed_Asm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPdWB_fixed_Asm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPdWB_register_Asm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPdWB_register_Asm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPdWB_register_Asm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPqAsm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPqAsm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPqAsm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPqWB_fixed_Asm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPqWB_fixed_Asm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPqWB_fixed_Asm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPqWB_register_Asm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPqWB_register_Asm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3DUPqWB_register_Asm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNdAsm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNdAsm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNdAsm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNdWB_fixed_Asm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNdWB_fixed_Asm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNdWB_fixed_Asm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNdWB_register_Asm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNdWB_register_Asm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNdWB_register_Asm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNqAsm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNqAsm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNqWB_fixed_Asm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNqWB_fixed_Asm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNqWB_register_Asm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3LNqWB_register_Asm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3dAsm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3dAsm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3dAsm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3dWB_fixed_Asm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3dWB_fixed_Asm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3dWB_fixed_Asm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3dWB_register_Asm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3dWB_register_Asm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3dWB_register_Asm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3qAsm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3qAsm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3qAsm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3qWB_fixed_Asm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3qWB_fixed_Asm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3qWB_fixed_Asm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3qWB_register_Asm_16, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3qWB_register_Asm_32, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD3qWB_register_Asm_8, ARM_INS_VLD3: vld3 */
	{ 0 }
},

{	/* ARM_VLD4DUPdAsm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPdAsm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPdAsm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPdWB_fixed_Asm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPdWB_fixed_Asm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPdWB_fixed_Asm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPdWB_register_Asm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPdWB_register_Asm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPdWB_register_Asm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPqAsm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPqAsm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPqAsm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPqWB_fixed_Asm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPqWB_fixed_Asm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPqWB_fixed_Asm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPqWB_register_Asm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPqWB_register_Asm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4DUPqWB_register_Asm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNdAsm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNdAsm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNdAsm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNdWB_fixed_Asm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNdWB_fixed_Asm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNdWB_fixed_Asm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNdWB_register_Asm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNdWB_register_Asm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNdWB_register_Asm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNqAsm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNqAsm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNqWB_fixed_Asm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNqWB_fixed_Asm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNqWB_register_Asm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4LNqWB_register_Asm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4dAsm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4dAsm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4dAsm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4dWB_fixed_Asm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4dWB_fixed_Asm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4dWB_fixed_Asm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4dWB_register_Asm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4dWB_register_Asm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4dWB_register_Asm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4qAsm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4qAsm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4qAsm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4qWB_fixed_Asm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4qWB_fixed_Asm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4qWB_fixed_Asm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4qWB_register_Asm_16, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4qWB_register_Asm_32, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VLD4qWB_register_Asm_8, ARM_INS_VLD4: vld4 */
	{ 0 }
},

{	/* ARM_VST1LNdAsm_16, ARM_INS_VST1: vst1 */
	{ 0 }
},

{	/* ARM_VST1LNdAsm_32, ARM_INS_VST1: vst1 */
	{ 0 }
},

{	/* ARM_VST1LNdAsm_8, ARM_INS_VST1: vst1 */
	{ 0 }
},

{	/* ARM_VST1LNdWB_fixed_Asm_16, ARM_INS_VST1: vst1 */
	{ 0 }
},

{	/* ARM_VST1LNdWB_fixed_Asm_32, ARM_INS_VST1: vst1 */
	{ 0 }
},

{	/* ARM_VST1LNdWB_fixed_Asm_8, ARM_INS_VST1: vst1 */
	{ 0 }
},

{	/* ARM_VST1LNdWB_register_Asm_16, ARM_INS_VST1: vst1 */
	{ 0 }
},

{	/* ARM_VST1LNdWB_register_Asm_32, ARM_INS_VST1: vst1 */
	{ 0 }
},

{	/* ARM_VST1LNdWB_register_Asm_8, ARM_INS_VST1: vst1 */
	{ 0 }
},

{	/* ARM_VST2LNdAsm_16, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST2LNdAsm_32, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST2LNdAsm_8, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST2LNdWB_fixed_Asm_16, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST2LNdWB_fixed_Asm_32, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST2LNdWB_fixed_Asm_8, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST2LNdWB_register_Asm_16, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST2LNdWB_register_Asm_32, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST2LNdWB_register_Asm_8, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST2LNqAsm_16, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST2LNqAsm_32, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST2LNqWB_fixed_Asm_16, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST2LNqWB_fixed_Asm_32, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST2LNqWB_register_Asm_16, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST2LNqWB_register_Asm_32, ARM_INS_VST2: vst2 */
	{ 0 }
},

{	/* ARM_VST3LNdAsm_16, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3LNdAsm_32, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3LNdAsm_8, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3LNdWB_fixed_Asm_16, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3LNdWB_fixed_Asm_32, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3LNdWB_fixed_Asm_8, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3LNdWB_register_Asm_16, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3LNdWB_register_Asm_32, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3LNdWB_register_Asm_8, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3LNqAsm_16, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3LNqAsm_32, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3LNqWB_fixed_Asm_16, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3LNqWB_fixed_Asm_32, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3LNqWB_register_Asm_16, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3LNqWB_register_Asm_32, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3dAsm_16, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3dAsm_32, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3dAsm_8, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3dWB_fixed_Asm_16, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3dWB_fixed_Asm_32, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3dWB_fixed_Asm_8, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3dWB_register_Asm_16, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3dWB_register_Asm_32, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3dWB_register_Asm_8, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3qAsm_16, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3qAsm_32, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3qAsm_8, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3qWB_fixed_Asm_16, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3qWB_fixed_Asm_32, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3qWB_fixed_Asm_8, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3qWB_register_Asm_16, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3qWB_register_Asm_32, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST3qWB_register_Asm_8, ARM_INS_VST3: vst3 */
	{ 0 }
},

{	/* ARM_VST4LNdAsm_16, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4LNdAsm_32, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4LNdAsm_8, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4LNdWB_fixed_Asm_16, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4LNdWB_fixed_Asm_32, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4LNdWB_fixed_Asm_8, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4LNdWB_register_Asm_16, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4LNdWB_register_Asm_32, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4LNdWB_register_Asm_8, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4LNqAsm_16, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4LNqAsm_32, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4LNqWB_fixed_Asm_16, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4LNqWB_fixed_Asm_32, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4LNqWB_register_Asm_16, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4LNqWB_register_Asm_32, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4dAsm_16, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4dAsm_32, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4dAsm_8, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4dWB_fixed_Asm_16, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4dWB_fixed_Asm_32, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4dWB_fixed_Asm_8, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4dWB_register_Asm_16, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4dWB_register_Asm_32, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4dWB_register_Asm_8, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4qAsm_16, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4qAsm_32, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4qAsm_8, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4qWB_fixed_Asm_16, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4qWB_fixed_Asm_32, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4qWB_fixed_Asm_8, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4qWB_register_Asm_16, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4qWB_register_Asm_32, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_VST4qWB_register_Asm_8, ARM_INS_VST4: vst4 */
	{ 0 }
},

{	/* ARM_t2LDRBpcrel, ARM_INS_LDRB: ldrb */
	{ 0 }
},

{	/* ARM_t2LDRConstPool, ARM_INS_LDR: ldr */
	{ 0 }
},

{	/* ARM_t2LDRHpcrel, ARM_INS_LDRH: ldrh */
	{ 0 }
},

{	/* ARM_t2LDRSBpcrel, ARM_INS_LDRSB: ldrsb */
	{ 0 }
},

{	/* ARM_t2LDRSHpcrel, ARM_INS_LDRSH: ldrsh */
	{ 0 }
},

{	/* ARM_t2LDRpcrel, ARM_INS_LDR: ldr */
	{ 0 }
},

{	/* ARM_t2MOVSsi, ARM_INS_MOVS: movs */
	{ 0 }
},

{	/* ARM_t2MOVSsr, ARM_INS_MOVS: movs */
	{ 0 }
},

{	/* ARM_t2MOVsi, ARM_INS_MOV: mov */
	{ 0 }
},

{	/* ARM_t2MOVsr, ARM_INS_MOV: mov */
	{ 0 }
},

{	/* ARM_tLDRConstPool, ARM_INS_LDR: ldr */
	{ 0 }
},

{	/* ARM_ADCri, ARM_INS_ADC: adc */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_ADCrr, ARM_INS_ADC: adc */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_ADCrsi, ARM_INS_ADC: adc */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_ADCrsr, ARM_INS_ADC: adc */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_ADDri, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_ADDrr, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_ADDrsi, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_ADDrsr, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_ADR, ARM_INS_ADR: adr */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_AESD, ARM_INS_AESD: aesd */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_AESE, ARM_INS_AESE: aese */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_AESIMC, ARM_INS_AESIMC: aesimc */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_AESMC, ARM_INS_AESMC: aesmc */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_ANDri, ARM_INS_AND: and */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_ANDrr, ARM_INS_AND: and */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_ANDrsi, ARM_INS_AND: and */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_ANDrsr, ARM_INS_AND: and */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_BFC, ARM_INS_BFC: bfc */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_BFI, ARM_INS_BFI: bfi */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_BICri, ARM_INS_AND: and */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_BICrr, ARM_INS_BIC: bic */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_BICrsi, ARM_INS_BIC: bic */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_BICrsr, ARM_INS_BIC: bic */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_BKPT, ARM_INS_BKPT: bkpt */
	{ 0 }
},

{	/* ARM_BL, ARM_INS_BL: bl */
	{ 0 }
},

{	/* ARM_BLX, ARM_INS_BLX: blx */
	{ CS_AC_READ, 0 }
},

{	/* ARM_BLX_pred, ARM_INS_BLX: blx */
	{ CS_AC_READ, 0 }
},

{	/* ARM_BLXi, ARM_INS_BLX: blx */
	{ 0 }
},

{	/* ARM_BL_pred, ARM_INS_BL: bl */
	{ 0 }
},

{	/* ARM_BX, ARM_INS_BX: bx */
	{ CS_AC_READ, 0 }
},

{	/* ARM_BXJ, ARM_INS_BXJ: bxj */
	{ CS_AC_READ, 0 }
},

{	/* ARM_BX_RET, ARM_INS_BX: bx */
	{ 0 }
},

{	/* ARM_BX_pred, ARM_INS_BX: bx */
	{ CS_AC_READ, 0 }
},

{	/* ARM_Bcc, ARM_INS_B: b */
	{ 0 }
},

{	/* ARM_CDP, ARM_INS_CDP: cdp */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_IGNORE, 0 }
},

{	/* ARM_CDP2, ARM_INS_CDP2: cdp2 */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_IGNORE, 0 }
},

{	/* ARM_CLREX, ARM_INS_CLREX: clrex */
	{ 0 }
},

{	/* ARM_CLZ, ARM_INS_CLZ: clz */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_CMNri, ARM_INS_CMN: cmn */
	{ CS_AC_READ, 0 }
},

{	/* ARM_CMNzrr, ARM_INS_CMN: cmn */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_CMNzrsi, ARM_INS_CMN: cmn */
	{ CS_AC_READ, 0 }
},

{	/* ARM_CMNzrsr, ARM_INS_CMN: cmn */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_CMPri, ARM_INS_CMN: cmn */
	{ CS_AC_READ, 0 }
},

{	/* ARM_CMPrr, ARM_INS_CMP: cmp */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_CMPrsi, ARM_INS_CMP: cmp */
	{ CS_AC_READ, 0 }
},

{	/* ARM_CMPrsr, ARM_INS_CMP: cmp */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_CPS1p, ARM_INS_CPS: cps */
	{ 0 }
},

{	/* ARM_CPS2p, ARM_INS_CPS: cps */
	{ 0 }
},

{	/* ARM_CPS3p, ARM_INS_CPS: cps */
	{ 0 }
},

{	/* ARM_CRC32B, ARM_INS_CRC32B: crc32b */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_CRC32CB, ARM_INS_CRC32CB: crc32cb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_CRC32CH, ARM_INS_CRC32CH: crc32ch */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_CRC32CW, ARM_INS_CRC32CW: crc32cw */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_CRC32H, ARM_INS_CRC32H: crc32h */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_CRC32W, ARM_INS_CRC32W: crc32w */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_DBG, ARM_INS_DBG: dbg */
	{ 0 }
},

{	/* ARM_DMB, ARM_INS_DMB: dmb */
	{ 0 }
},

{	/* ARM_DSB, ARM_INS_DFB: dfb */
	{ 0 }
},

{	/* ARM_EORri, ARM_INS_EOR: eor */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_EORrr, ARM_INS_EOR: eor */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_EORrsi, ARM_INS_EOR: eor */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_EORrsr, ARM_INS_EOR: eor */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_ERET, ARM_INS_ERET: eret */
	{ 0 }
},

{	/* ARM_FCONSTD, ARM_INS_FCONSTD: fconstd */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_FCONSTH, ARM_INS_VMOV: vmov */
	{ 0 }
},

{	/* ARM_FCONSTS, ARM_INS_FCONSTS: fconsts */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_FLDMXDB_UPD, ARM_INS_FLDMDBX: fldmdbx */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_FLDMXIA, ARM_INS_FLDMIAX: fldmiax */
	{ CS_AC_READ, 0 }
},

{	/* ARM_FLDMXIA_UPD, ARM_INS_FLDMIAX: fldmiax */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_FMSTAT, ARM_INS_FMSTAT: fmstat */
	{ 0 }
},

{	/* ARM_FSTMXDB_UPD, ARM_INS_FSTMDBX: fstmdbx */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_FSTMXIA, ARM_INS_FSTMIAX: fstmiax */
	{ CS_AC_READ, 0 }
},

{	/* ARM_FSTMXIA_UPD, ARM_INS_FSTMIAX: fstmiax */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_HINT, ARM_INS_CSDB: csdb */
	{ 0 }
},

{	/* ARM_HLT, ARM_INS_HLT: hlt */
	{ 0 }
},

{	/* ARM_HVC, ARM_INS_HVC: hvc */
	{ 0 }
},

{	/* ARM_ISB, ARM_INS_ISB: isb */
	{ 0 }
},

{	/* ARM_LDA, ARM_INS_LDA: lda */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDAB, ARM_INS_LDAB: ldab */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDAEX, ARM_INS_LDAEX: ldaex */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDAEXB, ARM_INS_LDAEXB: ldaexb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDAEXD, ARM_INS_LDAEXD: ldaexd */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDAEXH, ARM_INS_LDAEXH: ldaexh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDAH, ARM_INS_LDAH: ldah */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDC2L_OFFSET, ARM_INS_LDC2L: ldc2l */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDC2L_OPTION, ARM_INS_LDC2L: ldc2l */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDC2L_POST, ARM_INS_LDC2L: ldc2l */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDC2L_PRE, ARM_INS_LDC2L: ldc2l */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDC2_OFFSET, ARM_INS_LDC2: ldc2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDC2_OPTION, ARM_INS_LDC2: ldc2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDC2_POST, ARM_INS_LDC2: ldc2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDC2_PRE, ARM_INS_LDC2: ldc2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDCL_OFFSET, ARM_INS_LDCL: ldcl */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDCL_OPTION, ARM_INS_LDCL: ldcl */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDCL_POST, ARM_INS_LDCL: ldcl */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDCL_PRE, ARM_INS_LDCL: ldcl */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDC_OFFSET, ARM_INS_LDC: ldc */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDC_OPTION, ARM_INS_LDC: ldc */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDC_POST, ARM_INS_LDC: ldc */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDC_PRE, ARM_INS_LDC: ldc */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDMDA, ARM_INS_LDMDA: ldmda */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_LDMDA_UPD, ARM_INS_LDMDA: ldmda */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_LDMDB, ARM_INS_LDMDB: ldmdb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_LDMDB_UPD, ARM_INS_LDMDB: ldmdb */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_LDMIA, ARM_INS_LDM: ldm */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_LDMIA_UPD, ARM_INS_LDM: ldm */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_LDMIB, ARM_INS_LDMIB: ldmib */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_LDMIB_UPD, ARM_INS_LDMIB: ldmib */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_LDRBT_POST_IMM, ARM_INS_LDRBT: ldrbt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRBT_POST_REG, ARM_INS_LDRBT: ldrbt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRB_POST_IMM, ARM_INS_LDRB: ldrb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRB_POST_REG, ARM_INS_LDRB: ldrb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRB_PRE_IMM, ARM_INS_LDRB: ldrb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRB_PRE_REG, ARM_INS_LDRB: ldrb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRBi12, ARM_INS_LDRB: ldrb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRBrs, ARM_INS_LDRB: ldrb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRD, ARM_INS_LDRD: ldrd */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_LDRD_POST, ARM_INS_LDRD: ldrd */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRD_PRE, ARM_INS_LDRD: ldrd */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_LDREX, ARM_INS_LDREX: ldrex */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDREXB, ARM_INS_LDREXB: ldrexb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDREXD, ARM_INS_LDREXD: ldrexd */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDREXH, ARM_INS_LDREXH: ldrexh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRH, ARM_INS_LDRH: ldrh */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_LDRHTi, ARM_INS_LDRHT: ldrht */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRHTr, ARM_INS_LDRHT: ldrht */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDRH_POST, ARM_INS_LDRH: ldrh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRH_PRE, ARM_INS_LDRH: ldrh */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_LDRSB, ARM_INS_LDRSB: ldrsb */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_LDRSBTi, ARM_INS_LDRSBT: ldrsbt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRSBTr, ARM_INS_LDRSBT: ldrsbt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDRSB_POST, ARM_INS_LDRSB: ldrsb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRSB_PRE, ARM_INS_LDRSB: ldrsb */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_LDRSH, ARM_INS_LDRSH: ldrsh */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_LDRSHTi, ARM_INS_LDRSHT: ldrsht */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRSHTr, ARM_INS_LDRSHT: ldrsht */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_LDRSH_POST, ARM_INS_LDRSH: ldrsh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRSH_PRE, ARM_INS_LDRSH: ldrsh */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_LDRT_POST_IMM, ARM_INS_LDRT: ldrt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRT_POST_REG, ARM_INS_LDRT: ldrt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDR_POST_IMM, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDR_POST_REG, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDR_PRE_IMM, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDR_PRE_REG, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRcp,  ARM_INS_LDR: ldr${p}	$rt $addr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRi12, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_LDRrs, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_MCR, ARM_INS_MCR: mcr */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_IGNORE, 0 }
},

{	/* ARM_MCR2, ARM_INS_MCR2: mcr2 */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_IGNORE, 0 }
},

{	/* ARM_MCRR, ARM_INS_MCRR: mcrr */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_MCRR2, ARM_INS_MCRR2: mcrr2 */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_MLA, ARM_INS_MLA: mla */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_MLS, ARM_INS_MLS: mls */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_MOVPCLR, ARM_INS_MOV: mov */
	{ 0 }
},

{	/* ARM_MOVTi16, ARM_INS_MOVT: movt */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_MOVi, ARM_INS_MOV: mov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_MOVi16, ARM_INS_MOV: mov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_MOVr, ARM_INS_MOV: mov */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_MOVr_TC, ARM_INS_MOV: mov */
	{ 0 }
},

{	/* ARM_MOVsi, ARM_INS_MOV: mov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_MOVsr, ARM_INS_MOV: mov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_MRC, ARM_INS_MRC: mrc */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_IGNORE, 0 }
},

{	/* ARM_MRC2, ARM_INS_MRC2: mrc2 */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_IGNORE, 0 }
},

{	/* ARM_MRRC, ARM_INS_MRRC: mrrc */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_MRRC2, ARM_INS_MRRC2: mrrc2 */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_MRS, ARM_INS_MRS: mrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_MRSbanked, ARM_INS_MRS: mrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_MRSsys, ARM_INS_MRS: mrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_MSR, ARM_INS_MSR: msr */
	{ CS_AC_READ, 0 }
},

{	/* ARM_MSRbanked, ARM_INS_MSR: msr */
	{ CS_AC_READ, 0 }
},

{	/* ARM_MSRi, ARM_INS_MSR: msr */
	{ 0 }
},

{	/* ARM_MUL, ARM_INS_MUL: mul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_MVNi, ARM_INS_MOV: mov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_MVNr, ARM_INS_MVN: mvn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_MVNsi, ARM_INS_MVN: mvn */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_MVNsr, ARM_INS_MVN: mvn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_ORRri, ARM_INS_ORR: orr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_ORRrr, ARM_INS_ORR: orr */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_ORRrsi, ARM_INS_ORR: orr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_ORRrsr, ARM_INS_ORR: orr */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_PKHBT, ARM_INS_PKHBT: pkhbt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_PKHTB, ARM_INS_PKHTB: pkhtb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_PLDWi12, ARM_INS_PLDW: pldw */
	{ CS_AC_READ, 0 }
},

{	/* ARM_PLDWrs, ARM_INS_PLDW: pldw */
	{ CS_AC_READ, 0 }
},

{	/* ARM_PLDi12, ARM_INS_PLD: pld */
	{ CS_AC_READ, 0 }
},

{	/* ARM_PLDrs, ARM_INS_PLD: pld */
	{ CS_AC_READ, 0 }
},

{	/* ARM_PLIi12, ARM_INS_PLI: pli */
	{ CS_AC_READ, 0 }
},

{	/* ARM_PLIrs, ARM_INS_PLI: pli */
	{ CS_AC_READ, 0 }
},

{	/* ARM_QADD, ARM_INS_QADD: qadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_QADD16, ARM_INS_QADD16: qadd16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_QADD8, ARM_INS_QADD8: qadd8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_QASX, ARM_INS_QASX: qasx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_QDADD, ARM_INS_QDADD: qdadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_QDSUB, ARM_INS_QDSUB: qdsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_QSAX, ARM_INS_QSAX: qsax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_QSUB, ARM_INS_QSUB: qsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_QSUB16, ARM_INS_QSUB16: qsub16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_QSUB8, ARM_INS_QSUB8: qsub8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_RBIT, ARM_INS_RBIT: rbit */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_REV, ARM_INS_REV: rev */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_REV16, ARM_INS_REV16: rev16 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_REVSH, ARM_INS_REVSH: revsh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_RFEDA, ARM_INS_RFEDA: rfeda */
	{ CS_AC_READ, 0 }
},

{	/* ARM_RFEDA_UPD, ARM_INS_RFEDA: rfeda */
	{ CS_AC_READ, 0 }
},

{	/* ARM_RFEDB, ARM_INS_RFEDB: rfedb */
	{ CS_AC_READ, 0 }
},

{	/* ARM_RFEDB_UPD, ARM_INS_RFEDB: rfedb */
	{ CS_AC_READ, 0 }
},

{	/* ARM_RFEIA, ARM_INS_RFEIA: rfeia */
	{ CS_AC_READ, 0 }
},

{	/* ARM_RFEIA_UPD, ARM_INS_RFEIA: rfeia */
	{ CS_AC_READ, 0 }
},

{	/* ARM_RFEIB, ARM_INS_RFEIB: rfeib */
	{ CS_AC_READ, 0 }
},

{	/* ARM_RFEIB_UPD, ARM_INS_RFEIB: rfeib */
	{ CS_AC_READ, 0 }
},

{	/* ARM_RSBri, ARM_INS_NEG: neg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_RSBrr, ARM_INS_RSB: rsb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_RSBrsi, ARM_INS_RSB: rsb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_RSBrsr, ARM_INS_RSB: rsb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_RSCri, ARM_INS_RSC: rsc */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_RSCrr, ARM_INS_RSC: rsc */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_RSCrsi, ARM_INS_RSC: rsc */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_RSCrsr, ARM_INS_RSC: rsc */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SADD16, ARM_INS_SADD16: sadd16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SADD8, ARM_INS_SADD8: sadd8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SASX, ARM_INS_SASX: sasx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SBCri, ARM_INS_ADC: adc */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_SBCrr, ARM_INS_SBC: sbc */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SBCrsi, ARM_INS_SBC: sbc */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_SBCrsr, ARM_INS_SBC: sbc */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SBFX, ARM_INS_SBFX: sbfx */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_SDIV, ARM_INS_SDIV: sdiv */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SEL, ARM_INS_SEL: sel */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SETEND, ARM_INS_SETEND: setend */
	{ 0 }
},

{	/* ARM_SETPAN, ARM_INS_SETPAN: setpan */
	{ 0 }
},

{	/* ARM_SHA1C, ARM_INS_SHA1C: sha1c */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SHA1H, ARM_INS_SHA1H: sha1h */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_SHA1M, ARM_INS_SHA1M: sha1m */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SHA1P, ARM_INS_SHA1P: sha1p */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SHA1SU0, ARM_INS_SHA1SU0: sha1su0 */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SHA1SU1, ARM_INS_SHA1SU1: sha1su1 */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_SHA256H, ARM_INS_SHA256H: sha256h */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SHA256H2, ARM_INS_SHA256H2: sha256h2 */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SHA256SU0, ARM_INS_SHA256SU0: sha256su0 */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_SHA256SU1, ARM_INS_SHA256SU1: sha256su1 */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SHADD16, ARM_INS_SHADD16: shadd16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SHADD8, ARM_INS_SHADD8: shadd8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SHASX, ARM_INS_SHASX: shasx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SHSAX, ARM_INS_SHSAX: shsax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SHSUB16, ARM_INS_SHSUB16: shsub16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SHSUB8, ARM_INS_SHSUB8: shsub8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMC, ARM_INS_SMC: smc */
	{ 0 }
},

{	/* ARM_SMLABB, ARM_INS_SMLABB: smlabb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLABT, ARM_INS_SMLABT: smlabt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLAD, ARM_INS_SMLAD: smlad */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLADX, ARM_INS_SMLADX: smladx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLAL, ARM_INS_SMLAL: smlal */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLALBB, ARM_INS_SMLALBB: smlalbb */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLALBT, ARM_INS_SMLALBT: smlalbt */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLALD, ARM_INS_SMLALD: smlald */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLALDX, ARM_INS_SMLALDX: smlaldx */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLALTB, ARM_INS_SMLALTB: smlaltb */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLALTT, ARM_INS_SMLALTT: smlaltt */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLATB, ARM_INS_SMLATB: smlatb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLATT, ARM_INS_SMLATT: smlatt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLAWB, ARM_INS_SMLAWB: smlawb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLAWT, ARM_INS_SMLAWT: smlawt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLSD, ARM_INS_SMLSD: smlsd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLSDX, ARM_INS_SMLSDX: smlsdx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLSLD, ARM_INS_SMLSLD: smlsld */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMLSLDX, ARM_INS_SMLSLDX: smlsldx */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMMLA, ARM_INS_SMMLA: smmla */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMMLAR, ARM_INS_SMMLAR: smmlar */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMMLS, ARM_INS_SMMLS: smmls */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMMLSR, ARM_INS_SMMLSR: smmlsr */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMMUL, ARM_INS_SMMUL: smmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMMULR, ARM_INS_SMMULR: smmulr */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMUAD, ARM_INS_SMUAD: smuad */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMUADX, ARM_INS_SMUADX: smuadx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMULBB, ARM_INS_SMULBB: smulbb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMULBT, ARM_INS_SMULBT: smulbt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMULL, ARM_INS_SMULL: smull */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMULTB, ARM_INS_SMULTB: smultb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMULTT, ARM_INS_SMULTT: smultt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMULWB, ARM_INS_SMULWB: smulwb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMULWT, ARM_INS_SMULWT: smulwt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMUSD, ARM_INS_SMUSD: smusd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SMUSDX, ARM_INS_SMUSDX: smusdx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SRSDA, ARM_INS_SRSDA: srsda */
	{ 0 }
},

{	/* ARM_SRSDA_UPD, ARM_INS_SRSDA: srsda */
	{ 0 }
},

{	/* ARM_SRSDB, ARM_INS_SRSDB: srsdb */
	{ 0 }
},

{	/* ARM_SRSDB_UPD, ARM_INS_SRSDB: srsdb */
	{ 0 }
},

{	/* ARM_SRSIA, ARM_INS_SRSIA: srsia */
	{ 0 }
},

{	/* ARM_SRSIA_UPD, ARM_INS_SRSIA: srsia */
	{ 0 }
},

{	/* ARM_SRSIB, ARM_INS_SRSIB: srsib */
	{ 0 }
},

{	/* ARM_SRSIB_UPD, ARM_INS_SRSIB: srsib */
	{ 0 }
},

{	/* ARM_SSAT, ARM_INS_SSAT: ssat */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_SSAT16, ARM_INS_SSAT16: ssat16 */
	{ CS_AC_WRITE, CS_AC_IGNORE, CS_AC_READ, 0 }
},

{	/* ARM_SSAX, ARM_INS_SSAX: ssax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SSUB16, ARM_INS_SSUB16: ssub16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SSUB8, ARM_INS_SSUB8: ssub8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STC2L_OFFSET, ARM_INS_STC2L: stc2l */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STC2L_OPTION, ARM_INS_STC2L: stc2l */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STC2L_POST, ARM_INS_STC2L: stc2l */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STC2L_PRE, ARM_INS_STC2L: stc2l */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STC2_OFFSET, ARM_INS_STC2: stc2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STC2_OPTION, ARM_INS_STC2: stc2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STC2_POST, ARM_INS_STC2: stc2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STC2_PRE, ARM_INS_STC2: stc2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STCL_OFFSET, ARM_INS_STCL: stcl */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STCL_OPTION, ARM_INS_STCL: stcl */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STCL_POST, ARM_INS_STCL: stcl */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STCL_PRE, ARM_INS_STCL: stcl */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STC_OFFSET, ARM_INS_STC: stc */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STC_OPTION, ARM_INS_STC: stc */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STC_POST, ARM_INS_STC: stc */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STC_PRE, ARM_INS_STC: stc */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STL, ARM_INS_STL: stl */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STLB, ARM_INS_STLB: stlb */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STLEX, ARM_INS_STLEX: stlex */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STLEXB, ARM_INS_STLEXB: stlexb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STLEXD, ARM_INS_STLEXD: stlexd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STLEXH, ARM_INS_STLEXH: stlexh */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STLH, ARM_INS_STLH: stlh */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STMDA, ARM_INS_STMDA: stmda */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STMDA_UPD, ARM_INS_STMDA: stmda */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_STMDB, ARM_INS_STMDB: stmdb */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STMDB_UPD, ARM_INS_PUSH: push */
	{ CS_AC_READ, 0 }
},

{	/* ARM_STMIA, ARM_INS_STM: stm */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STMIA_UPD, ARM_INS_STM: stm */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_STMIB, ARM_INS_STMIB: stmib */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STMIB_UPD, ARM_INS_STMIB: stmib */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_STRBT_POST_IMM, ARM_INS_STRBT: strbt */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STRBT_POST_REG, ARM_INS_STRBT: strbt */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STRB_POST_IMM, ARM_INS_STRB: strb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STRB_POST_REG, ARM_INS_STRB: strb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STRB_PRE_IMM, ARM_INS_STRB: strb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STRB_PRE_REG, ARM_INS_STRB: strb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STRBi12, ARM_INS_STRB: strb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STRBrs, ARM_INS_STRB: strb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STRD, ARM_INS_STRD: strd */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STRD_POST, ARM_INS_STRD: strd */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STRD_PRE, ARM_INS_STRD: strd */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STREX, ARM_INS_STREX: strex */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STREXB, ARM_INS_STREXB: strexb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STREXD, ARM_INS_STREXD: strexd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STREXH, ARM_INS_STREXH: strexh */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STRH, ARM_INS_STRH: strh */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STRHTi, ARM_INS_STRHT: strht */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STRHTr, ARM_INS_STRHT: strht */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STRH_POST, ARM_INS_STRH: strh */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_STRH_PRE, ARM_INS_STRH: strh */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STRT_POST_IMM, ARM_INS_STRT: strt */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STRT_POST_REG, ARM_INS_STRT: strt */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STR_POST_IMM, ARM_INS_STR: str */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STR_POST_REG, ARM_INS_STR: str */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STR_PRE_IMM, ARM_INS_STR: str */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STR_PRE_REG, ARM_INS_STR: str */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STRi12, ARM_INS_STR: str */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_STRrs, ARM_INS_STR: str */
	{ CS_AC_READ, 0 }
},

{	/* ARM_SUBri, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_SUBrr, ARM_INS_SUB: sub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SUBrsi, ARM_INS_SUB: sub */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_SUBrsr, ARM_INS_SUB: sub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SVC, ARM_INS_SVC: svc */
	{ 0 }
},

{	/* ARM_SWP, ARM_INS_SWP: swp */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SWPB, ARM_INS_SWPB: swpb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_SXTAB, ARM_INS_SXTAB: sxtab */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_SXTAB16, ARM_INS_SXTAB16: sxtab16 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_SXTAH, ARM_INS_SXTAH: sxtah */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_SXTB, ARM_INS_SXTB: sxtb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_SXTB16, ARM_INS_SXTB16: sxtb16 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_SXTH, ARM_INS_SXTH: sxth */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_TEQri, ARM_INS_TEQ: teq */
	{ CS_AC_READ, 0 }
},

{	/* ARM_TEQrr, ARM_INS_TEQ: teq */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_TEQrsi, ARM_INS_TEQ: teq */
	{ CS_AC_READ, 0 }
},

{	/* ARM_TEQrsr, ARM_INS_TEQ: teq */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_TRAP, ARM_INS_TRAP: trap */
	{ 0 }
},

{	/* ARM_TRAPNaCl, ARM_INS_TRAP: trap */
	{ 0 }
},

{	/* ARM_TSB, ARM_INS_TSB: tsb */
	{ 0 }
},

{	/* ARM_TSTri, ARM_INS_TST: tst */
	{ CS_AC_READ, 0 }
},

{	/* ARM_TSTrr, ARM_INS_TST: tst */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_TSTrsi, ARM_INS_TST: tst */
	{ CS_AC_READ, 0 }
},

{	/* ARM_TSTrsr, ARM_INS_TST: tst */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UADD16, ARM_INS_UADD16: uadd16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UADD8, ARM_INS_UADD8: uadd8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UASX, ARM_INS_UASX: uasx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UBFX, ARM_INS_UBFX: ubfx */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_UDF, ARM_INS_UDF: udf */
	{ 0 }
},

{	/* ARM_UDIV, ARM_INS_UDIV: udiv */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UHADD16, ARM_INS_UHADD16: uhadd16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UHADD8, ARM_INS_UHADD8: uhadd8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UHASX, ARM_INS_UHASX: uhasx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UHSAX, ARM_INS_UHSAX: uhsax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UHSUB16, ARM_INS_UHSUB16: uhsub16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UHSUB8, ARM_INS_UHSUB8: uhsub8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UMAAL, ARM_INS_UMAAL: umaal */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UMLAL, ARM_INS_UMLAL: umlal */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UMULL, ARM_INS_UMULL: umull */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UQADD16, ARM_INS_UQADD16: uqadd16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UQADD8, ARM_INS_UQADD8: uqadd8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UQASX, ARM_INS_UQASX: uqasx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UQSAX, ARM_INS_UQSAX: uqsax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UQSUB16, ARM_INS_UQSUB16: uqsub16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UQSUB8, ARM_INS_UQSUB8: uqsub8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_USAD8, ARM_INS_USAD8: usad8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_USADA8, ARM_INS_USADA8: usada8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_USAT, ARM_INS_USAT: usat */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_USAT16, ARM_INS_USAT16: usat16 */
	{ CS_AC_WRITE, CS_AC_IGNORE, CS_AC_READ, 0 }
},

{	/* ARM_USAX, ARM_INS_USAX: usax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_USUB16, ARM_INS_USUB16: usub16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_USUB8, ARM_INS_USUB8: usub8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_UXTAB, ARM_INS_UXTAB: uxtab */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_UXTAB16, ARM_INS_UXTAB16: uxtab16 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_UXTAH, ARM_INS_UXTAH: uxtah */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_UXTB, ARM_INS_UXTB: uxtb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_UXTB16, ARM_INS_UXTB16: uxtb16 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_UXTH, ARM_INS_UXTH: uxth */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VABALsv2i64, ARM_INS_VABAL: vabal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABALsv4i32, ARM_INS_VABAL: vabal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABALsv8i16, ARM_INS_VABAL: vabal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABALuv2i64, ARM_INS_VABAL: vabal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABALuv4i32, ARM_INS_VABAL: vabal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABALuv8i16, ARM_INS_VABAL: vabal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABAsv16i8, ARM_INS_VABA: vaba */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABAsv2i32, ARM_INS_VABA: vaba */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABAsv4i16, ARM_INS_VABA: vaba */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABAsv4i32, ARM_INS_VABA: vaba */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABAsv8i16, ARM_INS_VABA: vaba */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABAsv8i8, ARM_INS_VABA: vaba */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABAuv16i8, ARM_INS_VABA: vaba */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABAuv2i32, ARM_INS_VABA: vaba */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABAuv4i16, ARM_INS_VABA: vaba */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABAuv4i32, ARM_INS_VABA: vaba */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABAuv8i16, ARM_INS_VABA: vaba */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABAuv8i8, ARM_INS_VABA: vaba */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDLsv2i64, ARM_INS_VABDL: vabdl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDLsv4i32, ARM_INS_VABDL: vabdl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDLsv8i16, ARM_INS_VABDL: vabdl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDLuv2i64, ARM_INS_VABDL: vabdl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDLuv4i32, ARM_INS_VABDL: vabdl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDLuv8i16, ARM_INS_VABDL: vabdl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDfd, ARM_INS_VABD: vabd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDfq, ARM_INS_VABD: vabd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDhd, ARM_INS_VABD: vabd */
	{ 0 }
},

{	/* ARM_VABDhq, ARM_INS_VABD: vabd */
	{ 0 }
},

{	/* ARM_VABDsv16i8, ARM_INS_VABD: vabd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDsv2i32, ARM_INS_VABD: vabd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDsv4i16, ARM_INS_VABD: vabd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDsv4i32, ARM_INS_VABD: vabd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDsv8i16, ARM_INS_VABD: vabd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDsv8i8, ARM_INS_VABD: vabd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDuv16i8, ARM_INS_VABD: vabd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDuv2i32, ARM_INS_VABD: vabd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDuv4i16, ARM_INS_VABD: vabd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDuv4i32, ARM_INS_VABD: vabd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDuv8i16, ARM_INS_VABD: vabd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABDuv8i8, ARM_INS_VABD: vabd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VABSD, ARM_INS_VABS: vabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VABSH, ARM_INS_VABS: vabs */
	{ 0 }
},

{	/* ARM_VABSS, ARM_INS_VABS: vabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VABSfd, ARM_INS_VABS: vabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VABSfq, ARM_INS_VABS: vabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VABShd, ARM_INS_VABS: vabs */
	{ 0 }
},

{	/* ARM_VABShq, ARM_INS_VABS: vabs */
	{ 0 }
},

{	/* ARM_VABSv16i8, ARM_INS_VABS: vabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VABSv2i32, ARM_INS_VABS: vabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VABSv4i16, ARM_INS_VABS: vabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VABSv4i32, ARM_INS_VABS: vabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VABSv8i16, ARM_INS_VABS: vabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VABSv8i8, ARM_INS_VABS: vabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VACGEfd, ARM_INS_VACGE: vacge */
	{ 0 }
},

{	/* ARM_VACGEfq, ARM_INS_VACGE: vacge */
	{ 0 }
},

{	/* ARM_VACGEhd, ARM_INS_VACGE: vacge */
	{ 0 }
},

{	/* ARM_VACGEhq, ARM_INS_VACGE: vacge */
	{ 0 }
},

{	/* ARM_VACGTfd, ARM_INS_VACGT: vacgt */
	{ 0 }
},

{	/* ARM_VACGTfq, ARM_INS_VACGT: vacgt */
	{ 0 }
},

{	/* ARM_VACGThd, ARM_INS_VACGT: vacgt */
	{ 0 }
},

{	/* ARM_VACGThq, ARM_INS_VACGT: vacgt */
	{ 0 }
},

{	/* ARM_VADDD, ARM_INS_FADDD: faddd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDH, ARM_INS_VADD: vadd */
	{ 0 }
},

{	/* ARM_VADDHNv2i32, ARM_INS_VADDHN: vaddhn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDHNv4i16, ARM_INS_VADDHN: vaddhn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDHNv8i8, ARM_INS_VADDHN: vaddhn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDLsv2i64, ARM_INS_VADDL: vaddl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDLsv4i32, ARM_INS_VADDL: vaddl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDLsv8i16, ARM_INS_VADDL: vaddl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDLuv2i64, ARM_INS_VADDL: vaddl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDLuv4i32, ARM_INS_VADDL: vaddl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDLuv8i16, ARM_INS_VADDL: vaddl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDS, ARM_INS_FADDS: fadds */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDWsv2i64, ARM_INS_VADDW: vaddw */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDWsv4i32, ARM_INS_VADDW: vaddw */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDWsv8i16, ARM_INS_VADDW: vaddw */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDWuv2i64, ARM_INS_VADDW: vaddw */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDWuv4i32, ARM_INS_VADDW: vaddw */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDWuv8i16, ARM_INS_VADDW: vaddw */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDfd, ARM_INS_VADD: vadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDfq, ARM_INS_VADD: vadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDhd, ARM_INS_VADD: vadd */
	{ 0 }
},

{	/* ARM_VADDhq, ARM_INS_VADD: vadd */
	{ 0 }
},

{	/* ARM_VADDv16i8, ARM_INS_VADD: vadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDv1i64, ARM_INS_VADD: vadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDv2i32, ARM_INS_VADD: vadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDv2i64, ARM_INS_VADD: vadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDv4i16, ARM_INS_VADD: vadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDv4i32, ARM_INS_VADD: vadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDv8i16, ARM_INS_VADD: vadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VADDv8i8, ARM_INS_VADD: vadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VANDd, ARM_INS_VAND: vand */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VANDq, ARM_INS_VAND: vand */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VBICd, ARM_INS_VBIC: vbic */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VBICiv2i32, ARM_INS_VAND: vand */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VBICiv4i16, ARM_INS_VAND: vand */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VBICiv4i32, ARM_INS_VAND: vand */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VBICiv8i16, ARM_INS_VAND: vand */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VBICq, ARM_INS_VBIC: vbic */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VBIFd, ARM_INS_VBIF: vbif */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VBIFq, ARM_INS_VBIF: vbif */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VBITd, ARM_INS_VBIT: vbit */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VBITq, ARM_INS_VBIT: vbit */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VBSLd, ARM_INS_VBSL: vbsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VBSLq, ARM_INS_VBSL: vbsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCADDv2f32, ARM_INS_VCADD: vcadd */
	{ 0 }
},

{	/* ARM_VCADDv4f16, ARM_INS_VCADD: vcadd */
	{ 0 }
},

{	/* ARM_VCADDv4f32, ARM_INS_VCADD: vcadd */
	{ 0 }
},

{	/* ARM_VCADDv8f16, ARM_INS_VCADD: vcadd */
	{ 0 }
},

{	/* ARM_VCEQfd, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCEQfq, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCEQhd, ARM_INS_VCEQ: vceq */
	{ 0 }
},

{	/* ARM_VCEQhq, ARM_INS_VCEQ: vceq */
	{ 0 }
},

{	/* ARM_VCEQv16i8, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCEQv2i32, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCEQv4i16, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCEQv4i32, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCEQv8i16, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCEQv8i8, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCEQzv16i8, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCEQzv2f32, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCEQzv2i32, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCEQzv4f16, ARM_INS_VCEQ: vceq */
	{ 0 }
},

{	/* ARM_VCEQzv4f32, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCEQzv4i16, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCEQzv4i32, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCEQzv8f16, ARM_INS_VCEQ: vceq */
	{ 0 }
},

{	/* ARM_VCEQzv8i16, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCEQzv8i8, ARM_INS_VCEQ: vceq */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGEfd, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGEfq, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGEhd, ARM_INS_VCGE: vcge */
	{ 0 }
},

{	/* ARM_VCGEhq, ARM_INS_VCGE: vcge */
	{ 0 }
},

{	/* ARM_VCGEsv16i8, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGEsv2i32, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGEsv4i16, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGEsv4i32, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGEsv8i16, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGEsv8i8, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGEuv16i8, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGEuv2i32, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGEuv4i16, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGEuv4i32, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGEuv8i16, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGEuv8i8, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGEzv16i8, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGEzv2f32, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGEzv2i32, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGEzv4f16, ARM_INS_VCGE: vcge */
	{ 0 }
},

{	/* ARM_VCGEzv4f32, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGEzv4i16, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGEzv4i32, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGEzv8f16, ARM_INS_VCGE: vcge */
	{ 0 }
},

{	/* ARM_VCGEzv8i16, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGEzv8i8, ARM_INS_VCGE: vcge */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGTfd, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGTfq, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGThd, ARM_INS_VCGT: vcgt */
	{ 0 }
},

{	/* ARM_VCGThq, ARM_INS_VCGT: vcgt */
	{ 0 }
},

{	/* ARM_VCGTsv16i8, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGTsv2i32, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGTsv4i16, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGTsv4i32, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGTsv8i16, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGTsv8i8, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGTuv16i8, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGTuv2i32, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGTuv4i16, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGTuv4i32, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGTuv8i16, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGTuv8i8, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCGTzv16i8, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGTzv2f32, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGTzv2i32, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGTzv4f16, ARM_INS_VCGT: vcgt */
	{ 0 }
},

{	/* ARM_VCGTzv4f32, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGTzv4i16, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGTzv4i32, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGTzv8f16, ARM_INS_VCGT: vcgt */
	{ 0 }
},

{	/* ARM_VCGTzv8i16, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCGTzv8i8, ARM_INS_VCGT: vcgt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLEzv16i8, ARM_INS_VCLE: vcle */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLEzv2f32, ARM_INS_VCLE: vcle */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLEzv2i32, ARM_INS_VCLE: vcle */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLEzv4f16, ARM_INS_VCLE: vcle */
	{ 0 }
},

{	/* ARM_VCLEzv4f32, ARM_INS_VCLE: vcle */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLEzv4i16, ARM_INS_VCLE: vcle */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLEzv4i32, ARM_INS_VCLE: vcle */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLEzv8f16, ARM_INS_VCLE: vcle */
	{ 0 }
},

{	/* ARM_VCLEzv8i16, ARM_INS_VCLE: vcle */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLEzv8i8, ARM_INS_VCLE: vcle */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLSv16i8, ARM_INS_VCLS: vcls */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLSv2i32, ARM_INS_VCLS: vcls */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLSv4i16, ARM_INS_VCLS: vcls */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLSv4i32, ARM_INS_VCLS: vcls */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLSv8i16, ARM_INS_VCLS: vcls */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLSv8i8, ARM_INS_VCLS: vcls */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLTzv16i8, ARM_INS_VCLT: vclt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLTzv2f32, ARM_INS_VCLT: vclt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLTzv2i32, ARM_INS_VCLT: vclt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLTzv4f16, ARM_INS_VCLT: vclt */
	{ 0 }
},

{	/* ARM_VCLTzv4f32, ARM_INS_VCLT: vclt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLTzv4i16, ARM_INS_VCLT: vclt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLTzv4i32, ARM_INS_VCLT: vclt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLTzv8f16, ARM_INS_VCLT: vclt */
	{ 0 }
},

{	/* ARM_VCLTzv8i16, ARM_INS_VCLT: vclt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLTzv8i8, ARM_INS_VCLT: vclt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLZv16i8, ARM_INS_VCLZ: vclz */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLZv2i32, ARM_INS_VCLZ: vclz */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLZv4i16, ARM_INS_VCLZ: vclz */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLZv4i32, ARM_INS_VCLZ: vclz */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLZv8i16, ARM_INS_VCLZ: vclz */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCLZv8i8, ARM_INS_VCLZ: vclz */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCMLAv2f32, ARM_INS_VCMLA: vcmla */
	{ 0 }
},

{	/* ARM_VCMLAv2f32_indexed, ARM_INS_VCMLA: vcmla */
	{ 0 }
},

{	/* ARM_VCMLAv4f16, ARM_INS_VCMLA: vcmla */
	{ 0 }
},

{	/* ARM_VCMLAv4f16_indexed, ARM_INS_VCMLA: vcmla */
	{ 0 }
},

{	/* ARM_VCMLAv4f32, ARM_INS_VCMLA: vcmla */
	{ 0 }
},

{	/* ARM_VCMLAv4f32_indexed, ARM_INS_VCMLA: vcmla */
	{ 0 }
},

{	/* ARM_VCMLAv8f16, ARM_INS_VCMLA: vcmla */
	{ 0 }
},

{	/* ARM_VCMLAv8f16_indexed, ARM_INS_VCMLA: vcmla */
	{ 0 }
},

{	/* ARM_VCMPD, ARM_INS_VCMP: vcmp */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCMPED, ARM_INS_VCMPE: vcmpe */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCMPEH, ARM_INS_VCMPE: vcmpe */
	{ 0 }
},

{	/* ARM_VCMPES, ARM_INS_VCMPE: vcmpe */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCMPEZD, ARM_INS_VCMPE: vcmpe */
	{ CS_AC_READ, 0 }
},

{	/* ARM_VCMPEZH, ARM_INS_VCMPE: vcmpe */
	{ 0 }
},

{	/* ARM_VCMPEZS, ARM_INS_VCMPE: vcmpe */
	{ CS_AC_READ, 0 }
},

{	/* ARM_VCMPH, ARM_INS_VCMP: vcmp */
	{ 0 }
},

{	/* ARM_VCMPS, ARM_INS_VCMP: vcmp */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VCMPZD, ARM_INS_FCMPZD: fcmpzd */
	{ CS_AC_READ, 0 }
},

{	/* ARM_VCMPZH, ARM_INS_VCMP: vcmp */
	{ 0 }
},

{	/* ARM_VCMPZS, ARM_INS_FCMPZS: fcmpzs */
	{ CS_AC_READ, 0 }
},

{	/* ARM_VCNTd, ARM_INS_VCNT: vcnt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCNTq, ARM_INS_VCNT: vcnt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTANSDf, ARM_INS_VCVTA: vcvta */
	{ 0 }
},

{	/* ARM_VCVTANSDh, ARM_INS_VCVTA: vcvta */
	{ 0 }
},

{	/* ARM_VCVTANSQf, ARM_INS_VCVTA: vcvta */
	{ 0 }
},

{	/* ARM_VCVTANSQh, ARM_INS_VCVTA: vcvta */
	{ 0 }
},

{	/* ARM_VCVTANUDf, ARM_INS_VCVTA: vcvta */
	{ 0 }
},

{	/* ARM_VCVTANUDh, ARM_INS_VCVTA: vcvta */
	{ 0 }
},

{	/* ARM_VCVTANUQf, ARM_INS_VCVTA: vcvta */
	{ 0 }
},

{	/* ARM_VCVTANUQh, ARM_INS_VCVTA: vcvta */
	{ 0 }
},

{	/* ARM_VCVTASD, ARM_INS_VCVTA: vcvta */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTASH, ARM_INS_VCVTA: vcvta */
	{ 0 }
},

{	/* ARM_VCVTASS, ARM_INS_VCVTA: vcvta */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTAUD, ARM_INS_VCVTA: vcvta */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTAUH, ARM_INS_VCVTA: vcvta */
	{ 0 }
},

{	/* ARM_VCVTAUS, ARM_INS_VCVTA: vcvta */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTBDH, ARM_INS_VCVTB: vcvtb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTBHD, ARM_INS_VCVTB: vcvtb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTBHS, ARM_INS_VCVTB: vcvtb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTBSH, ARM_INS_VCVTB: vcvtb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTDS, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTMNSDf, ARM_INS_VCVTM: vcvtm */
	{ 0 }
},

{	/* ARM_VCVTMNSDh, ARM_INS_VCVTM: vcvtm */
	{ 0 }
},

{	/* ARM_VCVTMNSQf, ARM_INS_VCVTM: vcvtm */
	{ 0 }
},

{	/* ARM_VCVTMNSQh, ARM_INS_VCVTM: vcvtm */
	{ 0 }
},

{	/* ARM_VCVTMNUDf, ARM_INS_VCVTM: vcvtm */
	{ 0 }
},

{	/* ARM_VCVTMNUDh, ARM_INS_VCVTM: vcvtm */
	{ 0 }
},

{	/* ARM_VCVTMNUQf, ARM_INS_VCVTM: vcvtm */
	{ 0 }
},

{	/* ARM_VCVTMNUQh, ARM_INS_VCVTM: vcvtm */
	{ 0 }
},

{	/* ARM_VCVTMSD, ARM_INS_VCVTM: vcvtm */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTMSH, ARM_INS_VCVTM: vcvtm */
	{ 0 }
},

{	/* ARM_VCVTMSS, ARM_INS_VCVTM: vcvtm */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTMUD, ARM_INS_VCVTM: vcvtm */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTMUH, ARM_INS_VCVTM: vcvtm */
	{ 0 }
},

{	/* ARM_VCVTMUS, ARM_INS_VCVTM: vcvtm */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTNNSDf, ARM_INS_VCVTN: vcvtn */
	{ 0 }
},

{	/* ARM_VCVTNNSDh, ARM_INS_VCVTN: vcvtn */
	{ 0 }
},

{	/* ARM_VCVTNNSQf, ARM_INS_VCVTN: vcvtn */
	{ 0 }
},

{	/* ARM_VCVTNNSQh, ARM_INS_VCVTN: vcvtn */
	{ 0 }
},

{	/* ARM_VCVTNNUDf, ARM_INS_VCVTN: vcvtn */
	{ 0 }
},

{	/* ARM_VCVTNNUDh, ARM_INS_VCVTN: vcvtn */
	{ 0 }
},

{	/* ARM_VCVTNNUQf, ARM_INS_VCVTN: vcvtn */
	{ 0 }
},

{	/* ARM_VCVTNNUQh, ARM_INS_VCVTN: vcvtn */
	{ 0 }
},

{	/* ARM_VCVTNSD, ARM_INS_VCVTN: vcvtn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTNSH, ARM_INS_VCVTN: vcvtn */
	{ 0 }
},

{	/* ARM_VCVTNSS, ARM_INS_VCVTN: vcvtn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTNUD, ARM_INS_VCVTN: vcvtn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTNUH, ARM_INS_VCVTN: vcvtn */
	{ 0 }
},

{	/* ARM_VCVTNUS, ARM_INS_VCVTN: vcvtn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTPNSDf, ARM_INS_VCVTP: vcvtp */
	{ 0 }
},

{	/* ARM_VCVTPNSDh, ARM_INS_VCVTP: vcvtp */
	{ 0 }
},

{	/* ARM_VCVTPNSQf, ARM_INS_VCVTP: vcvtp */
	{ 0 }
},

{	/* ARM_VCVTPNSQh, ARM_INS_VCVTP: vcvtp */
	{ 0 }
},

{	/* ARM_VCVTPNUDf, ARM_INS_VCVTP: vcvtp */
	{ 0 }
},

{	/* ARM_VCVTPNUDh, ARM_INS_VCVTP: vcvtp */
	{ 0 }
},

{	/* ARM_VCVTPNUQf, ARM_INS_VCVTP: vcvtp */
	{ 0 }
},

{	/* ARM_VCVTPNUQh, ARM_INS_VCVTP: vcvtp */
	{ 0 }
},

{	/* ARM_VCVTPSD, ARM_INS_VCVTP: vcvtp */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTPSH, ARM_INS_VCVTP: vcvtp */
	{ 0 }
},

{	/* ARM_VCVTPSS, ARM_INS_VCVTP: vcvtp */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTPUD, ARM_INS_VCVTP: vcvtp */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTPUH, ARM_INS_VCVTP: vcvtp */
	{ 0 }
},

{	/* ARM_VCVTPUS, ARM_INS_VCVTP: vcvtp */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTSD, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTTDH, ARM_INS_VCVTT: vcvtt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTTHD, ARM_INS_VCVTT: vcvtt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTTHS, ARM_INS_VCVTT: vcvtt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTTSH, ARM_INS_VCVTT: vcvtt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTf2h, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTf2sd, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTf2sq, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTf2ud, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTf2uq, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTf2xsd, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTf2xsq, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTf2xud, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTf2xuq, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTh2f, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTh2sd, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTh2sq, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTh2ud, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTh2uq, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTh2xsd, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTh2xsq, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTh2xud, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTh2xuq, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTs2fd, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTs2fq, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTs2hd, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTs2hq, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTu2fd, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTu2fq, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTu2hd, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTu2hq, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTxs2fd, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTxs2fq, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTxs2hd, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTxs2hq, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTxu2fd, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTxu2fq, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VCVTxu2hd, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VCVTxu2hq, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VDIVD, ARM_INS_VDIV: vdiv */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VDIVH, ARM_INS_VDIV: vdiv */
	{ 0 }
},

{	/* ARM_VDIVS, ARM_INS_VDIV: vdiv */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VDUP16d, ARM_INS_VDUP: vdup */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VDUP16q, ARM_INS_VDUP: vdup */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VDUP32d, ARM_INS_VDUP: vdup */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VDUP32q, ARM_INS_VDUP: vdup */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VDUP8d, ARM_INS_VDUP: vdup */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VDUP8q, ARM_INS_VDUP: vdup */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VDUPLN16d, ARM_INS_VDUP: vdup */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VDUPLN16q, ARM_INS_VDUP: vdup */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VDUPLN32d, ARM_INS_VDUP: vdup */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VDUPLN32q, ARM_INS_VDUP: vdup */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VDUPLN8d, ARM_INS_VDUP: vdup */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VDUPLN8q, ARM_INS_VDUP: vdup */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VEORd, ARM_INS_VEOR: veor */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VEORq, ARM_INS_VEOR: veor */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VEXTd16, ARM_INS_VEXT: vext */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VEXTd32, ARM_INS_VEXT: vext */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VEXTd8, ARM_INS_VEXT: vext */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VEXTq16, ARM_INS_VEXT: vext */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VEXTq32, ARM_INS_VEXT: vext */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VEXTq64, ARM_INS_VEXT: vext */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VEXTq8, ARM_INS_VEXT: vext */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VFMAD, ARM_INS_VFMA: vfma */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VFMAH, ARM_INS_VFMA: vfma */
	{ 0 }
},

{	/* ARM_VFMAS, ARM_INS_VFMA: vfma */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VFMAfd, ARM_INS_VFMA: vfma */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VFMAfq, ARM_INS_VFMA: vfma */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VFMAhd, ARM_INS_VFMA: vfma */
	{ 0 }
},

{	/* ARM_VFMAhq, ARM_INS_VFMA: vfma */
	{ 0 }
},

{	/* ARM_VFMSD, ARM_INS_VFMS: vfms */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VFMSH, ARM_INS_VFMS: vfms */
	{ 0 }
},

{	/* ARM_VFMSS, ARM_INS_VFMS: vfms */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VFMSfd, ARM_INS_VFMS: vfms */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VFMSfq, ARM_INS_VFMS: vfms */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VFMShd, ARM_INS_VFMS: vfms */
	{ 0 }
},

{	/* ARM_VFMShq, ARM_INS_VFMS: vfms */
	{ 0 }
},

{	/* ARM_VFNMAD, ARM_INS_VFNMA: vfnma */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VFNMAH, ARM_INS_VFNMA: vfnma */
	{ 0 }
},

{	/* ARM_VFNMAS, ARM_INS_VFNMA: vfnma */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VFNMSD, ARM_INS_VFNMS: vfnms */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VFNMSH, ARM_INS_VFNMS: vfnms */
	{ 0 }
},

{	/* ARM_VFNMSS, ARM_INS_VFNMS: vfnms */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VGETLNi32, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VGETLNs16, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VGETLNs8, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VGETLNu16, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VGETLNu8, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VHADDsv16i8, ARM_INS_VHADD: vhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHADDsv2i32, ARM_INS_VHADD: vhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHADDsv4i16, ARM_INS_VHADD: vhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHADDsv4i32, ARM_INS_VHADD: vhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHADDsv8i16, ARM_INS_VHADD: vhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHADDsv8i8, ARM_INS_VHADD: vhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHADDuv16i8, ARM_INS_VHADD: vhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHADDuv2i32, ARM_INS_VHADD: vhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHADDuv4i16, ARM_INS_VHADD: vhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHADDuv4i32, ARM_INS_VHADD: vhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHADDuv8i16, ARM_INS_VHADD: vhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHADDuv8i8, ARM_INS_VHADD: vhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHSUBsv16i8, ARM_INS_VHSUB: vhsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHSUBsv2i32, ARM_INS_VHSUB: vhsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHSUBsv4i16, ARM_INS_VHSUB: vhsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHSUBsv4i32, ARM_INS_VHSUB: vhsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHSUBsv8i16, ARM_INS_VHSUB: vhsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHSUBsv8i8, ARM_INS_VHSUB: vhsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHSUBuv16i8, ARM_INS_VHSUB: vhsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHSUBuv2i32, ARM_INS_VHSUB: vhsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHSUBuv4i16, ARM_INS_VHSUB: vhsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHSUBuv4i32, ARM_INS_VHSUB: vhsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHSUBuv8i16, ARM_INS_VHSUB: vhsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VHSUBuv8i8, ARM_INS_VHSUB: vhsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VINSH, ARM_INS_VINS: vins */
	{ 0 }
},

{	/* ARM_VJCVT, ARM_INS_VJCVT: vjcvt */
	{ 0 }
},

{	/* ARM_VLD1DUPd16, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPd16wb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPd16wb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPd32, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPd32wb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPd32wb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPd8, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPd8wb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPd8wb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPq16, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPq16wb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPq16wb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPq32, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPq32wb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPq32wb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPq8, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPq8wb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1DUPq8wb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1LNd16, ARM_INS_VLD1: vld1 */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1LNd16_UPD, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VLD1LNd32, ARM_INS_VLD1: vld1 */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VLD1LNd32_UPD, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VLD1LNd8, ARM_INS_VLD1: vld1 */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1LNd8_UPD, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VLD1d16, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d16Q, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d16Qwb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d16Qwb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d16T, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d16Twb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d16Twb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d16wb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d16wb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d32, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d32Q, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d32Qwb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d32Qwb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d32T, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d32Twb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d32Twb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d32wb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d32wb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d64, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d64Q, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d64Qwb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d64Qwb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d64T, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d64Twb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d64Twb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d64wb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d64wb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d8, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d8Q, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d8Qwb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d8Qwb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d8T, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d8Twb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d8Twb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d8wb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1d8wb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1q16, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1q16wb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1q16wb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1q32, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1q32wb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1q32wb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1q64, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1q64wb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1q64wb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD1q8, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1q8wb_fixed, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD1q8wb_register, ARM_INS_VLD1: vld1 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd16, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd16wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd16wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd16x2, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd16x2wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd16x2wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd32, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd32wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd32wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd32x2, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd32x2wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd32x2wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd8, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd8wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd8wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd8x2, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd8x2wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2DUPd8x2wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2LNd16,  ARM_INS_VLD2: vld2${p}.16	\{$vd[$lane] $dst2[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2LNd16_UPD,  ARM_INS_VLD2: vld2${p}.16	\{$vd[$lane] $dst2[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD2LNd32,  ARM_INS_VLD2: vld2${p}.32	\{$vd[$lane] $dst2[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2LNd32_UPD,  ARM_INS_VLD2: vld2${p}.32	\{$vd[$lane] $dst2[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD2LNd8,  ARM_INS_VLD2: vld2${p}.8	\{$vd[$lane] $dst2[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2LNd8_UPD,  ARM_INS_VLD2: vld2${p}.8	\{$vd[$lane] $dst2[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD2LNq16,  ARM_INS_VLD2: vld2${p}.16	\{$vd[$lane] $dst2[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2LNq16_UPD,  ARM_INS_VLD2: vld2${p}.16	\{$vd[$lane] $dst2[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD2LNq32,  ARM_INS_VLD2: vld2${p}.32	\{$vd[$lane] $dst2[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2LNq32_UPD,  ARM_INS_VLD2: vld2${p}.32	\{$vd[$lane] $dst2[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD2b16, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2b16wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2b16wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2b32, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2b32wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2b32wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2b8, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2b8wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2b8wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2d16, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2d16wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2d16wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2d32, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2d32wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2d32wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2d8, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2d8wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2d8wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2q16, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2q16wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2q16wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2q32, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2q32wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2q32wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD2q8, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2q8wb_fixed, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD2q8wb_register, ARM_INS_VLD2: vld2 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VLD3DUPd16, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3DUPd16_UPD, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3DUPd32, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3DUPd32_UPD, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3DUPd8, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3DUPd8_UPD, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3DUPq16, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3DUPq16_UPD, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3DUPq32, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3DUPq32_UPD, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3DUPq8, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3DUPq8_UPD, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3LNd16,  ARM_INS_VLD3: vld3${p}.16	\{$vd[$lane] $dst2[$lane] $dst3[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3LNd16_UPD,  ARM_INS_VLD3: vld3${p}.16	\{$vd[$lane] $dst2[$lane] $dst3[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD3LNd32,  ARM_INS_VLD3: vld3${p}.32	\{$vd[$lane] $dst2[$lane] $dst3[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3LNd32_UPD,  ARM_INS_VLD3: vld3${p}.32	\{$vd[$lane] $dst2[$lane] $dst3[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD3LNd8,  ARM_INS_VLD3: vld3${p}.8	\{$vd[$lane] $dst2[$lane] $dst3[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3LNd8_UPD,  ARM_INS_VLD3: vld3${p}.8	\{$vd[$lane] $dst2[$lane] $dst3[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD3LNq16,  ARM_INS_VLD3: vld3${p}.16	\{$vd[$lane] $dst2[$lane] $dst3[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3LNq16_UPD,  ARM_INS_VLD3: vld3${p}.16	\{$vd[$lane] $dst2[$lane] $dst3[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD3LNq32,  ARM_INS_VLD3: vld3${p}.32	\{$vd[$lane] $dst2[$lane] $dst3[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3LNq32_UPD,  ARM_INS_VLD3: vld3${p}.32	\{$vd[$lane] $dst2[$lane] $dst3[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD3d16, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3d16_UPD, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD3d32, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3d32_UPD, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD3d8, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3d8_UPD, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD3q16, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3q16_UPD, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD3q32, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3q32_UPD, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD3q8, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD3q8_UPD, ARM_INS_VLD3: vld3 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD4DUPd16, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4DUPd16_UPD, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4DUPd32, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4DUPd32_UPD, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4DUPd8, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4DUPd8_UPD, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4DUPq16, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4DUPq16_UPD, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4DUPq32, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4DUPq32_UPD, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4DUPq8, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4DUPq8_UPD, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4LNd16,  ARM_INS_VLD4: vld4${p}.16	\{$vd[$lane] $dst2[$lane] $dst3[$lane] $dst4[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4LNd16_UPD,  ARM_INS_VLD4: vld4${p}.16	\{$vd[$lane] $dst2[$lane] $dst3[$lane] $dst4[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD4LNd32,  ARM_INS_VLD4: vld4${p}.32	\{$vd[$lane] $dst2[$lane] $dst3[$lane] $dst4[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4LNd32_UPD,  ARM_INS_VLD4: vld4${p}.32	\{$vd[$lane] $dst2[$lane] $dst3[$lane] $dst4[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD4LNd8,  ARM_INS_VLD4: vld4${p}.8	\{$vd[$lane] $dst2[$lane] $dst3[$lane] $dst4[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4LNd8_UPD,  ARM_INS_VLD4: vld4${p}.8	\{$vd[$lane] $dst2[$lane] $dst3[$lane] $dst4[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD4LNq16,  ARM_INS_VLD4: vld4${p}.16	\{$vd[$lane] $dst2[$lane] $dst3[$lane] $dst4[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4LNq16_UPD,  ARM_INS_VLD4: vld4${p}.16	\{$vd[$lane] $dst2[$lane] $dst3[$lane] $dst4[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD4LNq32,  ARM_INS_VLD4: vld4${p}.32	\{$vd[$lane] $dst2[$lane] $dst3[$lane] $dst4[$lane]\} $rn */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4LNq32_UPD,  ARM_INS_VLD4: vld4${p}.32	\{$vd[$lane] $dst2[$lane] $dst3[$lane] $dst4[$lane]\} $rn$rm */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD4d16, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4d16_UPD, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD4d32, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4d32_UPD, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD4d8, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4d8_UPD, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD4q16, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4q16_UPD, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD4q32, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4q32_UPD, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLD4q8, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VLD4q8_UPD, ARM_INS_VLD4: vld4 */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VLDMDDB_UPD, ARM_INS_VLDMDB: vldmdb */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VLDMDIA, ARM_INS_VLDMIA: vldmia */
	{ CS_AC_READ, 0 }
},

{	/* ARM_VLDMDIA_UPD, ARM_INS_VLDMIA: vldmia */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VLDMSDB_UPD, ARM_INS_VLDMDB: vldmdb */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VLDMSIA, ARM_INS_VLDMIA: vldmia */
	{ CS_AC_READ, 0 }
},

{	/* ARM_VLDMSIA_UPD, ARM_INS_VLDMIA: vldmia */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VLDRD, ARM_INS_VLDR: vldr */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VLDRH, ARM_INS_VLDR: vldr */
	{ 0 }
},

{	/* ARM_VLDRS, ARM_INS_VLDR: vldr */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VLLDM, ARM_INS_VLLDM: vlldm */
	{ 0 }
},

{	/* ARM_VLSTM, ARM_INS_VLSTM: vlstm */
	{ 0 }
},

{	/* ARM_VMAXNMD, ARM_INS_VMAXNM: vmaxnm */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXNMH, ARM_INS_VMAXNM: vmaxnm */
	{ 0 }
},

{	/* ARM_VMAXNMNDf, ARM_INS_VMAXNM: vmaxnm */
	{ 0 }
},

{	/* ARM_VMAXNMNDh, ARM_INS_VMAXNM: vmaxnm */
	{ 0 }
},

{	/* ARM_VMAXNMNQf, ARM_INS_VMAXNM: vmaxnm */
	{ 0 }
},

{	/* ARM_VMAXNMNQh, ARM_INS_VMAXNM: vmaxnm */
	{ 0 }
},

{	/* ARM_VMAXNMS, ARM_INS_VMAXNM: vmaxnm */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXfd, ARM_INS_VMAX: vmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXfq, ARM_INS_VMAX: vmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXhd, ARM_INS_VMAX: vmax */
	{ 0 }
},

{	/* ARM_VMAXhq, ARM_INS_VMAX: vmax */
	{ 0 }
},

{	/* ARM_VMAXsv16i8, ARM_INS_VMAX: vmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXsv2i32, ARM_INS_VMAX: vmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXsv4i16, ARM_INS_VMAX: vmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXsv4i32, ARM_INS_VMAX: vmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXsv8i16, ARM_INS_VMAX: vmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXsv8i8, ARM_INS_VMAX: vmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXuv16i8, ARM_INS_VMAX: vmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXuv2i32, ARM_INS_VMAX: vmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXuv4i16, ARM_INS_VMAX: vmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXuv4i32, ARM_INS_VMAX: vmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXuv8i16, ARM_INS_VMAX: vmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMAXuv8i8, ARM_INS_VMAX: vmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINNMD, ARM_INS_VMINNM: vminnm */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINNMH, ARM_INS_VMINNM: vminnm */
	{ 0 }
},

{	/* ARM_VMINNMNDf, ARM_INS_VMINNM: vminnm */
	{ 0 }
},

{	/* ARM_VMINNMNDh, ARM_INS_VMINNM: vminnm */
	{ 0 }
},

{	/* ARM_VMINNMNQf, ARM_INS_VMINNM: vminnm */
	{ 0 }
},

{	/* ARM_VMINNMNQh, ARM_INS_VMINNM: vminnm */
	{ 0 }
},

{	/* ARM_VMINNMS, ARM_INS_VMINNM: vminnm */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINfd, ARM_INS_VMIN: vmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINfq, ARM_INS_VMIN: vmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINhd, ARM_INS_VMIN: vmin */
	{ 0 }
},

{	/* ARM_VMINhq, ARM_INS_VMIN: vmin */
	{ 0 }
},

{	/* ARM_VMINsv16i8, ARM_INS_VMIN: vmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINsv2i32, ARM_INS_VMIN: vmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINsv4i16, ARM_INS_VMIN: vmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINsv4i32, ARM_INS_VMIN: vmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINsv8i16, ARM_INS_VMIN: vmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINsv8i8, ARM_INS_VMIN: vmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINuv16i8, ARM_INS_VMIN: vmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINuv2i32, ARM_INS_VMIN: vmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINuv4i16, ARM_INS_VMIN: vmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINuv4i32, ARM_INS_VMIN: vmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINuv8i16, ARM_INS_VMIN: vmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMINuv8i8, ARM_INS_VMIN: vmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLAD, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLAH, ARM_INS_VMLA: vmla */
	{ 0 }
},

{	/* ARM_VMLALslsv2i32, ARM_INS_VMLAL: vmlal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLALslsv4i16, ARM_INS_VMLAL: vmlal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLALsluv2i32, ARM_INS_VMLAL: vmlal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLALsluv4i16, ARM_INS_VMLAL: vmlal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLALsv2i64, ARM_INS_VMLAL: vmlal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLALsv4i32, ARM_INS_VMLAL: vmlal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLALsv8i16, ARM_INS_VMLAL: vmlal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLALuv2i64, ARM_INS_VMLAL: vmlal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLALuv4i32, ARM_INS_VMLAL: vmlal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLALuv8i16, ARM_INS_VMLAL: vmlal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLAS, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLAfd, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLAfq, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLAhd, ARM_INS_VMLA: vmla */
	{ 0 }
},

{	/* ARM_VMLAhq, ARM_INS_VMLA: vmla */
	{ 0 }
},

{	/* ARM_VMLAslfd, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLAslfq, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLAslhd, ARM_INS_VMLA: vmla */
	{ 0 }
},

{	/* ARM_VMLAslhq, ARM_INS_VMLA: vmla */
	{ 0 }
},

{	/* ARM_VMLAslv2i32, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLAslv4i16, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLAslv4i32, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLAslv8i16, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLAv16i8, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLAv2i32, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLAv4i16, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLAv4i32, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLAv8i16, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLAv8i8, ARM_INS_VMLA: vmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSD, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSH, ARM_INS_VMLS: vmls */
	{ 0 }
},

{	/* ARM_VMLSLslsv2i32, ARM_INS_VMLSL: vmlsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLSLslsv4i16, ARM_INS_VMLSL: vmlsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLSLsluv2i32, ARM_INS_VMLSL: vmlsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLSLsluv4i16, ARM_INS_VMLSL: vmlsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLSLsv2i64, ARM_INS_VMLSL: vmlsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSLsv4i32, ARM_INS_VMLSL: vmlsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSLsv8i16, ARM_INS_VMLSL: vmlsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSLuv2i64, ARM_INS_VMLSL: vmlsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSLuv4i32, ARM_INS_VMLSL: vmlsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSLuv8i16, ARM_INS_VMLSL: vmlsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSS, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSfd, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSfq, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLShd, ARM_INS_VMLS: vmls */
	{ 0 }
},

{	/* ARM_VMLShq, ARM_INS_VMLS: vmls */
	{ 0 }
},

{	/* ARM_VMLSslfd, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLSslfq, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLSslhd, ARM_INS_VMLS: vmls */
	{ 0 }
},

{	/* ARM_VMLSslhq, ARM_INS_VMLS: vmls */
	{ 0 }
},

{	/* ARM_VMLSslv2i32, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLSslv4i16, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLSslv4i32, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLSslv8i16, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMLSv16i8, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSv2i32, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSv4i16, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSv4i32, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSv8i16, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMLSv8i8, ARM_INS_VMLS: vmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMOVD, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMOVDRR, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMOVH, ARM_INS_VMOVX: vmovx */
	{ 0 }
},

{	/* ARM_VMOVHR, ARM_INS_VMOV: vmov */
	{ 0 }
},

{	/* ARM_VMOVLsv2i64, ARM_INS_VMOVL: vmovl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMOVLsv4i32, ARM_INS_VMOVL: vmovl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMOVLsv8i16, ARM_INS_VMOVL: vmovl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMOVLuv2i64, ARM_INS_VMOVL: vmovl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMOVLuv4i32, ARM_INS_VMOVL: vmovl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMOVLuv8i16, ARM_INS_VMOVL: vmovl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMOVNv2i32, ARM_INS_VMOVN: vmovn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMOVNv4i16, ARM_INS_VMOVN: vmovn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMOVNv8i8, ARM_INS_VMOVN: vmovn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMOVRH, ARM_INS_VMOV: vmov */
	{ 0 }
},

{	/* ARM_VMOVRRD, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMOVRRS, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMOVRS, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMOVS, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMOVSR, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMOVSRR, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMOVv16i8, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMOVv1i64, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMOVv2f32, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMOVv2i32, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMOVv2i64, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMOVv4f32, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMOVv4i16, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMOVv4i32, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMOVv8i16, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMOVv8i8, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMRS, ARM_INS_VMRS: vmrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMRS_FPEXC, ARM_INS_VMRS: vmrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMRS_FPINST, ARM_INS_VMRS: vmrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMRS_FPINST2, ARM_INS_VMRS: vmrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMRS_FPSID, ARM_INS_VMRS: vmrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMRS_MVFR0, ARM_INS_VMRS: vmrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMRS_MVFR1, ARM_INS_VMRS: vmrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMRS_MVFR2, ARM_INS_VMRS: vmrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMSR, ARM_INS_VMSR: vmsr */
	{ CS_AC_IGNORE, CS_AC_READ, 0 }
},

{	/* ARM_VMSR_FPEXC, ARM_INS_VMSR: vmsr */
	{ CS_AC_IGNORE, CS_AC_READ, 0 }
},

{	/* ARM_VMSR_FPINST, ARM_INS_VMSR: vmsr */
	{ CS_AC_IGNORE, CS_AC_READ, 0 }
},

{	/* ARM_VMSR_FPINST2, ARM_INS_VMSR: vmsr */
	{ CS_AC_IGNORE, CS_AC_READ, 0 }
},

{	/* ARM_VMSR_FPSID, ARM_INS_VMSR: vmsr */
	{ CS_AC_IGNORE, CS_AC_READ, 0 }
},

{	/* ARM_VMULD, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULH, ARM_INS_VMUL: vmul */
	{ 0 }
},

{	/* ARM_VMULLp64, ARM_INS_VMULL: vmull */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULLp8, ARM_INS_VMULL: vmull */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULLslsv2i32, ARM_INS_VMULL: vmull */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMULLslsv4i16, ARM_INS_VMULL: vmull */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMULLsluv2i32, ARM_INS_VMULL: vmull */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMULLsluv4i16, ARM_INS_VMULL: vmull */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMULLsv2i64, ARM_INS_VMULL: vmull */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULLsv4i32, ARM_INS_VMULL: vmull */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULLsv8i16, ARM_INS_VMULL: vmull */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULLuv2i64, ARM_INS_VMULL: vmull */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULLuv4i32, ARM_INS_VMULL: vmull */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULLuv8i16, ARM_INS_VMULL: vmull */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULS, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULfd, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULfq, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULhd, ARM_INS_VMUL: vmul */
	{ 0 }
},

{	/* ARM_VMULhq, ARM_INS_VMUL: vmul */
	{ 0 }
},

{	/* ARM_VMULpd, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULpq, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULslfd, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMULslfq, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMULslhd, ARM_INS_VMUL: vmul */
	{ 0 }
},

{	/* ARM_VMULslhq, ARM_INS_VMUL: vmul */
	{ 0 }
},

{	/* ARM_VMULslv2i32, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMULslv4i16, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMULslv4i32, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMULslv8i16, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMULv16i8, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULv2i32, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULv4i16, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULv4i32, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULv8i16, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMULv8i8, ARM_INS_VMUL: vmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VMVNd, ARM_INS_VMVN: vmvn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMVNq, ARM_INS_VMVN: vmvn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VMVNv2i32, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMVNv4i16, ARM_INS_VMVN: vmvn */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMVNv4i32, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VMVNv8i16, ARM_INS_VMVN: vmvn */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_VNEGD, ARM_INS_VNEG: vneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VNEGH, ARM_INS_VNEG: vneg */
	{ 0 }
},

{	/* ARM_VNEGS, ARM_INS_VNEG: vneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VNEGf32q, ARM_INS_VNEG: vneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VNEGfd, ARM_INS_VNEG: vneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VNEGhd, ARM_INS_VNEG: vneg */
	{ 0 }
},

{	/* ARM_VNEGhq, ARM_INS_VNEG: vneg */
	{ 0 }
},

{	/* ARM_VNEGs16d, ARM_INS_VNEG: vneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VNEGs16q, ARM_INS_VNEG: vneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VNEGs32d, ARM_INS_VNEG: vneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VNEGs32q, ARM_INS_VNEG: vneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VNEGs8d, ARM_INS_VNEG: vneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VNEGs8q, ARM_INS_VNEG: vneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VNMLAD, ARM_INS_VNMLA: vnmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VNMLAH, ARM_INS_VNMLA: vnmla */
	{ 0 }
},

{	/* ARM_VNMLAS, ARM_INS_VNMLA: vnmla */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VNMLSD, ARM_INS_VNMLS: vnmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VNMLSH, ARM_INS_VNMLS: vnmls */
	{ 0 }
},

{	/* ARM_VNMLSS, ARM_INS_VNMLS: vnmls */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VNMULD, ARM_INS_VNMUL: vnmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VNMULH, ARM_INS_VNMUL: vnmul */
	{ 0 }
},

{	/* ARM_VNMULS, ARM_INS_VNMUL: vnmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VORNd, ARM_INS_VORN: vorn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VORNq, ARM_INS_VORN: vorn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VORRd, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VORRiv2i32, ARM_INS_VORR: vorr */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VORRiv4i16, ARM_INS_VORR: vorr */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VORRiv4i32, ARM_INS_VORR: vorr */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VORRiv8i16, ARM_INS_VORR: vorr */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VORRq, ARM_INS_VMOV: vmov */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPADALsv16i8, ARM_INS_VPADAL: vpadal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADALsv2i32, ARM_INS_VPADAL: vpadal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADALsv4i16, ARM_INS_VPADAL: vpadal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADALsv4i32, ARM_INS_VPADAL: vpadal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADALsv8i16, ARM_INS_VPADAL: vpadal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADALsv8i8, ARM_INS_VPADAL: vpadal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADALuv16i8, ARM_INS_VPADAL: vpadal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADALuv2i32, ARM_INS_VPADAL: vpadal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADALuv4i16, ARM_INS_VPADAL: vpadal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADALuv4i32, ARM_INS_VPADAL: vpadal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADALuv8i16, ARM_INS_VPADAL: vpadal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADALuv8i8, ARM_INS_VPADAL: vpadal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADDLsv16i8, ARM_INS_VPADDL: vpaddl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADDLsv2i32, ARM_INS_VPADDL: vpaddl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADDLsv4i16, ARM_INS_VPADDL: vpaddl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADDLsv4i32, ARM_INS_VPADDL: vpaddl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADDLsv8i16, ARM_INS_VPADDL: vpaddl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADDLsv8i8, ARM_INS_VPADDL: vpaddl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADDLuv16i8, ARM_INS_VPADDL: vpaddl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADDLuv2i32, ARM_INS_VPADDL: vpaddl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADDLuv4i16, ARM_INS_VPADDL: vpaddl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADDLuv4i32, ARM_INS_VPADDL: vpaddl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADDLuv8i16, ARM_INS_VPADDL: vpaddl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADDLuv8i8, ARM_INS_VPADDL: vpaddl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VPADDf, ARM_INS_VPADD: vpadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPADDh, ARM_INS_VPADD: vpadd */
	{ 0 }
},

{	/* ARM_VPADDi16, ARM_INS_VPADD: vpadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPADDi32, ARM_INS_VPADD: vpadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPADDi8, ARM_INS_VPADD: vpadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPMAXf, ARM_INS_VPMAX: vpmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPMAXh, ARM_INS_VPMAX: vpmax */
	{ 0 }
},

{	/* ARM_VPMAXs16, ARM_INS_VPMAX: vpmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPMAXs32, ARM_INS_VPMAX: vpmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPMAXs8, ARM_INS_VPMAX: vpmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPMAXu16, ARM_INS_VPMAX: vpmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPMAXu32, ARM_INS_VPMAX: vpmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPMAXu8, ARM_INS_VPMAX: vpmax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPMINf, ARM_INS_VPMIN: vpmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPMINh, ARM_INS_VPMIN: vpmin */
	{ 0 }
},

{	/* ARM_VPMINs16, ARM_INS_VPMIN: vpmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPMINs32, ARM_INS_VPMIN: vpmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPMINs8, ARM_INS_VPMIN: vpmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPMINu16, ARM_INS_VPMIN: vpmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPMINu32, ARM_INS_VPMIN: vpmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VPMINu8, ARM_INS_VPMIN: vpmin */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQABSv16i8, ARM_INS_VQABS: vqabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQABSv2i32, ARM_INS_VQABS: vqabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQABSv4i16, ARM_INS_VQABS: vqabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQABSv4i32, ARM_INS_VQABS: vqabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQABSv8i16, ARM_INS_VQABS: vqabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQABSv8i8, ARM_INS_VQABS: vqabs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQADDsv16i8, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDsv1i64, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDsv2i32, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDsv2i64, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDsv4i16, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDsv4i32, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDsv8i16, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDsv8i8, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDuv16i8, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDuv1i64, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDuv2i32, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDuv2i64, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDuv4i16, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDuv4i32, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDuv8i16, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQADDuv8i8, ARM_INS_VQADD: vqadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQDMLALslv2i32, ARM_INS_VQDMLAL: vqdmlal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQDMLALslv4i16, ARM_INS_VQDMLAL: vqdmlal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQDMLALv2i64, ARM_INS_VQDMLAL: vqdmlal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQDMLALv4i32, ARM_INS_VQDMLAL: vqdmlal */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQDMLSLslv2i32, ARM_INS_VQDMLSL: vqdmlsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQDMLSLslv4i16, ARM_INS_VQDMLSL: vqdmlsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQDMLSLv2i64, ARM_INS_VQDMLSL: vqdmlsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQDMLSLv4i32, ARM_INS_VQDMLSL: vqdmlsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQDMULHslv2i32, ARM_INS_VQDMULH: vqdmulh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQDMULHslv4i16, ARM_INS_VQDMULH: vqdmulh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQDMULHslv4i32, ARM_INS_VQDMULH: vqdmulh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQDMULHslv8i16, ARM_INS_VQDMULH: vqdmulh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQDMULHv2i32, ARM_INS_VQDMULH: vqdmulh */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQDMULHv4i16, ARM_INS_VQDMULH: vqdmulh */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQDMULHv4i32, ARM_INS_VQDMULH: vqdmulh */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQDMULHv8i16, ARM_INS_VQDMULH: vqdmulh */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQDMULLslv2i32, ARM_INS_VQDMULL: vqdmull */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQDMULLslv4i16, ARM_INS_VQDMULL: vqdmull */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQDMULLv2i64, ARM_INS_VQDMULL: vqdmull */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQDMULLv4i32, ARM_INS_VQDMULL: vqdmull */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQMOVNsuv2i32, ARM_INS_VQMOVUN: vqmovun */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQMOVNsuv4i16, ARM_INS_VQMOVUN: vqmovun */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQMOVNsuv8i8, ARM_INS_VQMOVUN: vqmovun */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQMOVNsv2i32, ARM_INS_VQMOVN: vqmovn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQMOVNsv4i16, ARM_INS_VQMOVN: vqmovn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQMOVNsv8i8, ARM_INS_VQMOVN: vqmovn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQMOVNuv2i32, ARM_INS_VQMOVN: vqmovn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQMOVNuv4i16, ARM_INS_VQMOVN: vqmovn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQMOVNuv8i8, ARM_INS_VQMOVN: vqmovn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQNEGv16i8, ARM_INS_VQNEG: vqneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQNEGv2i32, ARM_INS_VQNEG: vqneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQNEGv4i16, ARM_INS_VQNEG: vqneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQNEGv4i32, ARM_INS_VQNEG: vqneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQNEGv8i16, ARM_INS_VQNEG: vqneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQNEGv8i8, ARM_INS_VQNEG: vqneg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQRDMLAHslv2i32, ARM_INS_VQRDMLAH: vqrdmlah */
	{ 0 }
},

{	/* ARM_VQRDMLAHslv4i16, ARM_INS_VQRDMLAH: vqrdmlah */
	{ 0 }
},

{	/* ARM_VQRDMLAHslv4i32, ARM_INS_VQRDMLAH: vqrdmlah */
	{ 0 }
},

{	/* ARM_VQRDMLAHslv8i16, ARM_INS_VQRDMLAH: vqrdmlah */
	{ 0 }
},

{	/* ARM_VQRDMLAHv2i32, ARM_INS_VQRDMLAH: vqrdmlah */
	{ 0 }
},

{	/* ARM_VQRDMLAHv4i16, ARM_INS_VQRDMLAH: vqrdmlah */
	{ 0 }
},

{	/* ARM_VQRDMLAHv4i32, ARM_INS_VQRDMLAH: vqrdmlah */
	{ 0 }
},

{	/* ARM_VQRDMLAHv8i16, ARM_INS_VQRDMLAH: vqrdmlah */
	{ 0 }
},

{	/* ARM_VQRDMLSHslv2i32, ARM_INS_VQRDMLSH: vqrdmlsh */
	{ 0 }
},

{	/* ARM_VQRDMLSHslv4i16, ARM_INS_VQRDMLSH: vqrdmlsh */
	{ 0 }
},

{	/* ARM_VQRDMLSHslv4i32, ARM_INS_VQRDMLSH: vqrdmlsh */
	{ 0 }
},

{	/* ARM_VQRDMLSHslv8i16, ARM_INS_VQRDMLSH: vqrdmlsh */
	{ 0 }
},

{	/* ARM_VQRDMLSHv2i32, ARM_INS_VQRDMLSH: vqrdmlsh */
	{ 0 }
},

{	/* ARM_VQRDMLSHv4i16, ARM_INS_VQRDMLSH: vqrdmlsh */
	{ 0 }
},

{	/* ARM_VQRDMLSHv4i32, ARM_INS_VQRDMLSH: vqrdmlsh */
	{ 0 }
},

{	/* ARM_VQRDMLSHv8i16, ARM_INS_VQRDMLSH: vqrdmlsh */
	{ 0 }
},

{	/* ARM_VQRDMULHslv2i32, ARM_INS_VQRDMULH: vqrdmulh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQRDMULHslv4i16, ARM_INS_VQRDMULH: vqrdmulh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQRDMULHslv4i32, ARM_INS_VQRDMULH: vqrdmulh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQRDMULHslv8i16, ARM_INS_VQRDMULH: vqrdmulh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQRDMULHv2i32, ARM_INS_VQRDMULH: vqrdmulh */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRDMULHv4i16, ARM_INS_VQRDMULH: vqrdmulh */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRDMULHv4i32, ARM_INS_VQRDMULH: vqrdmulh */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRDMULHv8i16, ARM_INS_VQRDMULH: vqrdmulh */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLsv16i8, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLsv1i64, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLsv2i32, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLsv2i64, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLsv4i16, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLsv4i32, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLsv8i16, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLsv8i8, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLuv16i8, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLuv1i64, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLuv2i32, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLuv2i64, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLuv4i16, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLuv4i32, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLuv8i16, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHLuv8i8, ARM_INS_VQRSHL: vqrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHRNsv2i32, ARM_INS_VQRSHRN: vqrshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHRNsv4i16, ARM_INS_VQRSHRN: vqrshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHRNsv8i8, ARM_INS_VQRSHRN: vqrshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHRNuv2i32, ARM_INS_VQRSHRN: vqrshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHRNuv4i16, ARM_INS_VQRSHRN: vqrshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHRNuv8i8, ARM_INS_VQRSHRN: vqrshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHRUNv2i32, ARM_INS_VQRSHRUN: vqrshrun */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHRUNv4i16, ARM_INS_VQRSHRUN: vqrshrun */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQRSHRUNv8i8, ARM_INS_VQRSHRUN: vqrshrun */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsiv16i8, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsiv1i64, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsiv2i32, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsiv2i64, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsiv4i16, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsiv4i32, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsiv8i16, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsiv8i8, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsuv16i8, ARM_INS_VQSHLU: vqshlu */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsuv1i64, ARM_INS_VQSHLU: vqshlu */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsuv2i32, ARM_INS_VQSHLU: vqshlu */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsuv2i64, ARM_INS_VQSHLU: vqshlu */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsuv4i16, ARM_INS_VQSHLU: vqshlu */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsuv4i32, ARM_INS_VQSHLU: vqshlu */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsuv8i16, ARM_INS_VQSHLU: vqshlu */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsuv8i8, ARM_INS_VQSHLU: vqshlu */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsv16i8, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsv1i64, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsv2i32, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsv2i64, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsv4i16, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsv4i32, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsv8i16, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLsv8i8, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuiv16i8, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuiv1i64, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuiv2i32, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuiv2i64, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuiv4i16, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuiv4i32, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuiv8i16, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuiv8i8, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuv16i8, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuv1i64, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuv2i32, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuv2i64, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuv4i16, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuv4i32, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuv8i16, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHLuv8i8, ARM_INS_VQSHL: vqshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSHRNsv2i32, ARM_INS_VQSHRN: vqshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHRNsv4i16, ARM_INS_VQSHRN: vqshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHRNsv8i8, ARM_INS_VQSHRN: vqshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHRNuv2i32, ARM_INS_VQSHRN: vqshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHRNuv4i16, ARM_INS_VQSHRN: vqshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHRNuv8i8, ARM_INS_VQSHRN: vqshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHRUNv2i32, ARM_INS_VQSHRUN: vqshrun */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHRUNv4i16, ARM_INS_VQSHRUN: vqshrun */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSHRUNv8i8, ARM_INS_VQSHRUN: vqshrun */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBsv16i8, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBsv1i64, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBsv2i32, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBsv2i64, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBsv4i16, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBsv4i32, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBsv8i16, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBsv8i8, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBuv16i8, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBuv1i64, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBuv2i32, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBuv2i64, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBuv4i16, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBuv4i32, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBuv8i16, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VQSUBuv8i8, ARM_INS_VQSUB: vqsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRADDHNv2i32, ARM_INS_VRADDHN: vraddhn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRADDHNv4i16, ARM_INS_VRADDHN: vraddhn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRADDHNv8i8, ARM_INS_VRADDHN: vraddhn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRECPEd, ARM_INS_VRECPE: vrecpe */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRECPEfd, ARM_INS_VRECPE: vrecpe */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRECPEfq, ARM_INS_VRECPE: vrecpe */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRECPEhd, ARM_INS_VRECPE: vrecpe */
	{ 0 }
},

{	/* ARM_VRECPEhq, ARM_INS_VRECPE: vrecpe */
	{ 0 }
},

{	/* ARM_VRECPEq, ARM_INS_VRECPE: vrecpe */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRECPSfd, ARM_INS_VRECPS: vrecps */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRECPSfq, ARM_INS_VRECPS: vrecps */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRECPShd, ARM_INS_VRECPS: vrecps */
	{ 0 }
},

{	/* ARM_VRECPShq, ARM_INS_VRECPS: vrecps */
	{ 0 }
},

{	/* ARM_VREV16d8, ARM_INS_VREV16: vrev16 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VREV16q8, ARM_INS_VREV16: vrev16 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VREV32d16, ARM_INS_VREV32: vrev32 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VREV32d8, ARM_INS_VREV32: vrev32 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VREV32q16, ARM_INS_VREV32: vrev32 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VREV32q8, ARM_INS_VREV32: vrev32 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VREV64d16, ARM_INS_VREV64: vrev64 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VREV64d32, ARM_INS_VREV64: vrev64 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VREV64d8, ARM_INS_VREV64: vrev64 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VREV64q16, ARM_INS_VREV64: vrev64 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VREV64q32, ARM_INS_VREV64: vrev64 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VREV64q8, ARM_INS_VREV64: vrev64 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRHADDsv16i8, ARM_INS_VRHADD: vrhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRHADDsv2i32, ARM_INS_VRHADD: vrhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRHADDsv4i16, ARM_INS_VRHADD: vrhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRHADDsv4i32, ARM_INS_VRHADD: vrhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRHADDsv8i16, ARM_INS_VRHADD: vrhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRHADDsv8i8, ARM_INS_VRHADD: vrhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRHADDuv16i8, ARM_INS_VRHADD: vrhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRHADDuv2i32, ARM_INS_VRHADD: vrhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRHADDuv4i16, ARM_INS_VRHADD: vrhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRHADDuv4i32, ARM_INS_VRHADD: vrhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRHADDuv8i16, ARM_INS_VRHADD: vrhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRHADDuv8i8, ARM_INS_VRHADD: vrhadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRINTAD, ARM_INS_VRINTA: vrinta */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRINTAH, ARM_INS_VRINTA: vrinta */
	{ 0 }
},

{	/* ARM_VRINTANDf, ARM_INS_VRINTA: vrinta */
	{ 0 }
},

{	/* ARM_VRINTANDh, ARM_INS_VRINTA: vrinta */
	{ 0 }
},

{	/* ARM_VRINTANQf, ARM_INS_VRINTA: vrinta */
	{ 0 }
},

{	/* ARM_VRINTANQh, ARM_INS_VRINTA: vrinta */
	{ 0 }
},

{	/* ARM_VRINTAS, ARM_INS_VRINTA: vrinta */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRINTMD, ARM_INS_VRINTM: vrintm */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRINTMH, ARM_INS_VRINTM: vrintm */
	{ 0 }
},

{	/* ARM_VRINTMNDf, ARM_INS_VRINTM: vrintm */
	{ 0 }
},

{	/* ARM_VRINTMNDh, ARM_INS_VRINTM: vrintm */
	{ 0 }
},

{	/* ARM_VRINTMNQf, ARM_INS_VRINTM: vrintm */
	{ 0 }
},

{	/* ARM_VRINTMNQh, ARM_INS_VRINTM: vrintm */
	{ 0 }
},

{	/* ARM_VRINTMS, ARM_INS_VRINTM: vrintm */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRINTND, ARM_INS_VRINTN: vrintn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRINTNH, ARM_INS_VRINTN: vrintn */
	{ 0 }
},

{	/* ARM_VRINTNNDf, ARM_INS_VRINTN: vrintn */
	{ 0 }
},

{	/* ARM_VRINTNNDh, ARM_INS_VRINTN: vrintn */
	{ 0 }
},

{	/* ARM_VRINTNNQf, ARM_INS_VRINTN: vrintn */
	{ 0 }
},

{	/* ARM_VRINTNNQh, ARM_INS_VRINTN: vrintn */
	{ 0 }
},

{	/* ARM_VRINTNS, ARM_INS_VRINTN: vrintn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRINTPD, ARM_INS_VRINTP: vrintp */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRINTPH, ARM_INS_VRINTP: vrintp */
	{ 0 }
},

{	/* ARM_VRINTPNDf, ARM_INS_VRINTP: vrintp */
	{ 0 }
},

{	/* ARM_VRINTPNDh, ARM_INS_VRINTP: vrintp */
	{ 0 }
},

{	/* ARM_VRINTPNQf, ARM_INS_VRINTP: vrintp */
	{ 0 }
},

{	/* ARM_VRINTPNQh, ARM_INS_VRINTP: vrintp */
	{ 0 }
},

{	/* ARM_VRINTPS, ARM_INS_VRINTP: vrintp */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRINTRD, ARM_INS_VRINTR: vrintr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRINTRH, ARM_INS_VRINTR: vrintr */
	{ 0 }
},

{	/* ARM_VRINTRS, ARM_INS_VRINTR: vrintr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRINTXD, ARM_INS_VRINTX: vrintx */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRINTXH, ARM_INS_VRINTX: vrintx */
	{ 0 }
},

{	/* ARM_VRINTXNDf, ARM_INS_VRINTX: vrintx */
	{ 0 }
},

{	/* ARM_VRINTXNDh, ARM_INS_VRINTX: vrintx */
	{ 0 }
},

{	/* ARM_VRINTXNQf, ARM_INS_VRINTX: vrintx */
	{ 0 }
},

{	/* ARM_VRINTXNQh, ARM_INS_VRINTX: vrintx */
	{ 0 }
},

{	/* ARM_VRINTXS, ARM_INS_VRINTX: vrintx */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRINTZD, ARM_INS_VRINTZ: vrintz */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRINTZH, ARM_INS_VRINTZ: vrintz */
	{ 0 }
},

{	/* ARM_VRINTZNDf, ARM_INS_VRINTZ: vrintz */
	{ 0 }
},

{	/* ARM_VRINTZNDh, ARM_INS_VRINTZ: vrintz */
	{ 0 }
},

{	/* ARM_VRINTZNQf, ARM_INS_VRINTZ: vrintz */
	{ 0 }
},

{	/* ARM_VRINTZNQh, ARM_INS_VRINTZ: vrintz */
	{ 0 }
},

{	/* ARM_VRINTZS, ARM_INS_VRINTZ: vrintz */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLsv16i8, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLsv1i64, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLsv2i32, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLsv2i64, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLsv4i16, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLsv4i32, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLsv8i16, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLsv8i8, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLuv16i8, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLuv1i64, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLuv2i32, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLuv2i64, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLuv4i16, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLuv4i32, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLuv8i16, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHLuv8i8, ARM_INS_VRSHL: vrshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRNv2i32, ARM_INS_VRSHRN: vrshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRNv4i16, ARM_INS_VRSHRN: vrshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRNv8i8, ARM_INS_VRSHRN: vrshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRsv16i8, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRsv1i64, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRsv2i32, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRsv2i64, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRsv4i16, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRsv4i32, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRsv8i16, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRsv8i8, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRuv16i8, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRuv1i64, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRuv2i32, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRuv2i64, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRuv4i16, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRuv4i32, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRuv8i16, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSHRuv8i8, ARM_INS_VRSHR: vrshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSQRTEd, ARM_INS_VRSQRTE: vrsqrte */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSQRTEfd, ARM_INS_VRSQRTE: vrsqrte */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSQRTEfq, ARM_INS_VRSQRTE: vrsqrte */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSQRTEhd, ARM_INS_VRSQRTE: vrsqrte */
	{ 0 }
},

{	/* ARM_VRSQRTEhq, ARM_INS_VRSQRTE: vrsqrte */
	{ 0 }
},

{	/* ARM_VRSQRTEq, ARM_INS_VRSQRTE: vrsqrte */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSQRTSfd, ARM_INS_VRSQRTS: vrsqrts */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSQRTSfq, ARM_INS_VRSQRTS: vrsqrts */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSQRTShd, ARM_INS_VRSQRTS: vrsqrts */
	{ 0 }
},

{	/* ARM_VRSQRTShq, ARM_INS_VRSQRTS: vrsqrts */
	{ 0 }
},

{	/* ARM_VRSRAsv16i8, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAsv1i64, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAsv2i32, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAsv2i64, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAsv4i16, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAsv4i32, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAsv8i16, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAsv8i8, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAuv16i8, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAuv1i64, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAuv2i32, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAuv2i64, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAuv4i16, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAuv4i32, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAuv8i16, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSRAuv8i8, ARM_INS_VRSRA: vrsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VRSUBHNv2i32, ARM_INS_VRSUBHN: vrsubhn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSUBHNv4i16, ARM_INS_VRSUBHN: vrsubhn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VRSUBHNv8i8, ARM_INS_VRSUBHN: vrsubhn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSDOTD, ARM_INS_VSDOT: vsdot */
	{ 0 }
},

{	/* ARM_VSDOTDI, ARM_INS_VSDOT: vsdot */
	{ 0 }
},

{	/* ARM_VSDOTQ, ARM_INS_VSDOT: vsdot */
	{ 0 }
},

{	/* ARM_VSDOTQI, ARM_INS_VSDOT: vsdot */
	{ 0 }
},

{	/* ARM_VSELEQD, ARM_INS_VSELEQ: vseleq */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSELEQH, ARM_INS_VSELEQ: vseleq */
	{ 0 }
},

{	/* ARM_VSELEQS, ARM_INS_VSELEQ: vseleq */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSELGED, ARM_INS_VSELGE: vselge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSELGEH, ARM_INS_VSELGE: vselge */
	{ 0 }
},

{	/* ARM_VSELGES, ARM_INS_VSELGE: vselge */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSELGTD, ARM_INS_VSELGT: vselgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSELGTH, ARM_INS_VSELGT: vselgt */
	{ 0 }
},

{	/* ARM_VSELGTS, ARM_INS_VSELGT: vselgt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSELVSD, ARM_INS_VSELVS: vselvs */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSELVSH, ARM_INS_VSELVS: vselvs */
	{ 0 }
},

{	/* ARM_VSELVSS, ARM_INS_VSELVS: vselvs */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSETLNi16, ARM_INS_VMOV: vmov */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSETLNi32, ARM_INS_FMDHR: fmdhr */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSETLNi8, ARM_INS_VMOV: vmov */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLLi16, ARM_INS_VSHLL: vshll */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLLi32, ARM_INS_VSHLL: vshll */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLLi8, ARM_INS_VSHLL: vshll */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLLsv2i64, ARM_INS_VSHLL: vshll */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLLsv4i32, ARM_INS_VSHLL: vshll */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLLsv8i16, ARM_INS_VSHLL: vshll */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLLuv2i64, ARM_INS_VSHLL: vshll */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLLuv4i32, ARM_INS_VSHLL: vshll */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLLuv8i16, ARM_INS_VSHLL: vshll */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLiv16i8, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLiv1i64, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLiv2i32, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLiv2i64, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLiv4i16, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLiv4i32, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLiv8i16, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLiv8i8, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHLsv16i8, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLsv1i64, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLsv2i32, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLsv2i64, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLsv4i16, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLsv4i32, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLsv8i16, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLsv8i8, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLuv16i8, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLuv1i64, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLuv2i32, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLuv2i64, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLuv4i16, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLuv4i32, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLuv8i16, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHLuv8i8, ARM_INS_VSHL: vshl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSHRNv2i32, ARM_INS_VSHRN: vshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRNv4i16, ARM_INS_VSHRN: vshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRNv8i8, ARM_INS_VSHRN: vshrn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRsv16i8, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRsv1i64, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRsv2i32, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRsv2i64, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRsv4i16, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRsv4i32, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRsv8i16, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRsv8i8, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRuv16i8, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRuv1i64, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRuv2i32, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRuv2i64, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRuv4i16, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRuv4i32, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRuv8i16, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHRuv8i8, ARM_INS_VSHR: vshr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSHTOD, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VSHTOH, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VSHTOS, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VSITOD, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSITOH, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VSITOS, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSLIv16i8, ARM_INS_VSLI: vsli */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSLIv1i64, ARM_INS_VSLI: vsli */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSLIv2i32, ARM_INS_VSLI: vsli */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSLIv2i64, ARM_INS_VSLI: vsli */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSLIv4i16, ARM_INS_VSLI: vsli */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSLIv4i32, ARM_INS_VSLI: vsli */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSLIv8i16, ARM_INS_VSLI: vsli */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSLIv8i8, ARM_INS_VSLI: vsli */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSLTOD, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VSLTOH, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VSLTOS, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VSQRTD, ARM_INS_VSQRT: vsqrt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSQRTH, ARM_INS_VSQRT: vsqrt */
	{ 0 }
},

{	/* ARM_VSQRTS, ARM_INS_VSQRT: vsqrt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAsv16i8, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAsv1i64, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAsv2i32, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAsv2i64, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAsv4i16, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAsv4i32, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAsv8i16, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAsv8i8, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAuv16i8, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAuv1i64, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAuv2i32, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAuv2i64, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAuv4i16, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAuv4i32, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAuv8i16, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRAuv8i8, ARM_INS_VSRA: vsra */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRIv16i8, ARM_INS_VSRI: vsri */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRIv1i64, ARM_INS_VSRI: vsri */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRIv2i32, ARM_INS_VSRI: vsri */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRIv2i64, ARM_INS_VSRI: vsri */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRIv4i16, ARM_INS_VSRI: vsri */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRIv4i32, ARM_INS_VSRI: vsri */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRIv8i16, ARM_INS_VSRI: vsri */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VSRIv8i8, ARM_INS_VSRI: vsri */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VST1LNd16, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1LNd16_UPD, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, 0 }
},

{	/* ARM_VST1LNd32, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, 0 }
},

{	/* ARM_VST1LNd32_UPD, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, 0 }
},

{	/* ARM_VST1LNd8, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1LNd8_UPD, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, 0 }
},

{	/* ARM_VST1d16, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d16Q, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d16Qwb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d16Qwb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d16T, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d16Twb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d16Twb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d16wb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d16wb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d32, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d32Q, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d32Qwb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d32Qwb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d32T, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d32Twb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d32Twb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d32wb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d32wb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d64, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d64Q, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d64Qwb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d64Qwb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d64T, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d64Twb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d64Twb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d64wb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d64wb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d8, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d8Q, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d8Qwb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d8Qwb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d8T, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d8Twb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d8Twb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d8wb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1d8wb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1q16, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1q16wb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1q16wb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1q32, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1q32wb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1q32wb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1q64, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1q64wb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1q64wb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1q8, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1q8wb_fixed, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST1q8wb_register, ARM_INS_VST1: vst1 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2LNd16,  ARM_INS_VST2: vst2${p}.16	\{$vd[$lane] $src2[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2LNd16_UPD,  ARM_INS_VST2: vst2${p}.16	\{$vd[$lane] $src2[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2LNd32,  ARM_INS_VST2: vst2${p}.32	\{$vd[$lane] $src2[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2LNd32_UPD,  ARM_INS_VST2: vst2${p}.32	\{$vd[$lane] $src2[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2LNd8,  ARM_INS_VST2: vst2${p}.8	\{$vd[$lane] $src2[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2LNd8_UPD,  ARM_INS_VST2: vst2${p}.8	\{$vd[$lane] $src2[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2LNq16,  ARM_INS_VST2: vst2${p}.16	\{$vd[$lane] $src2[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2LNq16_UPD,  ARM_INS_VST2: vst2${p}.16	\{$vd[$lane] $src2[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2LNq32,  ARM_INS_VST2: vst2${p}.32	\{$vd[$lane] $src2[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2LNq32_UPD,  ARM_INS_VST2: vst2${p}.32	\{$vd[$lane] $src2[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2b16, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2b16wb_fixed, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2b16wb_register, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2b32, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2b32wb_fixed, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2b32wb_register, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2b8, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2b8wb_fixed, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2b8wb_register, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2d16, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2d16wb_fixed, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2d16wb_register, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2d32, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2d32wb_fixed, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2d32wb_register, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2d8, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2d8wb_fixed, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2d8wb_register, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2q16, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2q16wb_fixed, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2q16wb_register, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2q32, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2q32wb_fixed, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2q32wb_register, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2q8, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2q8wb_fixed, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST2q8wb_register, ARM_INS_VST2: vst2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3LNd16,  ARM_INS_VST3: vst3${p}.16	\{$vd[$lane] $src2[$lane] $src3[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3LNd16_UPD,  ARM_INS_VST3: vst3${p}.16	\{$vd[$lane] $src2[$lane] $src3[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3LNd32,  ARM_INS_VST3: vst3${p}.32	\{$vd[$lane] $src2[$lane] $src3[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3LNd32_UPD,  ARM_INS_VST3: vst3${p}.32	\{$vd[$lane] $src2[$lane] $src3[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3LNd8,  ARM_INS_VST3: vst3${p}.8	\{$vd[$lane] $src2[$lane] $src3[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3LNd8_UPD,  ARM_INS_VST3: vst3${p}.8	\{$vd[$lane] $src2[$lane] $src3[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3LNq16,  ARM_INS_VST3: vst3${p}.16	\{$vd[$lane] $src2[$lane] $src3[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3LNq16_UPD,  ARM_INS_VST3: vst3${p}.16	\{$vd[$lane] $src2[$lane] $src3[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3LNq32,  ARM_INS_VST3: vst3${p}.32	\{$vd[$lane] $src2[$lane] $src3[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3LNq32_UPD,  ARM_INS_VST3: vst3${p}.32	\{$vd[$lane] $src2[$lane] $src3[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3d16, ARM_INS_VST3: vst3 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3d16_UPD, ARM_INS_VST3: vst3 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3d32, ARM_INS_VST3: vst3 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3d32_UPD, ARM_INS_VST3: vst3 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3d8, ARM_INS_VST3: vst3 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3d8_UPD, ARM_INS_VST3: vst3 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3q16, ARM_INS_VST3: vst3 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3q16_UPD, ARM_INS_VST3: vst3 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3q32, ARM_INS_VST3: vst3 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3q32_UPD, ARM_INS_VST3: vst3 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3q8, ARM_INS_VST3: vst3 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST3q8_UPD, ARM_INS_VST3: vst3 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4LNd16,  ARM_INS_VST4: vst4${p}.16	\{$vd[$lane] $src2[$lane] $src3[$lane] $src4[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4LNd16_UPD,  ARM_INS_VST4: vst4${p}.16	\{$vd[$lane] $src2[$lane] $src3[$lane] $src4[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4LNd32,  ARM_INS_VST4: vst4${p}.32	\{$vd[$lane] $src2[$lane] $src3[$lane] $src4[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4LNd32_UPD,  ARM_INS_VST4: vst4${p}.32	\{$vd[$lane] $src2[$lane] $src3[$lane] $src4[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4LNd8,  ARM_INS_VST4: vst4${p}.8	\{$vd[$lane] $src2[$lane] $src3[$lane] $src4[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4LNd8_UPD,  ARM_INS_VST4: vst4${p}.8	\{$vd[$lane] $src2[$lane] $src3[$lane] $src4[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4LNq16,  ARM_INS_VST4: vst4${p}.16	\{$vd[$lane] $src2[$lane] $src3[$lane] $src4[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4LNq16_UPD,  ARM_INS_VST4: vst4${p}.16	\{$vd[$lane] $src2[$lane] $src3[$lane] $src4[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4LNq32,  ARM_INS_VST4: vst4${p}.32	\{$vd[$lane] $src2[$lane] $src3[$lane] $src4[$lane]\} $rn */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4LNq32_UPD,  ARM_INS_VST4: vst4${p}.32	\{$vd[$lane] $src2[$lane] $src3[$lane] $src4[$lane]\} $rn$rm */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4d16, ARM_INS_VST4: vst4 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4d16_UPD, ARM_INS_VST4: vst4 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4d32, ARM_INS_VST4: vst4 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4d32_UPD, ARM_INS_VST4: vst4 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4d8, ARM_INS_VST4: vst4 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4d8_UPD, ARM_INS_VST4: vst4 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4q16, ARM_INS_VST4: vst4 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4q16_UPD, ARM_INS_VST4: vst4 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4q32, ARM_INS_VST4: vst4 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4q32_UPD, ARM_INS_VST4: vst4 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4q8, ARM_INS_VST4: vst4 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VST4q8_UPD, ARM_INS_VST4: vst4 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSTMDDB_UPD, ARM_INS_VPUSH: vpush */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VSTMDIA, ARM_INS_VSTMIA: vstmia */
	{ CS_AC_READ, 0 }
},

{	/* ARM_VSTMDIA_UPD, ARM_INS_VSTMIA: vstmia */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VSTMSDB_UPD, ARM_INS_VPUSH: vpush */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VSTMSIA, ARM_INS_VSTMIA: vstmia */
	{ CS_AC_READ, 0 }
},

{	/* ARM_VSTMSIA_UPD, ARM_INS_VSTMIA: vstmia */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VSTRD, ARM_INS_VSTR: vstr */
	{ CS_AC_READ, 0 }
},

{	/* ARM_VSTRH, ARM_INS_VSTR: vstr */
	{ 0 }
},

{	/* ARM_VSTRS, ARM_INS_VSTR: vstr */
	{ CS_AC_READ, 0 }
},

{	/* ARM_VSUBD, ARM_INS_FSUBD: fsubd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBH, ARM_INS_VSUB: vsub */
	{ 0 }
},

{	/* ARM_VSUBHNv2i32, ARM_INS_VSUBHN: vsubhn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBHNv4i16, ARM_INS_VSUBHN: vsubhn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBHNv8i8, ARM_INS_VSUBHN: vsubhn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBLsv2i64, ARM_INS_VSUBL: vsubl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBLsv4i32, ARM_INS_VSUBL: vsubl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBLsv8i16, ARM_INS_VSUBL: vsubl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBLuv2i64, ARM_INS_VSUBL: vsubl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBLuv4i32, ARM_INS_VSUBL: vsubl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBLuv8i16, ARM_INS_VSUBL: vsubl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBS, ARM_INS_FSUBS: fsubs */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBWsv2i64, ARM_INS_VSUBW: vsubw */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBWsv4i32, ARM_INS_VSUBW: vsubw */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBWsv8i16, ARM_INS_VSUBW: vsubw */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBWuv2i64, ARM_INS_VSUBW: vsubw */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBWuv4i32, ARM_INS_VSUBW: vsubw */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBWuv8i16, ARM_INS_VSUBW: vsubw */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBfd, ARM_INS_VSUB: vsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBfq, ARM_INS_VSUB: vsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBhd, ARM_INS_VSUB: vsub */
	{ 0 }
},

{	/* ARM_VSUBhq, ARM_INS_VSUB: vsub */
	{ 0 }
},

{	/* ARM_VSUBv16i8, ARM_INS_VSUB: vsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBv1i64, ARM_INS_VSUB: vsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBv2i32, ARM_INS_VSUB: vsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBv2i64, ARM_INS_VSUB: vsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBv4i16, ARM_INS_VSUB: vsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBv4i32, ARM_INS_VSUB: vsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBv8i16, ARM_INS_VSUB: vsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSUBv8i8, ARM_INS_VSUB: vsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VSWPd, ARM_INS_VSWP: vswp */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VSWPq, ARM_INS_VSWP: vswp */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VTBL1, ARM_INS_VTBL: vtbl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VTBL2, ARM_INS_VTBL: vtbl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VTBL3, ARM_INS_VTBL: vtbl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VTBL4, ARM_INS_VTBL: vtbl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VTBX1, ARM_INS_VTBX: vtbx */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VTBX2, ARM_INS_VTBX: vtbx */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VTBX3, ARM_INS_VTBX: vtbx */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VTBX4, ARM_INS_VTBX: vtbx */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VTOSHD, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VTOSHH, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VTOSHS, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VTOSIRD, ARM_INS_VCVTR: vcvtr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VTOSIRH, ARM_INS_VCVTR: vcvtr */
	{ 0 }
},

{	/* ARM_VTOSIRS, ARM_INS_VCVTR: vcvtr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VTOSIZD, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VTOSIZH, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VTOSIZS, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VTOSLD, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VTOSLH, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VTOSLS, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VTOUHD, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VTOUHH, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VTOUHS, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VTOUIRD, ARM_INS_VCVTR: vcvtr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VTOUIRH, ARM_INS_VCVTR: vcvtr */
	{ 0 }
},

{	/* ARM_VTOUIRS, ARM_INS_VCVTR: vcvtr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VTOUIZD, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VTOUIZH, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VTOUIZS, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VTOULD, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VTOULH, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VTOULS, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VTRNd16, ARM_INS_VTRN: vtrn */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VTRNd32, ARM_INS_VTRN: vtrn */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VTRNd8, ARM_INS_VTRN: vtrn */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VTRNq16, ARM_INS_VTRN: vtrn */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VTRNq32, ARM_INS_VTRN: vtrn */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VTRNq8, ARM_INS_VTRN: vtrn */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VTSTv16i8, ARM_INS_VTST: vtst */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VTSTv2i32, ARM_INS_VTST: vtst */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VTSTv4i16, ARM_INS_VTST: vtst */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VTSTv4i32, ARM_INS_VTST: vtst */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VTSTv8i16, ARM_INS_VTST: vtst */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VTSTv8i8, ARM_INS_VTST: vtst */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_VUDOTD, ARM_INS_VUDOT: vudot */
	{ 0 }
},

{	/* ARM_VUDOTDI, ARM_INS_VUDOT: vudot */
	{ 0 }
},

{	/* ARM_VUDOTQ, ARM_INS_VUDOT: vudot */
	{ 0 }
},

{	/* ARM_VUDOTQI, ARM_INS_VUDOT: vudot */
	{ 0 }
},

{	/* ARM_VUHTOD, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VUHTOH, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VUHTOS, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VUITOD, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VUITOH, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VUITOS, ARM_INS_VCVT: vcvt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_VULTOD, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VULTOH, ARM_INS_VCVT: vcvt */
	{ 0 }
},

{	/* ARM_VULTOS, ARM_INS_VCVT: vcvt */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_VUZPd16, ARM_INS_VUZP: vuzp */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VUZPd8, ARM_INS_VUZP: vuzp */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VUZPq16, ARM_INS_VUZP: vuzp */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VUZPq32, ARM_INS_VUZP: vuzp */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VUZPq8, ARM_INS_VUZP: vuzp */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VZIPd16, ARM_INS_VZIP: vzip */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VZIPd8, ARM_INS_VZIP: vzip */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VZIPq16, ARM_INS_VZIP: vzip */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VZIPq32, ARM_INS_VZIP: vzip */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_VZIPq8, ARM_INS_VZIP: vzip */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_sysLDMDA, ARM_INS_LDMDA: ldmda */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_sysLDMDA_UPD, ARM_INS_LDMDA: ldmda */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_sysLDMDB, ARM_INS_LDMDB: ldmdb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_sysLDMDB_UPD, ARM_INS_LDMDB: ldmdb */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_sysLDMIA, ARM_INS_LDM: ldm */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_sysLDMIA_UPD, ARM_INS_LDM: ldm */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_sysLDMIB, ARM_INS_LDMIB: ldmib */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_sysLDMIB_UPD, ARM_INS_LDMIB: ldmib */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_sysSTMDA, ARM_INS_STMDA: stmda */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_sysSTMDA_UPD, ARM_INS_STMDA: stmda */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_sysSTMDB, ARM_INS_STMDB: stmdb */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_sysSTMDB_UPD, ARM_INS_STMDB: stmdb */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_sysSTMIA, ARM_INS_STM: stm */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_sysSTMIA_UPD, ARM_INS_STM: stm */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_sysSTMIB, ARM_INS_STMIB: stmib */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_sysSTMIB_UPD, ARM_INS_STMIB: stmib */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2ADCri, ARM_INS_ADC: adc */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2ADCrr, ARM_INS_ADC: adc */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2ADCrs, ARM_INS_ADC: adc */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2ADDri, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2ADDri12, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2ADDrr, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2ADDrs, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2ADR, ARM_INS_ADD: add */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2ANDri, ARM_INS_AND: and */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2ANDrr, ARM_INS_AND: and */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2ANDrs, ARM_INS_AND: and */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2ASRri, ARM_INS_ASR: asr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2ASRrr, ARM_INS_ASR: asr */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2B, ARM_INS_B: b */
	{ 0 }
},

{	/* ARM_t2BFC, ARM_INS_BFC: bfc */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_t2BFI, ARM_INS_BFI: bfi */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2BICri, ARM_INS_AND: and */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2BICrr, ARM_INS_BIC: bic */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2BICrs, ARM_INS_BIC: bic */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2BXJ, ARM_INS_BXJ: bxj */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2Bcc, ARM_INS_B: b */
	{ 0 }
},

{	/* ARM_t2CDP, ARM_INS_CDP: cdp */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_IGNORE, 0 }
},

{	/* ARM_t2CDP2, ARM_INS_CDP2: cdp2 */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_IGNORE, 0 }
},

{	/* ARM_t2CLREX, ARM_INS_CLREX: clrex */
	{ 0 }
},

{	/* ARM_t2CLZ, ARM_INS_CLZ: clz */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2CMNri, ARM_INS_CMN: cmn */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2CMNzrr, ARM_INS_CMN: cmn */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2CMNzrs, ARM_INS_CMN: cmn */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2CMPri, ARM_INS_CMN: cmn */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2CMPrr, ARM_INS_CMP: cmp */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2CMPrs, ARM_INS_CMP: cmp */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2CPS1p, ARM_INS_CPS: cps */
	{ 0 }
},

{	/* ARM_t2CPS2p, ARM_INS_CPS: cps */
	{ 0 }
},

{	/* ARM_t2CPS3p, ARM_INS_CPS: cps */
	{ 0 }
},

{	/* ARM_t2CRC32B, ARM_INS_CRC32B: crc32b */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2CRC32CB, ARM_INS_CRC32CB: crc32cb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2CRC32CH, ARM_INS_CRC32CH: crc32ch */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2CRC32CW, ARM_INS_CRC32CW: crc32cw */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2CRC32H, ARM_INS_CRC32H: crc32h */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2CRC32W, ARM_INS_CRC32W: crc32w */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2DBG, ARM_INS_DBG: dbg */
	{ 0 }
},

{	/* ARM_t2DCPS1, ARM_INS_DCPS1: dcps1 */
	{ 0 }
},

{	/* ARM_t2DCPS2, ARM_INS_DCPS2: dcps2 */
	{ 0 }
},

{	/* ARM_t2DCPS3, ARM_INS_DCPS3: dcps3 */
	{ 0 }
},

{	/* ARM_t2DMB, ARM_INS_DMB: dmb */
	{ 0 }
},

{	/* ARM_t2DSB, ARM_INS_DFB: dfb */
	{ 0 }
},

{	/* ARM_t2EORri, ARM_INS_EOR: eor */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2EORrr, ARM_INS_EOR: eor */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2EORrs, ARM_INS_EOR: eor */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2HINT, ARM_INS_CSDB: csdb */
	{ 0 }
},

{	/* ARM_t2HVC, ARM_INS_HVC: hvc */
	{ 0 }
},

{	/* ARM_t2ISB, ARM_INS_ISB: isb */
	{ 0 }
},

{	/* ARM_t2IT, ARM_INS_IT: it */
	{ 0 }
},

{	/* ARM_t2LDA, ARM_INS_LDA: lda */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDAB, ARM_INS_LDAB: ldab */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDAEX, ARM_INS_LDAEX: ldaex */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDAEXB, ARM_INS_LDAEXB: ldaexb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDAEXD, ARM_INS_LDAEXD: ldaexd */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDAEXH, ARM_INS_LDAEXH: ldaexh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDAH, ARM_INS_LDAH: ldah */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDC2L_OFFSET, ARM_INS_LDC2L: ldc2l */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDC2L_OPTION, ARM_INS_LDC2L: ldc2l */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDC2L_POST, ARM_INS_LDC2L: ldc2l */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDC2L_PRE, ARM_INS_LDC2L: ldc2l */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDC2_OFFSET, ARM_INS_LDC2: ldc2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDC2_OPTION, ARM_INS_LDC2: ldc2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDC2_POST, ARM_INS_LDC2: ldc2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDC2_PRE, ARM_INS_LDC2: ldc2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDCL_OFFSET, ARM_INS_LDCL: ldcl */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDCL_OPTION, ARM_INS_LDCL: ldcl */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDCL_POST, ARM_INS_LDCL: ldcl */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDCL_PRE, ARM_INS_LDCL: ldcl */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDC_OFFSET, ARM_INS_LDC: ldc */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDC_OPTION, ARM_INS_LDC: ldc */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDC_POST, ARM_INS_LDC: ldc */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDC_PRE, ARM_INS_LDC: ldc */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LDMDB, ARM_INS_LDMDB: ldmdb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDMDB_UPD, ARM_INS_LDMDB: ldmdb */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDMIA, ARM_INS_LDM: ldm */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDMIA_UPD, ARM_INS_LDM: ldm */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRBT, ARM_INS_LDRBT: ldrbt */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRB_POST, ARM_INS_LDRB: ldrb */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRB_PRE, ARM_INS_LDRB: ldrb */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRBi12, ARM_INS_LDRB: ldrb */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRBi8, ARM_INS_LDRB: ldrb */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRBpci, ARM_INS_LDRB: ldrb */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRBs, ARM_INS_LDRB: ldrb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRD_POST, ARM_INS_LDRD: ldrd */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRD_PRE, ARM_INS_LDRD: ldrd */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRDi8, ARM_INS_LDRD: ldrd */
	{ CS_AC_WRITE, CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDREX, ARM_INS_LDREX: ldrex */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDREXB, ARM_INS_LDREXB: ldrexb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDREXD, ARM_INS_LDREXD: ldrexd */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDREXH, ARM_INS_LDREXH: ldrexh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRHT, ARM_INS_LDRHT: ldrht */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRH_POST, ARM_INS_LDRH: ldrh */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRH_PRE, ARM_INS_LDRH: ldrh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRHi12, ARM_INS_LDRH: ldrh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRHi8, ARM_INS_LDRH: ldrh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRHpci, ARM_INS_LDRH: ldrh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRHs, ARM_INS_LDRH: ldrh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRSBT, ARM_INS_LDRSBT: ldrsbt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRSB_POST, ARM_INS_LDRSB: ldrsb */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRSB_PRE, ARM_INS_LDRSB: ldrsb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRSBi12, ARM_INS_LDRSB: ldrsb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRSBi8, ARM_INS_LDRSB: ldrsb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRSBpci, ARM_INS_LDRSB: ldrsb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRSBs, ARM_INS_LDRSB: ldrsb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRSHT, ARM_INS_LDRSHT: ldrsht */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRSH_POST, ARM_INS_LDRSH: ldrsh */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRSH_PRE, ARM_INS_LDRSH: ldrsh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRSHi12, ARM_INS_LDRSH: ldrsh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRSHi8, ARM_INS_LDRSH: ldrsh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRSHpci, ARM_INS_LDRSH: ldrsh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRSHs, ARM_INS_LDRSH: ldrsh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LDRT, ARM_INS_LDRT: ldrt */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDR_POST, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDR_PRE, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRi12, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRi8, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRpci, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2LDRs, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LSLri, ARM_INS_LSL: lsl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LSLrr, ARM_INS_LSL: lsl */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2LSRri, ARM_INS_LSR: lsr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2LSRrr, ARM_INS_LSR: lsr */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2MCR, ARM_INS_MCR: mcr */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_IGNORE, 0 }
},

{	/* ARM_t2MCR2, ARM_INS_MCR2: mcr2 */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, CS_AC_IGNORE, 0 }
},

{	/* ARM_t2MCRR, ARM_INS_MCRR: mcrr */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2MCRR2, ARM_INS_MCRR2: mcrr2 */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2MLA, ARM_INS_MLA: mla */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2MLS, ARM_INS_MLS: mls */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2MOVTi16, ARM_INS_MOVT: movt */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_t2MOVi, ARM_INS_MOV: mov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2MOVi16, ARM_INS_MOV: mov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2MOVr, ARM_INS_LSL: lsl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2MOVsra_flag,  ARM_INS_ASR: asrs${p}.w	$rd $rm #1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2MOVsrl_flag,  ARM_INS_LSR: lsrs${p}.w	$rd $rm #1 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2MRC, ARM_INS_MRC: mrc */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_IGNORE, 0 }
},

{	/* ARM_t2MRC2, ARM_INS_MRC2: mrc2 */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_IGNORE, 0 }
},

{	/* ARM_t2MRRC, ARM_INS_MRRC: mrrc */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2MRRC2, ARM_INS_MRRC2: mrrc2 */
	{ CS_AC_READ, CS_AC_IGNORE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2MRS_AR, ARM_INS_MRS: mrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2MRS_M, ARM_INS_MRS: mrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2MRSbanked, ARM_INS_MRS: mrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2MRSsys_AR, ARM_INS_MRS: mrs */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2MSR_AR, ARM_INS_MSR: msr */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2MSR_M, ARM_INS_MSR: msr */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2MSRbanked, ARM_INS_MSR: msr */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2MUL, ARM_INS_MUL: mul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2MVNi, ARM_INS_MOV: mov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2MVNr, ARM_INS_MVN: mvn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2MVNs, ARM_INS_MVN: mvn */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2ORNri, ARM_INS_ORN: orn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2ORNrr, ARM_INS_ORN: orn */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2ORNrs, ARM_INS_ORN: orn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2ORRri, ARM_INS_ORN: orn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2ORRrr, ARM_INS_ORR: orr */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2ORRrs, ARM_INS_ORR: orr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2PKHBT, ARM_INS_PKHBT: pkhbt */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2PKHTB, ARM_INS_PKHTB: pkhtb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2PLDWi12, ARM_INS_PLDW: pldw */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2PLDWi8, ARM_INS_PLDW: pldw */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2PLDWs, ARM_INS_PLDW: pldw */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2PLDi12, ARM_INS_PLD: pld */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2PLDi8, ARM_INS_PLD: pld */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2PLDpci, ARM_INS_PLD: pld */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2PLDs, ARM_INS_PLD: pld */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2PLIi12, ARM_INS_PLI: pli */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2PLIi8, ARM_INS_PLI: pli */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2PLIpci, ARM_INS_PLI: pli */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2PLIs, ARM_INS_PLI: pli */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2QADD, ARM_INS_QADD: qadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2QADD16, ARM_INS_QADD16: qadd16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2QADD8, ARM_INS_QADD8: qadd8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2QASX, ARM_INS_QASX: qasx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2QDADD, ARM_INS_QDADD: qdadd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2QDSUB, ARM_INS_QDSUB: qdsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2QSAX, ARM_INS_QSAX: qsax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2QSUB, ARM_INS_QSUB: qsub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2QSUB16, ARM_INS_QSUB16: qsub16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2QSUB8, ARM_INS_QSUB8: qsub8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2RBIT, ARM_INS_RBIT: rbit */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2REV, ARM_INS_REV: rev */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2REV16, ARM_INS_REV16: rev16 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2REVSH, ARM_INS_REVSH: revsh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2RFEDB, ARM_INS_RFEDB: rfedb */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2RFEDBW, ARM_INS_RFEDB: rfedb */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2RFEIA, ARM_INS_RFEIA: rfeia */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2RFEIAW, ARM_INS_RFEIA: rfeia */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2RORri, ARM_INS_ROR: ror */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2RORrr, ARM_INS_ROR: ror */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2RRX, ARM_INS_RRX: rrx */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2RSBri, ARM_INS_NEG: neg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2RSBrr, ARM_INS_RSB: rsb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2RSBrs, ARM_INS_RSB: rsb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2SADD16, ARM_INS_SADD16: sadd16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SADD8, ARM_INS_SADD8: sadd8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SASX, ARM_INS_SASX: sasx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SBCri, ARM_INS_ADC: adc */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2SBCrr, ARM_INS_SBC: sbc */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SBCrs, ARM_INS_SBC: sbc */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2SBFX, ARM_INS_SBFX: sbfx */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2SDIV, ARM_INS_SDIV: sdiv */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SEL, ARM_INS_SEL: sel */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SETPAN, ARM_INS_SETPAN: setpan */
	{ 0 }
},

{	/* ARM_t2SG, ARM_INS_SG: sg */
	{ 0 }
},

{	/* ARM_t2SHADD16, ARM_INS_SHADD16: shadd16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SHADD8, ARM_INS_SHADD8: shadd8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SHASX, ARM_INS_SHASX: shasx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SHSAX, ARM_INS_SHSAX: shsax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SHSUB16, ARM_INS_SHSUB16: shsub16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SHSUB8, ARM_INS_SHSUB8: shsub8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMC, ARM_INS_SMC: smc */
	{ 0 }
},

{	/* ARM_t2SMLABB, ARM_INS_SMLABB: smlabb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLABT, ARM_INS_SMLABT: smlabt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLAD, ARM_INS_SMLAD: smlad */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLADX, ARM_INS_SMLADX: smladx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLAL, ARM_INS_SMLAL: smlal */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLALBB, ARM_INS_SMLALBB: smlalbb */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLALBT, ARM_INS_SMLALBT: smlalbt */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLALD, ARM_INS_SMLALD: smlald */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLALDX, ARM_INS_SMLALDX: smlaldx */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLALTB, ARM_INS_SMLALTB: smlaltb */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLALTT, ARM_INS_SMLALTT: smlaltt */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLATB, ARM_INS_SMLATB: smlatb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLATT, ARM_INS_SMLATT: smlatt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLAWB, ARM_INS_SMLAWB: smlawb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLAWT, ARM_INS_SMLAWT: smlawt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLSD, ARM_INS_SMLSD: smlsd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLSDX, ARM_INS_SMLSDX: smlsdx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLSLD, ARM_INS_SMLSLD: smlsld */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMLSLDX, ARM_INS_SMLSLDX: smlsldx */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMMLA, ARM_INS_SMMLA: smmla */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMMLAR, ARM_INS_SMMLAR: smmlar */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMMLS, ARM_INS_SMMLS: smmls */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMMLSR, ARM_INS_SMMLSR: smmlsr */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMMUL, ARM_INS_SMMUL: smmul */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMMULR, ARM_INS_SMMULR: smmulr */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMUAD, ARM_INS_SMUAD: smuad */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMUADX, ARM_INS_SMUADX: smuadx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMULBB, ARM_INS_SMULBB: smulbb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMULBT, ARM_INS_SMULBT: smulbt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMULL, ARM_INS_SMULL: smull */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMULTB, ARM_INS_SMULTB: smultb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMULTT, ARM_INS_SMULTT: smultt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMULWB, ARM_INS_SMULWB: smulwb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMULWT, ARM_INS_SMULWT: smulwt */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMUSD, ARM_INS_SMUSD: smusd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SMUSDX, ARM_INS_SMUSDX: smusdx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SRSDB, ARM_INS_SRSDB: srsdb */
	{ 0 }
},

{	/* ARM_t2SRSDB_UPD, ARM_INS_SRSDB: srsdb */
	{ 0 }
},

{	/* ARM_t2SRSIA, ARM_INS_SRSIA: srsia */
	{ 0 }
},

{	/* ARM_t2SRSIA_UPD, ARM_INS_SRSIA: srsia */
	{ 0 }
},

{	/* ARM_t2SSAT, ARM_INS_SSAT: ssat */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2SSAT16, ARM_INS_SSAT16: ssat16 */
	{ CS_AC_WRITE, CS_AC_IGNORE, CS_AC_READ, 0 }
},

{	/* ARM_t2SSAX, ARM_INS_SSAX: ssax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SSUB16, ARM_INS_SSUB16: ssub16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SSUB8, ARM_INS_SSUB8: ssub8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STC2L_OFFSET, ARM_INS_STC2L: stc2l */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STC2L_OPTION, ARM_INS_STC2L: stc2l */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STC2L_POST, ARM_INS_STC2L: stc2l */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STC2L_PRE, ARM_INS_STC2L: stc2l */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STC2_OFFSET, ARM_INS_STC2: stc2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STC2_OPTION, ARM_INS_STC2: stc2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STC2_POST, ARM_INS_STC2: stc2 */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STC2_PRE, ARM_INS_STC2: stc2 */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STCL_OFFSET, ARM_INS_STCL: stcl */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STCL_OPTION, ARM_INS_STCL: stcl */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STCL_POST, ARM_INS_STCL: stcl */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STCL_PRE, ARM_INS_STCL: stcl */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STC_OFFSET, ARM_INS_STC: stc */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STC_OPTION, ARM_INS_STC: stc */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STC_POST, ARM_INS_STC: stc */
	{ CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STC_PRE, ARM_INS_STC: stc */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STL, ARM_INS_STL: stl */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STLB, ARM_INS_STLB: stlb */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STLEX, ARM_INS_STLEX: stlex */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STLEXB, ARM_INS_STLEXB: stlexb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STLEXD, ARM_INS_STLEXD: stlexd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STLEXH, ARM_INS_STLEXH: stlexh */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STLH, ARM_INS_STLH: stlh */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STMDB, ARM_INS_STMDB: stmdb */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STMDB_UPD, ARM_INS_PUSH: push */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STMIA, ARM_INS_STM: stm */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STMIA_UPD, ARM_INS_STM: stm */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2STRBT, ARM_INS_STRBT: strbt */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2STRB_POST, ARM_INS_STRB: strb */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2STRB_PRE, ARM_INS_STRB: strb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2STRBi12, ARM_INS_STRB: strb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2STRBi8, ARM_INS_STRB: strb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2STRBs, ARM_INS_STRB: strb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2STRD_POST, ARM_INS_STRD: strd */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STRD_PRE, ARM_INS_STRD: strd */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STRDi8, ARM_INS_STRD: strd */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STREX, ARM_INS_STREX: strex */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2STREXB, ARM_INS_STREXB: strexb */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STREXD, ARM_INS_STREXD: strexd */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STREXH, ARM_INS_STREXH: strexh */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2STRHT, ARM_INS_STRHT: strht */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2STRH_POST, ARM_INS_STRH: strh */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2STRH_PRE, ARM_INS_STRH: strh */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2STRHi12, ARM_INS_STRH: strh */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2STRHi8, ARM_INS_STRH: strh */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2STRHs, ARM_INS_STRH: strh */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2STRT, ARM_INS_STRT: strt */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2STR_POST, ARM_INS_STR: str */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2STR_PRE, ARM_INS_STR: str */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2STRi12, ARM_INS_STR: str */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2STRi8, ARM_INS_STR: str */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2STRs, ARM_INS_STR: str */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_t2SUBS_PC_LR, ARM_INS_ERET: eret */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2SUBri, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2SUBri12, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2SUBrr, ARM_INS_SUB: sub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2SUBrs, ARM_INS_SUB: sub */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2SXTAB, ARM_INS_SXTAB: sxtab */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2SXTAB16, ARM_INS_SXTAB16: sxtab16 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2SXTAH, ARM_INS_SXTAH: sxtah */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2SXTB, ARM_INS_SXTB: sxtb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2SXTB16, ARM_INS_SXTB16: sxtb16 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2SXTH, ARM_INS_SXTH: sxth */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2TBB, ARM_INS_TBB: tbb */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2TBH, ARM_INS_TBH: tbh */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2TEQri, ARM_INS_TEQ: teq */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2TEQrr, ARM_INS_TEQ: teq */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2TEQrs, ARM_INS_TEQ: teq */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2TSB, ARM_INS_TSB: tsb */
	{ 0 }
},

{	/* ARM_t2TSTri, ARM_INS_TST: tst */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2TSTrr, ARM_INS_TST: tst */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2TSTrs, ARM_INS_TST: tst */
	{ CS_AC_READ, 0 }
},

{	/* ARM_t2TT, ARM_INS_TT: tt */
	{ 0 }
},

{	/* ARM_t2TTA, ARM_INS_TTA: tta */
	{ 0 }
},

{	/* ARM_t2TTAT, ARM_INS_TTAT: ttat */
	{ 0 }
},

{	/* ARM_t2TTT, ARM_INS_TTT: ttt */
	{ 0 }
},

{	/* ARM_t2UADD16, ARM_INS_UADD16: uadd16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UADD8, ARM_INS_UADD8: uadd8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UASX, ARM_INS_UASX: uasx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UBFX, ARM_INS_UBFX: ubfx */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2UDF, ARM_INS_UDF: udf */
	{ 0 }
},

{	/* ARM_t2UDIV, ARM_INS_UDIV: udiv */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UHADD16, ARM_INS_UHADD16: uhadd16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UHADD8, ARM_INS_UHADD8: uhadd8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UHASX, ARM_INS_UHASX: uhasx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UHSAX, ARM_INS_UHSAX: uhsax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UHSUB16, ARM_INS_UHSUB16: uhsub16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UHSUB8, ARM_INS_UHSUB8: uhsub8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UMAAL, ARM_INS_UMAAL: umaal */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UMLAL, ARM_INS_UMLAL: umlal */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UMULL, ARM_INS_UMULL: umull */
	{ CS_AC_WRITE, CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UQADD16, ARM_INS_UQADD16: uqadd16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UQADD8, ARM_INS_UQADD8: uqadd8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UQASX, ARM_INS_UQASX: uqasx */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UQSAX, ARM_INS_UQSAX: uqsax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UQSUB16, ARM_INS_UQSUB16: uqsub16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UQSUB8, ARM_INS_UQSUB8: uqsub8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2USAD8, ARM_INS_USAD8: usad8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2USADA8, ARM_INS_USADA8: usada8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2USAT, ARM_INS_USAT: usat */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_t2USAT16, ARM_INS_USAT16: usat16 */
	{ CS_AC_WRITE, CS_AC_IGNORE, CS_AC_READ, 0 }
},

{	/* ARM_t2USAX, ARM_INS_USAX: usax */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2USUB16, ARM_INS_USUB16: usub16 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2USUB8, ARM_INS_USUB8: usub8 */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_t2UXTAB, ARM_INS_UXTAB: uxtab */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2UXTAB16, ARM_INS_UXTAB16: uxtab16 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2UXTAH, ARM_INS_UXTAH: uxtah */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2UXTB, ARM_INS_UXTB: uxtb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2UXTB16, ARM_INS_UXTB16: uxtb16 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_t2UXTH, ARM_INS_UXTH: uxth */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tADC, ARM_INS_ADC: adc */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tADDhirr, ARM_INS_ADD: add */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tADDi3, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tADDi8, ARM_INS_ADD: add */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_tADDrSP, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_tADDrSPi, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tADDrr, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_tADDspi, ARM_INS_ADD: add */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_tADDspr, ARM_INS_ADD: add */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tADR, ARM_INS_ADR: adr */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_tAND, ARM_INS_AND: and */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tASRri, ARM_INS_ASR: asr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tASRrr, ARM_INS_ASR: asr */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tB, ARM_INS_B: b */
	{ 0 }
},

{	/* ARM_tBIC, ARM_INS_BIC: bic */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tBKPT, ARM_INS_BKPT: bkpt */
	{ 0 }
},

{	/* ARM_tBL, ARM_INS_BL: bl */
	{ 0 }
},

{	/* ARM_tBLXNSr, ARM_INS_BLXNS: blxns */
	{ 0 }
},

{	/* ARM_tBLXi, ARM_INS_BLX: blx */
	{ 0 }
},

{	/* ARM_tBLXr, ARM_INS_BLX: blx */
	{ CS_AC_READ, 0 }
},

{	/* ARM_tBX, ARM_INS_BX: bx */
	{ CS_AC_READ, 0 }
},

{	/* ARM_tBXNS, ARM_INS_BXNS: bxns */
	{ 0 }
},

{	/* ARM_tBcc, ARM_INS_B: b */
	{ 0 }
},

{	/* ARM_tCBNZ, ARM_INS_CBNZ: cbnz */
	{ CS_AC_READ, 0 }
},

{	/* ARM_tCBZ, ARM_INS_CBZ: cbz */
	{ CS_AC_READ, 0 }
},

{	/* ARM_tCMNz, ARM_INS_CMN: cmn */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_tCMPhir, ARM_INS_CMP: cmp */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_tCMPi8, ARM_INS_CMP: cmp */
	{ CS_AC_READ, 0 }
},

{	/* ARM_tCMPr, ARM_INS_CMP: cmp */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_tCPS, ARM_INS_CPS: cps */
	{ 0 }
},

{	/* ARM_tEOR, ARM_INS_EOR: eor */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tHINT, ARM_INS_HINT: hint */
	{ 0 }
},

{	/* ARM_tHLT, ARM_INS_HLT: hlt */
	{ 0 }
},

{	/* ARM_tLDMIA, ARM_INS_LDM: ldm */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_tLDRBi, ARM_INS_LDRB: ldrb */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_tLDRBr, ARM_INS_LDRB: ldrb */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_tLDRHi, ARM_INS_LDRH: ldrh */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_tLDRHr, ARM_INS_LDRH: ldrh */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_tLDRSB, ARM_INS_LDRSB: ldrsb */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_tLDRSH, ARM_INS_LDRSH: ldrsh */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_tLDRi, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tLDRpci, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tLDRr, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tLDRspi, ARM_INS_LDR: ldr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tLSLri, ARM_INS_LSL: lsl */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tLSLrr, ARM_INS_LSL: lsl */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tLSRri, ARM_INS_LSR: lsr */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tLSRrr, ARM_INS_LSR: lsr */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tMOVSr, ARM_INS_MOVS: movs */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tMOVi8, ARM_INS_MOV: mov */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_tMOVr, ARM_INS_MOV: mov */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tMUL, ARM_INS_MUL: mul */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_tMVN, ARM_INS_MVN: mvn */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tORR, ARM_INS_ORR: orr */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tPOP, ARM_INS_POP: pop */
	{ CS_AC_WRITE, 0 }
},

{	/* ARM_tPUSH, ARM_INS_PUSH: push */
	{ CS_AC_READ, 0 }
},

{	/* ARM_tREV, ARM_INS_REV: rev */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tREV16, ARM_INS_REV16: rev16 */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tREVSH, ARM_INS_REVSH: revsh */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tROR, ARM_INS_ROR: ror */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tRSB, ARM_INS_NEG: neg */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tSBC, ARM_INS_SBC: sbc */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tSETEND, ARM_INS_SETEND: setend */
	{ 0 }
},

{	/* ARM_tSTMIA_UPD, ARM_INS_STM: stm */
	{ CS_AC_READ | CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tSTRBi, ARM_INS_STRB: strb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_tSTRBr, ARM_INS_STRB: strb */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_tSTRHi, ARM_INS_STRH: strh */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_tSTRHr, ARM_INS_STRH: strh */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_tSTRi, ARM_INS_STR: str */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_tSTRr, ARM_INS_STR: str */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_tSTRspi, ARM_INS_STR: str */
	{ CS_AC_READ, CS_AC_WRITE, 0 }
},

{	/* ARM_tSUBi3, ARM_INS_ADD: add */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tSUBi8, ARM_INS_ADD: add */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_tSUBrr, ARM_INS_SUB: sub */
	{ CS_AC_WRITE, CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_tSUBspi, ARM_INS_ADD: add */
	{ CS_AC_READ | CS_AC_WRITE, 0 }
},

{	/* ARM_tSVC, ARM_INS_SVC: svc */
	{ 0 }
},

{	/* ARM_tSXTB, ARM_INS_SXTB: sxtb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tSXTH, ARM_INS_SXTH: sxth */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tTRAP, ARM_INS_TRAP: trap */
	{ 0 }
},

{	/* ARM_tTST, ARM_INS_TST: tst */
	{ CS_AC_READ, CS_AC_READ, 0 }
},

{	/* ARM_tUDF, ARM_INS_UDF: udf */
	{ 0 }
},

{	/* ARM_tUXTB, ARM_INS_UXTB: uxtb */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},

{	/* ARM_tUXTH, ARM_INS_UXTH: uxth */
	{ CS_AC_WRITE, CS_AC_READ, 0 }
},
